// *****************************************************************************
//
//  Routine:      acReadITR
//  Author:       T.Glebe, I.Kisel
//                Modified for the ITR: W. Gradl
//
//  Description:  call the ITR initialization to get the ITR constants
//                - if key not assigned, use KUIP variables as major,minor
//                - returns <0 error
//                           0 no ITR update necessary
//                           1 update done using key
//                           2 update done using date/time or run number
//                
//
//  Modification:	21.03.2000 - added code used by L2FORMARTE (fg sciacca)
//			02.04.2000 - fixed handling of keytable mechanism (fg sciacca)
//			11.04.2000 - bugs removed
//				   - L2FORMARTE code integrated with VDS code (fg sciacca)
//			30.06.2000 - bu fix: now ok with unsparsified data (fg sciacca)
//                      10.08.2000 - WGradl: adapted routines for the ITR
//                      13.11.2000 - TZ: new routine for update of ITR constants using
//                                   keytable mechanism
//                      15.11.2000 - AG: introduced ARTE_mark()
//                      28.11.2000 - WG: exit ARTE gracefully by calling ARTE_exit() in 
//                                   case of failure while reading the database
//                      12.12.2000 - WG: bugfix:  also in non-key-table-mode, the 
//                                   database server can now be set in the user's kumac
//                                   using the environment variables
//                                   Read ITR/Chanstat from monitor server (default /ITR_MONITOR)
//                                   Code cleanup.
//                      20.12.2000 - WG: evaluate return code of ITR_arte_updateGEDE()
//                      12.09.2001 - WG: incorporate changes to work also with Monte Carlo 
//                                   and keytables
//                      01.03.2002 - WG: include ITR/EffRes in list of tables read using
//                                   keytable mechanism.  Needed for MC and keytables.
//                      06.03.2002 - TZ: some changes to adapt digitisation to keytable
//                                   mechanism. ITR_MONITOR DB version can be steered by
//                                   kumac. Filling of OTDR table added.
// $Id: acReadITR.C,v 1.8 2002/03/29 17:37:16 irisha Exp $
//
// *****************************************************************************

using namespace std;


#include <stdio.h>

#include "arte/arte.hh"
#include "database/keytable.h"
#include "farm/farm.hh"
#include "grr.h"

#include "itrleda.h"
#include "ITR_marvin.h"

#include "itrarte/ITR_arte.h"
#include "itrarte/ITR_steer.h"


// sparse event definition
#include "event.h"
#include "etb.h"

// static variables global for this module
static int ITRserver =	-1;
static int ITRserver_mon = -1;
static int ITRserver_align = -1;
static int ITRtable_chamber =	-1;
static int ITRtable_chip =	-1;
static int ITRtable_nede =	-1;
static int ITRtable_nesl =	-1;
static int ITRtable_fed =	-1;
static int ITRtable_chanstat =	-1;
static int ITRtable_effres   =	-1;
static int ITRtable_ped =	-1;
static int ITRtable_noise =	-1;
static int ITRtable_align =	-1;

extern int ITR_MASK;

// *****************************************************************
// function
// *****************************************************************

int acReadITR(int KTstatus) 
{
  //==================================================================
  // no action for MC
  //==================================================================
#if 0
  if (ArteTable<MTRA>::size() > 0 ) return 0;
#endif
  ARTE_mark( "acReadITR 0", 0 );
    
  //==================================================================
  // database server (default: /ITR_Setup)
  //                  can be changed by setting DB_ITR
  //==================================================================
 
  char default_setup[]   = "/ITR_Setup";
  char default_monitor[] = "/ITR_MONITOR";
  char default_align[]   = "/ITR_Align";

  const char *kSetupEnv = "ITR_SETUPDB";
  const char *kAlignEnv = "ITR_ALIGNDB";
  const char *kMonitorEnv= "ITR_MONITORDB";

  char *dbserver, *dbserver_monitor, *dbserver_align;
  
  static bool firsttime = true;

  if (getenv(kSetupEnv) == NULL){
    dbserver = default_setup;
  } else {
    dbserver = getenv(kSetupEnv);
  }
 
  if (getenv (kAlignEnv) == NULL) {
    dbserver_align = default_align;
  } else {
    dbserver_align = getenv (kAlignEnv);
  }
  
  if (getenv (kMonitorEnv) == NULL) {
    dbserver_monitor = default_monitor;
  } else {
    dbserver_monitor = getenv (kMonitorEnv);
  }

  
  if (firsttime) {
    ITR_leda_param_init();
    ITR_marvin_param_init();
  }

  switch (KTstatus) {

    case 2: 
      // structure of keytable changed, so loop over the keytable and get 
      // your serverid and tableids 

      ITRserver       = KeyFindServer(dbserver); 
      ITRserver_mon   = KeyFindServer(dbserver_monitor); 
      ITRserver_align = KeyFindServer(dbserver_align); 

      if (ITR_DBVERS_MAJ != -1 && ITR_DBVERS_MIN != -1) {
	cout << "acReadITR: === Version of Setup Tables manually set to [" 
	     << ITR_DBVERS_MAJ << ", " << ITR_DBVERS_MIN << "]" << endl;
	cout << "           === Ignoring key table entry" << endl;
	break; // use old method
      }


      if (ITRserver == -1) { 
        cout<<"acReadITR: KeyTable Server " << dbserver << " not found!"<<endl; 
        break;  // use old method
      }
      if (ITRserver_mon == -1) { 
        cout<<"acReadITR: KeyTable Server " << dbserver_monitor << " not found!"<<endl; 
        break;  // use old method
      } 
      if (ITRserver == -1) { 
        cout<<"acReadITR: KeyTable Server " << dbserver_align << " not found!"<<endl; 
        break;  // use old method
      } 

      cout << "acReadITR: Looking for Keytable..." << endl;
      ITRtable_chamber =	KeyFindTable(ITRserver, "ITR/Chamber"); 
      ITRtable_chip =		KeyFindTable(ITRserver, "ITR/Chip"); 
      ITRtable_fed =		KeyFindTable(ITRserver, "ITR/Fed"); 
      ITRtable_nede =		KeyFindTable(ITRserver, "ITR/nede"); 
      ITRtable_nesl =		KeyFindTable(ITRserver, "ITR/nesl"); 
      ITRtable_chanstat =	KeyFindTable(ITRserver_mon, "ITR/Chanstat"); 
      ITRtable_effres  =	KeyFindTable(ITRserver_mon, "ITR/EffRes"); 
      ITRtable_ped =    	KeyFindTable(ITRserver_mon, "ITR/Ped"); 
      ITRtable_noise =  	KeyFindTable(ITRserver_mon, "ITR/Noise"); 
      ITRtable_align =  	KeyFindTable(ITRserver_align, "ITR/Align"); 

      if (ITRtable_effres == -1) {
	cout << "acReadITR: Warning:  Table ITR/EffRes not handled by keytable mechanism."
	     << endl;
      }


      if ((ITRtable_chamber == -1) || (ITRtable_chip == -1) || 
	  (ITRtable_fed == -1) || (ITRtable_chanstat == -1) || 
	  (ITRtable_nede == -1) || (ITRtable_nesl == -1)) { 
        cout<<"acReadITR: Keytable not found!" << endl; 
	break;  // use old method
      }; 
      // NO _BREAK_ HERE! 

    case 1: 
      // an update happened, look if your server is involved 

      if (!KeyServerNew(ITRserver) && !KeyServerNew(ITRserver_align) && !KeyServerNew(ITRserver_mon)) {
        cout<<"acReadITR: no ITR update necessary, server not involved"<<endl;      
        cout<<"acReadITR: returning with rc = 0"<<endl;
	ARTE_mark( "acReadITR 1", 0 );
      return 0; //no ITR update necessary
      };

      // now for the tables - look whether they have changed 
      if (KeyTableNew(ITRtable_chip) && ITR_DBVERS_MAJ == -1) { 
        // using LEDA to get the Tables x:
        MizziDatabase(mydb);
        Tables x;

	cout<<"acReadITR: keytable found: opening ITR database "<< ITRserver <<endl;
        mydb.open(KeyServerName(ITRserver),DBreadonly);
        x=mydb.get_id();

        ITR_Chamber	->book(mydb);
        ITR_Chip	->book(mydb);
        ITR_Fed		->book(mydb);
        // ITR_Chanstat	->book(mydb);
	ITR_nede	->book(mydb);
	ITR_nesl	->book(mydb);

        ITR_Chamber	->remove-> all();
        ITR_Chip	->remove-> all();
        ITR_Fed		->remove-> all();
        ITR_Chanstat	->remove-> all();
        ITR_EffRes	->remove-> all();
	ITR_nede	->remove-> all();
	ITR_nesl	->remove-> all();

	int rc;
	
	rc = true;
	
	cout << "acReadITR: reading ITR/Chamber [" 
	     << KeyTableMajor (ITRtable_chamber) << ", " 
	     << KeyTableMinor (ITRtable_chamber) << "]" << endl;
	KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_chamber));
	rc = rc && ITR_Chamber->fetch_data( DBeq );
	rc = rc && ITR_Chamber->set_keys();


	cout << "acReadITR: reading ITR/Chip [" 
	     << KeyTableMajor (ITRtable_chip) << ", " 
	     << KeyTableMinor (ITRtable_chip) << "]" << endl;
	KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_chip));
	rc = rc && ITR_Chip	->fetch_data( DBeq );
	rc = rc && ITR_Chip	->set_keys();
	  
	cout << "acReadITR: reading ITR/Fed [" 
	     << KeyTableMajor (ITRtable_fed) << ", " 
	     << KeyTableMinor (ITRtable_fed) << "]" << endl;
	KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_fed));
	rc = rc && ITR_Fed	->fetch_data( DBeq );
	rc = rc && ITR_Fed	->set_keys();

	cout << "acReadITR: reading ITR/nede [" 
	     << KeyTableMajor (ITRtable_nede) << ", " 
	     << KeyTableMinor (ITRtable_nede) << "]" << endl;
	KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_nede));
	rc = rc && ITR_nede	->fetch_data( DBeq );
	rc = rc && ITR_nede	->set_keys();

	cout << "acReadITR: reading ITR/nesl [" 
	     << KeyTableMajor (ITRtable_nesl) << ", " 
	     << KeyTableMinor (ITRtable_nesl) << "]" << endl;
	KeyTableStamp(x,KeyServerName(ITRserver),KeyTableName(ITRtable_nesl));
	rc = rc && ITR_nesl	->fetch_data( DBeq );
	rc = rc && ITR_nesl	->set_keys();
	  
	if (rc == false) {
	  ARTE_exit (GRR_SFATAL, "acReadITR: Error reading ITR setup database!", 0);
	  return (GRR_FISHY);
	}


	ITR_Chamber	->fetch_relations(DBeq); /* does not load any table */
	ITR_Fed		->fetch_relations(DBeq); /* does not load any table */
	ITR_nede	->fetch_relations(DBeq);
	ITR_nesl	->fetch_relations(DBeq);
	ITR_Chip	->fetch_relations(DBeq); /* loads fed and side relations */

        ITR_Chamber	->unbook();
        ITR_Chip	->unbook();
        ITR_Fed		->unbook();
        ITR_nede	->unbook();
        ITR_nesl	->unbook();

	cout<<"acReadITR: ITR tables loaded: closing ITR database "<< ITRserver <<endl;
      }  

      if (KeyTableNew (ITRtable_chanstat)) {
	// int monMajor, monMinor;
	MizziDatabase(mydb_mon);
        Tables x;

        mydb_mon.open(KeyServerName(ITRserver_mon),DBreadonly);
        x=mydb_mon.get_id();

	int rc = true;
	rc = rc && ITR_Chanstat->book (mydb_mon);

	if (rc == false) {
	  cout << "acReadITR: failed to book() table ITR_Chanstat" << endl;
	}

	cout << "acReadITR: reading ITR monitor information from Server " 
	     << dbserver_monitor << endl ;

	cout << "acReadITR: reading ITR/Chanstat [" 
	     << KeyTableMajor (ITRtable_chanstat) << ", " 
	     << KeyTableMinor (ITRtable_chanstat) << "]" << endl;
	KeyTableStamp(x,KeyServerName(ITRserver_mon),KeyTableName(ITRtable_chanstat));
	rc = rc && ITR_Chanstat	->fetch_data( DBeq );
	rc = rc && ITR_Chanstat	->set_keys();
	rc = rc && ITR_Chanstat->fetch_relations (DBeq);
	
	if (rc == false) {
	  ARTE_exit (GRR_SFATAL, "acReadITR: Error reading ITR monitor database!", 0);
	  return (GRR_FISHY);
	}
	
	ITR_Chanstat->unbook();
      }

      if (KeyTableNew (ITRtable_effres)) {
	// int monMajor, monMinor;
	MizziDatabase(mydb_mon);
        Tables x;

        mydb_mon.open(KeyServerName(ITRserver_mon),DBreadonly);
        x=mydb_mon.get_id();

	int rc = true;
	rc = rc && ITR_EffRes->book (mydb_mon);

	if (rc == false) {
	  cout << "acReadITR: failed to book() table ITR_EffRes" << endl;
	}


	if (KeyTableMajor (ITRtable_effres) > 0) {
	  cout << "acReadITR: reading ITR monitor information from Server " 
	       << dbserver_monitor << endl ;
	  
	  cout << "acReadITR: reading ITR/EffRes [" 
	       << KeyTableMajor (ITRtable_effres) << ", " 
	       << KeyTableMinor (ITRtable_effres) << "]" << endl;
	  KeyTableStamp(x,KeyServerName(ITRserver_mon),KeyTableName(ITRtable_effres));
	  rc = rc && ITR_EffRes	->fetch_data( DBeq );
	  rc = rc && ITR_EffRes	->set_keys();
	  rc = rc && ITR_EffRes   ->fetch_relations (DBeq);
	}
	else 
	  rc = false;

	if (rc == false) {
	  grr_error (GRR_SWARNING, 
		     "acReadITR: Error reading EffRes table from ITR monitor database!", 0);
	  ITR_leda_inventEffRes();
	}
	
	ITR_EffRes->unbook();
      }

      if (KeyTableNew(ITRtable_align) && ITR_ALIGN_MAJ == -1) {
	int alignMajor, alignMinor;
	alignMajor = KeyTableMajor (ITRtable_align);
	alignMinor = KeyTableMinor (ITRtable_align);
	cout << "acReadITR: reading ITR/Align information from Server " 
	     << dbserver_align << " for [" << alignMajor << "," << alignMinor << "]" << endl;
	ITR_leda_fetch_alignment (dbserver_align, alignMajor, alignMinor); 
      }
      else if (ITR_ALIGN_MAJ != -1) {
	cout << "acReadITR: === Version of alignment table set manually to [" 
	     << ITR_ALIGN_MAJ << "," << ITR_ALIGN_MIN << "]" << endl;
	cout << "           === Ignoring key table entry " << endl;
	
	ITR_leda_fetch_alignment (dbserver_align, ITR_ALIGN_MAJ, ITR_ALIGN_MIN);
      }


      if(firsttime) {	
	firsttime = false;
	ITR_leda_startprogram (-1, -1);
      }
      
      // Only update GEDEs and initialize Marvin in Data mode

      if (ArteTable<MTRA>::size() == 0 ){ 
	ITR_marvin_init();

	// ITR_arte_updateGEDE returns 
	//    -1 in case of failure
	//     0 in case of success
	int rc = ITR_arte_updateGEDE();
	
	if (rc < 0) {
	  char mess[] = "ITR_arte_updateGEDE failed: no alignment used for ITR";
	  ARTE_error(GRR_SWARNING, mess, 0);
	}

        if (ITR_MASK)
          ITR_fill_OTDR_table();  

	
      }
      else {  // MC:  pass channel status and efficiencies to digitisation

	// call MC functions here

        if (ITR_MASK)
          ITR_fill_OTDR_table();  

	int rc = setITRDIGI_from_DB(); 
        rc = setITRDIGIALIGN_from_DB();
   
      }

      cout<<"acReadITR: returning with rc = 1"<< dec << endl;
      ARTE_mark( "acReadITR 1", 0 );
      return 1; // update using keytable
      //break; 
  } // of switch 


  // No keytable information available, using old method...
  if(firsttime) {
    cout << "acReadITR: Key NOT assigned --> using old method" << endl; 
  }
  
  static int ITR_Chamber_major = -1;
  static bool findmajmin = true;
  static MizziDatabase( itrdb_setup );

  
  //    Get constants from database
  //
  // Time stamp is evaluated in this order:
  //  1. date/time   from itrleda IOL
  //  2. major/minor from itrleda IOL
  //  3. event time from DDAQ (if all IOL parameters are =0)
  int major = 0;
  int minor = 0;

  if (ITR_DBVERS_MAJ != -1 && ITR_DBVERS_MIN != -1) {
    findmajmin = false;
    major = ITR_DBVERS_MAJ;
    minor = ITR_DBVERS_MIN;
  } else {


    // set some default versions for DB 

    major = 11;
    minor = 1;
    ITR_MONITOR_MAJ = 11;
    ITR_MONITOR_MIN = 1;
    ITR_ALIGN_MAJ = 17137;
    ITR_ALIGN_MIN = 0;

    cout << "acReadITR:  *** using default DB versions Setup ["
	 << major << ", " << minor << "]" << "  MONITOR ["
         << ITR_MONITOR_MAJ << ", " << ITR_MONITOR_MIN << "]" << "  ALIGN ["
         << ITR_ALIGN_MAJ << ", " << ITR_ALIGN_MIN << "]"
	 << endl;
  } 

  /* now overwrite server names with environment variables, if they are set */
  int module_matches, verb_matches;
  if (getenv (kSetupEnv) != NULL) {
    iol_load_value_from_mod_verb(getenv (kSetupEnv), "itrleda",
				 "setup database", &module_matches, &verb_matches); 
  }

  if (getenv (kMonitorEnv) != NULL) {
    iol_load_value_from_mod_verb(getenv (kMonitorEnv), "itrleda",
				 "mon database", &module_matches, &verb_matches); 
  }

  if (getenv (kAlignEnv) != NULL) {
    iol_load_value_from_mod_verb(getenv (kAlignEnv), "itrleda",
				 "align database", &module_matches, &verb_matches); 
  }
  

  if(ITR_Chamber_major == -1 || 
     (major > ITR_Chamber_major && !findmajmin)) {

    // setup iol parameters for (old style) database access
    extern Itrleda_Iol_t itrleda_iol;

    /* decide which alignment/monitor version to take: 
     *  1. explicitly set with KUIP commands? 
     *  2. implicitly by setting setup major/minor? 
     *  3. no versions set -> take those of steering dbase read
     */
    if (ITR_ALIGN_MAJ != -1) {               // 1.
	cout << "acReadITR:  *** Version of alignment table set manually to [" 
	     << ITR_ALIGN_MAJ << "," << ITR_ALIGN_MIN << "]" << endl;

	itrleda_iol.align_major.value = ITR_ALIGN_MAJ;
	itrleda_iol.align_minor.value = ITR_ALIGN_MIN;
    } else if (major != 0 && minor != 0) {   // 2.
 	itrleda_iol.align_major.value = major;
	itrleda_iol.align_minor.value = minor;
    } else {                                 // 3.
      /* do nothing */
    }

    if (ITR_MONITOR_MAJ != -1) {               // 1.
	cout << "acReadITR:  *** Version of monitor DB set manually to [" 
	     << ITR_MONITOR_MAJ << "," << ITR_MONITOR_MIN << "]" << endl;

	itrleda_iol.monitor_major.value = ITR_MONITOR_MAJ;
	itrleda_iol.monitor_minor.value = ITR_MONITOR_MIN;
    } else if (major != 0 && minor != 0) {   // 2.
 	itrleda_iol.monitor_major.value = major;
	itrleda_iol.monitor_minor.value = minor;
    } else {                                 // 3.
      /* do nothing */
    }

    int rc;
    rc = ITR_leda_startprogram (major, minor);
    
    if (rc != GRR_OK) {
      cout <<"acReadITR: " << "Reading of setup database" << " not successful" << endl;

      ITR_HITPREP_FAILED=1;
      ARTE_mark( "acReadITR 1", 0 );
      ARTE_exit( GRR_SFATAL, "acReadITR: Error reading ITR setup database!", 0 );

      return (GRR_FISHY);
    }

    // Only update GEDEs and initialize Marvin in Data mode
    if (ArteTable<MTRA>::size() == 0 ){ 
      ITR_marvin_init();
      
      int rc = ITR_arte_updateGEDE();
      if (ITR_MASK)
        ITR_fill_OTDR_table();
        
      if(firsttime) {		//silraw and marvin init must occurr after DB access
	firsttime = false;
      }
    }
    else {  // MC:  pass channel status and efficiencies to digitisation

      // call MC functions here

      if (ITR_MASK)
        ITR_fill_OTDR_table();  

      int rc = setITRDIGI_from_DB(); 
      rc = setITRDIGIALIGN_from_DB();
    }

  } else {
    ARTE_mark( "acReadITR 1", 0 );
    return 0; // no update necessary
  }


  cout<<"acReadITR: returning with rc = 2"<<endl;
  ARTE_mark( "acReadITR 1", 0 );
  return 2;  // constants read in using old method 
}
