//                              -*- Mode: C++ -*-
//: EAHypAnalyzer.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 2/5/97
// Last Update ........
//

#define ESEM
#include "care/EMClusterTree.hh"
#include "care/EMTrackOnEcal.hh"
#include "care/EAHypAnalyzer.hh"
#include "arte/ArteMessage.hh"
#include "care/ECStrategyFactory.hh"
#include "care/ESCorCOGPos.hh"
#ifdef ESEM
#  include "care/ESEMSPos.hh"
#endif
#include "arte/ArteProto.hh"
#include "care/caredata.hh"
#include <math.h>
#include "care/ECCellID.hh"
#include "care/CmdCareLevel.hh"
#include "care/MMBank.hh"
#undef WRITE
#undef DEBUG
/* 
CLASS
  EAHypAnalyzer
  Class for finding the hypothetical showers.
  It stores a vector of pointers to the hyp showers.


*/

//////////////////// EAHypAnalyzer Class Definitions ////////////////

// No functions
EAHypAnalyzer::EAHypAnalyzer(std::vector<const EMCluster*> & hypSh) :
  hypShowers(&hypSh) {

  ECStrategyFactory  stratf;
  poscog = stratf.PositionStrategy(); // return current way to evaluate pos.
  // return current way to evaluate position
  poscor = (ESCorCOGPos*) stratf.PositionStrategy(E_CorCOG);
  // get the current PID strategy
  pidstra = stratf.PidStrategy();
  // initializa data for relational bank
  nrccl = 0;
}

EAHypAnalyzer::~EAHypAnalyzer(){
//  delete poscog;
//  delete poscor;
//  delete pidstra;
}

//--
//  Base function to be called with the TopCluster List
//
void EAHypAnalyzer::analyzeTopList(const EMClusterTree & topList){
  doAnalysis(&topList);
}

//--
//  Base function to be called with the TopCluster List
//
void EAHypAnalyzer::doAnalysis(const EMClusterTree * tree){
  if(hypShowers->begin()!=hypShowers->end()){
    // erase all data
    hypShowers->erase(hypShowers->begin(),hypShowers->end());
  }
# ifndef ONLINE
  //--
  //  test if RCCL bank is empty (it must be empty)
  if( ArtePointer<RCCL>::getLink()()!=0 ){
    int nrow = ArtePointer<RCCL>::nrow();
    // drop the bank
    MMBank<RCCL>::Drop();
    //
    static int EAHYPWARN = 0xECA0004;
    arteMsg<<"EAHypAnalyzer: RCCL bank not void: "
           <<nrow<<" entries ";
    arteMsg.warn(EAHYPWARN,40);
  }
# endif
  nrccl = 0;
  for(EMClusterMember* child = tree->begin(); child!=tree->end(); 
      child = child->next()){
    if( !child->isALeaf()){
      child->analyzeMe(*this); 
    }
  }
}

//--
// Cluster Analyzer
void EAHypAnalyzer::doAnalysis(const EMCluster * clus){
  if(clus->getHypotheses()==1){
    // Found one shower hypotheses: store it
    hypShowers->push_back(clus);
    //--
    //  Evaluate the information for the cluster and fill an entry 
    fillNewEntry(clus);  
  } else {
    // Analyze Cluster that is not a shower one
    // perform operation on all elements of the cluster
    for(EMClusterMember* child = clus->begin(); 
	child!=clus->end();  child = child->next())
      child->analyzeMe(*this);
  }
}

//--
// EMHitCell Analyzer
void EAHypAnalyzer::doAnalysis(const EMHitCell *){}

//--
// EMTrackOnEcal Analyzer
void EAHypAnalyzer::doAnalysis(const EMTrackOnEcal *){}

//--
// EMClusterMember Analyzer
void EAHypAnalyzer::doAnalysis(const EMClusterMember *){}

//--
//  Evaluate the information for the cluster and fill a new entry in the bank
//
void EAHypAnalyzer::fillNewEntry(const EMCluster * cluster){
  // get a table entry
# ifndef ONLINE
  MMBank<RCCL>::Book(1);
# else
  RCCL x;
  ArtePointer<RCCL>::append(x);
# endif

  ArtePointer<RCCL> rccl(ArtePointer<RCCL>::nrow());
  //
  // Now the entry exists and must be updated...
  updateOldEntry(cluster, rccl);
  // now fill the relations
  //rcLinks.fillNewEntry(cluster, rccl);
#ifdef WRITE
  // test links
  //testLinks(cluster, rccl);
#endif
}

//--
//  Evaluate the information for the cluster and 
//  fill an old entry in the bank
//
void EAHypAnalyzer::updateOldEntry(const EMCluster * clus, 
				   ArtePointer<RCCL> rccl){

  int allcells = 0;
  //
  // Evaluate all informations to store in a bank
  //
  poscog->evaluatePos(clus); // evaluate center of gravity

  // now writeout on bank
  const ESPosition * pos = & poscog->getData();
  // 
  // COG info
  rccl->xcog = pos->x();
  rccl->ycog = pos->y();
  for(int i=0; i<3; i++){
    rccl->vcog[i] = pos->var(i);
  }
  // 
  // Corrected COG info
  poscog->evaluatePos(clus); // evaluate center of gravity

  allcells = poscog->getData().width();

  if( CmdCareLevel::level()>2){
#   ifndef ESEM
      poscor->correct(poscog->getData()); //evaluate correction 
#   else
      poscor->evaluatePos(clus); // evaluate correction  
#   endif
    pos = & poscor->getData();
  } else {
    pos = & poscog->getData();  // No correction for level 2
  }

  ESPosition posi(*pos);

  // Energy corrections
  // Light Yield correction (RealData only)
  ECStrategyFactory  stratf;
  if( ArteTable<EVHD>::begin()->exp>0 ){
    //cout<<" RD: correcting for light yield"<<endl;
    ESEnCorStrategy* lightcorr = 
      stratf.EnCorStrategy(ECStrategyFactory::LightYield); 
    lightcorr->evaluateEnergy(clus, posi);
  }
  ESEnCorStrategy* deadcellcorr =
    stratf.EnCorStrategy(ECStrategyFactory::DeadCells); 
  if( deadcellcorr!=0){
    deadcellcorr->evaluateEnergy(clus, posi);
  }

  // write out on banks
 
  rccl->x = posi.x();
  rccl->y = posi.y();
  for(int j=0; j<3; j++){
    rccl->cve[j] = posi.var(j);
  }
  rccl->esum = posi.energy();
  rccl->cve[14] = posi.sqenerr();
  rccl->wid = posi.width();
  rccl->flag = (posi.cell()->ncl())+allcells*16;
 
			  // z dependence
  rccl->z = posi.z();
  if( CmdCareLevel::level()>2){
    // now likelyhood
    pidstra->beginAnalysis(clus, rccl, posi.cell());
  }
  
			  // Slopes with respect to point (0,0,0)
  float rz = 1./posi.z();
  rccl->tx = posi.x()*rz;
  rccl->ty = posi.y()*rz;
  rccl->cve[5] = rccl->cve[0]*rz*rz;  // covariance for tx tx
  rccl->cve[9] = rccl->cve[2]*rz*rz;  // covariance for ty ty
  rccl->cve[3] = rccl->cve[0]*rz;  // covariance for x tx
  rccl->cve[7] = rccl->cve[2]*rz;  // covariance for yty

  rccl->cve[13] = clus->maxEnergy()*1.0e-6;

}

//--
//  Evaluate the information for the cluster and 
//  fill an old entry in the bank
//
#ifndef DEBUG
void EAHypAnalyzer::testLinks(const EMCluster * , ArtePointer<RCCL> ){
#else
void EAHypAnalyzer::testLinks(const EMCluster * clus, ArtePointer<RCCL> rccl){
  std::cout<<"Testing cluster"<<std::endl<<std::endl;
  clus->dump();
  // test the link done
  std::cout<<"Cluster number:"<<rccl.index()
      <<" at "<<rccl->x<<","<<rccl->y
      <<" with "<<rccl->wid<<" cells and "<<rccl->esum<<" GeV"<<std::endl;
  // now loop on relations of this RCCL with HITC
  for( ArteRelationIterator<HITC> relIter = (rccl->hitc).begin();
       relIter!=(rccl->hitc).end(); ++relIter){
    ArtePointer<HITC> cell = relIter;
    std::cout<<"->Cell: ID="<<cell->gcce<<"  Energy = "<<cell->q
	<<"  Pos:("<<cell->x<<", "<<cell->y<<" ) ";
    std::cout<<std::endl;
  }
  // now loop on relations of this RCCL with RTRA
  for( ArteRelationIterator<RTRA> relIterat = (rccl->rtra).begin();
       relIterat!=(rccl->rtra).end(); ++relIterat){
    ArtePointer<RTRA> rtra(relIterat);
    std::cout<<"->Track: ID="<<rtra<<"  Momentum = "<< 1./rtra->pf;
    std::cout<<std::endl;
  }
#endif
}


