//  Author:   Dmitry Emeliyanov                Created:  01-Dec-01 ver. 1.00
//  Email:    emelian@mail.desy.de                       
//            HERA-B Collaboration                       
//            This routine is a part of OTR/ITR-CATS package, 
//            it includes new routine SltCatsEvent() called 
//            from ItrCatsEvent() and provides SLT-seeded 
//            track reconstruction in the OTR. 

#include "arte/RSEG_cmp.hh"
#include <map>
#include <math.h>
#include "cats/ItrCats.hh"
#include "cats/OtrCats.hh"
#include "arte/Chrono.hh"
#include <iostream.h>
#include "farm/farm.hh"
extern "C"{
#include "sltinfo.h"  
}
#include "rang/zfpar.hh"

#define DEBUG 0
#define NPLMAX 300

void SltCatsEvent(void);
extern void vier_dim_inv(double a[4][4]);
extern "C"{
  float prob_(const float & ccchi2, const int & ndeg);
}

int first_call=1;

static int NSLTSEEDS=0;
static int NSEEDRSEG=0;

int GetSltInfo(double par[])
{
  int n_track = 0;
  int Ntrack = 0;
  // Get slt info

  //------------------------------------------------------ Basic SLT info
  int daqcode=faDDAQFCS_tcode();
  
  // Detailed trigger definition is given at:
  // http://www-hera-b.desy.de/subgroup/shiftinfo/sub/DAQSHI.html
  
  if(DEBUG)
    {
      cout<<"Trigger Info:"<<endl;
      if (daqcode>=8) cout<<"Random (4Hz) triggered event"<<endl;
      if (daqcode==0||daqcode==8) cout<<"FLT triggered event"<<endl;
      if (daqcode==6||daqcode==14) cout<<"''Test triggger'' event"<<endl;
    }
   
  if (daqcode==9 || SLTReadDSLT()!=SLT_OK) 
    { 
      cout <<"No SLT information found"<<endl;
    } else 
      {
	SLT_STATUS stat;
	SLT_TRACK track;
	SLT_REFITTRACK refit;
        SLT_TRIGGER trig;

	if (SLTStatus(&stat)!=SLT_OK) 
	  {
	    if(DEBUG) cout <<"No status information"<<endl;
	    return 0;
	  }
	if (stat.status&SLTEVENT_BYPASS) 
	  {
	    if(DEBUG) cout <<"Event was a SLT BYPASS event!"<<endl;
	  }
	if (!(stat.status&SLTEVENT_ACCEPT)) 
	  {
	    if(DEBUG) cout <<"Event rejected"<<endl;
	    return 0;
	  }
	if(DEBUG)
	  {
	    cout<<"Event accepted with "<<stat.ntracks<<" tracks in ";
	    cout<<stat.ntriggers<<" trigger pairs"<<endl;
	  }
    
	if (SLTStatus(&stat)==SLT_OK && (stat.status&SLTEVENT_ACCEPT)) 
	  {
	    for(int trigId=0; trigId<stat.ntriggers; trigId++)
	      {
		if(SLTTrigInfo(trigId,&trig)==SLT_OK && 
		   trig.status!=SLTTRIG_REJECT &&
		   ((trig.type==SLTTYPE_DIELECTRON && 
		     (zfpar_.sltmode==4||zfpar_.sltmode==6)) 
		    || 
		    (trig.type==SLTTYPE_DIMUON &&
		     (zfpar_.sltmode==5||zfpar_.sltmode==6))))
		  {
		    // loop over all tracks in this trigger
		    for(int traId=0;traId<trig.ntracks;traId++) 
		      {
			if(SLTTrackInfo(trig.trackindex[traId],&track)==
			   SLT_OK)  
			  {
			    if(track.status != SLTTRACK_ERROR 
			       && track.status==SLTTRACK_ACCEPT) 
			      {
                
				if(SLTRefitTrackInfo(trig.trackindex[traId],
						     &refit)==SLT_OK
				   // && SLTSiliTrackInfo(trig.trackindex[traId],&sili)==SLT_OK &&
				   // && SLTSiliTrackCovInfo(trig.trackindex[traId],&cov)==SLT_OK &&
		   // && SLTSeedTrackInfo(trig.trackindex[traId],&seed)==SLT_OK 
				   ) 
				  {
				    // check if track is stored
				    int if_store = 0;
				    for(int nt=0;nt<=n_track;++nt)
				      {
					int L=nt*4;
					if(par[L]!=refit.bx) continue;
					if(par[L+1]!=refit.by) continue;
					if(par[L+2]!=refit.tx) continue;
					if(par[L+3]!=refit.ty) continue;
					if_store = 1;

				      }
				    if(if_store) continue;
				    int L = n_track * 4;
				    par[L    ] = refit.bx;
				    par[L + 1] = refit.by;
				    par[L + 2] = refit.tx;
				    par[L + 3] = refit.ty;
				    if(DEBUG)
				      cout<< "n_track"<< n_track 
					  <<" x "<<par[L] << " y "<<par[L+1]
					  <<" tx " <<par[L+2]<<" ty "<<
					par[L+3]<<endl;
				    ++n_track;
				    Ntrack = n_track;
				  }
			      }
			  }
		      }
		  }
	      }
	  }
      }
  return Ntrack;
}

typedef struct
{
  double Rk[4];
  int LR;
  ArtePointer<HITB> hitb;
} SLTHIT, *pSLTHIT;

typedef struct
{
  double z;
  double Rk[4];
  double Gk[4][4];
  vector<SLTHIT> h;
} PLANE_TYPE, *pPLANE_TYPE;

typedef struct
{
  vector<SLTHIT> h;
  double chi2,prob,pf,zf,ze,Frac;
  double Rf[4],Re[4];
  double cvf[4][4],cve[4][4];
  int nHit,flag;
} SLT_CANDIDATE, *pSLT_CANDIDATE;

SLT_CANDIDATE AS[20];
int nFound=0;

PLANE_TYPE pla[NPLMAX];
int nPlanes=NPLMAX;

bool SltHitCompareZ(const SLTHIT a, const SLTHIT b)
{
  return (a.hitb->z>b.hitb->z);
}

bool SltHitCompareX(const SLTHIT a, const SLTHIT b)
{
  return (a.hitb->x<b.hitb->x);
}

int CollectSltHits(double Rk[4],double z)
{
  const double uWindow=2.0;
  const double vWindow=5.0;
  const float minsep=0.2;

  int nPlane=0;
  vector<SLTHIT> full;
  SLTHIT buf;
  int i,k;
  float maxz;
  double xe,ye,zmid,u,v,sina,cosa;
  char hwc[12];
  
  if(DEBUG) 
    {
      cout<<"Seed at z="<<z<<" ";
      for(int v=0;v<4;v++) cout<<Rk[v]<<" ";
      cout<<endl;
    }
  maxz=-1.0;full.clear();k=0;
  for(ArteTable<GESL>::iterator geslIt=ArteTable<GESL>::begin();
      geslIt!=ArteTable<GESL>::end();++geslIt)
    {
      if((*geslIt)->cmp!=2) continue;
      for(ArteRelationIterator<GEDE> gedeIt=(*geslIt)->gede.begin();
	  gedeIt!=(*geslIt)->gede.end();++gedeIt)
	{
	  strncpy(hwc,(*gedeIt)->hwc,10);
	  if(hwc[1]!='p') continue;
	  zmid=0.5*((*gedeIt)->zmin+(*gedeIt)->zmax);
	  xe=Rk[0]+Rk[2]*(zmid-z);
	  ye=Rk[1]+Rk[3]*(zmid-z);
	  sina=(*gedeIt)->sin*(*gedeIt)->rot[4]-
	    (*gedeIt)->cos*(*gedeIt)->rot[3];
	  cosa=(*gedeIt)->cos*(*gedeIt)->rot[0]-
	    (*gedeIt)->sin*(*gedeIt)->rot[1];
	  u=xe*cosa-ye*sina;
	  v=xe*sina+ye*cosa;
	  if(((*gedeIt)->symx+vWindow<v)||((*gedeIt)->symn-vWindow>v))
	    continue;
	  if(((*gedeIt)->sxmx+uWindow<u)||((*gedeIt)->sxmn-uWindow>u))
	    continue;
	  ArteTableIterator<HITB> hitbItf =(*gedeIt)->hit1;
	  ArteTableIterator<HITB> hitbIte =(*gedeIt)->hitl;
	  if (distance(hitbItf, hitbIte)+1 == 0) continue;
	  hitbIte++;
	  for (ArteTableIterator<HITB> hitbIt=hitbItf; 
	       hitbIt!=hitbIte;++hitbIt)
	    {
	      xe=Rk[0]+Rk[2]*((*hitbIt)->z-z);
	      ye=Rk[1]+Rk[3]*((*hitbIt)->z-z);
	      u=xe*cosa-ye*sina;
	      //	      cout<<"u="<<u<<" x="<<(*hitbIt)->x<<endl;
	      if(fabs(u-(*hitbIt)->x)>uWindow) continue;
	      buf.hitb=(*hitbIt);buf.LR=0;
	      if(buf.hitb->z>maxz) maxz=buf.hitb->z;
	      full.push_back(buf);
	      //	      cout<<buf.hitb->z<<" ";
	      k++;
	    }
	}
    }
  sort(full.begin(),full.end(),SltHitCompareZ);
  i=0;
  // cout<<"maxz="<<maxz<<endl;
  for(int v=0;v<nPlanes;v++) pla[v].h.clear();
  for(vector<SLTHIT>::iterator mhitIt=full.begin();
      mhitIt!=full.end();++mhitIt)
    {
      if(maxz-(*mhitIt).hitb->z>minsep) 
	{
	  i++;maxz=(*mhitIt).hitb->z;
	}
      if(i>NPLMAX-1) 
	{
	  i=-1;break;
	}
      //  cout<<i<<" "<<(*mhitIt).z<<endl;
      pla[i].h.push_back((*mhitIt));pla[i].z=maxz;
    }
  nPlane=i+1;
  if(DEBUG) cout<<"Total "<<nPlane<<" z planes detected"<<endl;
  if(nPlane<5) return 0; 
  /* 
     for(i=0;i<nPlane;i++)
     {
     cout<<"Plane "<<i<<endl;
     for(vector<SLTHIT>::iterator mhitIt=pla[i].h.begin();
     mhitIt!=pla[i].h.end();++mhitIt)
     {
     cout<<(*mhitIt).hitb->z<<" ";
     }
     cout<<endl;
     }
  */
  return nPlane;
}

double TruncatedDensity(double zk, double Sk, double d)
{
  double sqrt2P=sqrt(2.0*3.14159*Sk),x;

  x=zk*zk/Sk;
  if(x>d) return 0.0;
  else return exp(-0.5*x)/sqrt2P;
}

int CollectCandidate(pSLT_CANDIDATE pSlc, double R0[4], double G0[4], 
		      double Z0, int flag)
{
  const double chi2cut=20.0;
  const double alpha1=0.001;
  const double alpha2=0.1;
  const double cludense=0.05;
  const double sigma_ms=1.0e-8;
  const double B    =0.65945;
  const double zmagn=450.0;
  const double maxchi2=5.0;

  int FPlane,LPlane,PLinc,i,nHits,GetFirst,GetLast;
  double xe,ye,sina,cosa,theta,u,tu;
  double Rk[4],Gk[4][4],zprev,Kk[4],Hk[4],Sk,bk,Gn[4][4],Ga[4][4],
    beta[100],fk[100],zeta[100],h[4],chi2,
    Wk,Vk,l2sqrt,q33,q34,q44,coeff,D_z,pinv,Sum,A[4][4],
    L[4][4],L_1[4][4],dG[4][4],CF[4][4];
  ArtePointer<HITB> hitb;
  ArtePointer<GEDE> gede;

  if(flag==0) 
    {
      FPlane=nPlanes-1;LPlane=-1;PLinc=-1;
    }
  else
    {
      LPlane=nPlanes;FPlane=0;PLinc=1;
    }
  i=FPlane;zprev=Z0;
  for(int v=0;v<4;v++) for(int vv=0;vv<4;vv++) Gk[v][vv]=0.0;
  for(int v=0;v<4;v++) 
    {
      Gk[v][v]=G0[v];Rk[v]=R0[v];
    }
  bk=cludense*(alpha1+alpha2-alpha1*alpha2)/
    ((1.0-alpha1)*(1.0-alpha2));
  coeff=(1.0+0.105*0.105*Rk[4]*Rk[4])*Rk[4]*Rk[4];
  pinv=1.0/(B*zmagn*sqrt(1.0+Rk[2]*Rk[2]+Rk[3]*Rk[3])/
	    ((Rk[0]-zprev*Rk[2])*sqrt(1.0+Rk[2]*Rk[2])));
  pSlc->pf=pinv;
  pinv = fabs(pinv);
  coeff=(1.0+0.105*0.105*pinv*pinv)*pinv*pinv; // muon mass correction
  l2sqrt=1.0+Rk[2]*Rk[2]+Rk[3]*Rk[3]; 
  l2sqrt=l2sqrt*sqrt(l2sqrt);
  q33=sigma_ms*coeff*l2sqrt*(1.0+Rk[2]*Rk[2]); 
  q44=sigma_ms*coeff*l2sqrt*(1.0+Rk[3]*Rk[3]); 
  q34=sigma_ms*coeff*l2sqrt*Rk[2]*Rk[3];
  while(i!=LPlane)
    {
      D_z=pla[i].z-zprev;zprev=pla[i].z;
      Rk[0]+=D_z*Rk[2];
      Rk[1]+=D_z*Rk[3];
      Gk[0][0]+=D_z*(2.0*Gk[0][2]+D_z*Gk[2][2]);
      Gk[0][1]+=D_z*(Gk[1][2]+Gk[0][3]+D_z*Gk[2][3]);
      Gk[0][2]+=D_z*Gk[2][2];
      Gk[0][3]+=D_z*Gk[2][3];
      Gk[1][1]+=D_z*(2.0*Gk[1][3]+D_z*Gk[3][3]);
      Gk[1][2]+=D_z*Gk[2][3];
      Gk[1][3]+=D_z*Gk[3][3];
      nHits=0;Sk=100.0;
      for(vector<SLTHIT>::iterator mhitIt=pla[i].h.begin();
	  mhitIt!=pla[i].h.end();++mhitIt)
	{
	  hitb=(*mhitIt).hitb;gede=hitb->gede;
	  sina=gede->sin*gede->rot[4]-gede->cos*gede->rot[3];
	  cosa=gede->cos*gede->rot[0]-gede->sin*gede->rot[1];
	  theta=Rk[2]*cosa-Rk[3]*sina;
	  tu=sqrt(1.0+theta*theta);
	  u=Rk[0]*cosa-Rk[1]*sina-hitb->x;
	  Vk=Rk[1]*cosa+Rk[0]*sina;
	  Hk[0]= cosa/tu;Hk[1]=-sina/tu;
	  Hk[2]=-Hk[0]*u*theta/(tu*tu);Hk[3]= Hk[1]*u*theta/(tu*tu);
	  Kk[0]=Hk[0]*Gk[0][0]+Hk[1]*Gk[0][1]+Hk[2]*Gk[0][2]+Hk[3]*Gk[0][3];
	  Kk[1]=Hk[0]*Gk[0][1]+Hk[1]*Gk[1][1]+Hk[2]*Gk[1][2]+Hk[3]*Gk[1][3];
	  Kk[2]=Hk[0]*Gk[0][2]+Hk[1]*Gk[1][2]+Hk[2]*Gk[2][2]+Hk[3]*Gk[2][3];
	  Kk[3]=Hk[0]*Gk[0][3]+Hk[1]*Gk[1][3]+Hk[2]*Gk[2][3]+Hk[3]*Gk[3][3];
	  Sk=hitb->r*hitb->r;for(int v=0;v<4;v++) Sk+=Hk[v]*Kk[v];      
	  for(int v=0;v<2;v++)
	    {
	      zeta[nHits+v]=(2*v-1)*fabs(hitb->q)-u/tu;
	      fk[nHits+v]=TruncatedDensity(zeta[nHits+v],Sk,chi2cut);
	      if((Vk>gede->symx)||(Vk<gede->symn)) fk[nHits+v]=0.0;
	    }
	  nHits+=2;
	}
      fk[nHits]=bk;
      Sum=0.0;Wk=0.0;
      for(int v=0;v<=nHits;v++) Sum+=fk[v];
      for(int v=0;v<=nHits;v++) beta[v]=fk[v]/Sum;
      for(int v=0;v<nHits;v++) Wk+=zeta[v]*beta[v];
      for(int v=0;v<4;v++)
	{
	  Rk[v]+=Kk[v]*Wk/Sk;
	  for(int vv=0;vv<=v;vv++) 
	    Gn[vv][v]=Gk[vv][v]-Kk[vv]*Kk[v]/Sk;
	}
      Vk=0.0;
      for(int v=0;v<nHits;v++) Vk+=beta[v]*zeta[v]*zeta[v];
      Vk-=Wk*Wk;Vk/=(Sk*Sk);
      for(int v=0;v<4;v++)
	{
	  for(int vv=0;vv<=v;vv++) 
	    Ga[vv][v]=Kk[vv]*Kk[v]*Vk;
	}
      for(int v=0;v<4;v++)
	{
	  for(int vv=0;vv<=v;vv++) 
	    Gk[vv][v]=beta[nHits]*Gk[vv][v]+(1-beta[nHits])*Gn[vv][v]+
	      Ga[vv][v];
	}
      for(int v=0;v<4;v++){
	  pla[i].Rk[v]=Rk[v];
	  for(int vv=0;vv<=v;vv++) 
	    pla[i].Gk[vv][v]=Gk[vv][v];
	}
      Gk[2][2]+=q33;Gk[3][3]+=q44;Gk[2][3]+=q34;
      /*    
	    cout<<i<<" : ";
	    for(int v=0;v<4;v++) cout<<Rk[v]<<" ";
	    cout<<endl;
      */
      i+=PLinc;
    }
  if(DEBUG)
    {
      cout<<"Last point: ";
      for(int v=0;v<4;v++) cout<<Rk[v]<<" ";
      cout<<endl;
    }
  i=LPlane-2*PLinc;zprev=pla[i+PLinc].z;
  while(i!=FPlane)
    {
      D_z=zprev-pla[i].z;zprev=pla[i].z;
      l2sqrt=1.0+pla[i].Rk[2]*pla[i].Rk[2]+pla[i].Rk[3]*pla[i].Rk[3]; 
      l2sqrt=sigma_ms*coeff*l2sqrt*sqrt(l2sqrt);
      q33=l2sqrt*(1.0+pla[i].Rk[2]*pla[i].Rk[2]); 
      q44=l2sqrt*(1.0+pla[i].Rk[3]*pla[i].Rk[3]); 
      q34=l2sqrt*pla[i].Rk[2]*pla[i].Rk[3];
      L[0][0]=pla[i].Gk[0][0]+D_z*(2.0*pla[i].Gk[0][1]+
				   D_z*pla[i].Gk[1][1]);
      L[0][1]=pla[i].Gk[0][1]+D_z*pla[i].Gk[1][1];L[1][0]=L[0][1];
      L[0][0]=pla[i].Gk[0][0]+D_z*(2.0*pla[i].Gk[0][2]+
				   pla[i].Gk[2][2]*D_z);
      L[0][1]=pla[i].Gk[0][1]+D_z*(pla[i].Gk[0][3]+pla[i].Gk[1][2]+
				   D_z*pla[i].Gk[2][3]);L[1][0]=L[0][1];
      L[0][2]=pla[i].Gk[0][2]+D_z*pla[i].Gk[2][2];L[2][0]=L[0][2];
      L[0][3]=pla[i].Gk[0][3]+D_z*pla[i].Gk[2][3];L[3][0]=L[0][3];
      L[1][1]=pla[i].Gk[1][1]+D_z*(2.0*pla[i].Gk[1][3]+pla[i].Gk[3][3]*D_z);
      L[1][2]=pla[i].Gk[1][2]+D_z*pla[i].Gk[2][3];L[2][1]=L[1][2];
      L[1][3]=pla[i].Gk[1][3]+D_z*pla[i].Gk[3][3];L[3][1]=L[1][3];      
      L[2][2]=pla[i].Gk[2][2]+q33;
      L[3][3]=pla[i].Gk[3][3]+q44;
      L[2][3]=pla[i].Gk[2][3]+q34;L[3][2]=L[2][3];
      L_1=L;vier_dim_inv(L_1);
      CF[0][0]=pla[i].Gk[0][0]+D_z*pla[i].Gk[0][2];
      CF[1][0]=pla[i].Gk[0][1]+D_z*pla[i].Gk[1][2];
      CF[2][0]=pla[i].Gk[0][2]+D_z*pla[i].Gk[2][2];
      CF[3][0]=pla[i].Gk[0][3]+D_z*pla[i].Gk[2][3];
      CF[0][1]=pla[i].Gk[0][1]+D_z*pla[i].Gk[0][3];
      CF[1][1]=pla[i].Gk[1][1]+D_z*pla[i].Gk[1][3];
      CF[2][1]=pla[i].Gk[1][2]+D_z*pla[i].Gk[2][3];
      CF[3][1]=pla[i].Gk[1][3]+D_z*pla[i].Gk[3][3];
      CF[0][2]=pla[i].Gk[0][2];CF[0][3]=pla[i].Gk[0][3];
      CF[1][2]=pla[i].Gk[1][2];CF[1][3]=pla[i].Gk[1][3];
      CF[2][2]=pla[i].Gk[2][2];CF[2][3]=pla[i].Gk[2][3];
      CF[3][2]=pla[i].Gk[2][3];CF[3][3]=pla[i].Gk[3][3];
      for(int v=0;v<4;v++) 
	for(int vv=0;vv<4;vv++)
	  {
	    A[v][vv]=0.0;
	    for(int m=0;m<4;m++) 
	      A[v][vv]+=CF[v][m]*L_1[m][vv];
	  }
      h[0]=pla[i+PLinc].Rk[0]-pla[i].Rk[0]-D_z*pla[i].Rk[2];
      h[1]=pla[i+PLinc].Rk[1]-pla[i].Rk[1]-D_z*pla[i].Rk[3];
      h[2]=pla[i+PLinc].Rk[2]-pla[i].Rk[2];
      h[3]=pla[i+PLinc].Rk[3]-pla[i].Rk[3];
      for(int v=0;v<4;v++)
	for(int vv=0;vv<4;vv++) 
	  pla[i].Rk[v]+=A[v][vv]*h[vv];
      for(int v=0;v<4;v++) 
	for(int vv=0;vv<=v;vv++) 
	  dG[vv][v]=pla[i+PLinc].Gk[vv][v]-L[vv][v];
      CF[0][0]=A[0][0]*dG[0][0]+A[0][1]*dG[0][1]+A[0][2]*dG[0][2]+A[0][3]*dG[0][3];
      CF[0][1]=A[0][0]*dG[0][1]+A[0][1]*dG[1][1]+A[0][2]*dG[1][2]+A[0][3]*dG[1][3];
      CF[0][2]=A[0][0]*dG[0][2]+A[0][1]*dG[1][2]+A[0][2]*dG[2][2]+A[0][3]*dG[2][3];
      CF[0][3]=A[0][0]*dG[0][3]+A[0][1]*dG[1][3]+A[0][2]*dG[2][3]+A[0][3]*dG[3][3];
      CF[1][0]=A[1][0]*dG[0][0]+A[1][1]*dG[0][1]+A[1][2]*dG[0][2]+A[1][3]*dG[0][3];
      CF[1][1]=A[1][0]*dG[0][1]+A[1][1]*dG[1][1]+A[1][2]*dG[1][2]+A[1][3]*dG[1][3];
      CF[1][2]=A[1][0]*dG[0][2]+A[1][1]*dG[1][2]+A[1][2]*dG[2][2]+A[1][3]*dG[2][3];
      CF[1][3]=A[1][0]*dG[0][3]+A[1][1]*dG[1][3]+A[1][2]*dG[2][3]+A[1][3]*dG[3][3];
      CF[2][0]=A[2][0]*dG[0][0]+A[2][1]*dG[0][1]+A[2][2]*dG[0][2]+A[2][3]*dG[0][3];
      CF[2][1]=A[2][0]*dG[0][1]+A[2][1]*dG[1][1]+A[2][2]*dG[1][2]+A[2][3]*dG[1][3];
      CF[2][2]=A[2][0]*dG[0][2]+A[2][1]*dG[1][2]+A[2][2]*dG[2][2]+A[2][3]*dG[2][3];
      CF[2][3]=A[2][0]*dG[0][3]+A[2][1]*dG[1][3]+A[2][2]*dG[2][3]+A[2][3]*dG[3][3];
      CF[3][0]=A[3][0]*dG[0][0]+A[3][1]*dG[0][1]+A[3][2]*dG[0][2]+A[3][3]*dG[0][3];
      CF[3][1]=A[3][0]*dG[0][1]+A[3][1]*dG[1][1]+A[3][2]*dG[1][2]+A[3][3]*dG[1][3];
      CF[3][2]=A[3][0]*dG[0][2]+A[3][1]*dG[1][2]+A[3][2]*dG[2][2]+A[3][3]*dG[2][3];
      CF[3][3]=A[3][0]*dG[0][3]+A[3][1]*dG[1][3]+A[3][2]*dG[2][3]+A[3][3]*dG[3][3];
      for(int v=0;v<4;v++) 
	for(int vv=0;vv<=v;vv++)
	  for(int m=0;m<4;m++) 
	    pla[i].Gk[vv][v]+=CF[vv][m]*A[v][m];
      /*
      cout<<i<<" smooth: ";
      for(int v=0;v<4;v++) cout<<pla[i].Rk[v]<<" ";
      cout<<endl;
      */
      i-=PLinc;
    }
  i=FPlane;pSlc->nHit=0;pSlc->chi2=0.0;pSlc->h.clear();
  GetFirst=-1;GetLast=-1;
  while(i!=LPlane)
    {
      for(int v=0;v<4;v++)
	{
	  Rk[v]=pla[i].Rk[v];
	  for(int vv=0;vv<=v;vv++) 
	    Gk[vv][v]=pla[i].Gk[vv][v];
	}
      for(vector<SLTHIT>::iterator mhitIt=pla[i].h.begin();
	  mhitIt!=pla[i].h.end();++mhitIt)
	{
	  hitb=(*mhitIt).hitb;gede=hitb->gede;
	  sina=gede->sin*gede->rot[4]-gede->cos*gede->rot[3];
	  cosa=gede->cos*gede->rot[0]-gede->sin*gede->rot[1];
	  xe=Rk[0]+Rk[2]*(hitb->z-pla[i].z);
	  ye=Rk[1]+Rk[3]*(hitb->z-pla[i].z);
	  theta=Rk[2]*cosa-Rk[3]*sina;
	  tu=sqrt(1.0+theta*theta);
	  u=xe*cosa-ye*sina-hitb->x;
	  Vk=ye*cosa+xe*sina;
	  Hk[0]= cosa/tu;Hk[1]=-sina/tu;
	  Hk[2]=-Hk[0]*u*theta/(tu*tu);Hk[3]= Hk[1]*u*theta/(tu*tu);
	  Kk[0]=Hk[0]*Gk[0][0]+Hk[1]*Gk[0][1]+Hk[2]*Gk[0][2]+Hk[3]*Gk[0][3];
	  Kk[1]=Hk[0]*Gk[0][1]+Hk[1]*Gk[1][1]+Hk[2]*Gk[1][2]+Hk[3]*Gk[1][3];
	  Kk[2]=Hk[0]*Gk[0][2]+Hk[1]*Gk[1][2]+Hk[2]*Gk[2][2]+Hk[3]*Gk[2][3];
	  Kk[3]=Hk[0]*Gk[0][3]+Hk[1]*Gk[1][3]+Hk[2]*Gk[2][3]+Hk[3]*Gk[3][3];
	  Sk=hitb->r*hitb->r;for(int v=0;v<4;v++) Sk+=Hk[v]*Kk[v];      
	  (*mhitIt).LR=(u<0.0)?-1:1;
	  u=fabs((*mhitIt).hitb->q)*(*mhitIt).LR-u/tu;
	  chi2=u*u/Sk;
	  if((Vk>gede->symx)||(Vk<gede->symn)) chi2=200.0;
	  if(chi2>maxchi2) 
	    {
	      (*mhitIt).LR=0;
	     
	      xe=R0[0]+R0[2]*((*mhitIt).hitb->z-Z0);
	      ye=R0[1]+R0[3]*((*mhitIt).hitb->z-Z0);
	      theta=R0[2]*cosa-R0[3]*sina;
	      u=(xe*cosa-ye*sina-(*mhitIt).hitb->x)/
		sqrt(1.0+theta*theta);
	    }
	  else
	    {
	      (*mhitIt).Rk[0]=xe;(*mhitIt).Rk[1]=ye;
	      (*mhitIt).Rk[2]=Rk[2];(*mhitIt).Rk[3]=Rk[3];
	      pSlc->h.push_back((*mhitIt));
	      pSlc->chi2+=chi2;pSlc->nHit++;
	      if(GetFirst==-1) GetFirst=i;
	      GetLast=i;
	    }
	}
      i+=PLinc;
    }
  if(GetLast==-1) return 0;
  if(pSlc->nHit>10)
    pSlc->prob=prob_(pSlc->chi2,pSlc->nHit-4);
  else return 0;
  if(pSlc->prob<0.005) return 0;
  if(pla[GetLast].z<pla[GetFirst].z)
    {
      i=GetLast;GetLast=GetFirst;GetFirst=i;
    }
  pSlc->zf=pla[GetFirst].z;
  for(int v=0;v<4;v++) 
    {
      pSlc->Rf[v]=pla[GetFirst].Rk[v];
      for(int vv=0;vv<4;vv++) pSlc->cvf[v][vv]=pla[GetFirst].Gk[v][vv];
    }
  pSlc->ze=pla[GetLast].z;
  for(int v=0;v<4;v++) 
    {
      pSlc->Re[v]=pla[GetLast].Rk[v];
      for(int vv=0;vv<4;vv++) pSlc->cve[v][vv]=pla[GetLast].Gk[v][vv];
    }
  pSlc->flag=1;
    
  return 1;
}

void StoreSltCandidate(pSLT_CANDIDATE pSlc)
{
  RSEG r;
  RHIT rh;
  RPNT rp;

  r.xf=pSlc->Rf[0];r.yf=pSlc->Rf[1];r.txf=pSlc->Rf[2];r.tyf=pSlc->Rf[3];    
  r.zf=pSlc->zf;      
  r.xe=pSlc->Re[0];r.ye=pSlc->Re[1];r.txe=pSlc->Re[2];r.tye=pSlc->Re[3];    
  r.ze=pSlc->ze;
  if(DEBUG)
    {
      cout<<"RSEG: zf="<<r.zf<<" ";
      for(int v=0;v<4;v++) cout<<pSlc->Rf[v]<<" ";
      cout<<endl;
      cout<<"RSEG: ze="<<r.ze<<" ";
      for(int v=0;v<4;v++) cout<<pSlc->Re[v]<<" ";
      cout<<endl;
    }
  r.cvf[ 0] = pSlc->cvf[0][0];      r.cvf[ 1] = pSlc->cvf[0][1];
  r.cvf[ 2] = pSlc->cvf[1][1];      r.cvf[ 3] = pSlc->cvf[0][2];
  r.cvf[ 4] = pSlc->cvf[1][2];      r.cvf[ 5] = pSlc->cvf[2][2];
  r.cvf[ 6] = pSlc->cvf[0][3];      r.cvf[ 7] = pSlc->cvf[1][3];
  r.cvf[ 8] = pSlc->cvf[2][3];      r.cvf[ 9] = pSlc->cvf[3][3];
  
  r.cvf[10] = r.cvf[11] = r.cvf[12] = r.cvf[13] = 0.0;
  r.cvf[14] = 100.0;
  
  r.cve[ 0] = pSlc->cve[0][0];      r.cve[ 1] = pSlc->cve[0][1];
  r.cve[ 2] = pSlc->cve[1][1];      r.cve[ 3] = pSlc->cve[0][2];
  r.cve[ 4] = pSlc->cve[1][2];      r.cve[ 5] = pSlc->cve[2][2];
  r.cve[ 6] = pSlc->cve[0][3];      r.cve[ 7] = pSlc->cve[1][3];
  r.cve[ 8] = pSlc->cve[2][3];      r.cve[ 9] = pSlc->cve[3][3];
  
  r.cve[10] = r.cve[11] = r.cve[12] = r.cve[13] = 0.0;
  r.cve[14] = 100.0;
  
  r.pf = r.pe = pSlc->pf;
  
  r.time = 0.0; r.dtim = 0.0;
  r.chi  = pSlc->chi2;
  
  r.fit  = 0;
  r.cmp  = Rsegc::patt | Rsegc::wch | Rsegc::bit16 | Rsegc::slt;
  
  ArteTable<RSEG>::push_back( r );
  ArtePointer<RSEG> rseg( ArteTable<RSEG>::size() );
  
  for (vector<SLTHIT>::reverse_iterator estIt = pSlc->h.rbegin();
       estIt != pSlc->h.rend(); ++estIt)
    {
      createRelation((*estIt).hitb, rseg, &RSEG::hitb );
      ArtePointer<RHIT> rhit = ArteTable<RHIT>::push_back(rh);
      createRelation((*estIt).hitb, rhit, &HITB::rhit );
      //  cout<<"z="<<(*estIt).hitb->z<<endl;
      if((*estIt).hitb->gede->cmp==2)
      {
	(*estIt).hitb->z*=-1.0;
      }
      createRelation( rseg, rhit, &RSEG::rhit );
      rhit->tx = (*estIt).Rk[2];
      rhit->ty = (*estIt).Rk[3];
      rhit->x  = (*estIt).Rk[0];
      rhit->y  = (*estIt).Rk[1];
      rhit->p  = rseg->pf;
      rhit->q  = (*estIt).hitb->q*(*estIt).LR;
     
    }
  rseg->nhit = pSlc->nHit;
  rp.xf  = rseg->xf;  rp.yf  = rseg->yf;  rp.zf = rseg->zf; 
  rp.txf = rseg->txf; rp.tyf = rseg->tyf; rp.pf = rseg->pf;
  for (int i=0; i<15; i++) rp.cvf[i] = rseg->cvf[i];
  ArtePointer<RPNT> rpntf = ArteTable<RPNT>::push_back( rp );
  createRelation( rseg, rpntf, &RSEG::rpnt );
  rp.xf  = rseg->xe;  rp.yf  = rseg->ye;  rp.zf = rseg->ze; 
  rp.txf = rseg->txe; rp.tyf = rseg->tye; rp.pf = rseg->pe;
  for (int i=0; i<15; i++) rp.cvf[i] = rseg->cve[i];
  ArtePointer<RPNT> rpnte = ArteTable<RPNT>::push_back( rp );
  createRelation( rseg, rpnte, &RSEG::rpnt ); 
}

void SltSeededTracking(double Rk[4])
{
  const double zPC1=702.0;
  const double zPC4=851.0;
  int nSeed,res;
  double G0[4],R0[4],zSeed;
  G0[0]=0.03;
  G0[1]=2.0;
  G0[2]=0.0000012;
  G0[3]=0.00012;
  nPlanes=CollectSltHits(Rk,0.0);
  if(nPlanes==0) return;
  for(nSeed=0;nSeed<2;nSeed++)
    {
      if(nSeed==0) zSeed=zPC1;
      else zSeed=zPC4;
      for(int v=0;v<4;v++) R0[v]=Rk[v];
      R0[0]+=Rk[2]*zSeed;
      R0[1]+=Rk[3]*zSeed;
      res=CollectCandidate(&AS[nFound],R0,G0,zSeed,nSeed);
      if(res!=0) nFound++;
    }
}

void SltCloneRemoval()
{
  int i,j,nZero,iBest;
  double F;
  for(i=0;i<nFound;i++)
    {
      if((AS[i].flag==0)||(AS[i].flag==2)) continue; 
      for(vector<SLTHIT>::iterator hitIt=AS[i].h.begin();
	  hitIt!=AS[i].h.end();++hitIt)
	{
	  if((*hitIt).hitb->gede->cmp==2)
	    {
	      (*hitIt).hitb->z*=-1.0;
	    }
	}
      for(j=i+1;j<nFound;j++)
	{
	  if((AS[j].flag==0)||(AS[j].flag==2)) continue; 
	  nZero=0;
	  for(vector<SLTHIT>::iterator hitIt=AS[j].h.begin();
	      hitIt!=AS[j].h.end();++hitIt)
	    if((*hitIt).hitb->z<0.0) nZero++;
	  AS[j].Frac=1.0*nZero/AS[j].nHit;
	}
      iBest=i;F=AS[i].prob;
      for(j=i+1;j<nFound;j++)
	{
	  if((AS[j].flag==0)||(AS[j].flag==2)) continue; 
	  if(AS[j].Frac<0.4) continue;
	  if(AS[j].prob>F)
	    {
	      iBest=j;F=AS[j].prob;
	    }
	}
      if(iBest!=i) AS[i].flag=0;
      for(j=i+1;j<nFound;j++)
	{
	  if((AS[j].flag==0)||(AS[j].flag==2)) continue; 
	  if(AS[j].Frac<0.4) continue;
	  if(j!=iBest) AS[j].flag=0;
	}
      AS[iBest].flag=2;
      for(vector<SLTHIT>::iterator hitIt=AS[i].h.begin();
	  hitIt!=AS[i].h.end();++hitIt)
	{
	  if((*hitIt).hitb->gede->cmp==2)
	    {
	      (*hitIt).hitb->z*=-1.0;
	    }
	}
    }
}

void SltCatsEvent()
{
  double parslt[400]; 
  double Rk[4];
  int nSeeds=0;
  int L;
  double eff;

  if((zfpar_.sltmode<4)||(zfpar_.sltmode>6)) return;
  if(first_call) 
    {
      first_call=0;
      cout<<"SLT-seeded Cats version 1.00 started ..."<<endl;
    }
  nSeeds=GetSltInfo(parslt);
  NSLTSEEDS+=nSeeds;
  if(DEBUG) cout<<nSeeds<<" Slt seeds detected ..."<<endl;
  if(nSeeds==0) return;
  nFound=0;
  for(int i=0;i<nSeeds;i++)
    {
      L=i*4;
      for(int v=0;v<4;v++) Rk[v]=parslt[L+v];
      SltSeededTracking(Rk);
    }
  if(DEBUG) cout<<"Total "<<nFound<<" SLT candidates collected ..."<<endl;
  
  SltCloneRemoval();
  for(int i=0;i<nFound;i++)
    if(AS[i].flag==2) 
      {
	StoreSltCandidate(&AS[i]);
	NSEEDRSEG++;
      }
  eff=1.0*NSEEDRSEG/NSLTSEEDS;
  if(DEBUG) cout<<NSLTSEEDS<<" seeds detected "<<NSEEDRSEG<<" segments found "
		<<" mean efficiency="<<eff<<endl;
}


