// -*- C++ -*-
//
// This file is a part of what might become the CLHEP -
// a Class Library for High Energy Physics.
// 
// 
// Copyright Cornell University 1993, 1996, All Rights Reserved.
// 
// This software written by Nobu Katayama and Mike Smyth, Cornell University.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice and author attribution, this list of conditions and the
//    following disclaimer. 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice and author attribution, this list of conditions and the
//    following disclaimer in the documentation and/or other materials
//    provided with the distribution.
// 3. Neither the name of the University nor the names of its contributors
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
// 
// Creation of derivative forms of this software for commercial
// utilization may be subject to restriction; written permission may be
// obtained from Cornell University.
// 
// CORNELL MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  By way
// of example, but not limitation, CORNELL MAKES NO REPRESENTATIONS OR
// WARRANTIES OF MERCANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT
// THE USE OF THIS SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS,
// COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.  Cornell University shall not be
// held liable for any liability with respect to any claim by the user or any
// other party arising from use of the program.
//

#ifdef GNUPRAGMA
#pragma implementation
#endif

// //#include <sys/types.h>
#include <iostream.h>
#include <math.h>
#include <string.h>

#include "CLHEP/config/CLHEP.h"
#ifdef HEP_USE_RANDOM
#include "CLHEP/Random/Random.h"
#endif
#ifndef HEP_NO_PERSISTENT_STREAMS
#include "CLHEP/PersistentIO/PIStream.h"
#include "CLHEP/PersistentIO/POStream.h"
#endif

#include "CLHEP/Matrix/SymMatrix.h"
#include "CLHEP/Matrix/Matrix.h"
#include "CLHEP/Matrix/DiagMatrix.h"
#include "CLHEP/Matrix/Vector.h"

#ifdef HEP_DEBUG_INLINE
#include "CLHEP/Matrix/SymMatrix.icc"
#endif

// Simple operation for all elements

#define SIMPLE_UOP(OPER)          \
   register HepDouble *a=m;            \
   register HepDouble *e=m+num_size(); \
   for(;a<e; a++) (*a) OPER t;

#define SIMPLE_BOP(OPER)          \
   register HepDouble *a=m;            \
   register HepDouble *b=m2.m;         \
   register HepDouble *e=m+num_size(); \
   for(;a<e; a++, b++) (*a) OPER (*b);

#define SIMPLE_TOP(OPER)          \
   register HepDouble *a=m1.m;            \
   register HepDouble *b=m2.m;         \
   register HepDouble *t=mret.m;         \
   register HepDouble *e=m1.m+m1.num_size(); \
   for( ;a<e; a++, b++, t++) (*t) = (*a) OPER (*b);

#define CHK_DIM_2(r1,r2,c1,c2,fun) \
   if (r1!=r2 || c1!=c2)  { \
     HepGenMatrix::error("Range error in SymMatrix function " #fun "(1)."); \
   }

#define CHK_DIM_1(c1,r2,fun) \
   if (c1!=r2) { \
     HepGenMatrix::error("Range error in SymMatrix function " #fun "(2)."); \
   }

// Constructors. (Default constructors are inlined and in .icc file)

HepSymMatrix::HepSymMatrix(int p)
   : nrow(p)
{
   size = nrow * (nrow+1) / 2;
   m = new_m(size);
}

HepSymMatrix::HepSymMatrix(int p, int init)
   : nrow(p)
{
   size = nrow * (nrow+1) / 2;
   m = new_m(size);

   memset(m, 0, size * sizeof(HepDouble));
   switch(init)
   {
   case 0:
      break;
      
   case 1:
      {
	 HepDouble *a = m;
	 for(int i=1;i<=nrow;i++) {
	    *a = 1.0;
	    a += (i+1);
	 }
	 break;
      }
   default:
      error("SymMatrix: initialization must be either 0 or 1.");
   }
}

#ifdef HEP_USE_RANDOM
HepSymMatrix::HepSymMatrix(int p, HepRandom &r)
  : nrow(p)
{
   size = nrow * (nrow+1) / 2;
   m = new_m(size);
   HepDouble *a = m;
   HepDouble *b = m + size;
   for(;a<b;a++) *a = r();
}
#endif

//
// Destructor
//
HepSymMatrix::~HepSymMatrix() {
  delete_m(size, m);
}

HepSymMatrix::HepSymMatrix(const HepSymMatrix &m1)
   : nrow(m1.nrow), size(m1.size)
{
   m = new_m(size);
   memcpy(m, m1.m, size*sizeof(HepDouble));
}

HepSymMatrix::HepSymMatrix(const HepDiagMatrix &m1)
   : nrow(m1.nrow)
{
   size = nrow * (nrow+1) / 2;
   m = new_m(size);

   int n = num_row();
   memset(m, 0, size * sizeof(HepDouble));

   HepDouble *mrr = m;
   HepDouble *mr = m1.m;
   for(int r=1;r<=n;r++) {
      *mrr = *(mr++);
      mrr += (r+1);
   }
}

//
//
// Sub matrix
//
//

HepSymMatrix HepSymMatrix::sub(int min_row, int max_row) const
#ifdef HEP_GNU_OPTIMIZED_RETURN
return mret(max_row-min_row+1);
{
#else
{
  HepSymMatrix mret(max_row-min_row+1);
#endif
  if(max_row > num_row())
    error("HepSymMatrix::sub: Index out of range");
  HepDouble *a = mret.m;
  HepDouble *b1 = m + (min_row+2)*(min_row-1)/2;
  for(int irow=1; irow<=mret.num_row(); irow++) {
    HepDouble *b = b1;
    for(int icol=1; icol<=irow; icol++) {
      *(a++) = *(b++);
    }
    b1 += irow+min_row-1;
  }
  return mret;
}

#ifdef HEP_CC_NEED_SUB_WITHOUT_CONST
HepSymMatrix HepSymMatrix::sub(int min_row, int max_row) 
{
  HepSymMatrix mret(max_row-min_row+1);
  if(max_row > num_row())
    error("HepSymMatrix::sub: Index out of range");
  HepDouble *a = mret.m;
  HepDouble *b1 = m + (min_row+2)*(min_row-1)/2;
  for(int irow=1; irow<=mret.num_row(); irow++) {
    HepDouble *b = b1;
    for(int icol=1; icol<=irow; icol++) {
      *(a++) = *(b++);
    }
    b1 += irow+min_row-1;
  }
  return mret;
}
#endif

void HepSymMatrix::sub(int row,const HepSymMatrix &m1)
{
  if(row <1 || row+m1.num_row()-1 > num_row() )
    error("HepSymMatrix::sub: Index out of range");
  HepDouble *a = m1.m;
  HepDouble *b1 = m + (row+2)*(row-1)/2;
  for(int irow=1; irow<=m1.num_row(); irow++) {
    HepDouble *b = b1;
    for(int icol=1; icol<=irow; icol++) {
      *(b++) = *(a++);
    }
    b1 += irow+row-1;
  }
}

//
// Direct sum of two matricies
//

HepSymMatrix dsum(const HepSymMatrix &m1,
				     const HepSymMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
  return mret(m1.num_row() + m2.num_row(), 0);
{
#else
{
  HepSymMatrix mret(m1.num_row() + m2.num_row(),
				       0);
#endif
  mret.sub(1,m1);
  mret.sub(m1.num_row()+1,m2);
  return mret;
}

/* -----------------------------------------------------------------------
   This section contains support routines for matrix.h. This section contains
   The two argument functions +,-. They call the copy constructor and +=,-=.
   ----------------------------------------------------------------------- */
HepSymMatrix HepSymMatrix::operator- () const 
#ifdef HEP_GNU_OPTIMIZED_RETURN
      return m2(nrow);
{
#else
{
   HepSymMatrix m2(nrow);
#endif
   register HepDouble *a=m;
   register HepDouble *b=m2.m;
   register HepDouble *e=m+num_size();
   for(;a<e; a++, b++) (*b) = -(*a);
   return m2;
}

   

HepMatrix operator+(const HepMatrix &m1,const HepSymMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1);
{
#else
{
  HepMatrix mret(m1);
#endif
  CHK_DIM_2(m1.num_row(),m2.num_row(), m1.num_col(),m2.num_col(),+);
  mret += m2;
  return mret;
}
HepMatrix operator+(const HepSymMatrix &m1,const HepMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m2);
{
#else
{
  HepMatrix mret(m2);
#endif
  CHK_DIM_2(m1.num_row(),m2.num_row(),m1.num_col(),m2.num_col(),+);
  mret += m1;
  return mret;
}

HepSymMatrix operator+(const HepSymMatrix &m1,const HepSymMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.nrow);
{
#else
{
  HepSymMatrix mret(m1.nrow);
#endif
  CHK_DIM_1(m1.nrow, m2.nrow,+);
  SIMPLE_TOP(+)
  return mret;
}

//
// operator -
//

HepMatrix operator-(const HepMatrix &m1,const HepSymMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1);
{
#else
{
  HepMatrix mret(m1);
#endif
  CHK_DIM_2(m1.num_row(),m2.num_row(),
			 m1.num_col(),m2.num_col(),-);
  mret -= m2;
  return mret;
}
HepMatrix operator-(const HepSymMatrix &m1,const HepMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1);
{
#else
{
  HepMatrix mret(m1);
#endif
  CHK_DIM_2(m1.num_row(),m2.num_row(),
			 m1.num_col(),m2.num_col(),-);
  mret -= m2;
  return mret;
}

HepSymMatrix operator-(const HepSymMatrix &m1,const HepSymMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.num_row());
{
#else
{
  HepSymMatrix mret(m1.num_row());
#endif
  CHK_DIM_1(m1.num_row(),m2.num_row(),-);
  SIMPLE_TOP(-)
  return mret;
}

/* -----------------------------------------------------------------------
   This section contains support routines for matrix.h. This file contains
   The two argument functions *,/. They call copy constructor and then /=,*=.
   ----------------------------------------------------------------------- */

HepSymMatrix operator/(
const HepSymMatrix &m1,HepDouble t)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1);
{
#else
{
  HepSymMatrix mret(m1);
#endif
  mret /= t;
  return mret;
}

HepSymMatrix operator*(const HepSymMatrix &m1,HepDouble t)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1);
{
#else
{
  HepSymMatrix mret(m1);
#endif
  mret *= t;
  return mret;
}

HepSymMatrix operator*(HepDouble t,const HepSymMatrix &m1)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1);
{
#else
{
  HepSymMatrix mret(m1);
#endif
  mret *= t;
  return mret;
}

HepMatrix operator*(const HepMatrix &m1,const HepSymMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.num_row(),m2.num_col());
{
#else
  {
    HepMatrix mret(m1.num_row(),m2.num_col());
#endif
    int m1numcol = m1.num_col() ;
    int m1numrow = m1.num_row() ;
    int m2numrow = m2.num_row() ;
    CHK_DIM_1(m1numcol,m2numrow,*);
    HepDouble *mit1,*mit2=0,*sp,*snp,temp;
    HepDouble *mir=mret.m;
    int step,stept;
    for(mit1=m1.m;mit1<m1.m+m1numrow*m1numcol;mit1 = mit2)
      for(step=1,snp=m2.m;step<=m2numrow;)
	{
	  mit2=mit1;
	  sp=snp;
	  snp+=step;
	  temp=0;
	  while(sp<snp)
	    temp+=*(sp++)*(*(mit2++));
	  sp+=step-1;
	  for(stept=++step;stept<=m2numrow;stept++)
	    {
	      temp+=*sp*(*(mit2++));
	      sp+=stept;
	    }
	  *(mir++)=temp;
	}
    return mret;
  }

HepMatrix operator*(const HepSymMatrix &m1,const HepMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.num_row(),m2.num_col());
{
#else
{
  HepMatrix mret(m1.num_row(),m2.num_col());
#endif
  int m1numcol= m1.num_col() ;
  int m1numrow = m1.num_row() ;
  int m2numcol = m2.num_col() ;
  int m2numrow = m2.num_row() ;

  CHK_DIM_1(m1numcol,m2numrow,*);
  int step,stept;
  HepDouble *mit1,*mit2,*sp,*snp;
  HepDouble temp;
  HepDouble *mir=mret.m;
  for(step=1,snp=m1.m;step<=m1numrow;snp+=step++)
    for(mit1=m2.m;mit1<m2.m+m2numcol;mit1++)
      {
	mit2=mit1;
	sp=snp;
	temp=0;
	while(sp<snp+step)
	  {
	    temp+=*mit2*(*(sp++));
	    mit2+=m2.num_col();
	  }
	sp+=step-1;
	for(stept=step+1;stept<=m1numrow;stept++)
	  {
	    temp+=*mit2*(*sp);
	    mit2+=m2.num_col();
	    sp+=stept;
	  }
	*(mir++)=temp;
      }
  return mret;
}

HepMatrix operator*(const HepSymMatrix &m1,const HepSymMatrix &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.num_row(),m1.num_row());
{
#else
{
  HepMatrix mret(m1.num_row(),m1.num_row());
#endif
  int m1numrow = m1.num_row() ;
  int m2numrow = m2.num_row() ;
  int m1numcol = m1.num_col() ;
  CHK_DIM_1(m1numcol,m2numrow,*);
  int step1,stept1,step2,stept2;
  HepDouble *snp1,*sp1,*snp2,*sp2,temp;
  HepDouble *mr = mret.m; 
  for(step1=1,snp1=m1.m;step1<=m1numrow;snp1+=step1++)
    for(step2=1,snp2=m2.m;step2<=m2numrow;)
      {
	sp1=snp1;
	sp2=snp2;
	snp2+=step2;
	temp=0;
	if(step1<step2)
	  {
	    while(sp1<snp1+step1)
	      temp+=(*(sp1++))*(*(sp2++));
	    sp1+=step1-1;
	    for(stept1=step1+1;stept1!=step2+1;sp1+=stept1++)
	      temp+=(*sp1)*(*(sp2++));
	    sp2+=step2-1;
	   for(stept2=++step2;stept2<=m2numrow;sp1+=stept1++,sp2+=stept2++)
	     temp+=(*sp1)*(*sp2);
	  }
	else
	  {
	    while(sp2<snp2)
	      temp+=(*(sp1++))*(*(sp2++));
	    sp2+=step2-1;
	    for(stept2=++step2;stept2!=step1+1;sp2+=stept2++)
	      temp+=(*(sp1++))*(*sp2);
	    sp1+=step1-1;
	   for(stept1=step1+1;stept1<=m1numrow;sp1+=stept1++,sp2+=stept2++)
	     temp+=(*sp1)*(*sp2);
	  }
	*(mr++)=temp;
      }
  return mret;
}

HepVector operator*(const HepSymMatrix &m1,const HepVector &m2)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.num_row());
{
#else
{
  HepVector mret(m1.num_row());
#endif
  int m1numrow = m1.num_row() ;
  CHK_DIM_1(m1.num_col(),m2.num_row(),*);
  HepDouble *sp,*snp,*vpt,temp;
  int step,stept;
  HepDouble *vrp=mret.m;
  for(step=1,snp=m1.m;step<=m1numrow;)
    {
      sp=snp;
      vpt=m2.m;
      snp+=step;
      temp=0;
      while(sp<snp)
	temp+=*(sp++)*(*(vpt++));
      sp+=step-1;
      for(stept=++step;stept<=m1numrow;stept++)
	{ 
	  temp+=*sp*(*(vpt++));
	  sp+=stept;
	}
      *(vrp++)=temp;
    }
  return mret;
}

HepSymMatrix vT_times_v(const HepVector &v)
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(v.num_row());
{
#else
{
  HepSymMatrix mret(v.num_row());
#endif
  HepDouble *mr=mret.m;
  HepDouble *vt1,*vt2;
  for(vt1=v.m;vt1<v.m+v.num_row();vt1++)
    for(vt2=v.m;vt2<=vt1;vt2++)
      *(mr++)=(*vt1)*(*vt2);
  return mret;
}

/* -----------------------------------------------------------------------
   This section contains the assignment and inplace operators =,+=,-=,*=,/=.
   ----------------------------------------------------------------------- */

HepMatrix & HepMatrix::operator+=(const HepSymMatrix &m2)
{
  CHK_DIM_2(num_row(),m2.num_row(),num_col(),m2.num_col(),+=);
  int n = num_col();
  HepDouble *sjk = m2.m;
  HepDouble *m1j = m;
  HepDouble *mj = m;
  // j >= k
  for(int j=1;j<=num_row();j++) {
    HepDouble *mjk = mj;
    HepDouble *mkj = m1j;
    for(int k=1;k<=j;k++) {
      *(mjk++) += *sjk;
      if(j!=k) *mkj += *sjk;
      sjk++;
      mkj += n;
    }
    mj += n;
    m1j++;
  }
  return (*this);
}

HepSymMatrix & HepSymMatrix::operator+=(const HepSymMatrix &m2)
{
  CHK_DIM_2(num_row(),m2.num_row(),num_col(),m2.num_col(),+=);
  SIMPLE_BOP(+=)
  return (*this);
}

HepMatrix & HepMatrix::operator-=(const HepSymMatrix &m2)
{
  CHK_DIM_2(num_row(),m2.num_row(),num_col(),m2.num_col(),-=);
  int n = num_col();
  HepDouble *sjk = m2.m;
  HepDouble *m1j = m;
  HepDouble *mj = m;
  // j >= k
  for(int j=1;j<=num_row();j++) {
    HepDouble *mjk = mj;
    HepDouble *mkj = m1j;
    for(int k=1;k<=j;k++) {
      *(mjk++) -= *sjk;
      if(j!=k) *mkj -= *sjk;
      sjk++;
      mkj += n;
    }
    mj += n;
    m1j++;
  }
  return (*this);
}

HepSymMatrix & HepSymMatrix::operator-=(const HepSymMatrix &m2)
{
  CHK_DIM_2(num_row(),m2.num_row(),num_col(),m2.num_col(),-=);
  SIMPLE_BOP(-=)
  return (*this);
}

HepSymMatrix & HepSymMatrix::operator/=(HepDouble t)
{
  SIMPLE_UOP(/=)
  return (*this);
}

HepSymMatrix & HepSymMatrix::operator*=(HepDouble t)
{
  SIMPLE_UOP(*=)
  return (*this);
}

HepMatrix & HepMatrix::operator=(const HepSymMatrix &m1)
{
   if(m1.nrow*m1.nrow != size)
   {
      delete_m(size,m);
      size = m1.nrow * m1.nrow;
      m = new_m(size);
   }
   nrow = m1.nrow;
   ncol = m1.nrow;
   int n = ncol;
   HepDouble *sjk = m1.m;
   HepDouble *m1j = m;
   HepDouble *mj = m;
   // j >= k
   for(int j=1;j<=num_row();j++) {
      HepDouble *mjk = mj;
      HepDouble *mkj = m1j;
      for(int k=1;k<=j;k++) {
	 *(mjk++) = *sjk;
	 if(j!=k) *mkj = *sjk;
	 sjk++;
	 mkj += n;
      }
      mj += n;
      m1j++;
   }
   return (*this);
}

HepSymMatrix & HepSymMatrix::operator=(const HepSymMatrix &m1)
{
   if(m1.nrow != nrow)
   {
      delete_m(size,m);
      nrow = m1.nrow;
      size = m1.size;
      m = new_m(size);
   }
   memcpy(m, m1.m, size*sizeof(HepDouble));
   return (*this);
}

HepSymMatrix & HepSymMatrix::operator=(const HepDiagMatrix &m1)
{
   if(m1.nrow != nrow)
   {
      delete_m(size, m);
      nrow = m1.nrow;
      size = nrow * (nrow+1) / 2;
      m = new_m(size);
   }

   memset(m, 0, size * sizeof(HepDouble));
   HepDouble *mrr = m;
   HepDouble *mr = m1.m;
   for(int r=1; r<=nrow; r++) {
      *mrr = *(mr++);
      mrr += (r+1);
   }
   return (*this);
}

// Print the Matrix.

ostream& operator<<(ostream &s, const HepSymMatrix &q)
{
  s << "\n";
/* Fixed format needs 3 extra characters for field, while scientific needs 7 */
  int width;
  if(s.flags()&ios::fixed)
    width = s.precision()+3;
  else
    width = s.precision()+7;
  for(int irow = 1; irow<= q.num_row(); irow++)
    {
      for(int icol = 1; icol <= q.num_col(); icol++)
	{
	  s.width(width);
	  s << q(irow,icol) << " ";
	}
      s<< endl;
    }
  return s;
}

HepSymMatrix HepSymMatrix::
apply(HepDouble (*f)(HepDouble, int, int)) const
#ifdef HEP_GNU_OPTIMIZED_RETURN
return mret(num_row());
{
#else
{
  HepSymMatrix mret(num_row());
#endif
  HepDouble *a = m;
  HepDouble *b = mret.m;
  for(int ir=1;ir<=num_row();ir++) {
    for(int ic=1;ic<=ir;ic++) {
      *(b++) = (*f)(*(a++), ir, ic);
    }
  }
  return mret;
}

void HepSymMatrix::assign (const HepMatrix &m1)
{
   if(m1.nrow != nrow)
   {
      delete_m(size, m);
      nrow = m1.nrow;
      size = nrow * (nrow+1) / 2;
      m = new_m(size);
   }
   HepDouble *a = m1.m;
   HepDouble *b = m;
   for(int r=1;r<=nrow;r++) {
      HepDouble *d = a;
      for(int c=1;c<=r;c++) {
	 *(b++) = *(d++);
      }
      a += nrow;
   }
}

HepSymMatrix HepSymMatrix::similarity(const HepMatrix &m1) const
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.num_row());
{
#else
{
  HepSymMatrix mret(m1.num_row());
#endif
  HepMatrix temp = m1*(*this);
// If m1*(*this) has correct dimensions, then so will the m1.T multiplication.
// So there is no need to check dimensions again.
  int n = m1.num_col();
  HepDouble *mr = mret.m;
  HepDouble *tempr1 = temp.m;
  int m1numcol = m1.num_col() ;
  for(int r=1;r<=mret.num_row();r++) {
    HepDouble *m1c1 = m1.m;
    for(int c=1;c<=r;c++) {
      register HepDouble tmp = 0.0;
      HepDouble *tempri = tempr1;
      HepDouble *m1ci = m1c1;
      for(int i=1;i<=m1numcol;i++) {
	tmp+=(*(tempri++))*(*(m1ci++));
      }
      *(mr++) = tmp;
      m1c1 += n;
    }
    tempr1 += n;
  }
  return mret;
}

HepSymMatrix HepSymMatrix::similarity(const HepSymMatrix &m1) const
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.num_row());
{
#else
{
  HepSymMatrix mret(m1.num_row());
#endif
  HepMatrix temp = m1*(*this);
  int n = m1.num_col();
  HepDouble *mr = mret.m;
  HepDouble *tempr1 = temp.m;
  int m1numcol = m1.num_col() ;
  for(int r=1;r<=mret.num_row();r++) {
    HepDouble *m1c1 = m1.m;
    int c;
    for(c=1;c<=r;c++) {
      register HepDouble tmp = 0.0;
      HepDouble *tempri = tempr1;
      HepDouble *m1ci = m1c1;
      int i;
      for(i=1;i<c;i++) {
	tmp+=(*(tempri++))*(*(m1ci++));
      }
      for(i=c;i<=m1numcol;i++) {
	tmp+=(*(tempri++))*(*(m1ci));
	m1ci += i;
      }
      *(mr++) = tmp;
      m1c1 += c;
    }
    tempr1 += n;
  }
  return mret;
}

HepDouble HepSymMatrix::similarity(const HepVector &m1)
const {
  register HepDouble mret = 0.0;
  HepVector temp = (*this) *m1;
// If m1*(*this) has correct dimensions, then so will the m1.T multiplication.
// So there is no need to check dimensions again.
  HepDouble *a=temp.m;
  HepDouble *b=m1.m;
  HepDouble *e=a+m1.num_row();
  for(;a<e;) mret += (*(a++)) * (*(b++));
  return mret;
}

HepSymMatrix HepSymMatrix::similarityT(const HepMatrix &m1) const
#ifdef HEP_GNU_OPTIMIZED_RETURN
     return mret(m1.num_col());
{
#else
{
  HepSymMatrix mret(m1.num_col());
#endif
  HepMatrix temp = (*this)*m1;
  int n = m1.num_col();
  HepDouble *mrc = mret.m;
  HepDouble *temp1r = temp.m;
  int m1numrow = m1.num_row() ;
  for(int r=1;r<=mret.num_row();r++) {
    HepDouble *m11c = m1.m;
    for(int c=1;c<=r;c++) {
      register HepDouble tmp = 0.0;
      register HepDouble *tempir = temp1r;
      register HepDouble *m1ic = m11c;
      for(int i=1;i<=m1numrow;i++) {
	tmp+=(*(tempir))*(*(m1ic));
	tempir += n;
	m1ic += n;
      }
      *(mrc++) = tmp;
      m11c++;
    }
    temp1r++;
  }
  return mret;
}

void HepSymMatrix::invert(int &ifail) {

  static const HepDouble epsilon = 1.0e-150;

  static int max_array = 20; 
  static int *r = new int [max_array+1];
  static HepDouble *p = new HepDouble [max_array+1];
  static HepDouble *q = new HepDouble [max_array+1];
  
  ifail = 0;

  HepDouble det, temp, s;
  switch(nrow) {
  case 3:
    HepDouble t1, t2, t3;
    HepDouble c11,c12,c13,c22,c23,c33;
    c11 = (*(m+2)) * (*(m+5)) - (*(m+4)) * (*(m+4));
    c12 = (*(m+4)) * (*(m+3)) - (*(m+1)) * (*(m+5));
    c13 = (*(m+1)) * (*(m+4)) - (*(m+2)) * (*(m+3));
    c22 = (*(m+5)) * (*m) - (*(m+3)) * (*(m+3));
    c23 = (*(m+3)) * (*(m+1)) - (*(m+4)) * (*m);
    c33 = (*m) * (*(m+2)) - (*(m+1)) * (*(m+1));
    t1 = fabs(*m);
    t2 = fabs(*(m+1));
    t3 = fabs(*(m+3));
    if (t1 >= t2) {
      if (t3 >= t1) {
      temp = *(m+3);
      det = c23*c12-c22*c13;
      } else {
	temp = *m;
	det = c22*c33-c23*c23;
      }
    } else if (t3 >= t2) {
      temp = *(m+3);
      det = c23*c12-c22*c13;
    } else {
      temp = *(m+1);
      det = c13*c23-c12*c33;
    }
    if (det==0) {
      ifail = 1;
      return;
    }
    {
      HepDouble s = temp/det;
      HepDouble *mm = m;
      *(mm++) = s*c11;
      *(mm++) = s*c12;
      *(mm++) = s*c22;
      *(mm++) = s*c13;
      *(mm++) = s*c23;
      *(mm) = s*c33;
    }
    break;
    case 2:
    det = (*m)*(*(m+2)) - (*(m+1))*(*(m+1));
    if (det==0) {
      ifail = 1;
      return;
    }
    s = 1.0/det;
    *(m+1) *= -s;
    temp = s*(*(m+2));
    *(m+2) = s*(*m);
    *m = temp;
    break;
  case 1:
    if ((*m)==0) {
      ifail = 1;
      return;
    }
    *m = 1.0/(*m);
    break;
  default:
  { //Indentation is off.
    //The original algorithm
  if(nrow>max_array) {
    delete [] r;
    delete [] p;
    delete [] q;
    max_array = nrow;
    r = new int [max_array+1];
    p = new HepDouble [max_array+1];
    q = new HepDouble [max_array+1];
  }
  int i;
  for (i=1; i<=nrow;i++)r[i] = 1;
  
  HepDouble pivot1 = 0;
  for (i=1; i<=nrow;i++) {
    int k = 0;
    HepDouble big = 0;
    HepDouble *testp = m;
    int j;
    for (j=1;j<=nrow;j++) {
      HepDouble test = fabs(*testp);
      testp += (j+1);
      if (test > big) {
	if(r[j]>0) {
	  big = test;
	  k = j;
	}
      }
    }
    if (k<=0) {
      ifail = 1;
      return;
    }
    HepDouble *mk = m + k * (k-1) / 2;
    HepDouble *mkk = mk + (k-1);
    if (i==1) 
       pivot1 = *mkk;
    else if ((fabs((*mkk)/pivot1))<epsilon) {
      cerr << "matrix inv " << *mkk << " " << pivot1 << " "
	   << epsilon << endl; 
      ifail = 1;
      return;
    }
    r[k] = 0;
    p[k] = 1.0;
    q[k] = 1.0/(*mkk);
    *mkk = 0.0;
    if (k>1) {
      HepDouble *mkj = mk;
      for (j=1;j<k;j++) {
	p[j] = *mkj;
	q[j] = (*mkj)*q[k];
	if (r[j]>0) q[j] = -q[j];
	*(mkj++) = 0.0;
      }
    }
    if (k<nrow) {
      HepDouble *mjk = mkk + k;
      for (j=k+1;j<=nrow;j++) {
	p[j] = *mjk;
	q[j] = -(*mjk)*q[k];
	if (r[j]==0) p[j] = - p[j];
	*(mjk) = 0.0;
	mjk += j;
      }
    }
    HepDouble *mjj = m;
    for (j=1;j<=nrow;j++) {
      HepDouble *mkj = mjj;
      for (k=j;k<=nrow;k++) {
	*mkj += p[j] * q[k];
	mkj += k;
      }
      mjj += (j+1);
    }
  }
  }
  ifail = 0;
  return;
  }
}

HepDouble HepSymMatrix::determinant() const {
  static int max_array = 20;
  static int *ir = new int [max_array+1];

  if (nrow > max_array) {
    delete [] ir;
    max_array = nrow;
    ir = new int [max_array+1];
  }

  HepDouble det;
  HepMatrix mt(*this);
  int i = mt.dfact(det, ir);
  if(i==0) return det;
  return 0.0;
}

HepDouble HepSymMatrix::trace() const {
   HepDouble t = 0.0;
   for (int i=0; i<nrow; i++) 
      t += *(m + (i+3)*i/2);
   return t;
}
   
#ifndef HEP_NO_PERSISTENT_STREAMS

HepPIStream & operator >> ( HepPIStream &is,
			    HepSymMatrix &m) {
  int p;
  is >> p;
  if(m.nrow!=p)
  {
     m.delete_m(m.size, m.m);
     m.nrow = p;
     m.size = p * (p+1) / 2;
     m.m = m.new_m(m.size);
  }
  HepDouble *a = m.m;
  HepDouble *e = a + m.size;
  for (;a<e;a++) {
     is >> *a;
  }
  return is;
}

HepPOStream & operator << ( HepPOStream &os,
			    const HepSymMatrix &m) {
   os << m.num_row();
   HepDouble *a = m.m;
   HepDouble *e = a + m.num_size();
   for (;a<e;a++) {
      os << *a;
   }
   return os;
}

#endif
