// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: ARTE constructor for RecoTrack
//
// changes:
// 27 Jan 1997 creation by I. Siccama, C. Borgmeier
// 22 Nov 2000 (TG) added namespace CLUE
// 07 Dec 2000 (TG) added ClueRelation container initialization
// 15 Dec 2000 (TG) bug fix in constructors
// 12 Jan 2001 (TG) removed obsolete include files
// 19 Apr 2001 (TG) added filling of mommask and RICH likelihoods from RTRI bank
// 11 Jul 2001 (TG) added ehit initialization
// 19 Jul 2001 (TG) ehit initialization in 2nd CTOR added
// 27 Sep 2001 (TG) removed all but global likelihoods, replaced particle
//                  identification flags by Pid.
// 10 Jan 2002 (TG) set default particle likelihoods to 0.2, to
//                  avoid division by zero problems.
// 31 Jan 2002 (MSy) added all subsystem likelihoods again
//                   global likelihoods still filled like before (0.2, checkPID)
//                   RICH likelihoods from RTRI table (RITER lrXX and RISE lsXX) 
// ********************************************************************

#include <algo.h>
#include <arte/artefunc.h>      // for getEntry() call
#include "clue/cluearte.hh"
#include "clue/RecoTrack.hh"
#include "cluearte/checkPID.hh"

float rc99Bremsstrahlung(ArtePointer<RTRA> rtra);
const PID_data& checkPID(const ArtePointer<RTRA>& rtra);
int dummyFLS, dummyFLT;


// taken from copyRTRAtoRTRI.C
#define FLR_TEST  0x80000000
// bit compression for Rise:
#define FLR_NEW0   0x80000800
#define FLR_MASK0  0x7FFFF
#define FLR_SHIFT0 12
#define HRI_MASK0  0x7FF
// bit compression for Riter:
#define FLR_NEW1   0x80008000
#define FLR_MASK1  0x7FFF
#define FLR_SHIFT1 16
#define HRI_MASK1  0x7FFF


#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // Constructor
  //==============================================================================
  RecoTrack::RecoTrack( const ArtePointer<RTRA>& rtra, double mass ):
    RecoSegment( rtra ),
    mtrack (),
    clones (),
    rvertex(),
    ehit   (),
    Mass   (mass),
    Ebrems (0),
    iTopp  (0),
    Pid    (0),

    Le     (0.2),
    Lmu    (0.2),
    Lpi    (0.2),
    Lk     (0.2),
    Lp     (0.2),


    Flt    (0),
    Fle    (0),
    Flm    (0),

    Lre   (0),
    Lrmu  (0),
    Lrpi  (0),
    Lrk   (0),
    Lrp   (0),

    Lte    (0),
    Lth    (0),

    Lee    (0),
    Lemi   (0),
    Leh    (0), 

    Lmmu   (0),
    Lmh    (0), 

    Fit    (0),
    Hsi    (0),
    Hin    (0),
    Hou    (0),
    Hri    (0),
    Hca    (0),
    Htr    (0),
    Hmu    (0),
    Mommask(0),

    Rls   (0),
    Rlt   (0),


    Lse   (0),
    Lsmu  (0),
    Lspi  (0),
    Lsk   (0),
    Lsp   (0),
    Lsme  (0)
  {
    



    if( Mass == 0 && rtra->topp != 0)
      Mass = rtra->topp->mass; 
    // fix misuse of Lee: reconstr. JPsi mass is filled in lee
    if( rtra->fit == 2) {
      Mass = rtra->lee;
    }
    
    Ebrems = rc99Bremsstrahlung(rtra);
    
    if( rtra->topp != 0 ) {
      iTopp = getIDX<TOPP>( rtra->topp );
    }
    
    //   else {
    //        if( rtra->fit == 2 ) { // special case for J/Psi
    //  	set_topp(Jpsi.topp());
    //        } else {
    //  	iTopp = 0;
    //        }
    //      }
    Cer  = rtra->cer;
    
    const PID_data& pid = checkPID(rtra);
    
    Pid  = pid.PID;
    
    Le   = pid.Lelectron;
    Lmu  = pid.Lmuon;
    Lpi  = pid.Lpion;
    Lk   = pid.Lkaon;
    Lp   = pid.Lproton;
      

    // TRD likelihoods
    Lte  = rtra->lte;
    Lth  = rtra->lth;
    
    // ECAL
    // fix misuse of Lee: reconstr. JPsi mass is filled in lee (TG)
    if( rtra->fit == 2) {
      Lee = 0.;
      Mass = rtra->lee;
    } else {
      Lee  = rtra->lee;
    }
    Lemi = rtra->lemi;
    Leh  = rtra->leh;

    // MUON likelihoods
    Lmmu = rtra->lmmu;
    Lmh  = rtra->lmh;

    const ArtePointer<RTRI> rtri = getPTR<RTRI,RTRA>(rtra);

    // RICH (RISE) from RTRI
    Lre  = rtri->lte;
    Lrmu = rtri->ltmu;
    Lrpi = rtri->ltpi;
    Lrk  = rtri->ltk;
    Lrp  = rtri->ltp;

    // RICH (RITER) from RTRI
    Lse  = rtri->lse;
    Lsmu = rtri->lsmu;
    Lspi = rtri->lspi;
    Lsk  = rtri->lsk;
    Lsp  = rtri->lsp;
    Lsme  = rtri->lsme;

    
    //PID flags

    //    dummyFLT=rtri->flt;    
    //    dummyFLS=rtri->fls;

    //     cout << dummyFLT << "     " << dummyFLS << endl;
    
//     if ( dummyFLT & FLR_TEST ) { dummyFLT = dummyFLT | FLR_NEW0 ; }
//     if ( dummyFLS & FLR_TEST ) { dummyFLS = dummyFLS | FLR_NEW1 ; }

//     if ( dummyFLT & FLR_NEW0 ) { Rlt = ( dummyFLT) >> FLR_SHIFT0 & FLR_MASK0; }
//     if ( dummyFLS & FLR_NEW1 ) { Rls = ( dummyFLS) >> FLR_SHIFT1 & FLR_MASK1; }

//     cout << dummyFLT << "     " << dummyFLS << endl;
    
    // a simple copy doesn't extract the flag
    Rls  = rtri->fls; //RITER
    Rlt  = rtri->flt; //RISE

    Flt  = rtra->flt;
    Fle  = rtra->fle;
    Flm  = rtra->flm;
      

    Fit  = rtra->fit;
    
    Hsi  = rtra->hsi;
    Hin  = rtra->hin;
    Hou  = rtra->hou;
    Hri  = rtra->hri;
    //    Hca  = rtra->hca; // at the moment this is nonsense
    Hca  = distance( rtra->rccl.begin(), rtra->rccl.end() );
    Htr  = rtra->htr;
    Hmu  = rtra->hmu;
    
    Mommask = rtra->mommask;
    
    set_nf( Hsi + Hin + Hou - 5 ); // degrees of freedom
  }


  //==============================================================================
  // Constructor
  //==============================================================================
  RecoTrack::RecoTrack( const ArtePointer<RSEG>& rseg ) :
    RecoSegment(rseg),
    mtrack (),
    clones (),
    rvertex(),
    ehit   (),
    Mass   (0),
    Ebrems (0),
    Cer    (0),
    iTopp  (0),
    Pid    (0),
    Le     (0.2),
    Lmu    (0.2),
    Lpi    (0.2),
    Lk     (0.2),
    Lp     (0.2),


    Flt    (0),
    Fle    (0),
    Flm    (0),

    Lre   (0),
    Lrmu  (0),
    Lrpi  (0),
    Lrk   (0),
    Lrp   (0),

    Lte    (0),
    Lth    (0),

    Lee    (0),
    Lemi   (0),
    Leh    (0),
 
    Lmmu   (0),
    Lmh    (0), 


    Fit    ( rseg!=0 ? rseg->fit : 0 ),

    //
    // sorry for this wild piece. should be changed in future.
    //
    Hsi( count_if( rseg->hitb.begin(), rseg->hitb.end(),
		   compose1( bind2nd( equal_to<int>(),
				      1 ),
			     compose1( getEntry( &GEDE::cmp ), 
				       getEntry( &HITB::gede ) )
			     )
		   )
	 ),
    Hin( count_if( rseg->hitb.begin(), rseg->hitb.end(),
		   compose1( bind2nd( equal_to<int>(),
				      3 ),
			     compose1( getEntry( &GEDE::cmp ), 
				       getEntry( &HITB::gede ) )
			     )
		   )
	 ),
    Hou( count_if( rseg->hitb.begin(), rseg->hitb.end(),
		   compose1( bind2nd( equal_to<int>(),
				      2 ),
			     compose1( getEntry( &GEDE::cmp ), 
				       getEntry( &HITB::gede ) )
			     )
		   )
	 ),
    Hri( count_if( rseg->hitb.begin(), rseg->hitb.end(),
		   compose1( bind2nd( equal_to<int>(),
				      5 ),
			     compose1( getEntry( &GEDE::cmp ), 
				       getEntry( &HITB::gede ) )
			     )
		   )
	 ),
    Hca( count_if( rseg->hitb.begin(), rseg->hitb.end(),
		   compose1( bind2nd( equal_to<int>(),
				      7 ),
			     compose1( getEntry( &GEDE::cmp ), 
				       getEntry( &HITB::gede ) )
			     )
		   )
	 ),
    Htr( count_if( rseg->hitb.begin(), rseg->hitb.end(),
		   compose1( bind2nd( equal_to<int>(),
				      6 ),
			     compose1( getEntry( &GEDE::cmp ), 
					 getEntry( &HITB::gede ) )
			     )
		   )
	 ),
    Hmu( count_if( rseg->hitb.begin(), rseg->hitb.end(),
		   compose1( bind2nd( equal_to<int>(),
				      4 ),
			     compose1( bind2nd( divides<int>(),
						2 ),
				       compose1( getEntry( &GEDE::cmp ), 
						 getEntry( &HITB::gede ) )
				       )
			     )
		   )
	 ),
    Mommask(0), 
    Rls    (0),
    Rlt    (0),

    Lse   (0),
    Lsmu  (0),
    Lspi  (0),
    Lsk   (0),
    Lsp   (0),
    Lsme  (0)


{

    set_nf(Hsi + Hin + Hou - 5);
  }
  
#ifdef NAMESPACE
} // namespace CLUE
#endif
