// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   02. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: default class to fill a BEE event
//
// changes:
// 02 Jan 2001 (TG) creation
// 09 Feb 2001 (TG) fillWire(): wire filled from RTAR now 
// 11 Jul 2001 (TG) new filling of RecoTrack-EcalHit relation, Fill(): call
//                  FillMisc() first
// 11 Aug 2001 (TG) simplified FillMCtracks()
// 10 Sep 2001 (TG) added CTOR, added min_energy, added FillSelectedMCtracks(),
//                  FillSelectedMCvertex()
// 13 Sep 2001 (TG) improved FillMCvertex(), FillSelectedMCvertex()
// 03 Dec 2001 (TG) bug fix in FillSelectedMCvertex: added missing relation
//
// *****************************************************************************

#include <map>
#include "cluearte/TableFill.hh"
#include "cluearte/CloneMap.hh"

#include "clue/cluearte.hh"

#include "clue/Wire.hh"
#include "clue/RecoTrack.hh"
#include "clue/RecoVertex.hh"
#include "clue/MC_Track.hh"
#include "clue/MC_Vertex.hh"
#include "clue/EcalHit.hh"
#include "clue/CustomTrack.hh"

typedef map<ArtePointer<MTRA>,ArtePointer<MC_Track> >  MTMAP; 
typedef map<ArtePointer<MVER>,ArtePointer<MC_Vertex> > MVMAP; 
static  MTMAP mtramap;
static  MVMAP mvermap;

//==============================================================================
// Constructors
//==============================================================================
TableFill::TableFill() :
  min_energy(0.)
{}

TableFill::TableFill(double e) :
  min_energy(e)
{}

TableFill::~TableFill() {}

//==============================================================================
// EventSelect
//==============================================================================
bool TableFill::EventSelect() const {
  return true;
}

//==============================================================================
// Fill
//==============================================================================
bool TableFill::Fill() const {

  FillMisc();
  FillWire();
  FillRecoTrack();
  FillRecoVertex();

  if(min_energy != 0.) {
    FillSelectedMCtrack();
    FillSelectedMCvertex();
  } else {
    FillMCtrack();
    FillMCvertex();
  }

  return true;
}

//==============================================================================
// FillRSEG
//==============================================================================
bool TableFill::FillRSEG() const {
  const ArteTableIterator<RSEG> end = ArteTable<RSEG>::end();
  for( ArteTableIterator<RSEG> rsit = ArteTable<RSEG>::begin();
       rsit != end; ++rsit ) {
    const ArtePointer<RSEG> rseg = *rsit;

    if( rseg->cmp & Rsegc::vxd ) {
      ArteTable<VDS_Track>::push_back( VDS_Track(rseg) );
    } else if( (rseg->cmp & Rsegc::patt || rseg->cmp & Rsegc::trit ) 
	       && (rseg->cmp & Rsegc::rich || rseg->cmp & Rsegc::ecal)) {
      // save only OTR/RICH matches!!
      ArteTable<OTR_SpacePoint>::push_back( OTR_SpacePoint(rseg) );
    } else if( rseg->cmp & Rsegc::rich ) {
      ArteTable<RICH_Track>::push_back( RICH_Track(rseg) );
    } else if( rseg->cmp & Rsegc::muon ) {
      ArteTable<MuonTrack>::push_back( MuonTrack(rseg) );
    }
  } // for RSEG

  return true;
} 

//==============================================================================
// FillWire
//==============================================================================
bool TableFill::FillWire() const {

  const ArteTableIterator<RTAR> end = ArteTable<RTAR>::end();
  for( ArteTableIterator<RTAR> rtar = ArteTable<RTAR>::begin();
       rtar != end; ++rtar ) {
      ArteTable<Wire>::push_back( Wire(*rtar) );
  }

#ifdef XXX
  /// fill from GTAR
  const ArteTableIterator<GTAR> end = ArteTable<GTAR>::end();
  for( ArteTableIterator<GTAR> gtar = ArteTable<GTAR>::begin();
       gtar != end; ++gtar ) {

    // select inserted wires
    if( fabs(gtar->x) < 1.2 && fabs(gtar->y) < 1.2) {
      ArteTable<Wire>::push_back( Wire(*gtar) );
    }
  }
#endif

#ifdef ClueDEBUG
  cout << "Wires bank:" << endl;
  for(ArteTable<Wire>::iterator it = ArteTable<Wire>::begin(); 
      it != ArteTable<Wire>::end(); ++it) {
    cout << **it << endl;
  }
#endif

  return true;
}

//==============================================================================
// FillRecoTrack
//==============================================================================
bool TableFill::FillRecoTrack() const {

  // define clone maps
  typedef R_trait<RecoTrack>::Cl_map Cl_map;
  Cl_map svdclone;
  Cl_map mainclone;
  Cl_map richclone;
  Cl_map ecalclone;
  Cl_map muonclone;

  const ArteTableIterator<RTRA> end = ArteTable<RTRA>::end();
  for( ArteTableIterator<RTRA> rtit = ArteTable<RTRA>::begin();
       rtit != end; ++rtit ) {
    const ArtePointer<RTRA> rtra = *rtit;

#ifdef XXX
    // the following is not true any more

    // do not save reconstructed particles here - they are saved
    // in RecoVertex array
    const unsigned int D = rtra->fit%100;
    if( D == 2 || D == 3 ) { 
      cout << " Skip rtra " << rtra << endl;
      continue;
    }
#endif

    const ArtePointer<RecoTrack> recotr = 
      ArteTable<RecoTrack>::push_back( RecoTrack(rtra) );

#ifdef XXX
    // set relation to ECAL cluster
    if(distance(rtra->rccl.begin(),rtra->rccl.end()) > 0) {
      recotr->set_ecal( getPTR<EcalHit,RCCL>( *rtra->rccl.begin() ));
    }
    // set relation to Ecal cluster
    const ArteRelationIterator<RCCL> rcend = rtra->rccl.end();
    for(ArteRelationIterator<RCCL> rcit = rtra->rccl.begin(); rcit != rcend;
	++rcit) {
      createRelation(recotr, getPTR<EcalHit,RCCL>( *rcit ));
    }
#endif

    // fill SVD clone map
    const int csvd = rtra->csvd;
    if( csvd > 0) {
      fillCloneMap<RecoTrack>(svdclone, csvd, recotr);
    }

    // fill OTR clone map
    const int cmain = rtra->cmain;
    if( cmain > 0 ) {
      fillCloneMap<RecoTrack>(mainclone, cmain, recotr);
    }

    // fill RICH clone map
    const int crich = rtra->crich;
    if( crich > 0 ) {
      fillCloneMap<RecoTrack>(richclone, crich, recotr);
    }

    // fill ECAL clone map
    const int cecal = rtra->cecal;
    if( cecal > 0 ) {
      fillCloneMap<RecoTrack>(ecalclone, cecal, recotr);
    }

    // fill Muon clone map
    const int cmuon = rtra->cmuon;
    if( cmuon > 0) {
      fillCloneMap<RecoTrack>(muonclone, cmuon, recotr);
    }

  } // for RTRA
  

  //----------------------------------------------------------------------------
  // Create Clone Relations
  //----------------------------------------------------------------------------
  createCloneRelation<RecoTrack>(svdclone,  Rsegc::vxd );
  createCloneRelation<RecoTrack>(mainclone, Rsegc::patt | Rsegc::magt );
  createCloneRelation<RecoTrack>(richclone, Rsegc::rich );
  createCloneRelation<RecoTrack>(ecalclone, Rsegc::ecal );
  createCloneRelation<RecoTrack>(muonclone, Rsegc::muon );

#ifdef ClueDEBUG
  cout << " clones again: " << endl;
  for(ArteTableIterator<RecoTrack> it = ArteTable<RecoTrack>::begin(); it !=
	ArteTable<RecoTrack>::end(); ++it) {
    if(it->clones.size() > 0) {
      cout << *it << " rtrack->clones: " << it->clones << endl;
    }
  }
#endif

  return true;
}

//==============================================================================
// FillRecoVertex
//==============================================================================
bool TableFill::FillRecoVertex() const {

  const ArteTableIterator<RVER> end = ArteTable<RVER>::end();
  for( ArteTableIterator<RVER> rvit = ArteTable<RVER>::begin();
       rvit != end; ++rvit ) {
    const ArtePointer<RVER> rver = *rvit;

    const ArtePointer<RecoVertex> rVertex = 
      ArteTable<RecoVertex>::push_back( RecoVertex(rver) );

    //    double chi2 = 0;
    const ArteRelationIterator<RREL> rend = rver->rrel.end();
    for( ArteRelationIterator<RREL> rrit = rver->rrel.begin();
	 rrit != rend; ++rrit ) {
      const ArtePointer<RREL> rrel = *rrit;

      if( rrel->rtra != 0 ) {
	const ArtePointer<RecoTrack> rTrack = getPTR<RecoTrack,RTRA>( rrel->rtra );
#ifdef XXX
	cout << "rrel->rtra: " << rrel->rtra 
	     << " ArteRTRA size: " << ArteTable<RTRA>::size()
	     << " ArteRECT size: " << ArteTable<RecoTrack>::size()
	     << " ArteRVER size: " << ArteTable<RVER>::size()
	     << " ArteRECV size: " << ArteTable<RecoVertex>::size()
	     << " rVertex: " << rVertex 
	     << " rTrack: " << rTrack << endl;
#endif
	createRelation( rVertex, rTrack );
      }
      //      chi2 += rrel->chi;
    }
    //    rVertex->set_chi2(chi2);
  }

  return true;
}


//==============================================================================
// FillMCtrack
//==============================================================================
bool TableFill::FillMCtrack() const {

  const ArteTableIterator<MTRA> end = ArteTable<MTRA>::end();
  for( ArteTableIterator<MTRA> mtit = ArteTable<MTRA>::begin();
       mtit != end; ++mtit ) {
    const ArtePointer<MTRA> mtra = *mtit;

    const ArtePointer<MC_Track> mcTrack = ArteTable<MC_Track>::push_back( MC_Track(mtra) );
    
    // create relations to reconstructed tracks
    const ArteRelationIterator<RTRA> rend = mtra->rtra.end();
    for( ArteRelationIterator<RTRA> rtra = mtra->rtra.begin();
	 rtra != rend; ++rtra ) {

      const ArtePointer<RecoTrack> rTrack = getPTR<RecoTrack,RTRA>( *rtra );
      createRelation( rTrack, mcTrack );
    } // for RTRA

  } // for MTRA

  return true;
}

//==============================================================================
// FillMCvertex
//==============================================================================
bool TableFill::FillMCvertex() const {

  const ArteTableIterator<MVER> end = ArteTable<MVER>::end();
  for( ArteTableIterator<MVER> mvit = ArteTable<MVER>::begin();
       mvit != end; ++mvit ) {
    const ArtePointer<MVER> mver = *mvit;

    const ArtePointer<MC_Vertex> mcvtx = ArteTable<MC_Vertex>::push_back( MC_Vertex(mver) );

    // could happen for primary vertex...
    if(mver->btra == 0) {

      // set relations to outgoing tracks
      const ArteRelationIterator<MTRA> mend = mver->etra.end();
      for( ArteRelationIterator<MTRA> mtit = mver->etra.begin();
	   mtit != mend; ++mtit ) {
	
	const ArtePointer<MC_Track> mtra = getPTR<MC_Track,MTRA>( *mtit );
	createSingleRelation( mcvtx, mtra );
	mtra->set_bVertex(mcvtx);
      } // for MTRA

    } else { // mtra->btra
      // find mother track
      const ArtePointer<MC_Track> mother = getPTR<MC_Track,MTRA>(mver->btra);
      // relation to endvertex
      createSingleRelation( mother, mcvtx );

      // set relations to outgoing tracks
      const ArteRelationIterator<MTRA> mend = mver->etra.end();
      for( ArteRelationIterator<MTRA> mtit = mver->etra.begin();
	   mtit != mend; ++mtit ) {
	
	const ArtePointer<MC_Track> mtra = getPTR<MC_Track,MTRA>( *mtit );
	createSingleRelation( mcvtx, mtra );
	mtra->set_bVertex(mcvtx);
	// set endTrack relations
	createSingleRelation( mother, mtra );
      } // for MTRA

    } // if mtra->bver
  } // for MVER


  // MC_Tracks: set relations to outgoing MC vertices
  const ArteTableIterator<MTRA> mtend = ArteTable<MTRA>::end();
  for( ArteTableIterator<MTRA> mtit = ArteTable<MTRA>::begin();
       mtit != mtend; ++mtit ) {
    const ArtePointer<MTRA> mtra = *mtit;

    // in case no end vertices exist, create relation from etras
    if(distance(mtra->ever.begin(),mtra->ever.end()) == 0) {

      const ArtePointer<MC_Track> mcTrack = getPTR<MC_Track,MTRA>(mtra);

      const ArteRelationIterator<MTRA> mrend = mtra->etra.end();
      for( ArteRelationIterator<MTRA> mrit = mtra->etra.begin();
	   mrit != mrend; ++mrit ) {
	const ArtePointer<MTRA> mtout = *mrit;

	const ArtePointer<MC_Track> mcEndTrack = getPTR<MC_Track,MTRA>(mtout);
	createSingleRelation(mcTrack, mcEndTrack);
      } // for MTRA
    }

#ifdef XXX
    copy(mcTrack->endVtx.begin(), mcTrack->endVtx.end(), ostream_iterator<ArtePointer<MC_Vertex> >(cout, " "));
    cout << "end vertices: " << mcTrack->endVtx.size() << endl;
#endif
  } // for MTRA

  return true;
}

//==============================================================================
// FillMisc
//==============================================================================
bool TableFill::FillMisc() const {

  //----------------------------------------------------------------------------
  // EcalHit
  //----------------------------------------------------------------------------
  const ArteTableIterator<RCCL> eend = ArteTable<RCCL>::end();
  for( ArteTableIterator<RCCL> rccl = ArteTable<RCCL>::begin();
       rccl != eend; ++rccl ) {

    ArteTable<EcalHit>::push_back( EcalHit(*rccl) );
  }

  return true;
}


//==============================================================================
// FillSelectedMCtrack
//==============================================================================
bool TableFill::FillSelectedMCtrack() const {

  mtramap.clear();

  // 1st loop: fill MC_Track table
  const ArteTableIterator<MTRA> end = ArteTable<MTRA>::end();
  for( ArteTableIterator<MTRA> mtit = ArteTable<MTRA>::begin();
       mtit != end; ++mtit ) {
    const ArtePointer<MTRA> mtra = *mtit;

    if(sqrt(square(mtra->px)+square(mtra->py)+square(mtra->pz)) < min_energy)
      continue;

    const ArtePointer<MC_Track> mcTrack = ArteTable<MC_Track>::push_back( MC_Track(mtra) );

    mtramap.insert(MTMAP::value_type(mtra,mcTrack));
    
    // create relations to reconstructed tracks
    const ArteRelationIterator<RTRA> rend = mtra->rtra.end();
    for( ArteRelationIterator<RTRA> rtra = mtra->rtra.begin();
	 rtra != rend; ++rtra ) {

      const ArtePointer<RecoTrack> rTrack = getPTR<RecoTrack,RTRA>( *rtra );
      createRelation( rTrack, mcTrack );
    } // for RTRA

  } // for MTRA


  // 2nd loop: set mother tracks
  const MTMAP::iterator mtend = mtramap.end();
  for( MTMAP::iterator it = mtramap.begin(); it != mtend; ++it ) {

    const ArtePointer<MTRA> moth = it->first->btra;
    const ArtePointer<MC_Track> mctrack = it->second;
    const MTMAP::iterator m = mtramap.find(moth);
    if(m != mtend) {
      mctrack->set_mother(m->second);
    }
  }
  
  return true;
}

//==============================================================================
// FillSelectedMCvertex
//==============================================================================
bool TableFill::FillSelectedMCvertex() const {

  mvermap.clear();

  const ArteTableIterator<MVER> end = ArteTable<MVER>::end();
  for( ArteTableIterator<MVER> mvit = ArteTable<MVER>::begin();
       mvit != end; ++mvit ) {
    const ArtePointer<MVER> mver = *mvit;

    // check whether this vertex has an outgoing track with E>=min_energy
    unsigned int ne = 0;
    const ArteRelationIterator<MTRA> mend = mver->etra.end();
    for( ArteRelationIterator<MTRA> mtit = mver->etra.begin();
	 mtit != mend; ++mtit ) {
      const ArtePointer<MTRA> mtra = *mtit;
      if(sqrt(square(mtra->px) + square(mtra->py) + square(mtra->pz)) >= min_energy) {
	++ne;
      }
    }

    if(ne == 0) continue;

    const ArtePointer<MTRA> btra = mver->btra;
    bool bexist = (btra != 0 && sqrt(square(btra->px) +
				     square(btra->py) +
				     square(btra->pz)) >=
		   min_energy) ? true : false;

    ArtePointer<MC_Track> mcbtra; // the begin track
    if(bexist) {
      mcbtra = mtramap[btra];
    } else {
      continue;
    }

    // now save MC_Vertex
    const ArtePointer<MC_Vertex> mcvtx = ArteTable<MC_Vertex>::push_back( MC_Vertex(mver) );
    mcvtx->set_mother(mcbtra); // set relation to mother track
    createSingleRelation( mcbtra, mcvtx ); // set relation to end vertices

    mvermap.insert(MVMAP::value_type(mver, mcvtx));

    // set relations to outgoing tracks
    for( ArteRelationIterator<MTRA> mtit = mver->etra.begin();
	 mtit != mend; ++mtit ) {
      const ArtePointer<MTRA> mtra = *mtit;

      if(sqrt(square(mtra->px) + square(mtra->py) + square(mtra->pz)) < min_energy)
	continue;

      const ArtePointer<MC_Track> mtrack = mtramap[*mtit];

      createSingleRelation( mcvtx, mtrack );
      mtrack->set_bVertex(mcvtx);

      if(bexist) {
	// MC_Track: create endTrack relations
	createSingleRelation(mcbtra, mtrack);
      }
    } // for MTRA
  } // for MVER

  return true;
}
