#ifndef __TABLEFILL_HH
#define __TABLEFILL_HH
// *****************************************************************************
//
// source:
//
// type:      header file
//
// created:   02. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: default class to fill a BEE event
//
// changes:
// 02 Jan 2001 (TG) creation
// 10 Sep 2001 (TG) added CTOR, added min_energy, added FillSelectedMCtracks(),
//                  FillSelectedMCvertex(), set_min_energy()
//
// *****************************************************************************

#include "cluearte/TableFillBase.hh"

/** Class which defines a way to fill a BEE event. The TableFillBase interface
    class is used. The TableFill class performs a 1:1 copy of ARTE tables to
    BEE tables. In case of writing MC truth information, an energy threshold
    for the MC tracks (MTRA in ARTE) can be set. Then only those MTRA tracks
    are stored in BEE, which have an energy above the threshold. This gives the
    user the possiblity to optionally avoid low-energetic tracks from
    shower simulations. This can reduce the size of an BEE file considerably.

    @memo BEE event filling: 1 to 1 copy
    @author T. Glebe
 */
//============================================================================
// TableFill
//============================================================================
class TableFill : public TableFillBase {
public:
  /** @name --- Constructors --- */
  ///
  TableFill();
  /// set track energy threshold
  TableFill(double e);
  ///
  virtual ~TableFill();

  /** @name --- Functions --- */
  /// returns false in case event should be skipped
  virtual bool EventSelect()    const;
  /// fill all tables, called in clueWriteEvent()
  virtual bool Fill()           const;
  /// fill from RSEG tables
  virtual bool FillRSEG()       const;
  /// fill Wire table
  virtual bool FillWire()       const;
  /// fill RecoTrack table
  virtual bool FillRecoTrack()  const;
  /// fill RecoVertex table
  virtual bool FillRecoVertex() const;
  /// fill MC tracks
  virtual bool FillMCtrack()    const;
  /// fill MC vertices
  virtual bool FillMCvertex()   const;
  /// fill other information
  virtual bool FillMisc()       const;
  /// fill MC tracks if above energy threshold
  virtual bool FillSelectedMCtrack()    const;
  /// fill MC vertices if tracks above threshold
  virtual bool FillSelectedMCvertex()   const;

  /** @name --- Access methods --- */
  /// MC: set min. energy of particles to be copied
  inline void set_min_energy(double e) { min_energy = e; }

private:
  double min_energy;   // in case of MC: min. Energy of particles
}; // class TableFill
#endif
