// *****************************************************************************
//
// source:    daq_unsparse_if.C
//
// type:      source code
//
// created:   13 Mar 2001
//
// author:    Lars Soezueer
//            email: soezueer@mail.desy.de
//
// purpose:   functions for unsparsified components
//            called by daqif.C
//
// changes:   
//
// *****************************************************************************
#include "arte/ArteMessage.hh"

extern "C" {
#include "esp.h"
	   }

#include "daqif/daq_unsparse_if.hh"

DAQ_UNSPARSE_IF::DAQ_UNSPARSE_IF(int component_id) : DAQ_MC_IF(component_id) {;}

/***
   *
   * The default method not only returns the number of SLBs, it also determines that number
   * and allocates FEDs to SLBs. 
   * getN_SLBs is only needed for unsparsified components.
   *
   ***/

unsigned int DAQ_UNSPARSE_IF::getN_SLBs(void)
{
  if (setSLBsCalled == 0)
    setSLBs();

  return nSLBs;
}


/***
   *
   * This asks for the FED ids, then attempts to allocate them to SLBs, thus
   * setting up FED_ids and SLB arrays.
   *
   ***/

void DAQ_UNSPARSE_IF::setSLBs(void)
{
  setSLBsCalled = 1;

  if(nSLBs != 0)
    return;

  unsigned short ids[MAX_FEDS];

  // first set up the FED array
  unsigned int nFEDs = getFED_IDs(ids, MAX_FEDS);


  if(nFEDs == 0)
    return;

  fedIDs = new unsigned short[nFEDs];

  for(unsigned int i = 0; i < nFEDs; i++)
    fedIDs[i] = ids[i];

  SLB_struct slbs[MAX_SLBS];

  // Assign FEDs to SLBs consecutively. Require that total SLB length be less than MAX_SLB_SIZE

  int status;
  unsigned int length = sizeof(SLB_HDR) + sizeof(SLB_TAIL); // running length
  short fed_len;
  unsigned int i_slb = 0;
  slbs[i_slb].first = fedIDs;
  slbs[i_slb].nFEDs = 0;

  FED_HDR *fhd = (FED_HDR*) fedBuf;
  for(unsigned int i = 0; i < nFEDs; i++) {
    if( (status = getFED(fedIDs[i], fhd, MAX_FED_SIZE)) != 0) {
      arteMsg << "DAQ_UNSPARSE_IF::setSLBs: error return from getFED, component: "
              << compid << "fed: " << fedIDs[i] ;
      arteMsg.warn();
      return; // effectively ignoring this component
    }

    ESP_get_short(&fhd->length, &fed_len, 1);

    if(length + fed_len < MAX_SLB_SIZE) {
      slbs[i_slb].nFEDs++;
      length += fed_len;
    } else {
      if( ++i_slb >= MAX_SLBS) {
        arteMsg  << "DAQ_MC_IF::setSLBs: not enough SLBs for component "
                 << compid << ", truncating";
        arteMsg.warn();
	break;
      }
      slbs[i_slb].first = &fedIDs[i];
      slbs[i_slb].nFEDs = 1;
      length = sizeof(SLB_HDR) + sizeof(SLB_TAIL) + fed_len;
    }
  }

  if(slbs[i_slb].nFEDs != 0)
    i_slb++;

  nSLBs = i_slb;
  SLB = new SLB_struct[nSLBs];

  if(SLB == 0) {
    arteMsg << "DAQ_MC_IF::setSLBs: no room for SLB array, component "
            << compid << ", length: " << nSLBs ;
    arteMsg.warn();
    nSLBs = 0;
    delete[] fedIDs;
    return;
  }

  for(int i = 0; i < nSLBs; i++)
    SLB[i] = slbs[i];

}

int DAQ_UNSPARSE_IF::prepare(void)
{
 nSLBs = getN_SLBs();
 return (nSLBs<=0);
}
