// *****************************************************************************
//
// source:    fillSEVTfromHITB.C
//
// type:      source code
//
// created:   14 Mar 2001
//
// author:    B. Schwingenheuer, Lars Soezueer
//            email: soezueer@mail.desy.de
//
// purpose:   fill Marvin Sparse Event Buffer from HITB for SVD
//            called by daq_svd_if.C
//
// changes:   
//
// *****************************************************************************
#include "arte/ArteMessage.hh"
#include "marvin.h"
#include "arte/arte.hh"

// memory allocation for Marvin Sparse Event: need 8 bytes/cluster
// 40000 bytes correspond to 5000 clusters
#define SEVT_INIT_SIZE 40000 
#define MAXHIT ((SEVT_INIT_SIZE-20)/8)

int fillSEVTfromHITB(int major, int minor)
{
  static Marvin_Sevt_t  *sevt;

  int cell = 0;
  int format = 3;
  int ncl = 0;
  int sideid;
  int width = 2;


  int isize = sizeof(int);
  int csize = sizeof(char);

  float  cog, sn, charge;
  float  f4sn;
  unsigned char bside, bcharge, b4sn;
  signed char   bwidth;

  static int first_marvin = 1;

  char *routine="fillSEVTfromHITB";

  /* initialize marvin and allocate some space for sparsified event */
  if ( first_marvin ) {

    marvin_init();

    sevt = marvin_fetch_sevt();
    if ( !sevt ) {
      arteMsg << routine
              << "no Marvin_Sevt_t available (marvin not initialized?)";
      arteMsg.warn();
      return 1;
    }

    sevt->data   = (char*) malloc(SEVT_INIT_SIZE); 
    sevt->size   = SEVT_INIT_SIZE;
    sevt->format = format;

    first_marvin = 0;
 
  }
 
  sevt->next = sevt->data;

  /* initialize an sevt-header */
  ESP_put_int( &major, (int*) sevt->next, 1);
  sevt->next += isize;
  ESP_put_int( &minor, (int*) sevt->next, 1); 
  sevt->next += isize;
  ESP_put_int( &cell,  (int*) sevt->next, 1); 
  sevt->next += isize;
  ESP_put_int( &format, (int*) sevt->next, 1); 
  sevt->next += isize;
  ESP_put_int( &ncl,    (int*) sevt->next, 1); 
  sevt->next += isize;


  // loop over all VDS hits, store them im sparse event format 3

  for ( ArteTable<HITB>::iterator hitbIt = ArteTable<HITB>::begin();
        hitbIt != ArteTable<HITB>::end(); hitbIt++ ) {
    ArtePointer<HITB> hitb = *hitbIt;

    ArtePointer<GEDE> gede = hitb->gede;
   
    // process VDS hits only 

    if(gede->cmp != 1) continue;

    // check for enough space in Sparse Event buffer 
    if(ncl+1 > MAXHIT) {
      arteMsg << routine
              << " Sparse Event Buffer full!";
      arteMsg.warn();
      return 1;
    }

    // calculate plane number in Mavin naming scheme 
    ArtePointer<GESL> gesl = gede->gesl;
    if(gesl->nsl < 4)
	sideid = 10000 + 10*(gesl->nsl - 1);
    else
	sideid = 10000*(gesl->nsl - 2);

    sideid +=  300*(gede->sec - 1) + gede->lay - 1;

    cog    = (hitb->x -  gede->xfw[0])/gede->pit; 
    sn     = hitb->y;
    charge = hitb->q;
 
    f4sn = 4.*sn;
    if(charge>0) {
        bcharge = charge > 254. ? 255 : (int)charge;
        bwidth  = width  > 127  ? 127 : width;
        b4sn    = f4sn   > 254. ? 255 : (int)f4sn;
      } 
      else {
        bcharge = -charge > 254. ?  255 : (int)(-charge);             
        bwidth  =  width  > 127  ? -127 : -width;
        b4sn    = -f4sn   > 254. ?  255 : (int)(-f4sn);
      }
      bside = marvin_pot2sl(sideid);
      sevt->next = (char *)memcpy(sevt->next, &bside  , csize) + csize;
      sevt->next = (char *)memcpy(sevt->next, &bcharge, csize) + csize;       
      sevt->next = (char *)memcpy(sevt->next, &b4sn   , csize) + csize;
      sevt->next = (char *)memcpy(sevt->next, &bwidth , csize) + csize;
      ESP_put_int( (int*) (&cog), (int*) sevt->next, 1);
      sevt->next += isize;
      ncl++;
   }

  // update cluster count in sparse event header
  ESP_put_int( &ncl, (int*) sevt->data + 4, 1);

  return(0);
}
