// *****************************************************************************
//
// source:    daq_mc_if.hh
//
// type:      header file
//
// created:   15 Mar 2001
//
// author:    Mike Medinnis, Lars Soezueer
//            email: soezueer@mail.desy.de
//
// purpose:   virtual base class for daqif
//
// changes:   
//
// *****************************************************************************
#ifndef _DAQ_MC_IF_
#define _DAQ_MC_IF_

#include <string>

extern "C" {
#include "component.h"
#include "event.h"
           }

/***
   * This is a virtual base class. 
   * It is intended for generating daq-format DSTs from monte carlo
   * inside of Arte.
   *
   * For each component id, a subclass is needed. 
   * This subclass can either be derived directly from the base class
   * or preferably from one of the intermediate virtual classes
   * DAQ_UNSPARSE_IF, DAQ_SPARSE_IF if it fits in there.
   * Two or three virtual methods must be
   * defined for each subclass. Just to avoid having to couple the base class in with all of
   * the various Arte libraries, I'll keep the subclasses together and ask each subsystem to
   * provide me with two functions which allow easy construction of the needed methods. The
   * two functions (1) give a list of all FED ids for the component and (2) assemble a single
   * FED record. For subclasses of DAQ_SPARSE_IF, the virtual method 'prepare' calls 
   * a third function 'prepareEvent' to prepare the sparsification for each event.
   *
   * One and only one instance of the subclass should be allocated on the heap
   * (via call to new) at initialization time.
   * (If more than one is instantiated, all but the first are ignored.)
   *
   * At least for now, getN_SLBs should be defined in the base class only:
   *
   * getN_SLBs returns the number of SLBs for this component. 
   *           The base class defines a method which returns 1.
   *           This is adequate for most purposes. The SLB so defined is not "real" in
   *           that it isn't limited to just six sharc links. 
   *           The default method "allocates" enough
   *           SLBs to keep the maximum size of the SLB records below a maximum value.
   *
   * The two virtual methods needed for each detector system:
   *
   * (1) getFED_IDs  builds a list of all FED ids for this component in buf,
   *                 for the given slb.
   *                 returns the number of FEDs.
   *
   * (2) getFED  returns an unsparsified FED record in export format for the given fedid. 
   *             Note that this should return an FED formatted record
   *             (FED_HDR, as defined in event.h) as opposed to an SLB_SUBHDR
   *             (also defined in event.h).
   *
   * The base class does all the other work of putting an event together.
   *
   ***/

class DAQ_MC_IF {

public:

  static int initialize(unsigned int compMask); 
  static bool initialized;
  static EVENT_HDR* getEvent(unsigned int *nBytes); // build the current event in memory
  static int init_pipe(void);
  static int init_file(char* filename);
  static int writeEvent(EVENT_HDR*); // build the current event then write it out

  static unsigned int buildEvent(char *buf, unsigned int maxSize); // do not override
  static int buildFED(unsigned int component, unsigned short fedid, FED_HDR *buf, unsigned int maxSize);
  static const DAQ_MC_IF *getDAQ_IF(int component_id);

  virtual unsigned int buildComponent(char *buf, unsigned int maxSize); // overridden only for DAQ and sparsified SVD & ITR
  unsigned int buildSLB(unsigned int slbid, char *buf, unsigned int maxSize); // do not override

  static short get_fltevt(void); //reads the FLT evt #

  // to be defined for each component (subclass):
  // Return the total number of FEDs for component
  virtual unsigned int getFED_IDs(unsigned short *buf, unsigned int maxIDs) { return 0;}

  virtual int prepare(void) {return -1;} //status return
    
protected:

  DAQ_MC_IF(int component_id = -1); // not meant to be instantiated

  // to be defined for each component (subclass):
  // assemble an unsparsified FED record in buf in export format. Return 0 on success.
  virtual int getFED(unsigned short fedid, FED_HDR *buf, unsigned int maxSize) {return -1;}

  int compid; // (corresponds to COMP_N_xxxx --- not COMP_ID_xxxx)

  //this is virtual for differences in the two intermediate classes
  virtual void setSLBs(void){;} // called on first call to getN_SLBs

  int nSLBs;
  int setSLBsCalled; 

  const static unsigned int MAX_SLBS = 500;          // per component
  const static unsigned int MAX_FEDS = MAX_SLBS * 6; // per component
  const static unsigned int MAX_SLB_SIZE = 1200;     // may need to be no more than an RPS message minus overhead
  const static unsigned int MAX_FED_SIZE = MAX_SLB_SIZE - sizeof(SLB_HDR) - sizeof(SLB_TAIL);

  static char fedBuf[MAX_FED_SIZE];                // a scratch area to receive FED records during build

  unsigned short *fedIDs;

  class SLB_struct {
  public:
    SLB_struct(unsigned short *f=0, int n=0) : first(f), nFEDs(n) {}
    unsigned short *first;    // points to first FED of SLB in fedIDs 
    int nFEDs;                // number of FEDs in this SLB
  };

  SLB_struct *SLB;

private:
  
  static void init_interfaces(void);

  unsigned short* getFEDs_for_SLB(int slbID, int *n_feds);

  static int nInstances[COMP_NUMBER]; // number of times a constructor is called (should be one or zero)
  static DAQ_MC_IF *ifs[COMP_NUMBER]; // pointers to an instance for each component

  int nFEDs;

  static int fd; // file descriptor of output file or pipe

  static unsigned int bufLen;  // length of evtBuf
  static char *evtBuf;         // the event is built here

  static int compMask;

};

#endif
