#include "gean/gcflag.h"
#include "gean/ptlset.hh"
#include "ptset/detector_menu.hh"
#include "arte/ArteMessage.hh"
#include <fstream.h>
//#include <cstdlib> // for getenv(..)

// Author V.Balagura (balagura@mail.desy.de)   (27.10.98).

// Modification made on 12.11.2001:
// hpt_setup object is used (see ptset/high_pt.hh,C) instead of
// Detector* HERA_B_detector. It is filled with hpt_setup.read(),
// detector_menu is not used any more.
// In addition clearing the hpt_setup structure is controlled now
// via KUIP command.

/* The description of hpt_setup structure can be found in ptset/high_pt.hh,C.
   If you use only ARTE and do not need C++ hpt_setup structure
   you may clear() it to save some memory. This is done by
   default. If you'd like to change this behaviour your kumac file should
   contain the line
   HBGEAN/PTL/CLEAR_HPT_SETUP_STRUCTURE NO
*/

static bool hpt_setup_will_be_cleared = true;
// for KUIP interface
extern "C" {
  char* ku_getc();
  void clear_hpt_setup_or_not_() {
    char* type = ku_getc();
    if ( string(type) == "NO" ) hpt_setup_will_be_cleared = false;
    else                        hpt_setup_will_be_cleared = true;
  }
}

extern "C" {
  void ptlset_() {
    create_hpt_setup();
    if ( global_error() ) {
      arteMsg  << "can not create high-pt setup";   arteMsg.error();
    }
    fill_ARTE_hpt_geometry_table();
    if ( hpt_setup_will_be_cleared == true ) hpt_setup.clear();
    if(gcflag_.idebug) {
      cout << " === C++ Setup structure is converted to PTGE bank and ";
      if ( hpt_setup.empty() ) cout << "CLEARED\n";
      else                     cout << "RETAINED\n";
    }
  }
}

#include "gean/ZGESL.h"                 // to define LGESL_PTC;
#include "gean/zgvers.hh"               // to find the (HPT) directory name
#include <stdio.h>                      // for struct FILE and fileno()
FILE* utfilp(const char*, const char*); // where the file with the setup
                                        // geometry description is placed
// utfilp is defined in gean/utfilp.C; there is no corresponding header

FILE* get_setup_geometry_description_FILE
(const string& file_name_without_dir) {
  zgvers ZGVERS;
  string dir_name(ZGVERS.Get_vers_datafile(LGESL_PTC));
  return utfilp(file_name_without_dir.c_str(),dir_name.c_str()); 
}
  
void create_hpt_setup() {
  // Get geometry file:
  string geom_file_name = "hpt_geo.dat";
  ifstream is(fileno(get_setup_geometry_description_FILE(geom_file_name)));
  // here fileno(FILE*) returns file descriptor,
  // ifstream(int file_descriptor) associates istream with it
  if ( !is )
    { global_error()("can not open file " + geom_file_name); return; }
  hpt_setup.read(is);
  if ( global_error() ) return;
  // print out the number of HPT Setups, chambers and cells
  if(gcflag_.idebug) {
    cout << " === High-Pt setup C++ structure created:\n";
    unsigned int nchambers[2] = {0,0}, ncells[2] = {0,0};
    for (unsigned int il=0; il<hpt_setup.ichambers().size(); ++il) {
      if ( hpt_setup.ichambers()[il].size() > 2 ) {
	cerr << "There are more than 2 parts (inner/outer) in layer " << il
	     << " of hpt_setup\n";
	continue;
      }
      for (unsigned int ip=0; ip<hpt_setup.ichambers()[il].size(); ++ip) {
	nchambers[ip] += hpt_setup.ichambers()[il][ip].size();
	for (unsigned int ich=0; ich<hpt_setup.ichambers()[il][ip].size();
	     ++ich) {
	  HPT_Chamber& ch = (*hpt_setup.ichambers()[il][ip][ich]).second;
	  for (HPT_Setup::ISection isect=ch.begin(); isect!=ch.end(); 
	       ++isect) ncells[ip] += (*isect).size();
	}
      }
    }
    cout <<"> Inner: " <<nchambers[0] <<" chambers, " <<ncells[0] <<" cells;\n"
	 <<"> Outer: " <<nchambers[1] <<" chambers, " <<ncells[1] <<" cells.\n";
  }
}
    
#include <arte/arte.hh>
#include "ptset/setup_algorithms.hh"
#include "ptset/setup_tools.hh"
#include <string.h>
#include <strings.h>

void fill_ARTE_hpt_geometry_table() {
  double chamber_width[2] = {3.5, 1.0};
  string part_name[2], section_name[2], material_name[2];
  part_name    [0] = "inner";        part_name    [1] = "outer";
  section_name [0] = "PTC";          section_name [1] = "PTO";
  material_name[0] = "PT:InnGas+Fe"; material_name[1] = "PT:OutGasCMW";

  for (unsigned int ilayer=0; ilayer<hpt_setup.ichambers().size(); ++ilayer) {
    unsigned int n_parts = hpt_setup.ichambers()[ilayer].size();
    if ( n_parts > 2 ) {
      if(gcflag_.idebug) {
	cerr << "There are " << n_parts << " parts in layer #" << ilayer
	     << " of hpt_setup\nIt should be not more than 2 (inner + outer).\n"
	  "Only first two will be processed\n";
	n_parts = 2;
      }
    }      
    for (unsigned int ipart=0; ipart<n_parts; ++ipart) {
      unsigned int i_rect = 1;
      for (unsigned int inside_layer=0;
	   inside_layer<hpt_setup.ichambers()[ilayer][ipart].size();
	   ++inside_layer) {
	HPT_Chamber& ch =
	  (*hpt_setup.ichambers()[ilayer][ipart][inside_layer]).second;
	vector<ChamberRectangularPartDescription> rectangles;
	make_chamber_from_rectangles(ch, &rectangles);
	
	string superlayer_name = "PTL";
	superlayer_name << ilayer+1;
	
	for (unsigned int i=0; i<rectangles.size(); ++i, ++i_rect) {
	  ChamberRectangularPartDescription& rp = rectangles[i];
	  Rectangle r(rp.ind1 * rp.grid, rp.ind2 * rp.grid);
	  
	  PTGE x;
	  
	  x.flag = 1;
	  
	  HPT_Setup::ISection isect = ch.isection(r.center());
	  unsigned int section_number = isect - ch.begin();
	  
	  string name = section_name[ipart];
	  if (section_number<10) name << section_number;
	  else if (section_number-10<=int('Z')-int('A'))
	    name << char(int('A')+section_number-10);
	  else name << '*';
	  
	  memcpy(x.name, name.c_str(), 4);
	  memcpy(x.supl, superlayer_name.c_str(), 4);
	  memcpy(x.moth, x.supl, 4);
	  memcpy(x.shap, "BOX ", 4);
	  memcpy(x.mate, material_name[ipart].c_str(), 12);
	  memcpy(x.maon, "ONLY", 4);
	  x.irot = 0;
	  bzero(x.rotm, sizeof(float)*9);
	  x.colo = section_number+2;
	  x.fill = 1;
	  x.seen = 1;
	  Grid& grid = (*isect).grid;
	  string comment = "HPT_";
	  comment << part_name[ipart] << "_#" << i_rect << ".Cell_"
		  << grid.x << "x" << grid.y;
	  memcpy(x.comm, comment.c_str(), comment.size());
	  memset(x.comm+comment.size(), ' ', 80-comment.size()); // fill the 
	  // rest with periods
	  
	  x.zmin = ch.z - chamber_width[ipart] / 2.;
	  x.zmax = ch.z + chamber_width[ipart] / 2.;
	  
	  x.dx = grid.x;
	  x.dy = grid.y;
	  
	  x.xmin = r.a.x;
	  x.xmax = r.b.x;
	  x.ymin = r.a.y;
	  x.ymax = r.b.y;
	  
	  Index ind = rp.ind2 - rp.ind1;
	  x.nx = ind.ix;
	  x.ny = ind.iy;
	  ArtePointer<PTGE>::append(x);
	}
      }
    }
  }
}
