#include <iostream>
#include <algorithm>

#ifndef GROVER_WITHOUT_ARTE
#include "arte/arte.hh"
#include "arte/RSEG_cmp.hh"
#endif

#include "grover/grover.hh"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 01-03-13  Max      Added preprocessor directives and updated to use 
//                    GroverParticle::operator==
//

int GroverJP(GroverVertex*       jPsi,
	     GroverParticleList* allParticles,
	     GroverVertexList*   allPrimaries = 0,
	     const GroverCuts*   usercuts = 0)
{
  ////////////////////////////////////////////////////////////////////
  //
  //                       GroverJP
  //
  // Finds secondary vertices which consist of a J/psi candidate
  // plus zero or more charged particles
  //
  // input: jPsi points to the J/psi vertex. allParticles points to
  //        a list of GroverParticles; by default all these particles
  //        are tried unless they are flagged as already used in a 
  //        secondary or primary.
  //
  // output: jPsi may have zero or more tracks added when the function
  //         returns. Returns SUCCESS if no errors.
  //
  ////////////////////////////////////////////////////////////////////

  const float CHI2MAXUNCO = 50.0;
  const float CHI2MAXCO   = 50.0;
  const float MAXC2INCREASE = 2.0;

  // Counter and error flag
  static int EventNumber = 0;
  static bool firstcall = true;
  EventNumber++;
  firstcall = (EventNumber==1);

  // Non-default cuts?
  static GroverCuts localcuts;
  const GroverCuts* cuts = &localcuts;
  if (usercuts != 0) cuts = usercuts;
  bool verbose = (cuts->Verbose==1);



  // Initialisation
  if (firstcall || verbose)
    cout << "*** GroverJP (Grover v."
	 << GroverVersion 
	 << "): first call" << endl;



  // some basic checks
  if (allParticles == 0) return FAILURE;
  if (allParticles->size() == 0) return FAILURE;
  if (jPsi->Size() != 2) return FAILURE;
  jPsi->Fit(-1.0);
  if (jPsi->Chi2() > CHI2MAXUNCO) return FAILURE;
  jPsi->Fit(3.097);
  if (jPsi->Chi2() > CHI2MAXCO)   return FAILURE;



  // Label the J/psi vertex if the user hasn't already.
  if (jPsi->Index() <= 0) jPsi->SetIndex(42);



  // Label particles with an index, flag the leptons
  // and move to local VDS coordinates
  int idx = 0;
  GroverParticle* lep1 = &(jPsi->particles[0]);
  GroverParticle* lep2 = &(jPsi->particles[1]);
  int jpindex = jPsi->Index();
  for (GroverParticleList::iterator ipar = allParticles->begin();
       ipar != allParticles->end(); ++ipar) {
    ipar->SetIndex(++idx);
    ipar->GlobalToLocal();
    if (*ipar == *lep1 || *ipar == *lep2) ipar->SetUsedInSecondary(jpindex); 
  }

  

  // Iteratively add best tracks 
  float thelimit = MAXC2INCREASE; // -> Maximum amount by which the last added 
                                  //    track can increase the chi2 before stopping.
  int method = 0; // -> Set to 1 for slower method.
  bool keep_looking = true;
  while (keep_looking) {
    float chi2b4 = jPsi->Chi2();
    int   sizeb4 = jPsi->Size();
    GroverAddBest(jPsi, allParticles, method);
    if (jPsi->Size() == sizeb4) keep_looking = false;
    else if ( (jPsi->Chi2()-chi2b4) > thelimit) {
      jPsi->particles.pop_back(); jPsi->ResetParticles();
      jPsi->Fit(-1.0);
      keep_looking = false;
    }
  }

  if (allPrimaries != 0) {
    if (allPrimaries->size() != 0){
      // here we might remove tracks from primary and refit. not yet.
    }
  }


  // done
  if (verbose) cout << "*** GroverJP: finished!" << endl;
  return SUCCESS;
  
}
