/*
  This file describes the "Setup" structure. 
  It can be used for the general description of various "experimental
  setups" which can be a group of chambers or a particular 
  subdetector.
  Chambers may have any number of "parts" (which are called "sections"),
  every part may have any geometrical form. 
  The "sections" are made of "cells" of the same size.
  It is required that cells are rectangular. 

  Where this structure can be used.
  ---------------------------------
  The structure is quite general. If the internal structure of the
  subdetector can be represented like
     setup -> chambers -> sections -> rectangular cells
  then it can be described using this Setup package. The information
  relevant for the particular subdetector is supplied as template
  arguments to the Setup classes. Using appropriate arguments one
  can describe any subdetector. 
  From the other hand the Setup package is able to describe chambers
  with any complex internal geometry. This is possible since every 
  chamber "section" can have any geometrical shape. It is determined
  using the general function object which takes a point and returns
  true/false depending whether the point is inside the "section". 
  
  Thus the advantage of the Setup package is its generality.
  Different subdetectors can be described in the same way.
  It is possible to write general algorithms which work with any
  subdetectors. 
  For example there is an algorithm which breaks the "sections"
  into rectangles. This can be done for any "section" geometry.
  These rectangles can be put then in ARTE geometry banks. 

  Currently the Setup package is in use inside the high-pt group.
  Other subdetectors are welcome to use it too.

  Why not use the Setup package.
  ------------------------------
  1) When the cells in your subdetector are not rectangular.

  2) The advantage of describing the chambers using Setup package can
  be seen if the chamber geometry is complex. However if the chamber
  is a pure rectangle then probably the usage of Setup package can
  be a bit inefficient.
  The reason is obvious: methods developed for a specific problem
  are usually more efficient than general methods. 
  For example, let's consider the problem to find the cell knowing
  the hit point in the chamber. If the chamber is rectangular, one
  just calculates the cell number, e.g. along X as 
    ix = int( (hit_point_x - left_chamber_corner_x) / cell_size_x );
  In case of Setup package it is assumed that the section can have any
  geometrical shape. If one adds new cell to the existing section 
  (or erase one cell inside the section) this produces the new section
  with different geometrical shape. But one should be able to describe
  any such shape. This means that the section should be considered 
  as a general container of cells. One may construct the section with
  any shape using the elementary operations of inserting/erasing 
  particular cells.
  Keeping all that in mind the section was made as a so called 
  associative array (map in C++ Standard Template Library terminology)
  of cells. Every element in this array is a pair (key, value). 
  Value is just Cell. The key is a pair of two integers: ix, iy which
  are responsible for cell numbering along x, y directions. 
  Knowing the key one can very fast find the corresponding value (Cell)
  since all elements in the container are already sorted.
  The time for this operation is proportional to log_2 (N) where N is
  the total number of cells in the section.
  Let's come back to our problem of finding the cell knowing the hit point
  in the section. Again as in the case of the rectangle one can 
  immediately calculate the "key", i.e. the pair (ix, iy) knowing the 
  hit point, the cell size and the cells offset from the point (x=0, y=0).
  Then one should perform the finding operation. It takes log_2(N)
  iterations. If the section contains not more than several hundreds
  of cells the number of iterations is not more than 10. (2^10=1024).

  3) The sections are now included in the chamber as in Russian 
  "matreshka". I.e. the next one is inserted into the previous. The 
  first section defines the boundary of the chamber, i.e. it cuts
  the whole x-y plane into points inside/outside the chamber.
  The second section cuts all points which have been remained after 
  the previous cut (by 1st section) into the points belonging 
  to the 1st section itself and all other points in the chamber. 
  The 3d section in turn cuts these "all other points" into points
  belonging to the 2d section and "the rest" etc..
  Graphically:
x-y plane -> 1+2+3+...sections -> 2+3+... sections -> 3+... sections ->...
                            |                   |                 |
                             -> 1 section        -> 2 section      ->...
  This can be convenient for the chambers with the such geometry that
  the next sections (usually closer to the beam) are "surrounded"
  by the previous sections. 
  If it is not the case the Setup package still will work. 
  But the geometry definition will not be so obvious. 
  E.g. suppose one needs to describe hypothetical chamber with 2 sensitive
  volumes separated by "dead" zone:
  -----    -----
  |   |    |   |
  | A |    | B |
  |   |    |   |
  -----    -----
  In this case the function object (or just function) associated with the
  1st section should define the whole chamber area, i.e. the region A+B.
  For example: bool f1(Point p) { return A(p) || B(p); }
  where A(p), B(p) are the functions defining A, B regions. 
  (Like bool A(p) { return (x>xmin && x<xmax && y>ymin && y<ymax); }).
  The 2nd section should separate B from A:
  bool f2(Point p) { return B(p); }
  or shorter: bool f2(Point p) { return x>xmin_B; }

  Probably more obvious way here is to define the sections independently
  by A(p) and B(p). I.e. every section has associated function object 
  which completely (and independently from other sections) defines the 
  section's shape. This requires some changes in the Setup structure
  but still can be implemented without any problems. 
  
  But in this case the geometry definition for the "matreshka"-like
  chambers (with inner sections surrounded by outer ones) becomes more
  complex. 
  The "matreshka"-like chambers were considered to be more important case
  and thus currently Setup structure is "optimized" for this type of
  chambers.
----------------------------------------------------------------------
  Now more deteiled view on Setup structure.

   Setup = map<string, Chamber> + SetupType
        Chamber = vector<Section> + ChamberType
	     Section = map<Index, Cell> + SectionType
	           Cell = SetupCell(bool is_read_out) + CellType.
   Here SetupType, ChamberType, SectionType, CellType are template
   parameters.
   Index has ix, iy fields which are responsible for the cell numbering
   along x,y directions. (See point.hh).

   The idea was to separate the information relevant for the particular 
   subdetector into "geometrical" and "non geometrical" parts.

   Template parameters which can be freely changed represent "non
   geometrical" part. E.g. for High-Pt
   CellType holds boolean value (fired/non fired),
            number of times this cell was fired and occupancy,
	    state: Ok, BAD, NOISY or UNKNOWN. 

   SectionType  holds average occupancy for the section,
                      vector of fired cells in the event for fast
		      traversing over fired cells.

   These types can be chosen according to the specific needs.
   From the other hand, the "geometrical" part is fixed in the "Setup"
   structure (setup->chamber->section->rectangular cell).
   But it is flexible enough to describe chambers with any geometry. 
   The principle idea is that each section has a function 
   (or function object) associated
   with it. It takes a point as an argument and returns true/false: 
   whether the point is inside the section. Thus one can define
   any figures.
   In addition Section has "grid" member (of type Grid, defined in
   point.hh). It defines the grid so that all cells in the Section
   has corners at the nodes of this grid.
   When constructed the Section holds all such cells with centers inside
   the figure associated with this Section.

   In more detail about "figures" associated with Sections:
     Section holds PtrWrapper<SetupFigureBase> object.
     Classes derived from SetupFigureBase can define any figure.
     (See ptr_wrap.hh and setup_figure.hh).

   Chamber is a container of Sections. Geometrically one Section is 
   positioned inside the previous one (like in russian "matreshka").
   This means each Section "figure" is responsible only how to cut
   the needed shape from the larger figure defined by the previous
   Section. The outermost Section should cut the shape of the
   whole chamber from the whole plane. All the other Sections
   should cut the figure left from the previous Section cutting.

   Section is an associative array which is made of the elements 
   pair<const Index, Cell>. Knowing the Index one can very fast find
   the corresponding Cell. 

   There are some tools for defining the Setup geometry "interactively".
   See file setup_menu.hh (and detector_menu.hh).

   Iterators through the structure:
   ICell    = map    iterator inside Section
   ISection = vector iterator inside Chamber
   IChamber = map    iterator inside Setup
   
   ISection_Cell    =            isection + icell inside Chamber
   IChamber_Cell    = ichamber + isection + icell inside Setup
   IChamber_Section = ichamber + isection         inside Setup

   The real class names are: SetupCell, SetupSection, SetupChamber, Setup.
   As shorthands I'll use Cell, Section, Chamber instead of
   SetupCell, SetupSection, SetupChamber. But one should remember that
   the "global" names should have "Setup" prefix in front. If you are
   already in, e.g. SetupChamber, there are some typedef's
   available, so that you may call the corresponding Section as
   SetupChamber::Section without extra "Setup" prefix.

   Inside Setup one may use the typedef's:
     ICell, ISection, IChamber, 
     ISection_Cell, IChamber_Cell, IChamber_Section, 
     Cell, Section, Chamber, 
     CellType, SectionType, ChamberType, SetupType;
   inside Chamber:
     ICell, ISection, 
     ISection_Cell, 
     Cell, Section, Chamber, 
     CellType, SectionType, ChamberType;
   inside Section:
     ICell,

     Cell, Section, 
     CellType, SectionType;
   inside Cell:


     Cell, 
     CellType;
......................................................................
   Cell always has 
      bool is_read_out
   field;
   Section:
      Grid grid
      PtrWrapper<SetupFigureBase> ptr_figure;
   Chamber:
      double z;

   Constructors:
      SetupCell()

      SetupSection()
      SetupSection(Grid grid_of_cell_corners)

      SetupChamber()
      SetupChamber(double z_position)

      Setup()
......................................................................

   Methods:
     SetupChamber::isection_cell(Point p) 
   - returns ISection_Cell of the cell where the Point p is.
   If p is outside the chamber
     isection_cell(p).isection == chamber.end().
   (The rule: if something can go wrong one should check
   the "highest" iterator in such compound iterators. The other "lower"
   iterator can conatin any value).

   ISection SetupChamber::isection(Point p)
   - the same but determines only the ISection (without ICell)
  ISection where(Point p)
  synonym for isection(p) /for compatibility with old programs/

  
    ICell Section::icell(const Index& relative_to_bottom_left_corner) 
  - returns ICell for the cell which has 
  Index == The lowest Index in Section + relative_to_bottom_left_corner;
  note that Index is sorted by iy and then by ix (i.e. iy is more
  important). Therefore the "reference cell" (with the lowest Index)
  is the leftmost cell in the bottom row.
  This way is chosen since the absolute Index can be not informative.
  As usual if there is no such cell, section.end() is returned.

    Index Section::index0(ICell i_cell) 
  - reverse operation (Index relative to bottom left corner is returned).

    ISection_Cell Chamber::isection_cell
    (unsigned int i_section,
     const Index& relative_to_bottom_left_corner)
  - the same but for chamber and with an additional argument - section
  number. (ISection_Cell::isection is equal to chamber.end()
  in case of error).
     ISection Chamber::isection(unsigned int i_section)
  - returns only ISection (or chamber.end()).

  Similarly:
    IChamber Setup::ichamber(const string& chamber_name)
    IChamber_Section Setup::ichamber_section(const string& chamber_name, 
				             unsigned int i_section)
    IChamber_Cell Setup::ichamber_cell
    (string chamber_name, 
     unsigned int i_section,
     const Index& relative_to_bottom_left_corner)
  - inside Setup (on error ichamber is set to setup.end()).

  When all Section's figures and grids are defined one can construct
  all the Sections, i.e. fill them with corresponding Cells. This is done
  by the call
    void Chamber::make_cells(const Rectangle& cover);
  The Sections will have all the cells with the centers inside
  the "cover" and the corresponding Section figure.

--------------------------------------------------
Modification made on 12.11.2001:
  When the chamber is defined it can be shifted with
  SetupChamber<CellT,SectionT,ChamberT>::move(Point shift); or
  move_from_nominal_position(Point shift); 

  The first function adds a shift=(dx,dy) to base_offset of the grid and
  calls move(shift) for SetupSection::ptr_figure for all
  chamber sections. (See the definition of base_offset and the Grid itself in
  the point.hh). The second one returns the chamber back to the nominal
  position and then makes a shift=(dx,dy). For this purpose the shift
  from the nominal position is always kept in the alignmentShift protected
  member of the SetupChamber<CellT,SectionT,ChamberT>. It can be accessed with
  Point alignment_shift(); public function.
  With the help of these functions the chamber position can be shifted without
  changing the Index'es or iterators of individual cells.
--------------------------------------------------
  
  The function for debugging:
      void Setup::dump_structure(ostream& os);
  - prints the whole structure.

----------------------------------------------------------------------
  In SetupType class one should define setup_type() function, e.g.:
     static const string& setup_type() { 
       static string x("<your_type>"); return x; 
     }
  The string "<your_type>" should be unique for your Setup type.
  See below for details.
  ............................................................

  In detector.hh Setups of different types are combined together
  in a "Detector". 
  Since Detector is a heterogeneous container (elements can have 
  different types) this causes some problems.
  Detector is implemented as
  Detector = map<string, PtrWrapper<__SetupBase>, less<string> >,
  i.e. Setups are included via the PtrWrapper of some base class
  __SetupBase. All Setups are derived from __SetupBase. (This is done
  internally, the programmer should not derive his SetupType from
  __SetupBase).
  
   To find out the actual type of the Setup one needs to make an 
   explicit cast from __SetupBase. The standard way to do it is to use
   RTTI (Run Time Type Information) implemented
   in C++. But many compilers do not support this feature. Therefore
   this casting is made "by hand". Of course standard way would be more 
   efficient.
   
   __SetupBase has a pure virtual function 
     virtual const string& __SetupBase::type() const = 0;
   which returns unique string for each of possible Setup types derived
   from __SetupBase. Then one can check this function for a particular
   __SetupBase and knowing its return value (at run time) call the
   algorithm specific for this Setup. 
   The programmer should not define this
     virtual const string& Setup::type() const;
   directly. Instead he needs to supply 
     SetupType::setup_type();
   function with the return type convertible to string in his SetupType 
   class. This function will be called from the corresponding 
       virtual const string& Setup::type() const;
   One may make this function static, e.g.:
     static const string& setup_type() { 
       static string x("<your_type>"); return x; 
     }
   Then this "<your_type>" string will be returned by the corresponding
     virtual const string& Setup::type() const;

   See detector.hh for more details.
   Also see high_pt.hh as an example how to define CellType, SectionType,
   ChamberType, SetupType and the corresponding SetupCell==HPT_Cell,
   SetupSection==HPT_Section, SetupChamber==HPT_Chamber, Setup==HPT_Setup.

-------------------------------------------------------------------------
   One can use for_each_...(action, ...) methods, e.g. 
      for_each_cell(do_something, your_setup);
   which are defined in setup_tools.hh. Here do_something is a function
   object which is able to "do something" with ICell. 
   The operation defined by do_something will be performed under all cells
   in your_setup. You may substitute setup by chamber or section.
   If you defined Detector detector; you may iterate through it using
   the same
      for_each_cell(do_something, detector);
   see detector_tools.hh.
   Cell also can be substituted by section, chamber or setup.
-------------------------------------------------------------------------

   If the cells in your detector have two distinct states: 
     fired/non fired,
   you may use fired_cells.hh and fired_cell_tools.hh. In this case 
   1) inside the CellType two things should be defined:
       a) CellType::DataType       - type for the cell's data;
       b) CellType::DataType data; - member containing the cell's data.
   2) you should provide bool is_fired() method in your CellType;

   Another tool: reset_cell_data_in(..., default_value); defined in
   setup_tools.hh also requires item 1).
   (See example in high_pt.hh)

   There is a working example in ptset/geometry_to_ntuple.example_readme
   file. This file can be used to construct the tool for displaying
   the Setup geometry.
*/

// Author V.Balagura (balagura@mail.desy.de)   (09.04.98).

#ifndef _setup_hh
#define _setup_hh 1

#include "ptset/point.hh"
#include <map>
#include <vector>
#include "ptset/ptr_wrap.hh"
#include "ptset/setup_figure.hh"
#include <string>
#include "ptset/delimit.hh"

template<class CellT>
class SetupCell : public CellT {
public:
  typedef SetupCell<CellT> Cell;
  typedef CellT            CellType;
  bool is_read_out;
  SetupCell() : is_read_out(false) { }
};

// shorthands to shrink lenghty template expansions
struct __LessIndex {
  bool operator()(Index ind1, Index ind2) const { return ind1 < ind2; }
};
struct __LessString {
  bool operator()(const string& s1, const string& s2) const
  { return s1 < s2; }
};

template <class CellT>
class __SetupSection : public map<Index, SetupCell<CellT>, __LessIndex> {
public:
  PtrWrapper<SetupFigureBase> ptr_figure;
  Grid grid;

  __SetupSection() {}
  __SetupSection(const Grid& grid_of_cell_corners)
    : grid(grid_of_cell_corners) {}

  typedef SetupCell<CellT>                  Cell;
  typedef CellT                             CellType;
  /* ICell is a usual map iterator */

#if (__GNUC__ > 2) || (__GNUC__==2 && __GNUC_MINOR__>90 && __GNUC_MINOR__<95)
  typedef iterator                             ICell;
  friend bool operator<(const ICell& icell1, const ICell& icell2) {
    return &(*icell1) < &(*icell2);
  }
#else
  /* It is made as a new class because of the fact that the compiler can not 
     handle long template expansions. map::iterator expansion is quite 
     long, so I shrink it to __SetupSection<CellT>::ICell */

  struct ICell : public iterator {
    typedef SetupCell<CellT> Cell;
    ICell() {}
    ICell(const iterator& iter) : iterator(iter) {}
    ICell& operator=(const iterator& iter) { 
      *((iterator*)this) = iter;
      return *this;
    }

    friend bool operator==(const ICell& ic1, const ICell& ic2) {
      return *((iterator*)(&ic1)) == *((iterator*)(&ic2));
    }
    friend bool operator==(const iterator& iter1, const ICell& ic2) {
      return iter1 == *((iterator*)(&ic2));
    }
    friend bool operator==(const ICell& ic1, const iterator& iter2) {
      return *((iterator*)(&ic1)) == iter2;
    }
    friend bool operator<(const ICell& ic1, const ICell& ic2) {
      return &(*ic1) < &(*ic2);
    }
    ICell& operator++()   { ++( *((iterator*)this) ); return *this; }
    ICell operator++(int) 
    { ICell tmp = *this; ++( *((iterator*)this) ); return tmp; }
    ICell& operator--()   { --( *((iterator*)this) ); return *this; }
    ICell operator--(int) 
    { ICell tmp = *this; --( *((iterator*)this) ); return tmp; }
  };
#endif

  ICell icell(const Index& relative_to_bottom_left_corner) {
    return find( (*begin()).first + relative_to_bottom_left_corner );
  }
  Index index0(ICell i_cell) { return (*i_cell).first - (*begin()).first; }
};


template<class CellT, class SectionT>
class SetupSection : public SectionT,
		     public __SetupSection<CellT> {
public:
  typedef SectionT                         SectionType;
  typedef SetupSection<CellT, SectionT> Section;

  SetupSection() {}
  SetupSection(const Grid& grid_of_cell_corners)
    : __SetupSection<CellT>(grid_of_cell_corners) {}
};

template<class CellT, class SectionT>
class __SetupChamber : public vector<SetupSection<CellT, SectionT> > {
  typedef __SetupSection<CellT>          __Section;
 public:
  double z;
 protected:
  Point alignmentShift;
 public:
  Point alignment_shift() { return alignmentShift; }
  // this value is added to the nominal chamber position
  
  __SetupChamber() : alignmentShift(Point(0,0)) {}
  __SetupChamber(double z_position)
    : z(z_position), alignmentShift(Point(0,0))  {}

  typedef CellT                            CellType;
  typedef SectionT                         SectionType;
  typedef SetupSection<CellT, SectionT> Section;
  typedef Section::Cell                       Cell;
  typedef __Section::ICell                    ICell;
  typedef iterator                            ISection;

  void move(Point shift) {
    for (ISection  is=begin(); is!=end(); ++is) {
      (*is).grid.base_offset += shift;
      (*is).ptr_figure()->move(shift);
    }
    alignmentShift += shift;
  }
  void move_from_nominal_position(Point shift) { // back to nominal + new shift
    Point shift_from_nominal = shift - alignmentShift;
    for (ISection  is=begin(); is!=end(); ++is) {
      (*is).grid.base_offset += shift_from_nominal; 
      (*is).ptr_figure()->move(shift_from_nominal);
    }
    alignmentShift = shift;
  }
    
  ISection isection(Point p) {
    ISection is;
    for (is=begin(); is!=end(); ++is) 
      if ( !(*is).ptr_figure()->contains(p) ) break;
    if ( is == begin() ) return end();
    else                 return --is;
  }
  // For compatibility with old programs
  ISection where(Point p) { return isection(p); }

  struct ISection_Cell {
    typedef __SetupChamber ContainerType; // one can go like this: 
    // ISection_Cell::ContainerType::ISection
    ICell icell;
    ISection isection;
    ISection_Cell() { }
    ISection_Cell(ISection i_section, ICell i_cell) :
      icell(i_cell), isection(i_section)  { }
    bool operator==(const ISection_Cell& isc) const {
      return isection == isc.isection && icell == isc.icell;
    }
    bool operator<(const ISection_Cell& isc) const {
      return isection < isc.isection || 
	(isection == isc.isection && icell < isc.icell);
    }
  };
  ISection isection(unsigned int i_section) {
    if ( i_section >= size() ) return end();
    else return begin() + i_section;
  }
  ISection_Cell isection_cell(unsigned int i_section,
			      const Index& relative_to_bottom_left_corner) {
    if ( i_section >= size() ) {
      ISection_Cell isc;
      isc.isection = end();
      return isc;
    }
    ISection is = begin() + i_section;
    ICell ic = (*is).icell(relative_to_bottom_left_corner);
    if ( ic == (*is).end() ) {
      ISection_Cell isc;
      isc.isection = end();
      return isc;
    }
    return ISection_Cell(is, ic);
  }
  ISection_Cell isection_cell(Point p) {
    ISection_Cell isc;
    if ( (isc.isection = where(p)) == end() ) return isc;
    Section& sect = *isc.isection;
    isc.icell = sect.find(Index(p, sect.grid));
    return isc;
  }      
};

template<class CellT, class SectionT, class ChamberT>
class SetupChamber : public ChamberT,
		     public __SetupChamber<CellT, SectionT> {
public:
  typedef ChamberT                         ChamberType;
  typedef SetupChamber<CellT, SectionT, ChamberT> Chamber;

  SetupChamber() {}
  SetupChamber(double z_position)
    : __SetupChamber<CellT, SectionT>(z_position) {}

  void make_cells(const Rectangle& cover); /* cell centers inside cover */
};

template<class CellT, class SectionT, class ChamberT>
class __Setup : public map<string,
                           SetupChamber<CellT,SectionT,ChamberT>,
	                   __LessString> {
  typedef __SetupChamber<CellT, SectionT>          __Chamber;
  typedef __SetupSection<CellT>                       __Section;
public:
  typedef CellT                            CellType;
  typedef SectionT                         SectionType;
  typedef ChamberT                         ChamberType;

  typedef SetupChamber<CellT, SectionT, ChamberT> Chamber;
  typedef Chamber::Section                                 Section;
  typedef Chamber::Cell                                    Cell;
  typedef __Section::ICell                                 ICell;
  typedef __Chamber::ISection                              ISection;
  typedef __Chamber::ISection_Cell                         ISection_Cell;

#if (__GNUC__ > 2) || (__GNUC__==2 && __GNUC_MINOR__>90 && __GNUC_MINOR__<95)
  typedef iterator                                          IChamber;
  friend bool operator<(const IChamber& ich1, const IChamber& ich2) {
    return &(*ich1) < &(*ich2);
  }

#else

  struct IChamber : public iterator {
    IChamber() {}
    IChamber(const iterator& iter) : iterator(iter) {}
    IChamber& operator=(const iterator& iter) { 
      *((iterator*)this) = iter;
      return *this;
    }

    friend bool operator==(const IChamber& ic1, const IChamber& ic2) {
      return *((iterator*)(&ic1)) == *((iterator*)(&ic2));
    }
    friend bool operator==(const iterator& iter1, const IChamber& ic2) {
      return iter1 == *((iterator*)(&ic2));
    }
    friend bool operator==(const IChamber& ic1, const iterator& iter2) {
      return *((iterator*)(&ic1)) == iter2;
    }
    friend bool operator<(const IChamber& ic1, const IChamber& ic2) {
      return &(*ic1) < &(*ic2);
    }
    IChamber& operator++()   { ++( *((iterator*)this) ); return *this; }
    IChamber operator++(int) 
    { IChamber tmp = *this; ++( *((iterator*)this) ); return tmp; }
    IChamber& operator--()   { --( *((iterator*)this) ); return *this; }
    IChamber operator--(int) 
    { IChamber tmp = *this; --( *((iterator*)this) ); return tmp; }
  };
#endif

  struct IChamber_Section {
    typedef __Setup ContainerType; // one can go like this: 
    // IChamber_Section::ContainerType::IChamber_Cell

    ISection isection;
    IChamber ichamber;
    IChamber_Section() { }
    IChamber_Section(IChamber i_chamber, ISection i_section) :
      isection(i_section), ichamber(i_chamber) { }
    bool operator==(const IChamber_Section& ics) const {
      return ichamber == ics.ichamber && isection == ics.isection;
    }
    bool operator<(const IChamber_Section& ics) const {
      return ichamber < ics.ichamber || 
	(ichamber == ics.ichamber && isection < ics.isection);
    }
  };
  struct IChamber_Cell {
    typedef __Setup ContainerType; // one can go like this: 
    // IChamber_Section::ContainerType::IChamber_Section

    ICell    icell;
    ISection isection;
    IChamber ichamber;
    IChamber_Cell() { }
    IChamber_Cell(IChamber i_chamber, ISection i_section, ICell i_cell) :
      icell(i_cell), isection(i_section), ichamber(i_chamber) { }
    IChamber_Cell(IChamber i_chamber, ISection_Cell i_section_cell) :
      icell(i_section_cell.icell), isection(i_section_cell.isection), 
      ichamber(i_chamber) { }
    bool operator==(const IChamber_Cell& icc) const {
      return ichamber == icc.ichamber && isection == icc.isection
	&&	icell == icc.icell;
    }
    bool operator<(const IChamber_Cell& icc) const {
      return ichamber < icc.ichamber || 
	(ichamber == icc.ichamber && 
	 (isection < icc.isection ||
	  (isection == icc.isection && icell < icc.icell)));
    }
  };
  IChamber ichamber(const string& chamber_name) {
    return find(chamber_name);
  }
  IChamber_Section ichamber_section(const string& chamber_name, 
				    unsigned int i_section) {
    IChamber_Section ics;
    IChamber ich = find(chamber_name);
    if ( ich == end() ) {
      ics.ichamber = end();
      return ics;
    }
    Chamber& ch = (*ich).second;
    ISection is = ch.isection(i_section);
    if ( is == ch.end() ) {
      ics.ichamber = end();
      return ics;
    }
    ics.ichamber = ich;
    ics.isection = is;
    return ics;
  }
  IChamber_Cell ichamber_cell(string chamber_name, 
			      unsigned int i_section,
			      const Index& relative_to_bottom_left_corner) {
    IChamber_Cell icc;
    IChamber ich = find(chamber_name);
    if ( ich == end() ) {
      icc.ichamber = end();
      return icc;
    }
    Chamber& ch = (*ich).second;
    ISection is = ch.isection(i_section);
    if ( is == ch.end() ) {
      icc.ichamber = end();
      return icc;
    }
    ICell ic = (*is).icell(relative_to_bottom_left_corner);
    if ( ic == (*is).end() ) {
      icc.ichamber = end();
      return icc;
    }
    icc.ichamber = ich;
    icc.isection = is;
    icc.icell    = ic;
    return icc;
  }
};
struct __SetupBase {
  virtual const string& type() const = 0;
  virtual __SetupBase* clone() const = 0;
  virtual ~__SetupBase() {}
};

/* SetupT should have member function SetupT::setup_type() with
   the return type convertible to string;
   ( e.g. 
     static const string& setup_type() 
      { static string x("..."); return x; }
   )
   different SetupT types should return different setup_type()'s */

template<class CellT, class SectionT, class ChamberT, class SetupT>
class Setup : public SetupT,
              public __Setup<CellT, SectionT, ChamberT>,
              public __SetupBase {
public:
  typedef SetupT                           SetupType;

  virtual const string& type() const { return SetupT::setup_type(); }
  virtual __SetupBase* clone() const { return new Setup(*this); }
  void dump_structure(ostream& os);
};

template<class Chamber>
class __Make_Cell__ {
  typedef typename Chamber::ISection ISection;
  typedef typename Chamber::Section Section;
  Chamber* pchamber;
  ISection isection;
 public:
  __Make_Cell__(Chamber* ptr_chamber, ISection i_section) :
    pchamber(ptr_chamber), isection(i_section) { }
  void operator()(const Point& p) const {
    if (pchamber->where(p) == isection) {
      /* create cell if cell center belongs to this section */
      Section& section = *isection;
      section[Index(p, section.grid)];
    }
  }
};
template<class CellT, class SectionT, class ChamberT>
inline void SetupChamber<CellT,SectionT,ChamberT>::make_cells
(const Rectangle& cover) {
  ISection is;
  for (is=begin(); is!=end(); ++is) {
    for_each_node
      (cover, (*is).grid.shift_by_half(), 
       __Make_Cell__<SetupChamber<CellT,SectionT,ChamberT> >
       (this, is));
  }
}
template<class CellT, class SectionT, 
         class ChamberT, class SetupT>
inline void
Setup<CellT, SectionT, ChamberT, SetupT>::dump_structure
(ostream& os) {
  os << "setup type: " << type() << '\n';
  IChamber ichamber;
  ISection is;
  int n;
  ICell ic;
  for (ichamber=begin(); ichamber!=end(); ++ichamber) {
    Chamber& chamber = (*ichamber).second;
    os << "chamber " << (*ichamber).first <<" (at z = " << chamber.z << "):\n";
    for (is=chamber.begin(), n=0; is!=chamber.end(); ++is, ++n) {
      Section& section = *is;
      os << "section #" << n << ":  " << section.grid << '\n';
      Delimiter delimiter(" ", 4, "\n");
      for (ic=section.begin(); ic!=section.end(); ++ic) {
	os << grid_center((*ic).first, section.grid);
	if ( (*ic).second.is_read_out ) os << "*";
	else                            os << " ";
	os << delimiter;
      }
      delimiter.flush(os);
      os << "------------------------------------------------------------\n";
    }
    os << "============================================================\n";
  }
}

#endif
