// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
// Modifications:
// - In ::insertWalls: loop `active' volumes only
//                new  rangerTrack::fitGlobal() 
//                     trick for path in magnet
//                                      14-Aug-00 Wouter Hulsbergen 
//  check parameter fitmagn  in rangerTrack(rtra..        27/11/00 A. Spiridonov
//  upper limit for magmet < 660.                         04/12/00 A. Spiridonov
//  upper limit for magnet > 660. in rangerTrack::genCmp()21/01/01 A. Spiridonov 
// - added filling of rtra->ndof, 
//   removed chiSquare data member                        11/04/01 Wouter
//
// - insertWalls: use stl set to keep track of which wall 16/05/01 Wouter
//   have already been used
// - discard muon hits in rangerTrack(ArtePointer<RTRA>.. 20/05/01 A. Spiridonov
// - insertWalls: better procedure to find walls in magnet 04/07/01 Wouter
//   insertWallsV: removed/replaced by insertWalls
//   extrapolateDownstream/extrapolateUpstream: new routines
//     used by rangerHitPattern
//   fit: extra smoother/filter step before inserting walls
// - new routine rejectOutliers                           17/07/01 Wouter
//   removed <set> from insertWalls since it is too slow 
//   changed default flag fit(firstpass) to true for wall insertion
// - changed chi2 definition
//   debug printing                                      01/03/02 A.Spiridonov

#include <iostream>
#include <algorithm>
#include <arte/arte.hh>
#include <arte/RSEG_cmp.hh>
#include <arte/artefunc.h>
#include <arte/ArteMessage.hh>
#include "rang/Hit.hh"
#include "rang/Geom.hh"
#include "rang/Track.hh"
#include "rang/hfill.hh"
#include "rang/zfpar.hh"
#include "rang/RMap.hh"
#include "rang/HitMap.hh"
#include "rang/FLAGHist.hh"
#include "rang/HitPattern.hh"

#undef FULL
#undef SAVEWALLS
//#define FULL_FILTER

//-- Default constructor
rangerTrack::rangerTrack() {
  qpEstimate = 0;
  electron = 0;
  nIteration = 0;
  smoothed = 0;
  tFitMode = ModeP;
  tPartType = MUON;
  status = 0;
};

//-- constructor from RSEG
rangerTrack::rangerTrack(ArtePointer<RSEG> rseg, rangerFitMode fm,
			 rangerParticleType pType,
			 rangerRMap* rmap) {
  qpEstimate = rseg->pf;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z
  
  for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	rhitIt != rseg->rhit.end(); ++rhitIt ) {
    ArtePointer<RHIT> rhit(*rhitIt);
#ifdef FULL
    cout << "Current rhit=" << rhit << 
      " hitb=" << rhit->hitb << " z=" << rhit->hitb->z << endl;
#endif

    rangerNode* w = rangerNode::makeNode( rhit, fitMode() );
    if (w) {
      prepend(*w);
      if (rmap) {
	rangerHitMap hm( w,rhit );
	rmap->addHitMap( hm );
      }
    }
  }
  orderIfNeeded();
};

//-- constructor from RSEG vector
rangerTrack::rangerTrack(const vector<ArtePointer<RSEG> >& rsegVec, 
			 rangerFitMode fm,
			 rangerParticleType pType,
			 vector<rangerRMap>* rmapVec) {
  qpEstimate = 0;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z

  for (vector<ArtePointer<RSEG> >::const_iterator sIt=rsegVec.begin();
       sIt != rsegVec.end(); sIt++) {
    ArtePointer<RSEG> rseg = *sIt;
    if (rseg->pf != 0) qpEstimate = rseg->pf;
    rangerRMap rMap(rseg);
  
    vector<ArtePointer<RHIT> > rhits;
    rhits.reserve(256);
    for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	  rhitIt != rseg->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(rhitIt);
      rhits.push_back(rhit);
    }

    int correctOrder = 1;
    if (rhits.size() > 1) {  // determine correct order
      if ((*rhits.begin())->hitb->z > (*rhits.rbegin())->hitb->z) {
	correctOrder = 0;
      }
    }

    if (correctOrder) {
      for (vector<ArtePointer<RHIT> >::iterator rIt=rhits.begin();
	   rIt != rhits.end(); ++rIt) {
#ifdef FULL
	cout << "Track(Vec...): current rhit=" << *rIt << 
	  " hitb=" << (*rIt)->hitb << " z=" << (*rIt)->hitb->z << endl;
#endif

	rangerNode* w = rangerNode::makeNode( *rIt, fitMode() );
	if (w) {
	  prepend(*w);
	  rangerHitMap hm( w,*rIt );
	  rMap.addHitMap( hm );
	}
      }
      if (rmapVec) rmapVec->push_back(rMap);
    }
    else {
      for (vector<ArtePointer<RHIT> >::reverse_iterator rIt=rhits.rbegin();
	   rIt != rhits.rend(); ++rIt) {
#ifdef FULL
	cout << "Track(Vec...): current rhit=" << *rIt << 
	  " hitb=" << (*rIt)->hitb << " z=" << (*rIt)->hitb->z << endl;
#endif

	rangerNode* w = rangerNode::makeNode( *rIt, fitMode() );
	if (w) {
	  prepend(*w);
	  rangerHitMap hm( w,*rIt );
	  rMap.addHitMap( hm );
	}
      }
      if (rmapVec) rmapVec->push_back(rMap);
    }
  }
  orderIfNeeded();
};


//-- constructor from RSEG vector + RCCL
rangerTrack::rangerTrack(const vector<ArtePointer<RSEG> >& rsegVec, 
			 ArtePointer<RCCL> rccl,int charge,
			 rangerFitMode fm,
			 rangerParticleType pType,
			 vector<rangerRMap>* rmapVec) {
  qpEstimate = 0;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z

  for (vector<ArtePointer<RSEG> >::const_iterator sIt=rsegVec.begin();
       sIt != rsegVec.end(); sIt++) {
    ArtePointer<RSEG> rseg = *sIt;
    if (rseg->pf != 0) qpEstimate = rseg->pf;
    rangerRMap rMap(rseg);
  
    vector<ArtePointer<RHIT> > rhits;
    rhits.reserve(256);
    for ( ArteRelationIterator<RHIT> rhitIt = rseg->rhit.begin();
	  rhitIt != rseg->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(rhitIt);
      rhits.push_back(rhit);
    }

    int correctOrder = 1;
    if (rhits.size() > 1) {  // determine correct order
      if ((*rhits.begin())->hitb->z > (*rhits.rbegin())->hitb->z) {
	correctOrder = 0;
      }
    }

    if (correctOrder) {
      for (vector<ArtePointer<RHIT> >::iterator rIt=rhits.begin();
	   rIt != rhits.end(); ++rIt) {
#ifdef FULL
	cout << "Track(Vec...): current rhit=" << *rIt << 
	  " hitb=" << (*rIt)->hitb << " z=" << (*rIt)->hitb->z << endl;
#endif

	rangerNode* w = rangerNode::makeNode( *rIt, fitMode() );
	if (w) {
	  prepend(*w);
	  rangerHitMap hm( w,*rIt );
	  rMap.addHitMap( hm );
	}
      }
      if (rmapVec) rmapVec->push_back(rMap);
    }
    else {
      for (vector<ArtePointer<RHIT> >::reverse_iterator rIt=rhits.rbegin();
	   rIt != rhits.rend(); ++rIt) {
#ifdef FULL
	cout << "Track(Vec...): current rhit=" << *rIt << 
	  " hitb=" << (*rIt)->hitb << " z=" << (*rIt)->hitb->z << endl;
#endif

	rangerNode* w = rangerNode::makeNode( *rIt, fitMode() );
	if (w) {
	  prepend(*w);
	  rangerHitMap hm( w,*rIt );
	  rMap.addHitMap( hm );
	}
      }
      if (rmapVec) rmapVec->push_back(rMap);
    }
  }
  rangerNode* w = rangerNode::makeNode( rccl,charge,fitMode());
  if (w) prepend(*w);
  orderIfNeeded();
};


//-- constructor from RTRA
rangerTrack::rangerTrack(ArtePointer<RTRA> rtra, rangerFitMode fm,
			 rangerParticleType pType,
			 rangerRMap* rmap) {
  qpEstimate = rtra->pf;
  electron = 0;        // set according to electron hypothesis
  nIteration = 0;
  smoothed = 0;
  tFitMode = fm;
  tPartType = pType;
  status = 0;

  // the RHITs are assumed to be retrieved in the order of increasing z.
  // the nodes are stored such that the first node in the list 
  // points to highest z
  
  for ( ArteRelationIterator<HITB> hitbIt = rtra->hitb.begin();
	hitbIt != rtra->hitb.end(); ++hitbIt ) {
    ArtePointer<HITB> hitb(*hitbIt);

    // check hit in magnet
    if( !zfpar_.fitmagn )
      {
	if(hitb->z > 206. && hitb->z < 660.)
	  {
	    //cout << " Track.C rtra " << rtra << " hit " << hitb.index()
	    //	 << " z " << hitb->z << " discarded " << endl;

	    continue;
	  }
      }

    // discard muon hits, gede->cmp == 8  
    if(hitb->gede->cmp == 8 )
      {
	//cout << " Track.C hitb " << hitb << " cmp " << hitb->gede->cmp<< endl;
	continue;
      }
    for ( ArteRelationIterator<RHIT> rhitIt = hitb->rhit.begin();
	rhitIt != hitb->rhit.end(); ++rhitIt ) {
      ArtePointer<RHIT> rhit(*rhitIt);
      if (rhit->rtra == rtra) {
	rangerNode* w = rangerNode::makeNode( rhit, fitMode() );
	if (w) {
	  prepend(*w);
	  if (rmap) {
	    rangerHitMap hm( w,rhit );
	    rmap->addHitMap( hm );
	  }
	}
      }
    }
  }
  orderIfNeeded();
}


// Destructor
rangerTrack::~rangerTrack() {
  for (rangerNodeIterator nodeIt = begin();
       nodeIt != end();
       nodeIt++) {
    delete *nodeIt;
  }
}

ostream& operator<<(ostream &os,const rangerTrack &t) {
  os << " #nodes=" << (t.nodes).size() << " chiSquare=" << t.totalChiSq()
     << " qpEst=" << t.qpEstimate
     << " FMode=" << int(t.tFitMode)
     << " ParType=" << int(t.tPartType) << endl;
  for (list<rangerNode*>::const_iterator n = (t.nodes).begin(); 
       n != (t.nodes).end(); n++) {
    os << *n << '\t' << *(*n) << endl;
  }
  return os;
}
  
void rangerTrack::covInit() {
  (*begin())->covInit();
}

int rangerTrack::fit(const int firstPass,const int thirdPass,
		     const int insertMaterial,const int updateInternal) {  
  // PASS 1
  if (firstPass) 
    {
      status = filter();
      if (status < 0) return -10;
      status = smooth();
      if (status < 0) return -11;
      if (updateInternal) updInternal();
    }
  
  int nInsert = 0;
  if (insertMaterial) nInsert=insertWalls();
  
  // PASS 2
  if (nInsert || (!firstPass)) 
    {
      status = filter();
      if (status < 0) 
	{
	  static int id4;
	  arteMsg << "rangerTrack::fit: filter (pass 2) failed ! " << endl;
	  arteMsg.allp(id4,10); 
	  return -10;
	}
      status = smooth();
      if (status < 0) 
	{
	  static int id5; 
	  arteMsg << "rangerTrack::fit: smoother (pass 2) failed !" << endl; 
	  arteMsg.allp(id5,10); 
	  return -11;
	}
      // update left-right signs
      if (updateInternal) updInternal();
      // reject outliers 
      if ( zfpar_.outlmode==1 && rejectOutliers()>0) 
	{
	  if( filter()<0 ) return -10 ;
	  if( smooth()<0 ) return -11 ;
	}
    }
  // PASS 3
  if (thirdPass) 
    {
      status = filter();
      if (status < 0) 
	{
	  static int id6;
	  arteMsg << "rangerTrack::fit: filter (pass 3) failed ! " << endl;
	  arteMsg.allp(id6,10); 
	  return -10;
	}
      status = smooth();
      if (status < 0) 
	{
	  static int id7; 
	  arteMsg << "rangerTrack::fit: smoother (pass 3) failed !" << endl; 
	  arteMsg.allp(id7,10); 
	  return -11;
	}
      // update left-right signs
      if (updateInternal) updInternal();
      // reject outliers 
      if ( zfpar_.outlmode==1 && rejectOutliers()>0) 
	{
	  if( filter()<0 ) return -10 ;
	  if( smooth()<0 ) return -11 ;
	}
    }
  return 0;
}

int rangerTrack::fitFast() {

  status = filterDownStr();
  if (status < 0) {
    return -10;
  }
  status = filter();
  if (status < 0) {
    return -10;
  }
  return 0;
}




int rangerTrack::filter() { 

#ifdef FULL
  int counter=0;
#endif
  float chiSquare ;
  if (! empty()) {
    int stat = (*begin())->filter();
    chiSquare = (*begin())->getChiSq(); // chiSq contribution of first node should be 0 ??
    
#ifdef FULL
    cout << "first filter status stat=" << stat << endl;

    cout << "rangerTrack::filter: after filter step " << ++counter << endl;
    //char cc;
    //cc = cin.get();
#endif

    if (stat >= 0) {
      int last_stat = 0;
      rangerNodeIterator baseNode = begin();
      rangerNodeIterator nextNode = begin();
      nextNode++;
      for (rangerNodeIterator n = nextNode; 
	   n != end() && (last_stat >=0); 
	   n++) {

	rangerNodeIterator ntest = n;
	ntest++;

#ifdef FULL_FILTER
        //rangerState& state0 = (*n)->getState();
        cout << "rangerTrack::filter()   start node ---- " 
             << " zb " << (*baseNode)->zNode()
             << " zn " << (*n)->zNode() 
          //<< " QPExp " << (*n)->getQPExp() 
          //<< " qP " << state.qP() 
	  //<< " vec(5) " << state0.vec()(5)
	  //<< " cov(5,5) " << sqrt(state0.cov()(5,5))
	     << " hit " << (*n)->getHitb()
	     << " end " << (ntest != end())
             << endl;
#endif

	last_stat = (*n)->filter(*baseNode,tPartType);
	chiSquare += (*n)->getChiSq() ;
	baseNode = n;
   
#ifdef FULL_FILTER  
	rangerState& state = (*n)->getState();
	cout << "ranger::Track.filter(): _stat " 
	  //   << " zb " << (*baseNode)->zNode()
	     << " zn " << (*n)->zNode() 
	     << " l_state " << last_stat
	  //<< " QPExp " << (*n)->getQPExp() 
	  //<< " qP " << state.qP() 
	     << " vec(5) " << state.vec()(5)
	     << " cov(5,5) " << sqrt(state.cov()(5,5))
	     << endl;
	
	if(last_stat < 0) 
	  {
	    cout << "ranger::Track.filter(): last_stat " 
		 << last_stat << " z " << (*n)->zNode()
		 << " hit " << (*n)->getHitb() << endl; 
	  }
#endif 
      }

#ifdef FULL_FILTER  
      cout << "ranger::Track.filter(): return last_stat " << last_stat << endl;
#endif
      return last_stat;
    } 
    else {
      cout << "first filter status stat=" << stat << endl;
      return stat;
    }
  }
  static int id1;
  arteMsg << "Track.C : ranger::Track.filter(): no first node" << endl;
  arteMsg.allp(id1,1);

  return -10;
}

int rangerTrack::smooth() { 

  if ( empty() ) {
    cout << "rangerTrack::smooth(): empty track" << endl;
    return -10;
  }

  int last_stat = 0;
  for (rangerNodeReverseIterator n = rbegin(); 
       (n != rend()) && (last_stat >= 0); n++) {
    rangerNodeReverseIterator nk = n;
    nk++;
    if (nk != rend()) last_stat = (*nk)->smooth(*n);
  }
 
  return last_stat;
}

int rangerTrack::filterDownStr() { 

#ifdef FULL
  int counter=0;
#endif

  if (! empty()) {
    int stat = (*rbegin())->filter();

#ifdef FULL
    cout << "first filterDownStr status stat=" << stat << endl;

    cout << "rangerTrack::filterDownStr: after filter step " 
	 << ++counter << endl;
#endif

    if (stat >= 0) {
      int last_stat = 0;
      rangerNodeReverseIterator baseNode = rbegin();
      rangerNodeReverseIterator nextNode = rbegin();
      nextNode++;
      for (rangerNodeReverseIterator n = nextNode; 
	   n != rend() && (last_stat >=0); 
	   n++) {
	last_stat = (*n)->filter(*baseNode,tPartType);
	baseNode = n;
#ifdef FULL
	cout << "filter status last_stat=" << last_stat << endl;
	cout << "rangerTrack::filterDownStr: after filter step " 
	     << ++counter << endl;
	//cc = cin.get();
#endif
      }
      if (last_stat != 0) 
	cout << "rangerTrack::filterDownStr: filtering returns last_stat=" 
	     << last_stat << endl;
      return last_stat;
    } 
    else {
      cout << "rangerTrack::filterDownStr: first filtering returns stat=" 
	   << stat << endl;
      return stat;
    }
  }
#ifdef FULL 
  cout << "ranger::Track.filterDownStr(): no first node" << endl;
#endif
  static int id2; 
  arteMsg << "Track.C : ranger::Track.filterDownStr(): no first node" << endl; 
  arteMsg.allp(id2,1); 
  return -10;
}

inline void extrapolateState(const rangerState& s, 
			     const float& z, float& x, float& y, float& tx, float& ty)
{  
  // Utility function for insert wall routine. Extrapolates a state to
  // a new z-position. The magnettic field is assumed to be constant
  // between zfpar_.zminMag and zfpar_.zmaxMag and zero outside.
  double dz = z - s.z() ;
  y  = s.y() + dz*s.ty() ;
  ty = s.ty() ;
  if( (s.z()>zfpar_.zmaxMag && z>zfpar_.zmaxMag)
      || (s.z()<zfpar_.zminMag && z<zfpar_.zminMag) )  {
    x  = s.x() + dz*s.tx() ;
    tx = s.tx() ;
  } else { // through magnet
    //cout << "in magnet !" << endl ;
    x  = s.x() ;
    tx = s.tx() ; 
    double zthis = s.z() ;
    double znew  = s.z() ;
    // Step 1: from s.z() till the border of the magnet
    if( znew > zfpar_.zmaxMag ) znew = zfpar_.zmaxMag ;
    if( znew < zfpar_.zminMag ) znew = zfpar_.zminMag ;
    x += (znew-zthis)*tx ;
    zthis = znew ;
    // Step 2: through the magnet
    const float bfield = 0.633/(zfpar_.zmaxMag-zfpar_.zminMag) ; //1/[cm*GeV/c]
    double qpxz = s.qP() * sqrt((1+s.tx()*s.tx()+s.ty()*s.ty())/(1+s.tx()*s.tx())) ;
    double dtxdzmagn = qpxz * bfield ;
    znew = z ;
    if( znew > zfpar_.zmaxMag ) znew = zfpar_.zmaxMag ;
    if( znew < zfpar_.zminMag ) znew = zfpar_.zminMag ;
    x  += (znew-zthis)*(tx + 0.5*dtxdzmagn*(znew-zthis)) ;
    tx += dtxdzmagn*(znew-zthis) ;
    zthis = znew ;
    // Step 3: from border of magnet till z
    znew = z ;
    x += (znew-zthis)*tx ;
    zthis = znew ;
  }
}

int rangerTrack::insertWalls() {
  int nInserts = 0;
  if ( empty() ) return 0;
  
  // we keep track of which planes are already inserted.
  rangerNodeIterator lastNode(begin()),thisNode(end()) ;
  for (rangerNodeIterator nodeIt=begin(); nodeIt != end(); ++nodeIt) 
    if ( (*nodeIt)->plane() != 0 ) {
      lastNode = nodeIt ;
      if(thisNode==end()) thisNode = nodeIt ;
    }

  float z,x,y,tx,ty ;
  while( thisNode != lastNode ) {
    rangerNodeIterator nextNode = next(thisNode);
    while( (*nextNode)->plane()==0 ) ++nextNode ;
    
    // locate in-between nodes (we implicitly assume that nextnode is
    // an active node it self!)
    vector<rangerVolume*>::const_iterator volBeginIt, volEndIt ;
    volBeginIt = rangerGeom::firstUpstreamActiveVolN((*thisNode)->plane()) ;
    volEndIt   = rangerGeom::firstUpstreamActiveVolN((*nextNode)->plane()) ;
    
    // reset the state used for extrapolation
    z = -99 ;
    
    // loop over all in between volumes
    float sLeng=0 ;
    const float maxextrapolate = 10 ; //[cm]
    rangerNode* closestNode = *thisNode ;
    for (vector<rangerVolume*>::const_iterator volIt=volBeginIt ;
	 volIt > volEndIt+1 ; --volIt) {
      
      // calculate track parameters at this node
      // (note: also lineair extrapolation in Volume::sLength(...) )
      if( fabs((*volIt)->zMean()-z) > maxextrapolate ) {
	z = (*volIt)->zMean() ;
	closestNode = 
	  (fabs( (*nextNode)->zNode() - z ) <fabs( (*thisNode)->zNode() - z ) ) ?
	  *nextNode : *thisNode ;
	extrapolateState(closestNode->getState(),z,x,y,tx,ty) ;
      }

      if ( (sLeng = (*volIt)->sLengthFast(z,x,y,tx,ty)) > 0 ) {
	rangerNode* newNode = 0;
	if ((*volIt)->getGede() != 0)
	  newNode = rangerNode::makeNode((*volIt)->getGede(),closestNode,sLeng);
	else if ((*volIt)->getGwal() != 0)
	  newNode = rangerNode::makeNode((*volIt)->getGwal(),closestNode,sLeng);
	
	// insert node into track
	if (newNode) {
	  // filter from closest existing node
	  int stat = newNode->filter(closestNode,getParticleType()) ;
	  if(stat>=0) {
	    thisNode = insertBefore(nextNode,newNode);
	    nInserts++;
	    z = -99 ; // reset state
	  } else {
	    delete newNode ;
	  }
	}
      }
    }
    thisNode = nextNode ;
  }
  //  cout << "inserted walls: " << nInserts << endl ;
  return nInserts;
}


int rangerTrack::extrapolateUpstream()
{
  // Extrapolates the track through all upstream
  // volumes. Identification of these volumes is necessary for filling
  // the vds expected hit pattern

  int rc=0 ;
    
  int lastplane=0;
  for(rangerNodeConstIterator nodeIt = begin() ;
      nodeIt != end() ; nodeIt++) 
    if( (*nodeIt)->plane()!=0 ) lastplane = (*nodeIt)->plane() ;
    
  rangerNodeIterator lastNode = last() ;
  vector<rangerVolume*>::const_iterator volBeginIt, volEndIt ;
  volBeginIt = rangerGeom::firstUpstreamActiveVolN( lastplane ) ;
  volEndIt   = rangerGeom::getActiveVolumes().begin() ;
  volEndIt-- ;
  
  float sLeng=0 ;
  for (vector<rangerVolume*>::const_iterator volIt=volBeginIt ;
       volIt > volEndIt ; --volIt) 
    {
#ifdef FULL_VOL
      cout << "rangerTrack::extrapolateUpstream() volume " 
	   << " z " << (*volIt)->zMean() << " ms " << (*volIt)->ms()
	   << " dedx " << (*volIt)->dedx();       
      if ((*volIt)->getGede() != 0) 
	{
	  ArtePointer<GEDE> gede = (*volIt)->getGede();
	  cout << " gede " << gede << " cmp " << gede->cmp; 
	}
      if ((*volIt)->getGwal() != 0) 
	{
	  ArtePointer<GWAL> gwal = (*volIt)->getGwal();
	  cout << " gwal " << gwal << " shp " << gwal->shp; 
	}
      cout << endl;
#endif

      if( (sLeng = (*volIt)->sLengthFast((*lastNode)->getState()))  > 0 ) 
	{
	  rangerNode* newNode = 0;
	  if ((*volIt)->getGede() != 0)
	    newNode=rangerNode::makeNode((*volIt)->getGede(),*lastNode,sLeng);
	  else if ((*volIt)->getGwal() != 0)
	    newNode=rangerNode::makeNode((*volIt)->getGwal(),*lastNode,sLeng);
	  
	  if (newNode)
	    if(newNode->filter(*lastNode,getParticleType())>=0) 
	      {
		insertAfter(lastNode,newNode) ;
		lastNode = last() ;
		rc++ ;
	      } 
	    else {delete newNode ;}
	}
    }   
  return rc ;
}


int rangerTrack::extrapolateDownstream() 
{
  // Extrapolates the track through all downstream (untill the ECAL)
  // volumes. Identification of these volumes is necessary for filling
  // the otr/itr expected hit pattern

  int rc=0;
  
  int firstplane=0;
  for(rangerNodeConstIterator nodeIt = begin() ;
      nodeIt != end() && firstplane==0; nodeIt++) 
    if( (*nodeIt)->plane()!=0 ) firstplane =  (*nodeIt)->plane() ; 
  
  rangerNodeIterator firstNode = begin() ;
  
  vector<rangerVolume*>::const_iterator volBeginIt, volEndIt ;
  volBeginIt = rangerGeom::firstUpstreamActiveVolN( firstplane ) ;
  volEndIt   = rangerGeom::getActiveVolumes().end() ;
  
  float sLeng=0 ;
  const float zlast = 1330 ;
  for (vector<rangerVolume*>::const_iterator volIt=volBeginIt ;
       volIt != volEndIt && (*volIt)->zMean()<zlast ; ++volIt) 
    if( (sLeng = (*volIt)->sLengthFast((*firstNode)->getState()))>0 ) {   
      rangerNode* newNode = 0;
      if ((*volIt)->getGede() != 0)
	newNode = rangerNode::makeNode((*volIt)->getGede(),*firstNode,sLeng);
      else if ((*volIt)->getGwal() != 0)
	newNode = rangerNode::makeNode((*volIt)->getGwal(),*firstNode,sLeng);
      
      if (newNode) 
	if( newNode->filter(*firstNode,getParticleType()) >=0 ) { 
	  insertBefore(firstNode,newNode) ;
	  firstNode = begin() ;
	  rc++ ;
	} else {
	  delete newNode ;
	}
    }
  return rc ;
}



void rangerTrack::updInternal(){
  for (rangerNodeIterator nodeIt = begin();
       nodeIt != end();
       nodeIt++) {
    (*nodeIt)->updInternal();
  }
}



void rangerTrack::append(rangerNode& node) {
  nodes.push_back( &node );
}

void rangerTrack::prepend(rangerNode& node) {
  nodes.push_front( &node );
}


list<rangerNode*>::iterator rangerTrack::insertAfter
                                 (rangerNodeIterator baseNode, 
                                            rangerNode* newNode) { 
  if (baseNode != end()) {
    rangerNodeIterator insertPosition = baseNode;
    insertPosition++;
    rangerNodeIterator tmp = nodes.insert( insertPosition,newNode);
    return tmp;
  }
  else {
    cout << "rangerTrack::insertAfter: invalid base node " << endl;
    return baseNode;
  }
} 


list<rangerNode*>::iterator rangerTrack::insertBefore
(rangerNodeIterator baseNode,rangerNode* newNode) { 
  return nodes.insert(baseNode,newNode);

} 



void rangerTrack::trackList() {
  cout << "tFitMode=" << int(tFitMode) << endl;
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    (*n)->list();
  }
}


void rangerTrack::dump() {
  cout << "D u m p           tFitMode=" << int(tFitMode) << endl;
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    (*n)->dump();
  }
}

void rangerTrack::storeback(rangerRMap* rmap){

  if (!rmap) return;
  if (empty()) return;

  // Flag RHITS that have been removed
  if ( zfpar_.outlmode==1 ) rmap->flagOutliers(this) ;

  // Pass 0: prepare RPNTs for particle id
  rmap->prepareIdPnt();

  // for RTRA: prepare RPNT for high z point
  if ((rmap->getRtra()) != 0) {
    rmap->addPnt(*begin());
  }

  // Pass 1: store the RHITs
  for (rangerNodeIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      ArtePointer<RHIT> rhit = rmap->getRhit(*n);
      if (rhit != 0) (*n)->storeRhit(rhit);
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      if (rmap->getRpnt(*n) == 0) {
	ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  if ((rmap->getRseg()) != 0) {
	    createRelation( rmap->getRseg(), rp, &RSEG::rpnt );
	  }
	  else if ((rmap->getRtra()) != 0) {
	    createRelation( rmap->getRtra(), rp, &RTRA::rpnt );
	  }
	}
      }
    }
#endif
  }
 
  // Pass 2A: store the RSEG
  if ((rmap->getRseg()) != 0) {
    rangerNode* nLow = rmap->getNodeZLow();
    if (nLow) nLow->storeRseg(rmap->getRseg(),1);
    rangerNode* nHigh = rmap->getNodeZHigh();
    if (nHigh) nHigh->storeRseg(rmap->getRseg(),2);

    (rmap->getRseg())->chi = totalChiSq() ;
    (rmap->getRseg())->fit = (status<0)? -10: 0;

    if ((rmap->getRseg())->cmp == 0) (rmap->getRseg())->cmp = genCmp();
  }
  // Pass 2B: store the RTRA
  else if ((rmap->getRtra()) != 0) {
    ArtePointer<RTRA> rtra = rmap->getRtra() ;
    rangerNode* nLow = rmap->getNodeZLow();
    if (nLow) nLow->storeRtra(rtra,1);
    rtra->fit  = (status<0)? -10: 0;
    rtra->chi  = totalChiSq() ;
    rtra->ndof = totalNdof() ;
    rangerHitPattern::storeRtra(this,rtra) ;
  }

  // pass 3: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap->getHMList()).begin();
	HMIt !=(rmap->getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      if ((rmap->getRseg()) != 0) {
	int related = 0;
	related = (((*HMIt).getRpnt())->rseg == rmap->getRseg());

	if (! related) {  // set relation
	  createRelation( rmap->getRseg(), (*HMIt).getRpnt(), &RSEG::rpnt );
	}
      }
      if ((rmap->getRtra()) != 0) {
	if( find( (*HMIt).getRpnt()->rtra.begin(),
		  (*HMIt).getRpnt()->rtra.end(),
		  rmap->getRtra() ) == (*HMIt).getRpnt()->rtra.end() ) {
	  createRelation( rmap->getRtra(), (*HMIt).getRpnt(), &RTRA::rpnt );
	}
      }
    }
  }
}


void rangerTrack::storeback(vector<rangerRMap>* rmapVec) {
  if (! rmapVec) return;
  for (vector<rangerRMap>::iterator rIt=rmapVec->begin();
       rIt != rmapVec->end(); rIt++) {
    storeback( &(*rIt) );
  }
}


ArtePointer<RTRA> rangerTrack::storeRtra() const{

  if (empty()) return 0;

  // Pass 1: allocate and store RTRA entry
  mmbook_(ArtePointer<RTRA>::getLink().getRef(),1);
  ArtePointer<RTRA> rt( ArtePointer<RTRA>::nrow() );
  (*rbegin())->storeRtra(rt,1);   // store point of lowest z

  rt->fit  = (status<0)? -10: 0;
  rt->chi  = totalChiSq(); 
  rt->ndof = totalNdof() ;
  rt->hsi  = 0;
  rt->hin  = 0;
  rt->hou  = 0;
  rt->hri  = 0;
  rt->hca  = 0;
  rt->htr  = 0;
  rt->hmu  = 0;

  rangerRMap rmap(rt);  // create hit map to aid further operations

  // Pass 2: store the RHITs
  for (rangerNodeConstIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      mmbook_(ArtePointer<RHIT>::getLink().getRef(),1);
      ArtePointer<RHIT> rhit( ArtePointer<RHIT>::nrow() );
      if (rhit != 0) {
	createRelation( rt, (*n)->getHitb(), &RTRA::hitb );
	createRelation( rhit, (*n)->getHitb(), &HITB::rhit );
	(*n)->storeRhit(rhit);
	rhit->rtra = rt.index();
	rangerHitMap hm((*n),rhit);
	rmap.addHitMap( hm );
      }
      ArtePointer<GEDE> ge(((*n)->getHitb())->gede);  // count hits
      switch (ge->cmp) {
      case 1:
	rt->hsi++;
	break;
      case 2:
	rt->hou++;
	break;
      case 3:
	rt->hin++;
	break;
      case 4:
	rt->hri++;
	break;
      case 5:
	rt->hca++;
	break;
      case 6:
	rt->htr++;
	break;
      case 7:
	rt->hmu++;
	break;
      default:
	cout << " rangerTrack::storeRtra: bad cmp=" << ge->cmp << endl;
      }
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  createRelation( rt, rp, &RTRA::rpnt );
	}
    }
#endif
  }
  // Pass 3: prepare RPNTs for particle id
  rmap.prepareIdPnt();

  // prepare RPNT for high z point
  rmap.addPnt(*begin());

  // pass 4: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap.getHMList()).begin();
	HMIt !=(rmap.getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      // set relation
      createRelation( rt, (*HMIt).getRpnt(), &RTRA::rpnt );
    }
  }
  return rt;
}


ArtePointer<RSEG> rangerTrack::storeRseg() const{

  if (empty()) return 0;

  // Pass 1: allocate and store RSEG entry
  mmbook_(ArtePointer<RSEG>::getLink().getRef(),1);
  ArtePointer<RSEG> rs( ArtePointer<RSEG>::nrow() );
  (*rbegin())->storeRseg(rs,1);   // store point of lowest z
  (*begin())->storeRseg(rs,2);    // store point of high z

  rs->fit = (status<0)? -10: 0;
  rs->chi = totalChiSq();
  rs->cmp = genCmp();

  rangerRMap rmap(rs);  // create hit map to aid further operations

  // Pass 2: store the RHITs
  for (rangerNodeConstIterator n = begin(); n != end(); n++) {
    if ((*n)->hasHitb()) {
      mmbook_(ArtePointer<RHIT>::getLink().getRef(),1);
      ArtePointer<RHIT> rhit( ArtePointer<RHIT>::nrow() );
      if (rhit != 0) {
	createRelation( rs,   (*n)->getHitb(), &RSEG::hitb );
	createRelation( rhit, (*n)->getHitb(), &HITB::rhit );
	createRelation( rs,   rhit,            &RSEG::rhit );
	(*n)->storeRhit(rhit);
	rhit->rtra = 0;
	rangerHitMap hm((*n),rhit);
	rmap.addHitMap( hm );
      }
    }
#ifdef SAVEWALLS
    else {              // store wall RPNTs (for diagnosis only)
      ArtePointer<RPNT> rp = (*n)->storeRpnt(0);
	if (rp != 0) {  // new RPNT, set relation
	  createRelation( rs, rp, &RSEG::rpnt );
	}
    }
#endif
  }

  // Pass 3: prepare RPNTs for particle id
  rmap.prepareIdPnt();

  // pass 4: store the RPNT
  for ( list< rangerHitMap >::const_iterator 
	  HMIt=(rmap.getHMList()).begin();
	HMIt !=(rmap.getHMList()).end(); 
	HMIt++){
    if ((*HMIt).getRpnt() != 0) {   // store into assigned RPNT
      (*HMIt).getNode()->storeRpnt((*HMIt).getRpnt());

      // set relation
      createRelation( rs, (*HMIt).getRpnt(), &RSEG::rpnt );
    }
  }
  return rs;
}


double rangerTrack::getLowZ() const {
  if (empty()) return 0;
  else {
    return (*rbegin())->zNode();
  }
}

const HepVector* rangerTrack::getLowZVec() const {
  if (empty()) return 0;
  else {
    return &(((*rbegin())->getState()).vec());
  }
}

const HepSymMatrix* rangerTrack::getLowZCov() const {
  if (empty()) return 0;
  else {
    return &(((*rbegin())->getState()).cov());
  }
}

float rangerTrack::totalChiSq() const 
{
  // This is less trivial than it looks like. After discussions with
  // Rainer in the past and yesterday with Dmitri, I am quite
  // convinced that the chi2 that comes out of the Kalman filter is
  // not a reliable `global' chi2. It is for example very sensitive to
  // the initial state vector and cov matrix. If you look at it more
  // practically: It gives too many `very unlikely' chi2s and cannot
  // be used as a track quality criterium. So we need an alternative.

  // Let me first tell what are the possibilities:

  // 1) Rainer's chi2: This is the on-the-fly chi2 calculation of the
  // Kalman filter step. The chi2 contributions of the hits (`dchi2')
  // are normalized with the error in the residual.The first npar
  // dchi2 are (more or less) zero. The next (nhit-npar) dchi2 are
  // independent. Therefore the total chi2 has expectation value
  // (nhit-npar).

  // 2) The global chi2: This is the sum of all residuals after
  // smoothing normalized with the error in the hit. This has also
  // expectation value (nhit-npar).

  // 3) The `likelihood' chi2: If you define chi2 from the likelihood
  // distribution, it is something like res*W*res where res is the
  // vector of residuals (nhit-npar) `independent' residuals and W is
  // the inverse of the corresponding cov matrix. Does W contain
  // off-diagonal elements ? In the rangerfit these elements are
  // lost. Therefore, I wouldn't know how to calculate this
  // chi2. Closely related however, is:

  // 4) The `normalized' chi2: This is the sum of all residuals after
  // smoothing normalized with the error in the residual. This has
  // expectation value nhit, not nhit-npar. So it cann't be a chi2
  // contribution. However, if we scale with (nhit-npar)/nhit, we get
  // something which should be very close to (3). I have shown for
  // myself that for a Gaussian model this recipe gives a correct chi2
  // distribution.

  // Option (1) has the problems that Dmitri pointed out above. Option
  // (2) is incorrect if the process noise contributes considerably to
  // the errors in the residuals. (It becomes too small.) Option (3)
  // we cann't calculate. Option (4) seems the only left over
  // option.It looks fishy but I think it is correct. If you have a
  // better idea, tell me.
  //
  // Wouter, July 23rd 2001
  
  // Option 1: Rainer's definition (node->ChiSq is not smoothed)
  //   double chiSquare = 0;
  //   for(rangerNodeConstIterator n = begin(); n != end(); ++n) 
  //     chiSquare += (*n)->getChiSq() ;
  
  // Option (4):
  double chiSquare(0) ;
  unsigned int nhit(0) ;
  int ierr; 
  for(rangerNodeConstIterator n = begin(); n != end(); ++n)
    if((*n)->dim()>0) {
      rangerHitBase* hit = static_cast<rangerHitBase*>(*n) ;
      if( hit->getCovResidual()) {
	chiSquare += 
	  ((*hit->getCovResidual()).inverse(ierr)).similarity(*(hit->getResidual())) ;
	nhit += hit->dim()  ;
      }
    }
  unsigned int npar = (*begin())->getState().dim() ;  
  chiSquare *= double(nhit-npar)/double(nhit) ;
  
  return float(chiSquare) ;
}

int rangerTrack::totalNdof() const
{
  // returns the number of degrees of freedom in the fit
  int ndof = - (*begin())->getState().dim() ;
  for(rangerNodeConstIterator n = begin(); n != end(); ++n)
    ndof += (*n)->dim() ;
  return ndof ;
}

extern "C" float prob_(const float&, const int&) ;
float rangerTrack::chiSqProb() const
{
  int ndof    = totalNdof() ;
  float chiSq = totalChiSq() ;
  return ndof>0 ? prob_(chiSq,ndof) : -1 ;
}

double rangerTrack::totalRadThickness() const {
  
  double sum = 0;
  for (rangerNodeConstIterator n = begin(); n != end() ; n++) {
    sum += (*n)->radiationThickness();
  }
  return sum;
}

int rangerTrack::genCmp() const {
  int CMP = 0;
  for (rangerNodeConstIterator n = begin(); n != end() ; n++) {
    if ((*n)->hasHitb()) {
      if ((*n)->plane()>0) {
	ArtePointer<GEDE> ge((*n)->plane());
	float zsl = 0.5 * (((ArtePointer<GESL>) ge->gesl)->z1 
			   +((ArtePointer<GESL>) ge->gesl)->z2);

	if (               zsl <  190) CMP = CMP | Rsegc::vxd;  // set functional bits
	if (zsl >=  190 && zsl <  660) CMP = CMP | Rsegc::magt;
	if (zsl >=  660 && zsl < 1000) CMP = CMP | Rsegc::patt;
	if (zsl >= 1000 && zsl < 1330) CMP = CMP | Rsegc::trit;
	if (zsl >= 1330 && zsl < 1400) CMP = CMP | Rsegc::ecal;
	if (zsl >= 1400              ) CMP = CMP | Rsegc::muon;

	switch (ge->cmp) {
	case 1:
	  CMP = CMP | Rsegc::sil;
	  break;
	case 2:
	  CMP = CMP | Rsegc::wch;
	  break;
	case 3:
	  CMP = CMP | Rsegc::msg;
	  break;
	case 4:
	  CMP = CMP | Rsegc::ptc;
	  break;
	case 5:
	  CMP = CMP | Rsegc::rich;
	  break;
	case 6:
	  CMP = CMP | Rsegc::trd;
	  break;
	case 7:
	  CMP = CMP | Rsegc::bit23;
	  break;
	case 8:
	  CMP = CMP | Rsegc::bit22;
	  break;
	default:
	  break;
	}
      }
    }
  }
  return CMP;
}


void rangerTrack::lastNoise() {
  (*last())->processNoise(tPartType);
}

inline bool
compareNodeZ(const rangerNode* lhs, const rangerNode* rhs) 
{
  // see also Geom.C (need same order)
  double refzlhs = 1e-8*const_cast<rangerNode*>(lhs)->plane()+lhs->zNode() ; 
  double refzrhs = 1e-8*const_cast<rangerNode*>(rhs)->plane()+rhs->zNode() ; 
  return refzlhs>refzrhs ;
}


void rangerTrack::zOrder() {
  //nodes.sort( not2( compareCall( &rangerNode::zNode ) ) );
  nodes.sort(&compareNodeZ) ;
}

void rangerTrack::orderIfNeeded() {
  int nOrderFault = 0;
  rangerNodeIterator prevnodeIt = begin();
  rangerNodeIterator nodeIt = prevnodeIt;
  for (nodeIt++; nodeIt != end(); nodeIt++) {
    if ( !compareNodeZ(*prevnodeIt,*nodeIt) )  ++nOrderFault;
    prevnodeIt = nodeIt ;
  }

  if (nOrderFault > 0) {
    static int id;
#ifdef PRINT_ORDER_FAULTS
    cout << "rangerTrack::orderIfNeeded(): number of order faults on input "
	 << nOrderFault << " last at z=" << zFault << endl;
#endif
    arteMsg << "rangerTrack::orderIfNeed(): track required sort,"
	    << " zb=" << (*begin())->zNode() 
	    << " ze=" << (*last())->zNode() 
	    << " nfault=" << nOrderFault << endl;
    arteMsg.allp(id,10);
    zOrder();
  }
}

int rangerTrack::fitV() {

  int nInsert = 0;
  nInsert=insertWalls();
  status = filter();
  if (status < 0) { 
    return -10; 
  } 
  return 0; 
}

int rangerTrack::fitVf() {

  status = filter();
  if (status < 0) {  
    return -10;  
  }  
  return 0;  
}


int rangerTrack::fitVs() {

  int nInsert = 0;
  nInsert=insertWalls();
  status = filter();
  if (status < 0) {  
    return -10;  
  }  
  return 0;  
}

int rangerTrack::fitGlobal()
{
  const int maxnstep    = 5 ;
  const float dchi2conv = 0.1 ; 
  int stat = -1 ;
  if (! empty()) {
    rangerNode* baseNode = (*begin());
    stat = baseNode->filter();
    if (stat >= 0) {
      rangerState* globstate = &(baseNode->getState()) ;
      int dim = globstate->dim() ;
      int ndof = -dim;
      float chiSquare=0 ;
      for (rangerNodeIterator n = begin() ; n != end() && stat>=0 ; n++) {
	stat = (*n)->projectGlobal(globstate) ;
	ndof += (*n)->dim() ;
	chiSquare += (*n)->getChiSq() ;
      }
      if(stat>=0) {
	// check ndof
	if(ndof<=0) stat = -2;
	else {
	  HepVector hdChi2dQ(dim) ;
	  HepSymMatrix hd2Chi2dQ2(dim) ;
	  bool converged=false ;
	  for(int nstep=0; nstep<maxnstep && !converged && stat>=0; nstep++) {
	    //cout << "Step: " << nstep << endl ;
	    // get chi2 contribution
	    for(int row=1; row<=dim; row++) {
	      hdChi2dQ(row) = 0;
	      for(int col=1; col<=row; col++) 
		hd2Chi2dQ2(row,col) = 0;
	    }	  

	    for (rangerNodeIterator n = begin() ; n != end() && stat>=0 ; n++)
	      stat = (*n)->absorbGlobal(&hdChi2dQ,&hd2Chi2dQ2) ;

	    if(stat >=0) {
	      // invert covariance matrix
	      int ierr ;
	      hd2Chi2dQ2.invert(ierr) ;
	      if(ierr) stat = -3 ; 
	      else {
		// calc change in chi2
		float dchi2 = hd2Chi2dQ2.similarity(hdChi2dQ) ;
		if(dchi2 < dchi2conv/ndof) converged = true ;
		//cout << "dchi2: " << dchi2 << endl ;
		// update state
		globstate->cov() = hd2Chi2dQ2;
		globstate->vec() -= hd2Chi2dQ2*hdChi2dQ ;
		
		chiSquare = 0;
		for (rangerNodeIterator n = begin() ; n != end() ; n++) {
		  (*n)->projectGlobal(globstate) ;
		  chiSquare += (*n)->getChiSq() ;
		}
	      }
	    }
	  }
	}
      }
    }
  }
  return stat ;
}

// utility class for `removeBadHits'
class rangerNodeChiSq
{
public:
  rangerTrack::rangerNodeIterator nodeIt ;
  float chiSq ;
} ;

bool operator<(const rangerNodeChiSq& lhs, const rangerNodeChiSq& rhs) {
  return lhs.chiSq > rhs.chiSq ;
}


int
rangerTrack::rejectOutliers()
{
  // This routine rejects outliers in vds/main tracker tracks. 
  const float fracmaxreject=0.2 ;   // maximum fraction of hits to throw away
  const float maxchisq = 9 ;        // maximum chi2 of good hit
  const unsigned int minnvdshits=5 ; // minimum hits left in vds
  
  // first make a list of candidate `bad hits'
  int ierr;
  rangerNodeChiSq nchsq ;
  unsigned int nhits(0),nvdshits(0) ;
  vector<rangerNodeChiSq> badhits ;
  for(rangerTrack::rangerNodeIterator nodeIt = begin() ;
      nodeIt != end() ; nodeIt++) 
    if((*nodeIt)->hasHitb()) {
      int cmp = (*nodeIt)->getHitb()->gede->cmp ;
      rangerHit* hit = (rangerHit*)(*nodeIt) ;
      if( (cmp==1 || cmp==2 || cmp==3) && hit->getCovResidual() ) {
	float chiSq = 
          ((*hit->getCovResidual()).inverse(ierr)).similarity(*(hit->getResidual())) ;
        if( chiSq>maxchisq ) { // bad
	  // update internal and check chi2 again
	  hit->updInternal() ;
	  chiSq = 
	    ((*hit->getCovResidual()).inverse(ierr)).similarity(*(hit->getResidual())) ;
	  if( chiSq>maxchisq )  { //still bad
	    nchsq.nodeIt = nodeIt ;
	    nchsq.chiSq  = chiSq  ;
	    badhits.push_back(nchsq) ;
	  }
	}
      	if( cmp==1 ) nvdshits++ ;
	nhits++ ;
      }
    }

  // reject bad hits
  sort(badhits.begin(),badhits.end()) ;
  unsigned int nvdsrejected(0),nrejected(0) ;
  unsigned int maxnvdsreject = min(int(nvdshits*fracmaxreject),
				   int(nvdshits-minnvdshits)) ;
  unsigned int maxnreject = int(nhits*fracmaxreject) ;
  for(vector<rangerNodeChiSq>::iterator it = badhits.begin() ;
      it !=badhits.end() && nrejected<maxnreject; ++it) {
    rangerHit* hit = (rangerHit*)(*(it->nodeIt)) ;
    bool isvds = hit->getHitb()->gede->cmp==1 ;
    // replace hit by wall and delete hit
    if( !isvds || nvdsrejected<maxnvdsreject ) {
      if(isvds) ++nvdsrejected ;
      ++nrejected ;
      const rangerState& state = hit->getState() ;
      float sLeng = sqrt(1+state.tx()*state.tx()+state.ty()*state.ty()) ;
      *(it->nodeIt) = rangerNode::makeNode(ArtePointer<GEDE>(hit->plane()),hit,sLeng) ;
      delete hit;
    }
  }
  //cout << "rangerTrack: removed " << nrejected << " hits. " << endl ;
  
  return nrejected + 256*nvdsrejected ;
}
