//************************************************************************
//
//  Routine:      rcFitVdsPatMatch
//
//  Author:       W. Hulsbergen (november 2001)
//
//  In:   - vdsseg: ArtePointer to vds segment
//        - patseg: ArtePointer to pattern segment (RPNT, for now)
//  Out: - matchseg: pointer to matchsegment: contains fitted state
//         vectors at both sides (including momentum) and a chi2 with 3
//         (=2*4-5) degrees of freedom.
//
//  Description:
//  Performs a simple Kalman fit to match a vds and a pattern
//  segment. I use this for my private vds-pattern matching. The
//  calculated chi2 is very useful for getting rid of wrong
//  matches. The fitted momentum has a resolution of 1.6% which is as
//  good as the refit, if magnet hits are not used.
//
//  There are two versions of the routine. One accepts an rpnt, the other an
//  rseg as input for the pattern segment.
//
//
//************************************************************************
       
#include <iomanip>
#include <vector>
#include <CLHEP/Matrix/Vector.h>
#include <CLHEP/Matrix/Matrix.h>
#include <CLHEP/Matrix/SymMatrix.h>
#include <arte/arte.hh>
#include <arte/RSEG_cmp.hh>
#include <rang/rkmain.hh>

// interface to rkmain which makes it look like rk5clip:
static void rkmainint(double zin, double parin[5], double zout, double parout[5],
		      double rkd[25], int& ierr)
{
  // interface to rkmain (see also rang/StateP.C)
  double fstep = fabs(zout - zin);
  if (fstep<=0) fstep = 1;
  double pin[6];
  pin[0] = zin;
  for(int i=0; i<5; ++i) pin[i+1] = parin[i] ;
  double prec[4] = {0.001,0.001,0.0001,0.0001} ;
  double dzTrunc = fabs(zout-zin);
  if( dzTrunc>1 ) for(int i=0; i<4; ++i) prec[i] /= dzTrunc ;
  int icovar = 1;
  double pout[6];
  rkmain_(icovar,zout,fstep,pin,prec,pout,rkd,ierr);
  if (pout[0] != zout) ierr=3;
  for(int i=0; i<5; ++i) parout[i] = pout[i+1] ;
}

extern "C"  void rk5clip_(double& z_in ,double* p_in,
			  double& z_out,double* p_out,
			  double* rkd,int& ierror);

extern "C" void rcmesm_(int*, double*, double*);

int rcFitVdsPatMatch(ArtePointer<RSEG> vdsseg, ArtePointer<RPNT> patseg,
		     RSEG* matchseg) 
{
  int rc=0;

  // === STEP 1: copy measurement vectors and cov matrix
  const int dimvds = 4;
  HepVector    m_vds(dimvds) ;
  for(int row=1; row<=dimvds; row++) m_vds(row) = (&(vdsseg->xe))[row-1] ;
  
  HepSymMatrix V_vds(dimvds) ;
  int j=0;
  for(int row=1; row<=dimvds; row++)
    for(int col=1; col<=row; col++, j++) 
      V_vds.fast(row,col) = vdsseg->cve[j] ;
  double z_vds = vdsseg->ze ;
  
  const int dimpat = 4;
  HepVector m_pat(dimpat) ;
  for(int row=1; row<=dimpat; row++) m_pat(row) = (&(patseg->xf))[row-1] ;

  HepSymMatrix V_pat(dimpat) ;
  j=0;
  for(int row=1; row<=dimpat; row++)
    for(int col=1; col<=row; col++, j++) 
      V_pat.fast(row,col) = patseg->cvf[j] ;
  double z_pat = patseg->zf ;

  // === STEP 2: multiple scattering
  // use Olya's routine to calculate the total ms of all superlayers
  static vector<float> msmap ;
  if(msmap.empty())
    for(ArteTable<GESL>::iterator geslIt = ArteTable<GESL>::begin() ;
	geslIt != ArteTable<GESL>::end() ; geslIt++) {
      double ms(0),eloss(0) ;
      int igesl = geslIt.index();
      rcmesm_(&igesl,&ms,&eloss) ;
      msmap.resize(igesl+1) ;
      msmap[igesl] = ms ;
    }

  // sum up all ms in between these two states
  float mstot(0) ;
  for(ArteTable<GESL>::iterator geslIt = ArteTable<GESL>::begin() ;
      geslIt != ArteTable<GESL>::end() ; geslIt++) 
    if( geslIt->z1 > z_vds && geslIt->z2 < z_pat )
      mstot += msmap[geslIt.index()] ;

  // === STEP 3: create and initialize state vector
  HepVector state_vds(5),state_pat(5) ;
  HepSymMatrix cov_vds(5,0),cov_pat(5) ;
  
  // we have to initialize the momentum of the vds state:
  // calculate kick:
  const float zmagn = 455;
  float dtx = ((z_pat-z_vds)*m_vds[2] + m_vds[0] - m_pat[0])/(zmagn - z_pat) ;
  float qpkick = 455./288. * dtx * 
    sqrt( (1 + m_vds[2] * m_vds[2] )/
	  (1 + m_vds[2] * m_vds[2] + m_vds[3] * m_vds[3]) ) ;
  state_pat(5) = state_vds(5) = qpkick ;

  
  // === STEP 4: allocate some temproary matrices
  bool converged=false ;
  HepVector predstate(5) ;
  HepSymMatrix predcov(5) ;
  HepMatrix FT(5,5),A(5,5) ; 
  HepMatrix K(5,4,0) ;
  HepSymMatrix tmp(4) ;
  double chi2(-1) ;
  int ierr ;

  //HepMatrix H(4,5,0) ;
  //H(1,1) = H(2,2) = H(3,3) = H(4,4) = 1;

  // == STEP 5: filter
  const double thickness = fabs(z_pat - z_vds) ;
  double noise = mstot*state_vds(5)*state_vds(5) ;
  int step=0;
  while(!converged && step<3) {
    // process the first. node since this is the first node, we can
    // just copy the measurement. except for the momentum, which we
    // copy from the pat_state (that saves the smoothing step)
    state_vds.sub(1,m_vds) ;
    state_vds(5) = state_pat(5) ;
    cov_vds.sub(1,V_vds) ;
    for(int col=1; col<=4; ++col) cov_vds.fast(5,col) = 0 ;
    cov_vds.fast(5,5) = 0.1*0.1*state_vds(5)*state_vds(5) ;
    
    // extrapolate to the other side
    rkmainint(z_vds,&(state_vds(1)),z_pat,&(predstate(1)),&(FT(1,1)),ierr) ;
    //rk5clip_(z_vds,&(state_vds(1)),z_pat,&(predstate(1)),&(FT(1,1)),ierr) ;
    predcov = cov_vds.similarityT(FT) ;
    
    // apply multiple scattering
    predcov(1,1) += noise*(thickness*thickness)/3 ;
    predcov(1,3) += noise*thickness/2 ;
    predcov(3,3) += noise ;
    predcov(2,2) += noise*(thickness*thickness)/3 ;
    predcov(2,4) += noise*thickness/2 ;
    predcov(4,4) += noise ;
    
    // calculate the chi2 contribution (before filtering)
    float newchi2 = 
      ((predcov.sub(1,4)+V_pat).inverse(ierr)).
      similarity( m_pat - predstate.sub(1,4))  ;
    
    // calculate the gain matrix
    // K = predcov*H.T()* ( V_pat + predcov.similarity(H) ).inverse(ierr) ;
    for(int row=1; row<=4; ++row)
      for(int col=1; col<=row; ++col)
	tmp.fast(row,col) = V_pat.fast(row,col) + predcov.fast(row,col) ;
    tmp.invert(ierr) ;
    for(int row=1; row<=5; ++row)
      for(int col=1; col<=4; ++col) {
	K(row,col) = 0;
	for(int k=1; k<=4; ++k) 
	  K(row,col) += predcov(row,k) * tmp(k,col) ;
      }    
    
    // filter the tracker state
    state_pat = predstate ;
    for(int row=1; row<=5; ++row)
      for(int col=1; col<=4; ++col)
	state_pat(row) += K(row,col) * (m_pat(col) - predstate(col)) ;
    
    const int ndof = 3;
    converged = chi2>0 && (fabs(chi2-newchi2)/ndof<0.5 || newchi2>chi2) ;
    chi2 = newchi2 ;
    ++step ;
  }
  
  // do remaining part of filter
  cov_pat   = predcov ;
  for(int row=1; row<=5; ++row)
    for(int col=1; col<=row; ++col)
      for(int k=1; k<=4; ++k) 
	cov_pat.fast(row,col) -= K(row,k) * predcov(k,col) ;
  
  // smooth the vds state
  A = cov_vds*FT*predcov.inverse(ierr) ; 
  if(ierr) cout << "ierr in predcov.inverse(ierr)" << endl ;
  state_vds += A*(state_pat-predstate) ;
  cov_vds += ( cov_pat - predcov).similarity(A) ;
  
#ifdef CHISQCLASSIC
  // calculate the `classical chi2'. slow and not needed. 
  float chiA = V_otr.inverse(ierr).similarity(m_otr - state_otr.sub(1,4)) ;
  float chiB = V_vds.inverse(ierr).similarity(m_vds - state_vds.sub(1,4)) ;
  float chi2classic = chiA +  chiB ;
#endif

  // == STEP last: fill the matchsegment
  matchseg->zf  = z_vds ;
  matchseg->xf  = state_vds(1) ;
  matchseg->yf  = state_vds(2) ;
  matchseg->txf = state_vds(3) ;
  matchseg->tyf = state_vds(4) ;
  matchseg->pf  = state_vds(5) ;
  j=0;
  for(int row=1; row<=5; ++row)
    for(int col=1; col<=row; ++col, ++j)
      matchseg->cvf[j] = cov_vds.fast(row,col) ;
  
  matchseg->ze  = z_pat ;
  matchseg->xe  = state_pat(1) ;
  matchseg->ye  = state_pat(2) ;
  matchseg->txe = state_pat(3) ;
  matchseg->tye = state_pat(4) ;
  matchseg->pe  = state_pat(5) ;
  j=0;
  for(int row=1; row<=5; ++row)
    for(int col=1; col<=row; ++col, ++j)
      matchseg->cve[j] = cov_pat.fast(row,col) ; 
  matchseg->chi  = chi2 ;
  matchseg->fit = 7;
  matchseg->cmp = vdsseg->cmp | Rsegc::patt ;
#ifdef CHISQCLASSIC
  matchseg->time = chi2classic
#endif
  
  return rc;
}


int rcFitVdsPatMatch(ArtePointer<RSEG> vdsseg, ArtePointer<RSEG> patseg,
		      RSEG* matchseg) 
{
  // interface for RSEG: search for correct rpnt first ...
  ArtePointer<RPNT> rpnt ;
  for(ArteRelationIterator<RPNT> irpnt = patseg->rpnt.begin() ;
      irpnt != patseg->rpnt.end() ; irpnt++) 
    if((*irpnt)->zf>680)
      if(rpnt.index()==0 || rpnt->zf > (*irpnt)->zf) rpnt = *irpnt ;

  int rc =1;
  if(rpnt.index())
    rc = rcFitVdsPatMatch(vdsseg,rpnt,matchseg) ;

  return rc ;
}
