
// *****************************************************************************
//
// source:    Selection of B->J/Psi X, requires reconstructed
//            J/Psi and looks for a charge or Ks that points to
//            the same vertex
//            Returns:
//               0=no success, 
//               1=topology ok (sec. vertex found)
//               2=K found among prongs
//
//            Reconstruction of B-decays in channels
//               B+ :  J/Psi K+, J/Psi K+ Pi+ Pi-, ~D0 pi+
//               B- :  J/Psi K-, J/Psi K- Pi+ Pi-. D0 pi-
//               B0 :  J/Psi Ks, J/Psi K+ Pi-  and c.c.
//
//
// type:      source code
//
// created:   04 Aug 2000
//	
// author:    Denis Dujmic
//
// changes:
//  20 Aug 2000 (DD) debugging rhp histograms only in offline mode
//  21 Aug 2000 (DD) bug fix
//  06 Sep 2000 (DD) File description added
//  17 Sep 2000 (DD) Full reconstruction of B-decays
//  27 Oct 2000 (DD) D0 pi channels added
//  15 Dec 2000 (DD) Require RTRA-TOPP for J/Psi and D, older code removal
//  15 Dec 2000 (DD) Require unique vds-otr
//  03 Jan 2001 (DD) B__JPsi_X_MCeff not used by default
//  23 Oct 2001 (DS) fill RTRA rich likelihoods only if not filled before
//
// *****************************************************************************
#include "reco/rcX.h"
#include "reco/rcXkinema.h"
#include "reco/rcXsegment.h"
#include "rhp.h"
#include "dq/DQ.h"

#include "vector.h"
#include "multimap.h"

#ifdef USE_ROOT_DD
#include "reco/rcXdata.h"
#endif


#ifdef USE_ROOT_DD
#define MB_LO 3.0
#define MB_HI 10.0
#else
#define MB_LO 4.0
#define MB_HI 7.0
#endif

void B();

int B__JPsi_X_RC();
int B__JPsi_X_MCeff(vertexAttrib &vattrib, vertexmass &vert, int total);
int maximizeRTRA();
void rcResetRichRtra(float x=-1);


int Bplus__JPsi_Kplus();
int Bplus__JPsi_Kplus_Piplus_Piminus();

int Bminus__JPsi_Kminus();
int Bminus__JPsi_Kminus_Piplus_Piminus();

int B0__GOLDEN();
int B0__JPsi_Kplus_Piminus();
int B0__JPsi_Kminus_Piplus();

int B0__Dminus_Piplus();
int B0__Dplus_Piminus();

int Bplus__D0bar_Piplus();
int Bminus__D0_Piminus();

rcX *rcBp1=0,  *rcBm1=0;
rcX *rcBp2=0,  *rcBm2=0;
rcX *rcGold=0, *rcB0a=0, *rcB0b=0; 
rcX *rcBDpi1=0,*rcBDpi2=0;


/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  REAL RECO   /////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////
////////////////////////// RECONSTRUCTION ///////////////////////////////
/////////////////////////////////////////////////////////////////////////

//
//  B+ -> J/Psi K+
// 
int Bplus__JPsi_Kplus() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B+1  <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBp1) {
    cout<<"initializing B+1"<<endl;
    rcBp1=new rcX("Bp1", "B+ -> J/Psi K+", verb);

    decay bp;
    bp.fit         = 3;    // = short decay
    bp.book_tables = 1;    // 1/0
    bp.rlevel      = 0;    // 0=full, 1=rtra, 2=mtra
    bp.verb        = verb;
    rcBp1->setMother(bp);  // set
    vertexAttrib vbp;
    vbp.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbp.topp               = kinema::findTopp(521); // pointer to mother
    vbp.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vbp.momentum_flag      = 1;     // momentum in vertex fit?
    vbp.check_wire         = 1;     // wire constraint?
    vbp.verb               = verb;  // verbosity 
    vbp.min_mass           = MB_LO;// minimum mass
    vbp.max_mass           = MB_HI;// maximum mass
    vbp.min_dz_from_wire   = .25;    // z-cut from primary vertex
    vbp.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcBp1->setVertex(vbp);          // set
 
    attribute jp;
    jp.topp=kinema::findTopp(443);
    //jp.lund    = 443;// set MC id
    //jp.mother  = 521;// set MC id
    jp.fit     = 2;
    jp.mass    = jp.topp->mass;
    jp.matchTOPP =1;
    jp.minP    = 0;
    jp.maxP    = 200;
    jp.minMass = 2.7;
    jp.maxMass = 3.3;
    rcBp1->addDaughter(jp);
  
    attribute k;
    k.topp=kinema::findTopp(321);
    //k.lund         = 321;// set MC id
    //k.mother       = 521;// set MC id
    k.mass         = k.topp->mass;
    k.minP         = 0;
    k.maxP         = 200;
    k.minHIO       = 15;
    k.minHSI       = 6;
    k.minPt        = 0.3;
    k.minLRK      = .0001;    
    k.useVdsOnce   = 1;
    k.useMainOnce  = 1;
    rcBp1->addDaughter(k);

    if (verb) rcBp1->printDaughters();
  }
  rcBp1->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B+1 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


//
//  B- -> J/Psi K-
// 
int Bminus__JPsi_Kminus() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B-1  <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBm1) {
    cout<<"initializing B+1"<<endl;
    rcBm1=new rcX("Bm1", "B+ -> J/Psi K+", verb);

    decay bp;
    bp.fit         = 3;    // = short decay
    bp.book_tables = 1;    // 1/0
    bp.rlevel      = 0;    // 0=full, 1=rtra, 2=mtra
    bp.verb        = verb;
    rcBm1->setMother(bp);  // set
    vertexAttrib vbp;
    vbp.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbp.topp               = kinema::findTopp(-521); // pointer to mother
    vbp.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vbp.momentum_flag      = 1;     // momentum in vertex fit?
    vbp.check_wire         = 1;     // wire constraint?
    vbp.verb               = verb;  // verbosity 
    vbp.min_mass           = MB_LO;// minimum mass
    vbp.max_mass           = MB_HI;// maximum mass
    vbp.min_dz_from_wire   = .25;    // z-cut from primary vertex
    vbp.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcBm1->setVertex(vbp);          // set
 
    attribute jp;
    jp.topp=kinema::findTopp(443);
    //jp.lund    = 443;// set MC id
    //jp.mother  = -521;// set MC id
    jp.fit     = 2;
    jp.matchTOPP =1;
    jp.mass    = jp.topp->mass;
    jp.minP    = 0;
    jp.maxP    = 200;
    jp.minMass = 2.7;
    jp.maxMass = 3.3;
    rcBm1->addDaughter(jp);
  
    attribute k;
    k.topp=kinema::findTopp(-321);
    //k.lund         = -321;// set MC id
    //k.mother       = -521;// set MC id
    k.mass         = k.topp->mass;
    k.minP         = -200;
    k.maxP         = 0;
    k.minHIO       = 15;
    k.minHSI       = 6;
    k.minPt        = 0.3;
    k.minLRK       = .0001;    
    k.useVdsOnce   = 1;
    k.useMainOnce  = 1;
    rcBm1->addDaughter(k);

    if (verb) rcBm1->printDaughters();
  }
  rcBm1->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B-1 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}



//
//  B+ -> J/Psi K+ Pi+ Pi-
// 
int Bplus__JPsi_Kplus_Piplus_Piminus() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B+2  <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBp2) {
    cout<<"initializing B+2"<<endl;
    rcBp2=new rcX("Bp2", "B+ -> J/Psi K+ Pi+ Pi-", verb);

    decay bp;
    bp.fit         = 3;    // = short decay
    bp.book_tables = 1;    // 1/0
    bp.rlevel      = 0;    // 0=full, 1=rtra, 2=mtra
    bp.verb        = verb;
    rcBp2->setMother(bp);  // set
    vertexAttrib vbp;
    vbp.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbp.topp               = kinema::findTopp(521); // pointer to mother
    vbp.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vbp.momentum_flag      = 1;     // momentum in vertex fit?
    vbp.check_wire         = 1;     // wire constraint?
    vbp.verb               = verb;  // verbosity 
    vbp.min_mass           = MB_LO;// minimum mass
    vbp.max_mass           = MB_HI;// maximum mass
    vbp.min_dz_from_wire   = .25;    // z-cut from primary vertex
    vbp.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcBp2->setVertex(vbp);          // set
 
    attribute jp;
    jp.topp=kinema::findTopp(443);
    //jp.lund    = 443;// set MC id
    //jp.mother  = 521;// set MC id
    jp.fit     = 2;
    jp.matchTOPP=1;
    jp.mass    = jp.topp->mass;
    jp.minP    = 0;
    jp.maxP    = 200;
    jp.minMass = 2.7;
    jp.maxMass = 3.3;
    rcBp2->addDaughter(jp);
  
    attribute k;
    k.topp=kinema::findTopp(321);
    //k.lund         = 321;// set MC id
    //k.mother       = 521;// set MC id
    k.mass         = k.topp->mass;
    k.minP         = 0;
    k.maxP         = 200;
    k.minHIO       = 15;
    k.minHSI       = 6;
    k.minPt        = 0.3;
    k.minLRK      = .0001;    
    k.useVdsOnce   = 1;
    k.useMainOnce  = 1;
    rcBp2->addDaughter(k);

    attribute pi1;
    pi1.topp=kinema::findTopp(211);
    //pi1.lund         = 211;// set MC id
    //pi1.mother       = 521;// set MC id
    pi1.mass         = pi1.topp->mass;
    pi1.minP         = 0;
    pi1.maxP         = 200;
    pi1.minHIO       = 15;
    pi1.minHSI       = 6;
    pi1.minPt        = 0.15;
    pi1.minLRPI      = .0001;
    pi1.useVdsOnce   = 1;
    pi1.useMainOnce  = 1;
    rcBp2->addDaughter(pi1);

    attribute pi2;
    pi2.topp=kinema::findTopp(-211);
    //pi2.lund         = -211;// set MC id
    //pi2.mother       = 521;// set MC id
    pi2.mass         = pi2.topp->mass;
    pi2.minP         = -200;
    pi2.maxP         = 0;
    pi2.minHIO       = 15;
    pi2.minHSI       = 6;
    pi2.minPt        = 0.15;
    pi2.minLRPI      = .0001;
    pi2.useVdsOnce   = 1;
    pi2.useMainOnce  = 1;
    rcBp2->addDaughter(pi2);

    if (verb) rcBp2->printDaughters();
  }
  rcBp2->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B+2 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}



//
//  B- -> J/Psi K- Pi+ Pi-
// 
int Bminus__JPsi_Kminus_Piplus_Piminus() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B-2  <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBm2) {
    cout<<"initializing B-2"<<endl;
    rcBm2=new rcX("Bm2", "B- -> J/Psi K- Pi+ Pi-", verb);

    decay bm;
    bm.fit         = 3;    // = short decay
    bm.book_tables = 1;    // 1/0
    bm.rlevel      = 0;    // 0=full, 1=rtra, 2=mtra
    bm.verb        = verb;
    rcBm2->setMother(bm);  // set
    vertexAttrib vbm;
    vbm.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbm.topp               = kinema::findTopp(-521); // pointer to mother
    vbm.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vbm.momentum_flag      = 1;     // momentum in vertex fit?
    vbm.check_wire         = 1;     // wire constraint?
    vbm.verb               = verb;  // verbosity 
    vbm.min_mass           = MB_LO;// minimum mass
    vbm.max_mass           = MB_HI;// maximum mass
    vbm.min_dz_from_wire   = .25;    // z-cut from primary vertex
    vbm.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcBm2->setVertex(vbm);          // set
 
    attribute jp;
    jp.topp=kinema::findTopp(443);
    //jp.lund    = 443;// set MC id
    //jp.mother  = -521;// set MC id
    jp.fit     = 2;
    jp.matchTOPP=1;
    jp.mass    = jp.topp->mass;
    jp.minP    = 0;
    jp.maxP    = 200;
    jp.minMass = 2.7;
    jp.maxMass = 3.3;
    rcBm2->addDaughter(jp);
  
    attribute k;
    k.topp=kinema::findTopp(-321);
    //k.lund         = 321;// set MC id
    //k.mother       = -521;// set MC id
    k.mass         = k.topp->mass;
    k.minP         = -200;
    k.maxP         = 0;
    k.minHIO       = 15;
    k.minHSI       = 6;
    k.minPt        = 0.3;
    k.minLRK      = .0001;    
    k.useVdsOnce   = 1;
    k.useMainOnce  = 1;
    rcBm2->addDaughter(k);

    attribute pi1;
    pi1.topp=kinema::findTopp(211);
    //pi1.lund         = 211;// set MC id
    //pi1.mother       = -521;// set MC id
    pi1.mass         = pi1.topp->mass;
    pi1.minP         = 0;
    pi1.maxP         = 200;
    pi1.minHIO       = 15;
    pi1.minHSI       = 6;
    pi1.minPt        = 0.15;
    pi1.minLRPI      = .0001;    
    pi1.useVdsOnce   = 1;
    pi1.useMainOnce  = 1;
    rcBm2->addDaughter(pi1);

    attribute pi2;
    pi2.topp=kinema::findTopp(-211);
    //pi2.lund         = -211;// set MC id
    //pi2.mother       = -521;// set MC id
    pi2.mass         = pi2.topp->mass;
    pi2.minP         = -200;
    pi2.maxP         = 0;
    pi2.minHIO       = 15;
    pi2.minHSI       = 6;
    pi2.minPt        = 0.15;
    pi2.minLRPI      = .0001;    
    pi2.useVdsOnce   = 1;
    pi2.useMainOnce  = 1;
    rcBm2->addDaughter(pi2);

    if (verb) rcBm2->printDaughters();
  }
  rcBm2->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B-2 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}




//
//  B0 -> J/Psi K0
// 
int B0__GOLDEN() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  gold  <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcGold) {
    cout<<"initializing GOLD"<<endl;
    rcGold=new rcX("gold", "B0 -> J/Psi Ks", verb);

    decay gold;
    gold.fit         = 3;    // = short decay
    gold.book_tables = 1;    // 1/0
    gold.rlevel      = 0;    // 0=full, 1=rtra, 2=mtra
    gold.verb        = verb;
    rcGold->setMother(gold);  // set
    vertexAttrib vgold;
    vgold.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vgold.topp               = kinema::findTopp(511); // pointer to mother
    vgold.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vgold.momentum_flag      = 1;     // momentum in vertex fit?
    vgold.check_wire         = 1;     // wire constraint?
    vgold.verb               = verb;  // verbosity 
    vgold.min_mass           = MB_LO;// minimum mass
    vgold.max_mass           = MB_HI;// maximum mass
    vgold.min_dz_from_wire   = .25;    // z-cut from primary vertex
    vgold.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcGold->setVertex(vgold);          // set
 
    attribute jp;
    jp.topp=kinema::findTopp(443);
    //jp.lund    = 443;// set MC id
    //jp.mother  = 511;// set MC id
    jp.fit     = 2;
    jp.matchTOPP =1;
    jp.mass    = jp.topp->mass;
    jp.minP    = 0;
    jp.maxP    = 200;
    jp.minMass = 2.7;
    jp.maxMass = 3.3;
    rcGold->addDaughter(jp);
  
    attribute ks;
    ks.topp=kinema::findTopp(310);
    //ks.lund         = 310;// set MC id
    //ks.mother       = 511;// set MC id
    ks.mass         = ks.topp->mass;
    ks.fit          =3;
    ks.matchTOPP    =1;
    ks.minP         = 0;
    ks.maxP         = 200;
    ks.minMass      = .47;
    ks.maxMass      = .53;
    ks.minPt        = 0.3;    
    rcGold->addDaughter(ks);

    if (verb) rcGold->printDaughters();
  }
  rcGold->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  gold <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}





//
//  B0 -> J/Psi K+ Pi-
// 
int B0__JPsi_Kplus_Piminus() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B0a  <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcB0a) {
    cout<<"initializing B0a"<<endl;
    rcB0a=new rcX("B0a", "B0 -> J/Psi K+ Pi-", verb);

    decay b0a;
    b0a.fit         = 3;    // = short decay
    b0a.book_tables = 1;    // 1/0
    b0a.rlevel      = 0;    // 0=full, 1=rtra, 2=mtra
    b0a.verb        = verb;
    rcB0a->setMother(b0a);  // set
    vertexAttrib vb0a;
    vb0a.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vb0a.topp               = kinema::findTopp(511); // pointer to mother
    vb0a.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vb0a.momentum_flag      = 1;     // momentum in vertex fit?
    vb0a.check_wire         = 1;     // wire constraint?
    vb0a.verb               = verb;  // verbosity 
    vb0a.min_mass           = MB_LO;// minimum mass
    vb0a.max_mass           = MB_HI;// maximum mass
    vb0a.min_dz_from_wire   = .25;    // z-cut from primary vertex
    vb0a.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcB0a->setVertex(vb0a);          // set
 
    attribute jp;
    jp.topp=kinema::findTopp(443);
    //jp.lund    = 443;// set MC id
    //jp.mother  = 511;// set MC id
    jp.fit     = 2;
    jp.matchTOPP =1;
    jp.mass    = jp.topp->mass;
    jp.minP    = 0;
    jp.maxP    = 200;
    jp.minMass = 2.7;
    jp.maxMass = 3.3;
    rcB0a->addDaughter(jp);
  
    attribute k;
    k.topp=kinema::findTopp(321);
    //k.lund         = 321;// set MC id
    //k.mother       = 511;// set MC id
    k.mass         = k.topp->mass;
    k.minP         = 0;
    k.maxP         = 200;
    k.minHIO       = 15;
    k.minHSI       = 6;
    k.minPt        = 0.3;
    k.minLRK      = .0001;    
    k.useVdsOnce   = 1;
    k.useMainOnce  = 1;
    rcB0a->addDaughter(k);

    attribute pi;
    pi.topp=kinema::findTopp(-211);
    //pi.lund         = -211;// set MC id
    //pi.mother       = 511;// set MC id
    pi.mass         = pi.topp->mass;
    pi.minP         = -200;
    pi.maxP         = 0;
    pi.minHIO       = 15;
    pi.minHSI       = 6;
    pi.minPt        = 0.15;
    pi.minLRPI      = .0001;    
    pi.useVdsOnce   = 1;
    pi.useMainOnce  = 1;
    rcB0a->addDaughter(pi);

    if (verb) rcB0a->printDaughters();
  }
  rcB0a->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B0a <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}




//
//  ~B0 -> J/Psi K- Pi+
// 
int B0__JPsi_Kminus_Piplus() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B0b  <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcB0b) {
    cout<<"initializing B0b"<<endl;
    rcB0b=new rcX("B0b", "B0 -> J/Psi K- Pi+", verb);

    decay b0b;
    b0b.fit         = 3;    // = short decay
    b0b.book_tables = 1;    // 1/0
    b0b.rlevel      = 0;    // 0=full, 1=rtra, 2=mtra
    b0b.verb        = verb;
    rcB0b->setMother(b0b);  // set
    vertexAttrib vb0b;
    vb0b.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vb0b.topp               = kinema::findTopp(-511); // pointer to mother
    vb0b.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vb0b.momentum_flag      = 1;     // momentum in vertex fit?
    vb0b.check_wire         = 1;     // wire constraint?
    vb0b.verb               = verb;  // verbosity 
    vb0b.min_mass           = MB_LO;// minimum mass
    vb0b.max_mass           = MB_HI;// maximum mass
    vb0b.min_dz_from_wire   = .25;    // z-cut from primary vertex
    vb0b.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcB0b->setVertex(vb0b);          // set
 
    attribute jp;
    jp.topp=kinema::findTopp(443);
    //jp.lund    = 443;// set MC id
    //jp.mother  = 511;// set MC id
    jp.fit     = 2;
    jp.matchTOPP    =1;
    jp.mass    = jp.topp->mass;
    jp.minP    = 0;
    jp.maxP    = 200;
    jp.minMass = 2.7;
    jp.maxMass = 3.3;
    rcB0b->addDaughter(jp);
  
    attribute k;
    k.topp=kinema::findTopp(-321);
    //k.lund         = -321;// set MC id
    //k.mother       = -511;// set MC id
    k.mass         = k.topp->mass;
    k.minP         = -200;
    k.maxP         = 0;
    k.minHIO       = 15;
    k.minHSI       = 6;
    k.minPt        = 0.3;
    k.minLRK      = .0001;    
    k.useVdsOnce   = 1;
    k.useMainOnce  = 1;
    rcB0b->addDaughter(k);

    attribute pi;
    pi.topp=kinema::findTopp(211);
    //pi.lund         = 211;// set MC id
    //pi.mother       = -511;// set MC id
    pi.mass         = pi.topp->mass;
    pi.minP         = 0;
    pi.maxP         = 200;
    pi.minHIO       = 15;
    pi.minHSI       = 6;
    pi.minPt        = 0.15;
    pi.minLRPI      = .0001;   
    pi.useVdsOnce   = 1;
    pi.useMainOnce  = 1;
    rcB0b->addDaughter(pi);

    if (verb) rcB0b->printDaughters();
  }
  rcB0b->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  B0b <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


//int Bminus__D0_Piminus();
//
//  B+ -> ~D0 pi+
// 
int Bplus__D0bar_Piplus() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>  B+D0pi+  <<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBDpi1) {
    cout<<"initializing B0+D0pi+"<<endl;
    rcBDpi1=new rcX("BDpi1", "B+ -> ~D0 Pi+", verb);

    decay bp;
    bp.fit         = 3;      // = short decay
    bp.book_tables = 1;      // 1/0
    bp.rlevel      = 0;      // 0=full, 1=rtra, 2=mtra
    bp.verb        = verb;  
    rcBDpi1->setMother(bp);  // set
    vertexAttrib vbp;
    vbp.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbp.topp               = kinema::findTopp(521); // pointer to mother
    vbp.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vbp.momentum_flag      = 1;     // momentum in vertex fit?
    vbp.check_wire         = 1;     // wire constraint?
    vbp.verb               = verb;  // verbosity 
    vbp.min_mass           = MB_LO; // minimum mass
    vbp.max_mass           = MB_HI; // maximum mass
    vbp.min_dz_from_wire   = .25;   // z-cut from primary vertex
    vbp.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcBDpi1->setVertex(vbp);        // set
 
    attribute d0;
    d0.topp=kinema::findTopp(-421);
    //d0.lund       = -421; // set MC id
    //d0.mother     =  521; // set MC id
    d0.fit        = 3;
    d0.matchTOPP  = 1;
    d0.mass       = d0.topp->mass;
    d0.matchTOPP  = 1;
    d0.minMass    = 1.7;
    d0.maxMass    = 2.0;
    rcBDpi1->addDaughter(d0);
  
    attribute pi;
    pi.topp=kinema::findTopp(211);
    //pi.lund         = 211; // set MC id
    //pi.mother       = 521; // set MC id
    pi.mass         = pi.topp->mass;
    pi.minP         = 0;
    pi.maxP         = 200;
    pi.minHIO       = 15;
    pi.minHSI       = 6;
    pi.minPt        = 0.15;
    pi.minLRPI     = .0001;    
    pi.useVdsOnce   = 1;
    pi.useMainOnce  = 1;
    rcBDpi1->addDaughter(pi);

    if (verb) rcBDpi1->printDaughters();
  }
  rcBDpi1->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>  B+D0pi+  <<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


//
//  B- -> D0 pi-
// 
int Bminus__D0_Piminus() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>  B-D0pi-  <<<<<<<<<<<<<<<<<"<<endl;
  if (!rcBDpi2) {
    cout<<"initializing B0-D0pi-"<<endl;
    rcBDpi2=new rcX("BDpi2", "B- -> D0 pi-", verb);

    decay bp;
    bp.fit         = 3;      // = short decay
    bp.book_tables = 1;      // 1/0
    bp.rlevel      = 0;      // 0=full, 1=rtra, 2=mtra
    bp.verb        = verb;  
    rcBDpi2->setMother(bp);  // set
    vertexAttrib vbp;
    vbp.rlevel             = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vbp.topp               = kinema::findTopp(-521); // pointer to mother
    vbp.max_vertex_chi2    = 10.0;  // limit for vertex acceptance 
    vbp.momentum_flag      = 1;     // momentum in vertex fit?
    vbp.check_wire         = 1;     // wire constraint?
    vbp.verb               = verb;  // verbosity 
    vbp.min_mass           = MB_LO; // minimum mass
    vbp.max_mass           = MB_HI; // maximum mass
    vbp.min_dz_from_wire   = .25;   // z-cut from primary vertex
    vbp.max_drho_from_wire = 0.02;  // max impact parameter at primary vertex
    rcBDpi2->setVertex(vbp);        // set
 
    attribute d0;
    d0.topp=kinema::findTopp(421);
    //d0.lund       = 421; // set MC id
    //d0.mother     = 521; // set MC id
    d0.fit        = 3;
    d0.matchTOPP =1;
    d0.mass       = d0.topp->mass;
    d0.matchTOPP  = 1;
    d0.minMass    = 1.7;
    d0.maxMass    = 2.0;
    rcBDpi2->addDaughter(d0);
  
    attribute pi;
    pi.topp=kinema::findTopp(-211);
    //pi.lund         = -211; // set MC id
    //pi.mother       = 521; // set MC id
    pi.mass         = pi.topp->mass;
    pi.minP         = -200;
    pi.maxP         = 0;
    pi.minHIO       = 15;
    pi.minHSI       = 6;
    pi.minPt        = 0.15;
    pi.minLRPI     = .0001;    
    pi.useVdsOnce   = 1;
    pi.useMainOnce  = 1;
    rcBDpi2->addDaughter(pi);

    if (verb) rcBDpi2->printDaughters();
  }
  rcBDpi2->rc();

  if (verb) cout<<">>>>>>>>>>>>>>>>>>  B-D0pi-  <<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}



/////////////////////////////////////////////////////////////////////////
////////////////////////// SELECTION ////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

//
//  select displaced J/Psi with at least one additional prong
// 
int B__JPsi_X_RC() {

  int verb =0;
  int total=0;

  if (verb) cout<<">>>>>>>>>>>>>>>>>  B select <<<<<<<<<<<<<<<<<<"<<endl;

  static vertexmass   vert;
  static vertexAttrib vattrib;
  static attribute    jpsi;
  static attribute    x, ks;
  static bool         first=true;
  if (first) {
    ///////////////////////////// ALL CUTS HERE ///////////////////////////////
    // -- description of vertex fitting
    vattrib.rlevel           = 1;      // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vattrib.momentum_flag    = 0;      // momentum in fit?
    vattrib.max_vertex_chi2  = 3.0 ;   // upper limit for vertex acceptance 
    vattrib.verb             = 0;      // vertexing printout level
    vattrib.min_dz_from_wire = 0.25;   // minimum decay distance for B (cm)
    vattrib.check_wire       = 1;      // check primary vertex/wire
    vert.setVertex(vattrib);           // set

    // -- description of J/Psi
    jpsi.fit     = 2;                       // short decay
    jpsi.minMass = 2.5;                     // min reco mass
    jpsi.maxMass = 3.5;                     // max reco mass
    
    // -- description of candidates for common vertexing with jpsi
    x.fit        = 0; // charged track
    x.minHSI     = 5; // >>> require only vds <<< !!!

    // -- description of ks
    ks.fit     =    3; // long decay
    ks.minMass = .470; // min reco mass
    ks.maxMass = .530; // max reco mass
    ///////////////////////////// ALL CUTS HERE ///////////////////////////////
    first=false;
  }

  // -- find jpsi's and other candidates
  static vector<ArtePointer<RTRA> > jpsi_list(10);
  static vector<ArtePointer<RTRA> > x_list(100);
  static vector<ArtePointer<RTRA> >::iterator jpsi_it;
  static vector<ArtePointer<RTRA> >::iterator x_it;
  if (jpsi_list.size()) jpsi_list.clear();
  if (x_list.size())    x_list.clear();

  //static ArtePointer<MTRA> mB;
  //for (ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
  //irtra!=ArteTable<RTRA>::end(); ++irtra) {
  //if (!jpsi.checkAttributes(*irtra)) {
  //mB=(*(*irtra)->mtra.begin())->btra;
  //}
  //}


  for (ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
       irtra!=ArteTable<RTRA>::end(); ++irtra) {
    if (!jpsi.checkAttributes(*irtra)) {
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1000 ),   0., 1.); // jpsi found
      float fpar[]={(*irtra)->xf,(*irtra)->yf, (*irtra)->zf, (*irtra)->txf, (*irtra)->tyf};
      vert.checkWire(fpar);  // find used wire
      if ((*irtra)->zf-vert.getDecayWire()->z>vattrib.min_dz_from_wire)
	rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1000 ),  1., 1.); // jpsi with proper displacement
      float zjp=vert.vertexX(2);
      rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1005 ), zjp, 1.);
      jpsi_list.push_back(*irtra);
    }
    if (!x.checkAttributes(*irtra))  {
      //int lund=abs((*(*irtra)->mtra.begin())->topp->lund);
      //if (((*(*irtra)->mtra.begin())->btra==mB  ||
      //(*(*irtra)->mtra.begin())->btra->btra==mB &&
      //(abs((*(*irtra)->mtra.begin())->btra->topp->lund)==113 ||
      //abs((*(*irtra)->mtra.begin())->btra->topp->lund)==313)
      //)
      //&& (lund==211 || lund==321))
      x_list.push_back(*irtra);
    }
    if (!ks.checkAttributes(*irtra)) x_list.push_back(*irtra);
  }
  if (verb>1) {
    cout<<"Total Jpsi="<<jpsi_list.size();
    for (jpsi_it=jpsi_list.begin(); jpsi_it!=jpsi_list.end(); ++jpsi_it)  
      cout<<" "<<(*jpsi_it)<<" ";
    cout<<endl;
    cout<<"Total tracks="<<x_list.size();
    for (x_it=x_list.begin(); x_it!=x_list.end(); ++x_it)  
      cout<<" "<<(*x_it)<<" ";
    cout<<endl;
  }
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1018 ), float(jpsi_list.size()), 1.); 
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1019 ),    float(x_list.size()), 1.); 

  if (jpsi_list.size() && x_list.size()) {

    // -- search for multiprong vertex
    for (jpsi_it=jpsi_list.begin(); jpsi_it!=jpsi_list.end(); ++jpsi_it) {    

      // -- intialize search
      vert.removeAllVertexTracks();
      ArtePointer<RVER> ever=(*jpsi_it)->ever;

      // -- add J/Psi leptons to vertex track bank
      for (ArteRelationIterator<RREL> irrel=ever->rrel.begin();
	   irrel!=ever->rrel.end(); ++irrel) {
	vert.addVertexTrack((*irrel)->rtra);
      }
      if (vert.getNVertexTrack()!=2) { 
	cout<<"B__JPsi_X_RC: something wrong, jpsi without 2 tracks"<<endl; 
	continue; 
      }
      vert.fitVertex();
      float chi0=vert.vertexChi2();
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1002 ), chi0, 1.);

      // -- add vertex candidates into vertex track bank
      static vector<ArtePointer<RTRA> > prong_list(10);
      if (prong_list.size()) prong_list.clear();
      for (x_it=x_list.begin(); x_it!=x_list.end(); ++x_it) {
	// -- check if candidate uses any segment already used by jpsi leptons
	if (segment::TrackDiff((*x_it),(*jpsi_it))) continue;
	// -- add track and check vertex
	vert.addVertexTrack(*x_it);
	int   vflg=vert.fitVertex();
	float chi2=vert.vertexChi2();
	rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1003 ), chi2, 1.);
	// -- accept if:
	//        - chi2 improves
	//        - chi2 small
	if (chi2<chi0 || chi2<vattrib.max_vertex_chi2) { 
	  rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1004 ), chi2, 1.);
	  prong_list.push_back(*x_it); 
	  chi0=chi2; 
	}
	vert.removeLastVertexTrack();
      }

      // -- book multiprong vertex (RVER) and fill dq histos
      if (!prong_list.size()) continue;
      for (x_it=prong_list.begin(); x_it!=prong_list.end(); ++x_it) 
	vert.addVertexTrack((*x_it));
      vert.fitVertex();
      float zv=vert.vertexX(2);
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1011 ), zv, 1.);
      float dz0=vert.vertexX(2) - vert.getDecayWire()->z;
      if (dz0<vattrib.min_dz_from_wire) continue;
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1000 ),   2., 1.); // event selected
      float nt=float(vert.getNVertexTrack());
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1001 ),   nt, 1.);
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1006 ),   zv, 1.);
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1014 ),  dz0, 1.);
      float wx=vert.getDecayWire()->x;
      float wy=vert.getDecayWire()->y;
      float wz=vert.getDecayWire()->z;
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1015 ),    wz,     1.);
      rhp_hfill(DQ_ID( COMP_RECO,LVL_REC, 1016 ),  wx, wy, 1.);
      vert.bookRVER();
      total++;
      if (verb) {
	ArtePointer<RVER> Rver(ArteTable<RVER>::size());
	cout<<"found vertex="<<Rver<<": "<<Rver->xv<<","<<Rver->yv<<","<<Rver->zv<<endl;
	cout<<"RC: prongs found="<<Rver->nt<<endl;
      }
    }

  }
  
  // -- evaluate reco efficiency (MC only)
#ifdef USE_ROOT_DD
  if (ArteTable<MTRA>::size()) B__JPsi_X_MCeff(vattrib, vert, total);
#endif

  if (verb) cout<<">>>>>>>>>>>>>>>>>  B select <<<<<<<<<<<<<<<<<<"<<endl;

  return total;
}



//
//  MC: Monitor reco efficiency for  J/Psi + prongs 
//
int B__JPsi_X_MCeff(vertexAttrib &vattrib, vertexmass &vert, int total) {
  if (!ArteTable<MTRA>::size()) return 1;
  int verb=0;

#ifdef USE_ROOT_DD
  static bool first=true;
  if (first) {
    first=false;
    rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 1007 ),"B selection MC: Event efficiency",  6,  -.5,  5.5 ); 
    rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 1008 ),"B selection MC: B prongs in vds", 10,  -.5,  9.5 ); 
    rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 1009 ),"B selection MC: prongs/vertex", 10,  -.5,  9.5 ); 
    rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 1010 ),"B selection MC: B prongs in sel. events", 10,  -.5,  9.5 );
    rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 1012 ),"B selection MC: B vertex resoltion", 50,  -1,  1 ); 
    rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 1013 ),"B selection MC: Decay length from MC v0", 50,  -1,  5 ); 
    rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 1017 ),"B selection MC: Decay length (MC truth)", 50,0,5); 
    rhp_hbook1( DQ_ID( COMP_RECO,LVL_REC, 1020 ),"B selection MC: MC v0 - RC v0 (MC)", 50,  -5,  5 ); 
  }

  // -- check if J/Psi from B+, B0 exists
  ArteTable<MTRA>::iterator imtra=ArteTable<MTRA>::begin();
  ArteTable<MTRA>::iterator Bmtra=ArteTable<MTRA>::end();
  for (; imtra!=ArteTable<MTRA>::end(); ++imtra) {
    if ((*imtra)->topp->lund==443 &&
	(abs((*imtra)->btra->topp->lund)==521 || 
	 abs((*imtra)->btra->topp->lund)==511)) { 
      Bmtra=(*imtra)->btra; break; 
    }
    if ((*imtra)->topp->lund==443 &&
	(abs((*imtra)->btra->btra->topp->lund)==521 ||
	 abs((*imtra)->btra->btra->topp->lund)==511)) {
      Bmtra=(*imtra)->btra->btra; break;
    }
  }
  if (imtra==ArteTable<MTRA>::end()) return 0;
  ArteTable<MTRA>::iterator JPmtra=imtra;

  // -- check J/Psi decays to leptons
  int vdsflg[]={0,0}, i=0;
  for (ArteRelationIterator<MTRA> jmtra=(*JPmtra)->etra.begin();
       jmtra!=(*JPmtra)->etra.end(); ++jmtra) {
    int lund=abs((*jmtra)->topp->lund);
    if (lund!=11 && lund!=13) return 2;
    if (i==2)                 return 2;
    for (ArteRelationIterator<RTRA> irtra=(*jmtra)->rtra.begin();
	 irtra!=(*jmtra)->rtra.end(); ++irtra) {
      if ((*irtra)->hsi>0) vdsflg[i++]=1;
    }
  }
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1007 ), 0., 1.);

  // -- check if both leptons have reconstructed vds
  if (!vdsflg[0] || !vdsflg[1]) return 3;
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1007 ), 1., 1.);

  // -- check if booked in rtra (vertexing)
  if (JPmtra->rtra.begin()==JPmtra->rtra.end()) return 4;
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1007 ), 2., 1.); 

  // -- check if at least one additional prong with vds points to J/Psi vertex
  ArteRelationIterator<MTRA> jmtra=Bmtra->etra.begin();
  static vector<ArtePointer<MTRA> > prong_list(10); 
  if (prong_list.size()) prong_list.clear();
  for (; jmtra!=Bmtra->etra.end(); ++jmtra) {
    // -- reject jpsi
    if ((*jmtra)->topp->lund==443 || (*jmtra)->topp->lund==20443) continue;
    // -- check for nondecaying daughters
    int lund=abs((*jmtra)->topp->lund);
    if (lund==11 || lund==13 || lund==211 || lund==321 || lund==2212) {
      ArteRelationIterator<RTRA> irtra=(*jmtra)->rtra.begin();
      for (; irtra!=(*jmtra)->rtra.end(); ++irtra) if ((*irtra)->hsi>0) { prong_list.push_back(*jmtra); break; }
    }
    // -- check for Ks
    else if (lund==310) {
      for (ArteRelationIterator<MTRA> ipi=(*jmtra)->etra.begin();
	   ipi!=(*jmtra)->etra.end(); ++ipi) {
	if (abs((*ipi)->topp->lund)==211) 
	  for (ArteRelationIterator<RTRA> irtra=(*ipi)->rtra.begin(); 
	       irtra!=(*ipi)->rtra.end(); ++irtra) 
	    if ((*irtra)->hsi>0) {  prong_list.push_back(*ipi); break; };
      }
    }
    // -- check for resonance
    else if ((*jmtra)->flag==2) {
      ArteRelationIterator<MTRA> kmtra=(*jmtra)->etra.begin();
      for (; kmtra!=(*jmtra)->etra.end(); ++kmtra) {
	ArteRelationIterator<RTRA> irtra=(*kmtra)->rtra.begin();
	for (; irtra!=(*kmtra)->rtra.end(); ++irtra) if ((*irtra)->hsi>0) {  prong_list.push_back(*kmtra); break; }
      }
    }
  }
  int nprong=prong_list.size()+2;
  if (nprong==2) return 5;
  if (verb) cout<<"MC: prongs found = "<<prong_list.size()<<endl;
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1007 ),            3., 1.);  

  // -- check if real decay length (dz) passes cut
  float dz0_mc =(*JPmtra)->bver->z - (*Bmtra)->bver->z;
  if (dz0_mc>vattrib.min_dz_from_wire) {
    rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1007 ), 4., 1.);
    rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1008 ), float(nprong), 1.); 
  }


  // -- check if event selected
  if (!total) return 7;
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1007 ), 5., 1.); 
  int sprong=vert.getNVertexTrack();
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1009 ), float(sprong), 1.); 
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1010 ), float(nprong), 1.); 

  // -- plot vertex: z_RC - z_MC (vertex resolution)
  float dz= vert.vertexX(2) - (*(*Bmtra)->ever.begin())->z;
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1012 ), dz, 1.); 

  // -- plot vertex: z_RC - v0_MC
  float dz0_rc= vert.vertexX(2)   - (*Bmtra)->bver->z;
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1013 ), dz0_rc, 1.); 
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1017 ), dz0_mc, 1.);
  float dz0_mcrc =  (*Bmtra)->bver->z - vert.getDecayWire()->z;
  rhp_hf1( DQ_ID( COMP_RECO,LVL_REC, 1020 ), dz0_mcrc, 1.);

#endif

  return 0;
}




// 
//  returns number of vertices with a kaon
// 
int B__K(ArteTable<RVER>::iterator iv1, ArteTable<RVER>::iterator iv2) {
  int ret=0;
  ////////// cut //////////
  float minLRK=1e-4;
  ////////// cut //////////
  for (ArteTable<RVER>::iterator irver=iv1; irver!=iv2; ++irver) {
    int nk=0;
    for (ArteRelationIterator<RREL> irrel=(*irver)->rrel.begin();
	 irrel!=(*irver)->rrel.end(); ++irrel) {
      if ((*irrel)->rtra->lrk>minLRK) nk++;
      if ((*irrel)->rtra->fit%100==3) nk++;
    }
    rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1021 ),float(nk), 1.);
    if (nk) {
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1000 ), 3., 1.); // kaon found
      ret++;
    }
  }
  return ret;
}


//
// returns 1 if good vertex, 2 if kaon in vertex, 0 otherwise
//
int
selectB()
{
  ARTE_mark( "rcXB selection 0", 0 );
  int ret=0;

  //=== check if rich likelihoods in RTRA are filled,
  //  take maximal likelihoods from  RTRI otherwise

  int richpid=1;
  if (ArteTable<RTRA>::size()!=0){
    ArtePointer<RTRA> rtra(1);
    if (rtra->lre<0&&rtra->lrmu<0&&rtra->lrpi<0&&rtra->lrk<0&&rtra->lrp<0){
      int mr=maximizeRTRA();
      richpid=-1;
    }
  }

  ArteTable<RVER>::iterator iv1=ArteTable<RVER>::end();
  if (B__JPsi_X_RC())       ret=1;
  ArteTable<RVER>::iterator iv2=ArteTable<RVER>::end();
  if (ret && B__K(iv1,iv2)) ret=2; 
  //else                      ret=0;

  // -- make FULL reconstruction
  if (ret) B();


  //== reset rich likelihoods in RTRA to default value if no rich likelihood
  //   was filled before calling this routine

  if (richpid<0){
       rcResetRichRtra(-1);
  }

  ARTE_mark( "rcXB selection 1", 0 );
  //cout<<"selectB ret="<<ret<<endl;
  return ret;
}


//
//  call reconstruction channels
//
void 
B() 
{
//   Bplus__JPsi_Kplus();
//   Bplus__JPsi_Kplus_Piplus_Piminus();
  
//   Bminus__JPsi_Kminus();
//   Bminus__JPsi_Kminus_Piplus_Piminus();
  
//   B0__GOLDEN();
//   B0__JPsi_Kplus_Piminus();
//   B0__JPsi_Kminus_Piplus();

  Bplus__D0bar_Piplus();
  Bminus__D0_Piminus();
}



//
//  saving ROOT histograms (off-line only)
//
void 
stopB() 
{
  if (rcBp1)  delete rcBp1;
  if (rcBm1)  delete rcBm1;
  if (rcBp2)  delete rcBp2;
  if (rcBm2)  delete rcBm2;
  if (rcGold) delete rcGold;
  if (rcB0a)  delete rcB0a;
  if (rcB0b)  delete rcB0b;
  if (rcBDpi1)  delete rcBDpi1;
  if (rcBDpi2)  delete rcBDpi2;
}
