// *****************************************************************************
//
// source:    Selection of Ds using the following channels:
//               Ds+ :  K+ K- pi+, phi pi+, K* K+
//               Ds- :  K+ K- pi-, phi pi-, K* K-
//            Returns:
//               1 upon success, 0 otherwise
//
// dependences: 
//       - the following must be called BEFORE this routine:
//           selectKs(), selectK892(), selectPhi()
//       - RTRA must be filled
//
// type:      source code, selection/reconstruction of Ds 
//
// created:   26 Aug 2000
//	
// author:   Denis Dujmic
//
// changes:
//    26 Aug 2000 (DD) >4.5 mm cut from a prim. vertex for testing
//    06 Sep 2000 (DD) File description added
//    05 Dec 2000 (DD) COde polishing
//    15 Dec 2000 (DD) Call only pipiK channels, older code removal
//    23 Oct 2001 (DS) fill RTRA rich likelihoods only if not filled before
//
// *****************************************************************************
#include "reco/rcX.h"
#include "reco/rcXkinema.h"
#include "rhp.h"
#include "dq/DQ.h"


#ifdef USE_ROOT_DD
#define MDS_LO 1.0
#define MDS_HI 7.0
#else
#define MDS_LO 1.5
#define MDS_HI 2.5
#endif


rcX *rcDs1=0;
rcX *rcDs2=0;
rcX *rcDs3=0;
rcX *rcDs4=0;
rcX *rcDs5=0;
rcX *rcDs6=0;


int Dsplus__Kplus_Kminus_Piplus_RC();
int Dsplus__Phi_Piplus_RC();
int Dsplus__K892_Kplus_RC();
int Dsminus__Kplus_Kminus_Piminus_RC();
int Dsminus__Phi_Piminus_RC();
int Dsminus__K892_Kminus_RC();

int maximizeRTRA();
void rcResetRichRtra(float x=-1);


int Ds() {
  Dsplus__Kplus_Kminus_Piplus_RC();
  Dsminus__Kplus_Kminus_Piminus_RC();
  //Dsplus__Phi_Piplus_RC();
  //Dsminus__Phi_Piminus_RC();
  //Dsminus__K892_Kminus_RC();
  //Dsplus__K892_Kplus_RC();
  return 0;
}

#define VERT_RLEV 2
#define VERT_CHTR 1
#define VERT_MOFL 1




////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////   real reconstruction     //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////


int Dsplus__Kplus_Kminus_Piplus_RC() {
  int verb=0;

  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds1 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcDs1) {
    cout<<"initializing Ds"<<endl;
    rcDs1=new rcX("Ds1", "Ds+ -> K+K-pi+", verb);
    decay ds;
    ds.fit           = 3; // = long decay
    ds.book_tables   = 1; // 1/0
    ds.rlevel        = 0; // 0=full, 1=semi, 2=mtra
    ds.verb          = verb;
    rcDs1->setMother(ds);  // set
    vertexAttrib vds;
    vds.rlevel           = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vds.topp             = kinema::findTopp(431); // pointer to mother
    vds.momentum_flag    = 1;     // use momentum in fit
    vds.max_vertex_chi2  = 10.0;  // limit for vertex acceptance 
    vds.check_wire       = 1;     // require wire constraint
    vds.min_dz_from_wire = .15;   // z-cut from primary vertex
    vds.verb             = verb;
    vds.min_mass         = MDS_LO; // minimum mass
    vds.max_mass         = MDS_HI; // maximum mass
    rcDs1->setVertex(vds);        // set

    attribute k1;
    k1.topp=kinema::findTopp(-321);
    //k1.lund         = -321;
    //k1.mother       = 431;
    k1.mass         = k1.topp->mass;
    k1.minP         = -200;
    k1.maxP         = 0;
    k1.minPt        = 0.3;
    k1.minHSI       = 6;
    k1.minHIO       = 15;
    k1.minLRK       = .1;    
    k1.maxLRPI      = 0.01;
    k1.maxEoP       = 0.8;
    k1.maxCandidate = 10;
    k1.useVdsOnce   = 1;
    k1.useMainOnce  = 1;
    rcDs1->addDaughter(k1);

    attribute k2;
    k2.topp=kinema::findTopp(321);
    //k2.lund=321;
    //k2.mother=431;
    k2.mass         = k2.topp->mass;
    k2.minP         = 0;
    k2.maxP         = 200;
    k2.minHSI       = 6;
    k2.minHIO       = 15;
    k2.minPt        = 0.3;
    k2.minLRK       = .1;    
    k2.maxLRPI      = 0.01;
    k2.maxEoP       = 0.8;
    k2.maxCandidate = 10;
    k2.useVdsOnce   = 1;
    k2.useMainOnce  = 1;
    rcDs1->addDaughter(k2);
  
    attribute pi1;
    pi1.topp=kinema::findTopp(211);
    //pi1.lund         = 211;
    //pi1.mother       = 431;
    pi1.mass         = pi1.topp->mass;
    pi1.minP         = 0;
    pi1.maxP         = 200;
    pi1.minPt        = 0.15;
    pi1.minHSI       = 6;
    pi1.minHIO       = 15;
    pi1.minHRI       = 1;
    pi1.minLREMUPI   = 0.1;
    pi1.maxEoP       = 0.8;
    pi1.maxCandidate = 10;
    pi1.useVdsOnce   = 1;
    pi1.useMainOnce  = 1;
    rcDs1->addDaughter(pi1);

    if (verb) rcDs1->printDaughters();
  }
  rcDs1->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds1 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


int Dsminus__Kplus_Kminus_Piminus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds2 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcDs2) {
    cout<<"initializing Ds"<<endl;
    rcDs2=new rcX("Ds2", "Ds- -> K+K-pi-", verb);
    decay ds;
    ds.fit           = 3; // = short decay
    ds.book_tables   = 1; // 1/0
    ds.rlevel        = 0; // 0=full, 1=semi, 2=mtra
    ds.verb          = verb;
    rcDs2->setMother(ds);  // set
    vertexAttrib vds;
    vds.rlevel           = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vds.topp             = kinema::findTopp(-431); // pointer to mother
    vds.momentum_flag    = 1;     // use momentum in fit
    vds.max_vertex_chi2  = 10.0;  // limit for vertex acceptance 
    vds.momentum_flag    = 1;     //
    vds.check_wire       = 1;     // require wire constraint
    vds.min_dz_from_wire = .15;   // z-cut from primary vertex
    vds.verb             = verb;
    vds.min_mass         = MDS_LO; // minimum mass
    vds.max_mass         = MDS_HI; // maximum mass
    rcDs2->setVertex(vds);         // set


    attribute k1;
    k1.topp=kinema::findTopp(-321);
    //k1.lund         = -321;
    //k1.mother       = -431;
    k1.mass         = k1.topp->mass;
    k1.minP         = -200;
    k1.maxP         = 0;
    k1.minPt        = 0.3;
    k1.minHSI       = 6;
    k1.minHIO       = 15;
    k1.minLRK       = .1;    
    k1.maxLRPI      = 0.01;
    k1.maxEoP       = 0.8;
    k1.maxCandidate = 10;
    k1.useVdsOnce   = 1;
    k1.useMainOnce  = 1;
    rcDs2->addDaughter(k1);


    attribute k2;
    k2.topp=kinema::findTopp(321);
    //k2.lund         = 321;
    //k2.mother       = -431;
    k2.mass         = k2.topp->mass;
    k2.minP         = 0;
    k2.maxP         = 200;
    k2.minHSI       = 6;
    k2.minHIO       = 15;
    k2.minPt        = 0.3;
    k2.minLRK       = .1;    
    k2.maxLRPI      = 0.01;
    k2.maxEoP       = 0.8;
    k2.maxCandidate = 10;
    k2.useVdsOnce   = 1;
    k2.useMainOnce  = 1;
    rcDs2->addDaughter(k2);
  
    attribute pi1;
    pi1.topp=kinema::findTopp(-211);
    //pi1.lund         = -211;
    //pi1.mother       = -431;
    pi1.mass         = pi1.topp->mass;
    pi1.minP         = -200;
    pi1.maxP         = 0;
    pi1.minPt        = 0.15;
    pi1.minHSI       = 6;
    pi1.minHIO       = 15;
    pi1.minHRI       = 1;
    pi1.minLREMUPI   = 0.1;
    pi1.maxEoP       = 0.8;
    pi1.maxCandidate = 10;
    pi1.useVdsOnce   = 1;
    pi1.useMainOnce  = 1;
    rcDs2->addDaughter(pi1);

    if (verb) rcDs2->printDaughters();
  }
  rcDs2->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds2 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


int Dsplus__Phi_Piplus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds3 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcDs3) {
    cout<<"initializing Ds"<<endl;
    rcDs3=new rcX("Ds3", "Ds+ -> K+K-pi+", verb);
    decay ds;
    ds.fit           = 3; // = short decay
    ds.book_tables   = 1; // 1/0
    ds.rlevel        = 0; // 0=full, 1=semi, 2=mtra
    ds.verb          = verb;
    rcDs3->setMother(ds);  // set
    vertexAttrib vds;
    vds.rlevel           = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vds.topp             = kinema::findTopp(431); // pointer to mother
    vds.momentum_flag    = 1;
    vds.max_vertex_chi2  = 10.0;  // limit for vertex acceptance 
    vds.check_wire       = 0;
    vds.min_dz_from_wire = .15;   // z-cut from primary vertex
    vds.verb             = verb;
    vds.min_mass         = MDS_LO; // minimum mass
    vds.max_mass         = MDS_HI; // maximum mass
    rcDs3->setVertex(vds);        // set

    attribute phi;
    //phi.lund      = 333;
    phi.topp      = kinema::findTopp(333);
    phi.fit       = 2;
    phi.mass      = phi.topp->mass;
    phi.matchTOPP = 1;
    phi.maxMass   = 1.05;
    rcDs3->addDaughter(phi);
  
    attribute pi1;
    pi1.topp=kinema::findTopp(211);
    //pi1.lund         = 211;
    //pi1.mother       = 431;
    pi1.mass         = pi1.topp->mass;
    pi1.minP         = 0;
    pi1.maxP         = 200;
    pi1.minPt        = 0.15;
    pi1.minHIO       = 15;
    pi1.minHSI       = 6;
    pi1.minHRI       = 1;
    pi1.minLREMUPI   = 0.1;
    pi1.maxEoP       = 0.8;
    pi1.maxCandidate = 10;
    pi1.useVdsOnce   = 1;
    pi1.useMainOnce  = 1;
    rcDs3->addDaughter(pi1);

    if (verb) rcDs3->printDaughters();
  }
  rcDs3->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds3 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


int Dsminus__Phi_Piminus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds4 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcDs4) {
    cout<<"initializing Ds"<<endl;
    rcDs4=new rcX("Ds4", "Ds+ -> K+K-pi+", verb);
    decay ds;
    ds.fit           = 3; // = short decay
    ds.book_tables   = 1; // 1/0
    ds.rlevel        = 0; // 0=full, 1=semi, 2=mtra
    ds.verb          = verb;
    rcDs4->setMother(ds);  // set
    vertexAttrib vds;
    vds.rlevel           = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vds.topp             = kinema::findTopp(-431); // pointer to mother
    vds.momentum_flag    = 1;
    vds.max_vertex_chi2  = 10.0;  // limit for vertex acceptance 
    vds.check_wire       = 0;
    vds.min_dz_from_wire = .15;   // z-cut from primary vertex
    vds.verb             = verb;
    vds.min_mass         = MDS_LO; // minimum mass
    vds.max_mass         = MDS_HI; // maximum mass
    rcDs4->setVertex(vds);        // set

    attribute phi;
    //phi.lund      = 333;
    phi.topp      = kinema::findTopp(333);
    phi.fit       = 2;
    phi.mass      = phi.topp->mass;
    phi.matchTOPP = 1;
    phi.maxMass   = 1.05;
    rcDs4->addDaughter(phi);
  
    attribute pi1;
    pi1.topp=kinema::findTopp(-211);
    //pi1.lund         = -211;
    //pi1.mother       = -431;
    pi1.mass         = pi1.topp->mass;
    pi1.minP         = -200;
    pi1.maxP         = 0;
    pi1.minPt        = 0.15;
    pi1.minHIO       = 15;
    pi1.minHSI       = 6;
    pi1.minHRI       = 1;
    pi1.minLREMUPI   = 0.1;
    pi1.maxEoP       = 0.8;
    pi1.maxCandidate = 10;
    pi1.useVdsOnce   = 1;
    pi1.useMainOnce  = 1;
    rcDs4->addDaughter(pi1);

    if (verb) rcDs4->printDaughters();
  }
  rcDs4->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds4 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}



int Dsminus__K892_Kminus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds5 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcDs5) {
    cout<<"initializing Ds"<<endl;
    rcDs5=new rcX("Ds5", "Ds- -> K*K-", verb);
    decay ds;
    ds.fit           = 3; // = long decay
    ds.book_tables   = 1; // 1/0
    ds.rlevel        = 0; // 0=full, 1=semi, 2=mtra
    ds.verb          = verb;
    rcDs5->setMother(ds);  // set
    vertexAttrib vds;
    vds.rlevel           = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vds.topp             = kinema::findTopp(-431); // pointer to mother
    vds.momentum_flag    = 1;
    vds.max_vertex_chi2  = 10.0;  // limit for vertex acceptance 
    vds.check_wire       = 0;
    vds.min_dz_from_wire = .15;   // z-cut from primary vertex
    vds.verb             = verb;
    vds.min_mass         = MDS_LO; // minimum mass
    vds.max_mass         = MDS_HI; // maximum mass
    rcDs5->setVertex(vds);       // set

    attribute k892;
    //k892.lund=313;
    //k892.mother=-431;
    k892.topp      = kinema::findTopp(k892.lund);
    k892.fit       = 2;
    k892.matchTOPP = 1;
    k892.mass      = k892.topp->mass;
    k892.maxMass   = 1.2;
    k892.minMass   = 0.6;
    rcDs5->addDaughter(k892);
  
    attribute k1;
    k1.topp=kinema::findTopp(k1.lund);
    //k1.lund         = -321;
    //k1.mother       = -431;
    k1.mass         = k1.topp->mass;
    k1.minP         = -200;
    k1.maxP         = 0;
    k1.minHIO       = 15;
    k1.minHSI       = 6;
    k1.minPt        = 0.3;
    k1.minLRK       = .1;    
    k1.maxLRPI      = 0.01;
    k1.maxEoP       = 0.8;
    k1.maxCandidate = 10;
    k1.useVdsOnce   = 1;
    k1.useMainOnce  = 1;
    rcDs5->addDaughter(k1);

    if (verb) rcDs5->printDaughters();
  }
  rcDs5->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds5 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}



int Dsplus__K892_Kplus_RC() {
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds6 <<<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcDs6) {
    cout<<"initializing Ds"<<endl;
    rcDs6=new rcX("Ds6", "Ds+ -> K*K+", verb);
    decay ds;
    ds.fit           = 3; // = long decay
    ds.book_tables   = 1; // 1/0
    ds.rlevel        = 2; // 0=full, 1=semi, 2=mtra
    ds.verb          = verb;
    rcDs6->setMother(ds);  // set
    vertexAttrib vds;
    vds.rlevel           = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vds.topp             = kinema::findTopp(431); // pointer to mother
    vds.momentum_flag    = 1;
    vds.max_vertex_chi2  = 10.0;  // limit for vertex acceptance 
    vds.check_wire       = 0;
    vds.min_dz_from_wire = .15;   // z-cut from primary vertex
    vds.verb             = verb;
    vds.min_mass         = MDS_LO; // minimum mass
    vds.max_mass         = MDS_HI; // maximum mass
    rcDs6->setVertex(vds);        // set

    attribute k892;
    //k892.lund     = 313;
    //k892.mother   = 431;
    k892.topp      = kinema::findTopp(k892.lund);
    k892.fit       = 2;
    k892.matchTOPP = 1;
    k892.mass      = k892.topp->mass;
    k892.maxMass   = 1.2;
    k892.minMass   = 0.6;
    rcDs6->addDaughter(k892);
 
    attribute k1;
    //k1.lund         = 321;
    //k1.mother       = 431;
    k1.topp=kinema::findTopp(k1.lund);
    k1.mass         = k1.topp->mass;
    k1.minP         = 0;
    k1.maxP         = 200;
    k1.minHIO       = 15;
    k1.minHSI       = 6;
    k1.minPt        = 0.3;
    k1.minLRK       = .1;    
    k1.maxLRPI      = 0.01;
    k1.maxEoP       = 0.8;
    k1.maxCandidate = 10;
    k1.useVdsOnce   = 1;
    k1.useMainOnce  = 1;
    rcDs6->addDaughter(k1);

    if (verb) rcDs6->printDaughters();
  }
  rcDs6->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>>>  Ds6 <<<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}






//
//  Selection/reconstruction of Ds decays, returns 1 upon success
//
int
selectDs()
{
  ARTE_mark( "rcX_Ds 0", 0 );
  static int old_total1=0;
  static int old_total2=0;
  static int old_total3=0;
  static int old_total4=0;
  static int old_total5=0;
  static int old_total6=0;
  int ret=0;

  //=== check if rich likelihoods in RTRA are filled,
  //  take maximal likelihoods from  RTRI otherwise

  int richpid=1;
  if (ArteTable<RTRA>::size()!=0){
    ArtePointer<RTRA> rtra(1);
    if (rtra->lre<0&&rtra->lrmu<0&&rtra->lrpi<0&&rtra->lrk<0&&rtra->lrp<0){
      int mr=maximizeRTRA();
      richpid=-1;
    }
  }

  rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1070 ), 0., 1.); // new event
  ArteTable<RTRA>::iterator d1=ArteTable<RTRA>::end();
  Ds();
  ArteTable<RTRA>::iterator d2=ArteTable<RTRA>::end();
  if (rcDs1 && rcDs1->total>old_total1) { old_total1=rcDs1->total; ret=1; }
  if (rcDs2 && rcDs2->total>old_total2) { old_total2=rcDs2->total; ret=1; }
  if (rcDs3 && rcDs3->total>old_total3) { old_total3=rcDs3->total; ret=1; }
  if (rcDs4 && rcDs4->total>old_total4) { old_total4=rcDs4->total; ret=1; }
  if (rcDs5 && rcDs5->total>old_total5) { old_total5=rcDs5->total; ret=1; }
  if (rcDs6 && rcDs6->total>old_total6) { old_total6=rcDs6->total; ret=1; }
  if (ret) {
    rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1070 ), 1., 1.); // Ds found
    // -- fill mass histogram
    for (ArteTable<RTRA>::iterator id=d1; id!=d2; ++id) {
      ArtePointer<RVER> rver=(*id)->ever;
      float mass=rver->mass;
      rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1071 ), mass, 1.);
    }
  }

  //== reset rich likelihoods in RTRA to default value if no rich likelihood
  //   was filled before calling this routine

  if (richpid<0){
       rcResetRichRtra(-1);
  }

  ARTE_mark( "rcX_Ds 1", 0 );
  return ret;
}





void stopDs() {
  if (rcDs1) delete rcDs1;
  if (rcDs2) delete rcDs2;
  if (rcDs3) delete rcDs3;
  if (rcDs4) delete rcDs4;
  if (rcDs5) delete rcDs5;
  if (rcDs6) delete rcDs6;
}
