#ifndef CMEDIAN_ICC
#define CMEDIAN_ICC

#include "silarte/CMedian.h"

// Implementation file for the median class
// =======================================================================
// Author: Martin Braeuer 
//
//    From: 'Numerical Recipes in C: The Art of Scientific Computing'
//    (William  H.  Press, Brian P. Flannery, Saul A. Teukolsky,
//    William T.  Vetterling;  New  York:  Cambridge  University
//    Press,  1992 (2nd ed, p. 218))
//
// Modification history:
//
// MBr  10.10.2000: 1. Implementaion according to NumRec 
// MBr  31.10.2000: Bug-fix in reset treatment for repeated use!  
// MBr  12.01.2001: Extension for derived classes                   
// =======================================================================

#define CME_SWAP(a,b) temp=(a);(a)=(b);(b)=temp;

CMedian::CMedian() {
  iNumElements = 1000;
  Reset();
  pfStore     = new float[iNumElements];
}

CMedian::CMedian(int iNumber) {
  iNumElements = iNumber;
  Reset();
  pfStore     = new float[iNumElements];
}

CMedian::~CMedian() {
  iNumElements = 0;
  iCurrent    = 0;
  delete[] pfStore;
}

void  CMedian::Reset() {
  iDone       = 0;
  iStatDone   = 0;
  fMedian     = fMAD = 0.0;  
  iCurrent = 0; 
}

int CMedian::Insert(float fValue) {
  if (iCurrent >= iNumElements) return(0);
  pfStore[iCurrent] = fValue;
  ++iCurrent;
  iDone = 0;
  iStatDone = 0;
  return(1);
} // of Insert(1)

int CMedian::Insert(float         *pfValue,
		    unsigned long iFirst,
		    unsigned long iLast     ) {
  if (iFirst >= iLast)                            return(0);
  if (iCurrent + (iLast - iFirst) >= iNumElements) return(0);
  int i;
  for (i=iFirst; i<iLast; ++i) {
    if (iCurrent >=  iNumElements) break;
     pfStore[iCurrent] = pfValue[i];
     ++iCurrent;
  }
  iDone = 0;
  iStatDone = 0;
  return(i-iFirst);
} // Insert(2)


// From Press et al: BUT C-indexing 0..n-1 !!
float CMedian::Select(unsigned long iNumSmallest,
		      unsigned long iStart,
		      unsigned long iStop,
		      float         *arr          ) {
  if (iNumElements < 1) return(0.0);

  unsigned long i,ir,j,l,mid;
  float         a,temp;
  
  l  = iStart;
  ir = iStop-1;
  
  for (;;) {
    if (ir <= l+1) {
      if (ir == l+1 && arr[ir] < arr[l]) {
	CME_SWAP(arr[l],arr[ir])
      }
      return arr[iNumSmallest];
    } else {
      mid = (l+ir) >> 1; // choose meadin of left, center, right element as 
                         // partition and rearrange
      CME_SWAP(arr[mid],arr[l+1])
      if (arr[l+1] > arr[ir]) {
	CME_SWAP(arr[l+1],arr[ir])
      }
      if (arr[l] > arr[ir]) {
	CME_SWAP(arr[l],arr[ir])
      }
      if (arr[l+1] > arr[l]) {
	CME_SWAP(arr[l+1],arr[l])
      }
      i=l+1;
      j=ir;
      a=arr[l];
      for (;;) {
	do i++; while (arr[i] < a); // Scan for element > a
	do j--; while (arr[j] > a); // Scan down for < a element
	if (j < i) break;
	CME_SWAP(arr[i],arr[j])
      }
      arr[l] = arr[j];
      arr[j] = a;
      if (j >= iNumSmallest) ir = j-1;
      if (j <= iNumSmallest) l  = i;
    } // of else
  } // of for
  
} // of select

float CMedian::GetMedian() {
  if (iDone == 1)   return(fMedian); // computed before
  if (iCurrent < 1) return(0.0);

  if (iCurrent & 1) 
    fMedian = Select(((iCurrent+1) >> 1)-1, 0, iCurrent, pfStore);
  else {
    int j = iCurrent >> 1;
    fMedian = 0.5*(Select(j-1,0,iCurrent,pfStore) + Select(j,0,iCurrent,pfStore));
  }
  iDone = 1;
  return(fMedian);
}

float CMedian::GetMedian(float         *pfVals, 
			 unsigned long iSize   ) {
  if (iSize & 1) 
    return( Select(((iCurrent+1) >> 1)-1, 0, iSize, pfVals) );
  else {
    int j = iSize >> 1;
    //printf("SWARN\n");
    return( 0.5*(Select(j-1,0,iSize,pfVals) + Select(j,0,iSize,pfVals)) );
  }  
  return(0.0);
}

int CMedian::GetMedianStat(float *Median,
			   float *MedianAbsDeviation ) {
  int iRet = 1;
  if (iStatDone != 1)  iRet = GetMedianStat();
  
  *Median             = fMedian;
  *MedianAbsDeviation = fMAD;
  return(iRet); 
}

int CMedian::GetMedianStat() {
  if (iStatDone == 1) return(1); 
  int i;
  
  //  printf("nMed=%d\n",iCurrent);
  fMedian = 0.0;
  fMAD    = 0.0;
  
  if (iCurrent < 1) return(0);

  GetMedian();

  float *pfDiff = new float[iCurrent];
  for (i=0;i<iCurrent;++i) 
    pfDiff[i] = fabs(pfStore[i] - (fMedian));
  fMAD = GetMedian(pfDiff,iCurrent);
  //  printf("i=%d nMed=%d  MAD=%10.4f\n",i,iCurrent,fMAD);
  delete[] pfDiff;
  iStatDone = 1;

  return(1);
}

#endif // end of code

// =========== EOF ===== EOF ===== EOF ===== EOF ===== EOF ===== EOF =============
