// ********************************************************************
//
//  Routine:      acReadVDS
//  Author:       T.Glebe, I.Kisel
//
//  Description:  call the VDS initialization to get the VDS constants
//                - if key not assigned, use run number as major number
//                - returns <0 error
//                           0 no VDS update necessary
//                           1 update done using key
//                           2 update done using date/time or run number
//                
//
//  Modification:	21.03.2000 - added code used by L2FORMARTE (fg sciacca)
//			02.04.2000 - fixed handling of keytable mechanism (fg sciacca)
//			11.04.2000 - bugs removed
//				   - L2FORMARTE code integrated with VDS code (fg sciacca)	
//			30.06.2000 - bu fix: now ok with unsparsified data (fg sciacca)
//                      29.11.2000 - (TG) call ARTE_exit() in case DB could not
//                                   be read (for reprocessing, request by A. Gellrich)
//			14.12.2000 - DB access now allows writing (pwd hardcoded) (fg sciacca)
//				     (for new and old LEDA compatibility - cmp. V. Vagnoni)
//			04.04.2001 - GEDE update and SVDstartup skipped when in MC mode
//				     (for use within SLT simulation) (fg sciacca)
//			11.06.2001 - porting to TAXX (DG)
//			11.12.2001 - adapted to read VDSefficiency according to CnA scheme
//				     if in MC mode (fg sciacca)
//			29.01.2002 - started modification to read VDSchipmasking according
//				     to the CnA scheme (fg sciacca)	
//			07.02.2002 - can force reading geometry from file by setting in KUMAC:
//				     SETENV READ_VDSAL_FROM_FILE <filename>
//				     in this case rest of tables read from DB but not via keytable
// ********************************************************************
//!!!!!!!!!!!!!!!!!!!!!!
using namespace std;//!!
//!!!!!!!!!!!!!!!!!!!!!!

#include <stdio.h>

#include "arte/arte.hh"
#include "database/keytable.h"
#include "farm/farm.hh"
#include "grr.h"
#include "silarte/silarte.hh"
#include "vdsleda.h"

#include "alca/actaxxComm.hh"

// sparse event definition
#include "event.h"
#include "etb.h"

#define MAXPLANES 128
// prototypes
int SvdStartup();
//int  vdsleda_fetchtable( char *filename, int major, int minor);
int silarte_updateGEDEeff_CnA(char *dbserver, Tables y, char *tablename);
extern "C" {
  void fill_cats(void);
}

static int VDSserver =	-1;
static int VDStable =	-1;
static int VDStable_side =	-1;
static int VDStable_chip =	-1;
static int VDStable_bondmap =	-1;
static int VDStable_fed =	-1;
static int VDStable_chanstat =	-1;
static int VDStable_chipbond =	-1;
static int VDStable_chipchans =	-1;
static int VDStable_chipfedside =	-1;

static int VDStable_eff =      -1;  
static int VDStable_mask =      -1;

// *****************************************************************
// function
// *****************************************************************

int acReadVDS(int KTstatus) 
{
  char *message = 0;
  if (!message) message = new char[512];
  char *routine = "acReadVDS";
  Tables x;
  int t;
  int majormc, minormc = 0;
  static bool firsttime = true;

   cout<<"######## acReadVDS development version ########"<<endl;

//==================================================================
// if in MC mode: retrieve MC constants
//==================================================================
  
  //  if (ArteTable<MTRA>::size() > 0 ) return 0;
  if (ArteTable<MCIN>::size() > 0) {  //begin if MC

  cout<<"acReadVDS: MC mode"<<endl;

// ------------------------------------------------------------------
// database server (default: /MC_MASEF)
//                  can be changed by setting DB_VDS_EFF
// ------------------------------------------------------------------
  char defaultserver[11];
  sprintf(defaultserver,"/MC_MASEF");
  char * dbserver;
     
  if (getenv("DB_VDS_EFF") == NULL){
    dbserver = defaultserver;
  } else {
    dbserver = getenv("DB_VDS_EFF");
  };
        
// ------------------------------------------------------------------
// retrieve constants: mechanism chosen according to KTstatus
// ------------------------------------------------------------------
  switch (KTstatus) { //begin of switch

    case 2:
      // structure of keytable changed, so loop over the keytable and get
      // your serverid and tableids
  
      VDSserver = KeyFindServer(dbserver);
      if (VDSserver == -1) {
        ARTE_exit(GRR_SFATAL, "acReadVDS: Keytable not found!",0);
        cout<<"acReadVDS: KeyTable Server not found!"<<endl;
        //return -1;
        break;  // use old method
      };  

      cout << ""<< endl;
      cout << "acReadVDS: Looking for Keytable entries..." << endl;
      VDStable_eff	=	KeyFindTable(VDSserver, "VDSefficiency");
      VDStable_mask	=	KeyFindTable(VDSserver, "VDSchipmasking");

      if ((VDStable_eff == -1) || (VDStable_mask == -1)) {
        ARTE_exit(GRR_SFATAL, "acReadVDS: Keytable entries not found!",0);
        cout<<"acReadVDS: Keytable entries not found!"<<endl;
        break;  // use old method
      };
      // NO _BREAK_ HERE!

    case 1:
      // an update happened, look if your server is involved
      if (!KeyServerNew(VDSserver)) {
        cout<<"acReadVDS: no VDS update necessary, server not involved"<<endl;
        cout<<"acReadVDS: returning with rc = 0"<<endl;
      return 0; //no VDS update necessary
      };

      // now for the tables - look whether they have changed
      if (KeyTableNew(VDStable_eff) || KeyTableNew(VDStable_mask)) {

	// Opening tables
	sprintf(message,
	  "%s: Opening MC Masking & Efficiency DB server %s", routine,KeyServerName(VDSserver));
	cout<<message<<endl;

	if(!(x=OpenTables(KeyServerName(VDSserver),TABLEreadonly))) {
	  sprintf(message,"%s: Error opening DB server %s\n", routine,KeyServerName(VDSserver));
	  ARTE_exit(GRR_SFATAL, message, 1); //return 1
	}
	if (KeyTableNew(VDStable_eff)) { //if VDStable_eff has changed
	  if(silarte_updateGEDEeff_CnA(KeyServerName(VDSserver),x,KeyTableName(VDStable_eff))) {
            sprintf(message,"%s: Error loading table VDSefficiency\n", routine);
            ARTE_exit(GRR_SFATAL, message, 1); //return 1
	  }
	} //if VDStable_eff has changed

/*	if (KeyTableNew(VDStable_mask)) { //if VDStable_mask has changed
	  //NOW READ THE TABLE
          if(silarte_masking_CnA(KeyServerName(VDSserver),x,KeyTableName(VDStable_mask))) {
            sprintf(message,"%s: Error loading table VDSchipmasking\n", routine);
            ARTE_exit(GRR_SFATAL, message, 1); //return 1
          }
	} //if VDStable_mask has changed
*/
	// Closing tables
	if(!CloseTables(x)) {
	  sprintf(message,"%s: Error closing table VDSefficiency\n", routine);
	ARTE_error(GRR_SERROR, message, 0);
	}
        cout<<"acReadVDS: returning with rc = 1"<<endl;
        return 1; // update using keytable
      } // if one of the tables has changed

      //cout<<"acReadVDS: no VDS update necessary"<<endl;
      cout<<"acReadVDS: returning with rc = 0"<<endl;
      return 0;  // no VDS update necessary
        
      //break;
  } // of switch
        
  // No keytable information available, using old method...
    if(firsttime) {
      cout << "acReadVDS: Key NOT assigned --> using old method" << endl;
      firsttime = false;
    }
    // OLD METHOD HERE
    if(silarte_updateGEDEeff()) {
      sprintf(message,"%s: Error loading table VDSefficiency\n", routine);
      ARTE_exit(GRR_SFATAL, message, 1); //return 1
    }
/*    if(silarte_masking()) {
      sprintf(message,"%s: Error loading table VDSchipmasking\n", routine);
      ARTE_exit(GRR_SFATAL, message, 1); //return 1
    }
*/
    cout<<"acReadVDS: returning with rc = 2"<<endl;
    return 2;  // constants read in using old method
  } //if MC

//==================================================================
// if in REAL DATA mode: retrieve constants
//==================================================================
// ------------------------------------------------------------------
// database server (default: /VDS_CONFIG)
//                  can be changed by setting DB_VDS
// ------------------------------------------------------------------
 
  cout<<"acReadVDS: REAL DATA mode"<<endl;

  char defaultserver[13];
  sprintf(defaultserver,"/VDS_CONFIG");
  char * dbserver;
  
  static bool sidefromfile = false;

  if (getenv("DB_VDS") == NULL){
    dbserver = defaultserver;
  } else {
    dbserver = getenv("DB_VDS");
  };
 
  char * alignfile = 0;
  if (!alignfile) alignfile = new char[32];

// ------------------------------------------------------------------
// retrieve constants: mechanism chosen according to KTstatus
// ------------------------------------------------------------------
  switch (KTstatus) {

    case 2: 
      // structure of keytable changed, so loop over the keytable and get 
      // your serverid and tableids 

      // in case alignment file is given, skip Database (Side table only)!
      // WARNING: for experts only! Be sure you know what you're doing!!

      if (getenv("READ_VDSAL_FROM_FILE") != NULL) {
	if (firsttime) {
	  alignfile = getenv("READ_VDSAL_FROM_FILE");
	  printf("\n");
	  printf("acReadVDS: #### WARNING #### VDS alignment will be read from file <%s>, NOT from %s!!\n",
			alignfile, dbserver);
	  printf("acReadVDS: #### WARNING #### If this is NOT what you want check your KUMAC settings...\n");
	  printf("\n");
	  for (int n=0; n<32; n++) {
	    vdsleda_iol.side_file.value[n] = *alignfile;
	    alignfile++;
	  }
//	  printf("%%%% VDS side iol file specified: %s\n", vdsleda_iol.side_file.value);
//	  printf("%%%% VDS alignment file specified: %s\n", alignfile);
          break;  // use old method
//	  sidefromfile  = true;
        } else {
	    return 0; 	// YES, otherwise the globals will be applied!!
                	// => geometry will diverge!!!
        }
      } 

      VDSserver = KeyFindServer(dbserver); 
      if (VDSserver == -1) { 
	ARTE_exit(GRR_SFATAL, "acReadVDS: Keytable not found!",0);
        cout<<"acReadVDS: KeyTable Server not found!"<<endl; 
        //return -1; 
        break;  // use old method
      }; 

//      if(firsttime) {
//        SvdStartup();
//        firsttime = false;
//      }
      cout << "acReadVDS: Looking for Keytable entries..." << endl;
      VDStable_side =		KeyFindTable(VDSserver, "VDS/Side"); 
      VDStable_chip =		KeyFindTable(VDSserver, "VDS/Chip"); 
      VDStable_bondmap =	KeyFindTable(VDSserver, "VDS/Bondmap"); 
      VDStable_fed =		KeyFindTable(VDSserver, "VDS/Fed"); 
      VDStable_chanstat =	KeyFindTable(VDSserver, "VDS/Chanstat"); 
      VDStable_chipbond =       KeyFindTable(VDSserver, "VDS/ChipBondmap");
      VDStable_chipchans =      KeyFindTable(VDSserver, "VDS/ChipChanstat");
      VDStable_chipfedside =    KeyFindTable(VDSserver, "VDS/ChipFed");

      if ((VDStable_side == -1) || (VDStable_chip == -1) || 
	  (VDStable_bondmap == -1) || (VDStable_fed == -1) || 
          (VDStable_chanstat == -1) || (VDStable_chipbond  == -1) ||
          (VDStable_chipchans == -1) || (VDStable_chipfedside  == -1) ) { 
	ARTE_exit(GRR_SFATAL, "acReadVDS: Keytable entries not found!",0);
        cout<<"acReadVDS: Keytable entries not found!"<<endl; 
	break;  // use old method
      }; 
      // NO _BREAK_ HERE! 

    case 1: 
      // an update happened, look if your server is involved 

      if (!KeyServerNew(VDSserver)) {
        cout<<"acReadVDS: no VDS update necessary, server not involved"<<endl;      
        cout<<"acReadVDS: returning with rc = 0"<<endl;
      return 0; //no VDS update necessary
      };

      // now for the tables - look whether they have changed 
      if (KeyTableNew(VDStable)) { 
        // using LEDA to get the Tables x:
        MizziDatabase(mydb);
        Tables x;

	cout<<"acReadVDS: keytable found: opening VDS database "<< VDSserver <<endl;
        mydb.open(KeyServerName(VDSserver),21186); //hardcode password for new and old LEDA compatibility
        x=mydb.get_id();

        if(VDS_Side     ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Side table not found!",0); 
        }
        if(VDS_Chip     ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Chip table not found!",0);
        }
        if(VDS_Bondmap  ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Bondmap table not found!",0);
        }
        if(VDS_Fed      ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Fed table not found!",0);
        }
        if(VDS_Chanstat ->book(mydb) == 0) {
          ARTE_exit(GRR_SFATAL, "acReadVDS: VDS_Chanstat table not found!",0);
        }

        VDS_Side	->remove-> all();
        VDS_Chip	->remove-> all();
        VDS_Bondmap	->remove-> all();
        VDS_Fed		->remove-> all();
        VDS_Chanstat	->remove-> all();


        cout<<"acReadVDS: reading VDS/Side information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_side));
        VDS_Side	->fetch_data( DBeq );
        VDS_Side	->set_keys();

        cout<<"acReadVDS: reading VDS/Chip information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_chip));
        VDS_Chip	->fetch_data( DBeq );
        VDS_Chip	->set_keys();

        cout<<"acReadVDS: reading VDS/Bondmap information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_bondmap));
        VDS_Bondmap	->fetch_data( DBeq );
        VDS_Bondmap	->set_keys();
  
        cout<<"acReadVDS: reading VDS/Fed information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_fed));
        VDS_Fed		->fetch_data( DBeq );
        VDS_Fed		->set_keys();

        cout<<"acReadVDS: reading VDS/Chanstat information"<<endl;
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable_chanstat));
        VDS_Chanstat	->fetch_data( DBeq );
        VDS_Chanstat	->set_keys();

	VDS_Side	->fetch_relations(DBeq); /* does not load any table */
	VDS_Fed		->fetch_relations(DBeq); /* does not load any table */ 
        KeyTableStamp(x,KeyServerName(VDSserver),
                        KeyTableName(VDStable_chipbond ));
	VDS_Bondmap	->fetch_relations(DBeq);
        KeyTableStamp(x,KeyServerName(VDSserver),
                        KeyTableName(VDStable_chipchans ));
	VDS_Chanstat	->fetch_relations(DBeq);
        KeyTableStamp(x,KeyServerName(VDSserver),
                        KeyTableName(VDStable_chipfedside ));
	VDS_Chip	->fetch_relations(DBeq); /* loads fed and side relations */

        VDS_Side	->unbook();
        VDS_Chip	->unbook();
        VDS_Bondmap	->unbook();
        VDS_Fed		->unbook();
        VDS_Chanstat	->unbook();
        cout<<"acReadVDS: VDS tables loaded: closing VDS database "<< VDSserver <<endl;

/*	if(sidefromfile) {	// if forcing alignment from file with
				// SETENV READ_VDSAL_FROM_FILE <filename>
	  break;	// go read the file with old method
	}
*/
	// Only update GEDE and initialize Marvin in Data mode; skip if MC
	if (ArteTable<MTRA>::size() == 0 ){ 
	  silarte_updateGEDE();

	  fill_cats(); //fill cats with new geometry
	  if(firsttime) {		//silraw and marvin init must occurr after DB access
	    SvdStartup();
	    firsttime = false;
	  }
	}
	
	cout<<"acReadVDS: returning with rc = 1"<<endl;
        return 1; // update using keytable
      }  

      //cout<<"acReadVDS: no VDS update necessary"<<endl;
      cout<<"acReadVDS: returning with rc = 0"<<endl;
      return 0;  // no VDS update necessary

      //break; 
  } // of switch 

  // No keytable information available, using old method...

    if(firsttime) {
    cout << "acReadVDS: Key NOT assigned --> using old method" << endl; 
//    SvdStartup();
//    firsttime = false;
    }

  static int VDS_Side_major = -1;
  static int VDS_Side_minor = -1;
  static bool findmajmin = true;
  static MizziDatabase( vdsdb_setup );

  //    Get constants from database
  //
  // Time stamp is evaluated in this order:
  //  1. date/time   from vdsleda IOL
  //  2. major/minor from vdsleda IOL
  //  3. event time from DDAQ (if all IOL parameters are =0)
  int major = 0;
  int minor = 0;
  int ndaq;

  
  if ( (int)actaxx_Command.cmd <= 0 ) {
    EVENT_HDR *event;
    EVT_DATA_RECORD daq;

    // get event pointer
    event = ARTE_get_event();
    if ( !event ) {
      cout << "acReadVDS: Could not find event!" << endl;
      return -1; // error
    }      

    ETB_get_daqs( event, 1, &daq, &ndaq );
    if ( ndaq > 0 ) {
      major = daq.data.event_time;
      minor = 0;
    }
  } else {
    major = (ArtePointer<EVHD>(1))->dbmajor;
    minor = (ArtePointer<EVHD>(1))->dbminor;
  }



  if(VDS_Side_major == -1 || 
     (major > VDS_Side_major && !findmajmin)) {

    vdsleda_fetchdata(major, minor);
    // Only update GEDE and initialize Marvin in Data mode; skip if MC
    if (ArteTable<MTRA>::size() == 0 ){ 
      silarte_updateGEDE();
      
      fill_cats(); //fill cats with new geometry

 
      // DEBUG: 
      printf("\n========================================================\n");
      for (ArteTable<GEDE>::iterator gede = ArteTable<GEDE>::begin();
      gede != ArteTable<GEDE>::end(); ++gede)                     {
      if ((*gede)->cmp !=0x1) continue;
      if ((*gede)->gesl->name[3] == '8') 
      printf("%s Xmin=%8.4f Ymin=%8.4f zMean=%8.4f\n",
      (*gede)->gesl->name,
      (*gede)->xmin,
      (*gede)->ymin,
      ((*gede)->zmin+(*gede)->zmax)/2 );

      }
      printf("========================================================\n\n");

      if(firsttime) {		//silraw and marvin init must occurr after DB access
	SvdStartup();
	firsttime = false;
      }
    }
  } else {
      cout<<"acReadVDS: OLDMETHOD; no VDS update necessary"<<endl;
      cout<<"acReadVDS: OLDMETHOD; returning with rc = 0"<<endl;
      return 0; // no update necessary
  }

  // find major minor until which dataset is valid
  if(VDS_Side_major == -1  || 
     findmajmin            ||
     major > VDS_Side_major ) {

    //    vdsdb_setup.open( vdsleda_iol.setup_database.value, 0 );
    vdsdb_setup.open( dbserver, 0 );
    vdsdb_setup.stamp( major, minor );
    VDS_Side    ->book( vdsdb_setup );
    if( VDS_Side->match( DBgt ) == 1 ) {
      // successfull match
      VDS_Side_major = TableMajor( vdsdb_setup.get_id() );
      VDS_Side_minor = TableMinor( vdsdb_setup.get_id() );
      findmajmin = false;
    } else {
      //if there is no newer table, lookup would take place in each event!
      //      findmajmin = true;
      VDS_Side_major = major + 200;
      findmajmin = false;
    }

    VDS_Side->unbook();
    vdsdb_setup.close();
  }

  cout<<"acReadVDS: returning with rc = 2"<<endl;
  return 2;  // constants read in using old method 
}
