#ifndef CMEDIAN_HH
#define CMEDIAN_HH

// Implementation file for the median class
// =======================================================================
// Author: Martin Braeuer 
//
//    From: 'Numerical Recipes in C: The Art of Scientific Computing'
//    (William  H.  Press, Brian P. Flannery, Saul A. Teukolsky,
//    William T.  Vetterling;  New  York:  Cambridge  University
//    Press,  1992 (2nd ed, p. 218))
//
// Modification history:
//
// MBr  10.10.2000: 1. Implementaion according to NumRec 
// MBr  31.10.2000: Bug-fix in reset treatment for repeated use!                  
// MBr  00.00.2000:                   
// =======================================================================

#define MADSCALE (1.0/0.6745) // Scale to make MAD comp. wrt. sigma
                              // MADSCALE=1/(0.75-Quantile of N(0,1))

class CMedian {
 public: // ---------------------------------------- PUBLIC :

  /* ========================================================
   *  Def. Constructor, reserves space for 1000 elements
   * ========================================================
   */
  CMedian();

  /* ========================================================
   *  Constructor, reserves space for 'iNumber' elements
   * ========================================================
   */
  CMedian(int iNumber);

  /* ========================================================
   *  Def. Destructor, frees the space
   * ========================================================
   */  
   ~CMedian();


  /* ========================================================
   *  GetMedian : Returns the median of the distribution
   * ========================================================
   */
  float GetMedian();
  float GetMedian(float         *pfVals, 
		  unsigned long iSize   );
  
  /* ========================================================
   *  Reset  : Reset the entries for further use
   * ========================================================
   */
  void  Reset();

  /* ========================================================
   *  Insert : Inserts a new value, returns 1 if space is ok,
   *           0 else 
   * ========================================================
   */
  int   Insert(float fValue);

  /* ========================================================
   *  Insert : Inserts values, returns 1 if space is ok,
   *           0 else
   *
   *  Parameter: 
   *    pfValue     : Array from where to take vals
   *    iFirst      : First Val from pfValue to take
   *    iLast       : the first value NOT taken 
   *                  (must be >iFirst)
   *
   *  Return: Number of added elements 
   * ========================================================
   */
  int   Insert(float         *pfValue,
	       unsigned long iFirst,
	       unsigned long iLast     );
    
  /* ========================================================
   *  GetNumber : Returns the number of stored values
   * ========================================================
   */
  int    GetNumber() { return(iCurrent); };

  /* ========================================================
   *  GetMedianStat : Retruns the Median and the Median- 
   *                  Absolute-Deviation (MAD)
   *
   * Remark:
   * According to J.Huber: 'Robust Statistics', J. Wiley, 
   * 1981  is the MAD is a robust estimate for the scale of
   * the distribution. Ie. Comperable to the variance in case
   * of non-gaussian tails. It is the same as the Inter-
   * quartile distance for symetric distributions. In case of
   * eps contaminated samples, this goes up to eps<0.5 =>
   * higher breakdown-point.
   * Consistency with normal: divide by PHI^-1(3/4) = 0.6745
   * ========================================================
   */
  int    GetMedianStat(float *Median,
		       float *MedianAbsDeviation );
  // No public data


 protected: // ------------------------------------ PROTECTED :
  int    GetMedianStat();

  // data:
  float         fMedian,fMAD;
  char          iDone,iStatDone;
  unsigned long iNumElements;
  float         *pfStore;
  unsigned long iCurrent;
 

 private: // --------------------------------------- PRIVATE :
  float Select(unsigned long iNumSmallest,
	       unsigned long iStart,
	       unsigned long iStop,
	       float         *arr          );

  // data:
   
}; // of class def.

#endif // end of class definiton
// =========== EOF ===== EOF ===== EOF ===== EOF ===== EOF ===== EOF =============








