#ifndef CLASS_HCUT
#define CLASS_HCUT

// header file for the HCUT class library
// =========================================================
// Author: Martin Braeuer 
//
// NOTE: - This should run with the old and new gede versions
//       - One can cut on Helixes or regions (2do)
//
// Modification history:
// MBr  Feb.99    : Initial ideas of MBr and Stefan Scharein
// MBr  15.02.2000: Class Version 0.9
// MBr  01.03.2000: 0.9 finalized
// GS   02.07.2001: Class Version 1.0 (added new contsr HCUT_class(char *state, int dummy)               $
// MBr    .  .    :                                      
//                                  
// =========================================================


#include <stdio.h>
#include  <stdlib.h>

class HCUT_class {
private:
  // functions
  // Print an error message to the user:
  void fError_out(const char *);
   
  // data
  //       GEDE_class   *hc_GEDEp;           // Access pointer to GEDE

  char hstate[128][10];

public:
  // constructors
       HCUT_class(); 
  // constructor with parameters:
       /* ==================================================================
        * == Read cutter :  ==
        * == Parameter: Pointer to HOLMES structure                       ==
        * ==            A function taking usid, chip and value to store   ==
        * ==            occupancies in rhp objects                        ==
        * ==                                                              ==
        * ================================================================== 
        */
       HCUT_class(char   file[]        ); 
       HCUT_class(char   *state, int dummy	); 
					 
  // Functions
       /* ==================================================================
        * == fClear : Clear al HELIX states internally to H_WORKING       ==
        * ==                                                              ==
        * ================================================================== 
        */
        void fClear();

       /* ==================================================================
        * == fUnsin :  Calculate the 7bit ID of sides:                    ==
        * == needs  : --                                                  ==
        * == pars   : sl=superlayer, sec=sector, lay=layer (1..count)     ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fUnsin(int sl,int sec,int lay) { 
	 int i = (lay-1)+(4*(sec-1))+(16*(sl-1));
	 return(i); 
       }

       /* ==================================================================
        * == fSideUnsin :  Calculate the 7bit ID of side                  ==
        * == needs  : --                                                  ==
        * == pars   : side                                                ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fSideUnsin(int side) {
	 if (side < 20000)
	   return((side%10) + 4*((side%1000)/300) + 16*(  int(side/10)%10)); 
	 else 
	   return((side%10) + 4*((side%1000)/300) + 16*(1+int(side/10000))); 
       }

       /* ==================================================================
        * == fUnsinSide :  Calculate side no out of 7bit ID of side       ==
        * == needs  : --                                                  ==
        * == pars   : side                                                ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fUnsinSide(int unsin);

       /* ==================================================================
        * == fHelixid :  Sort a cog of a cluster into a HELIX ID          ==
        * == needs    : --                                                ==
        * == pars     : side=side=sideID to distinguish p/n, cog of clust ==
        * == return   : chip no (0..9)                                    ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fHelixid(int side,float cog) {
	 int ichip; 
	 if (((side%10) == 1) || ((side%10) == 2)) {
	   if (cog<966)
	     ichip = int((58+cog)/128);
	   else
	     ichip = 0;
	   return(ichip);
	 } else
	   return(int(cog/128));
       }

       /* ==================================================================
	* == fReadFile :  Read a simple text file with the bad chips      ==
	* == needs  : --                                                  ==
	* == pars   : char filenam[] Name of file                         ==
	* == return : 1 if ok and at least one found.                     ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fReadFile(char filenam[]);

	/* ==================================================================
 	* == fReadState :  Scan the hstate array for bad chips            ==
 	* == needs  : --                                                  ==
 	* == pars   : pointer to the array                                ==
 	* == return : 1 if ok and at least one found.                     ==
 	* ==                                                              ==
 	* ==================================================================
 	*/
	int fReadState(char *state);

       /* ==================================================================
	* == fDumpTable :  print out the internal Table                   ==
	* == needs  : Class inited                                        ==
	* == pars   : --                                                  ==
	* == return : 1                                                   ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fDumpTable();

       /* ==================================================================
        * == fCutSide :  check if hit is on a doubtfull                   ==
        * == needs  : Class inited                                        ==
        * == pars   : sl=superlayer, sec=sector, lay=layer (1..count)     ==
        * == return : 0 if hit to be rejected, 1 else                     ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fCutSide(int sl,int sec,int lay, float cog);

       /* ==================================================================
        * == fCutSide :  check if hit is on a doubtfull chip              ==
        * == needs  : Class inited                                        ==
        * == pars   : ARTE ID == 100*sl (1..7) +10*sect (1..4) + lay      ==
        * == return : 0 if hit to be rejected, 1 else                     ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fCutSide(int side, float cog);
       int fCutSide(int side, int iChip); // chip ID

       /* ==================================================================
        * == fCutARTE :  check if hit is on a doubtfull chip              ==
        * == needs  : Class inited                                        ==
        * == pars   : ARTE ID == 100*sl (1..7) +10*sect (1..4) + lay      ==
        * == return : 0 if hit to be rejected, 1 else                     ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fCutARTE(int arte, float cog);
       
  // Destructor:
       ~HCUT_class();

  // data public members
    // NONE there!

 private:

       int iFoundOne;

}; // ============== END of Class_HCUT def. ! ==============================

#endif
