#ifndef CLASS_HCheck_HH
#define CLASS_HCheck_HH


// header file for the HCheck class library
// ===================================================================
// Author: Martin Braeuer 
//
// NOTE: 
//
// Modification history:
// MBr  23.02.2000: Initial derive and testing form L4dqCheck
// MBr  24.04.2000: Added single event HELIX checks          
// MBr  09.01.2001: seH-checks debugged (access to state-table!)          
// ==================================================================

// Notation: all 'se' prefixes refer to single event cuts.

// includes :
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <math.h>
#include  <malloc.h>

// DQ stuff:

// descriptors:
#define H_WORKING   0 // flag : "Helix OK"
#define H_LOW      -1 // flag : "too few  cluster/chip"
#define H_MISSING  -2 // flag : "missing chip"
#define H_HIGH      1 // flag : "too many cluster/chip"

#define seH_High    1 // flag: too many hits/event seen

// size of occupancy checks:
#define MAXUNSIN 128
#define OCCCHANS 128
#define MAXHELIX 10

// min number of accumulated hits:
#define MIN_ACCU    10
#define NOT_READY   -2 // show that not enough stat. is there

// helixcuts (default)
// cluster < events/EVCUT => dead!
#define EVCUT     50
    
// cluster < MINCUT       => dead!
#define MINCUT      30    

// < (MAXCUT*meanhit) = max no of hits/chip
#define MAXCUT_N    1.8  

// < (MAXCUT*meanhit) = max no of hits/chip
#define MAXCUT_P    2.8  

// >HELIXCUT number of chips need to work 
#define HELIXCUT    1    

// a plane mus have > meanhit/minmeanhit hits!
#define MINMEANHIT  2.0    


// --------------------------------------------
//   Single Event stuff:
#define seMAXHITS 20 // default value for max num of hits/chip in single event
#define seMINACCU  1 // default value for single event min number of processed hits
#define seOK       1 // Chip OK in this event
#define seBAD      0 // Chip bad in this event
#define seINVALID -1 // Value invalid

/* Mind-map..
  se needed:
   state-table: constr. (1k) 
   hit/chip     constr. table
   
functs:
   seFillHit   / old fill hit?? (No: spoils the total histos!)
   seEvaluateEvent 0 put away, 1 OK
   
   seReset

general:
  kill total event !
  kill single chips: BAAD idea (2. 'check' step required)
  
  problem: coexistance? (2. run?? (=> hit vector or ARTE)
 */



// the class definition for HCeck  -------------------------------------------

class HCheck_class {
private:
  // functions
       // Print an error message to the user:
       void fError_out(const char *);
       void dump();
   
  // data
       // Tables: :----------------------------------------------------
       int         ocche[MAXUNSIN][MAXHELIX];
       int         helixstat[MAXUNSIN][MAXHELIX+1];
       int         o_helixstat[MAXUNSIN][MAXHELIX+1];
       int         mounted[MAXUNSIN];
       signed char  seOccHe[MAXUNSIN][MAXHELIX];
       signed char  seOccHeWrong[MAXUNSIN][MAXHELIX];
       //signed byte

       float globalmeanhit; // mean number of hits per plane
       int   ntrack;
       int   have_acc;      // number of accumulated hits (last) 
       int   have_events;   // number of accumulated events (last) 

       // the cuts ff:
       int   cEVCUT    ;  // cluster < events/EVCUT => dead!
       int   cMINCUT   ;  // cluster < MINCUT       => dead!
       float cMAXCUT_N ;  // < (MAXCUT*meanhit) = max no of hits/chip
       float cMAXCUT_P ;  // < (MAXCUT*meanhit) = max no of hits/chip
       int   cHELIXCUT ;  // >HELIXCUT number of chips need to work 
       float cMINMEANHIT; // a plane maust >minmeanhit hits
       int   iHi;         // count debug-histo ID
 
       int   seMaxHit;      // Current max. number of hits/event
       int   seMaxWrongHit; // Current max. number of wrong-sign hits/event
       int   seMinAccu;     // min number of hits stored with this event
       
       // the function to call for Occ storing:
       int (*Pstore)(int, int, int);    // pointer to store function ==0 -> int. store

       //------------------------------------------------------------------------------
       //   SINGLE EVENT cuts:

       int    prepare;      // 1 if prepare step was performed, 0 else
       int    accumulate;   // counts the number of accumulation steps so far
       int    evaluate;     // 1 if evaluation step was performed, 0 else

       int    seAccumulated; // number of accu steps for this event


public:
  // constructors
       HCheck_class(); 
  // constructor with parameters:
       /* ==================================================================
        * ==          : Prepare structures for HCeck class                ==
        * == Parameter: Pointer to a function taking usid, chip and value ==
	* ==            to store occupancies in rhp/hbook objects         ==
        * ==                                                              ==
        * ================================================================== 
        */
       HCheck_class( int (*)(int, int, int) ); 
					 
  // Functions
       /* ==================================================================
        * == fClear : Clear/unbook all allocated structures               ==
        * ==                                                              ==
        * ================================================================== 
        */
        void fClear();


       /* ==================================================================
        * == fPrepare : Reset structures for HCheck class                 ==
        * == Parameter: ----                                              ==
        * == needs    : Class exisiting                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fPrepare();

       /* ==================================================================
	* == fSetUnmounted : Set the non-mounted planes list              ==
	* == Parameter: *list : -1 terminated list of unmounted USID's    ==
	* == needs    : a prepared HCut object                            ==
	* == return   : 1 okay, -1 invalid planes                         ==
	* ================================================================== 
	*/
       int HCheck_class::fSetUnmounted(int *);


       /* ==================================================================
        * == fUnsin :  Calculate the 7bit ID of sides:                    ==
        * == needs  : --                                                  ==
        * == pars   : sl=superlayer, sec=sector, lay=layer (1..count)     ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fUnsin(int sl,int sec,int lay) { 
	 int i = (lay-1)+(4*(sec-1))+(16*(sl-1));
	 return(i); 
       }

       /* ==================================================================
        * == fSideUnsin :  Calculate the 7bit ID of side                  ==
        * == needs  : --                                                  ==
        * == pars   : side                                                ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fSideUnsin(int side) {
	 if (side < 20000)
	   return((side%10) + 4*((side%1000)/300) + 16*(  int(side/10)%10)); 
	 else 
	   return((side%10) + 4*((side%1000)/300) + 16*(1+int(side/10000))); 
       }

       /* ==================================================================
        * == fHelixid :  Sort a cog of a cluster into a HELIX ID          ==
        * == needs    : --                                                ==
        * == pars     : side=side=sideID to distinguish p/n, cog of clust ==
        * == return   : chip no (0..9)                                    ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fHelixid(int side,float cog) {
	 int ichip; 
	 if (((side%10) == 1) || ((side%10) == 2)) {
	   if (cog<966)
	     ichip = int((58+cog)/128);
	   else
	     ichip = 0;
	   return(ichip);
	 } else
	   return(int(cog/128));
       }

       /* ==================================================================
        * == fUHelixid :  Sort a cog of a cluster into a HELIX ID         ==
        * == needs    : --                                                ==
        * == pars     : takes ONLY USID name scheme !!!!                  ==
        * == return   : chip no (0..9)                                    ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fUHelixid(int side,float cog) {
	 int ichip; 
	 if (((side%4) == 1) || ((side%4) == 2)) {
	   if (cog<966)
	     ichip = int((58+cog)/128);
	   else
	     ichip = 0;
	   return(ichip);
	 } else
	   return(int(cog/128));
       }

       /* ==================================================================
        * == fSetPar : Set Cuts for helix states                          ==
        * == needs  : Class prepared                                      ==
        * == pars   : evcut     cluster < events/EVCUT => dead!           ==
        * ==          mincut    cluster < MINCUT       => dead!           ==
        * ==          maxn      (MAXN*meanhit) = max no of hits/chip  N   ==
        * ==          maxp      (MAXP*meanhit) = max no of hits/chip  P   ==
        * ==          minchips  >MINCHIPS number of chips need to work    ==
	* ==          minmeanhit > <hits all sides>/minmeanhit must       ==
	* ==                       be on a plane                          ==
        * == return : 1 if already inited                                 ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fSetPar(int evcut, int mincut, float maxn, float maxp, 
		   int minchips,float minmeanhit);

       /* ==================================================================
        * == fGetPar : Get Cuts for helix states                          ==
        * == needs  :  Class prepared                                     ==
        * == pars   : evcut     cluster < events/EVCUT => dead!           ==
        * ==          mincut    cluster < MINCUT       => dead!           ==
        * ==          maxn      (MAXN*meanhit) = max no of hits/chip  N   ==
        * ==          maxp      (MAXP*meanhit) = max no of hits/chip  P   ==
        * ==          minchips  >MINCHIPS number of chips need to work    ==
	* ==          minmeanhit > <hits all sides>/minmeanhit must       ==
	* ==                       be on a plane                          ==
        * == return : 1 if set correctly                                  ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fGetPar(int *evcut, int *mincut, float *maxn, float *maxp,
		   int *minchips,float *minmeanhit);
       

       /* ==================================================================
        * == fSetHelixState : Set states for the chips                    ==
        * == needs  : Class prepared                                      ==
        * == pars   : int UNSIN [0..MAXUNSIN], chip [0..MAXHELIX+1]       ==
        * == return : 1 if already inited                                 ==
        * == Note   : the 11th column of a side flags the whole side!     ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fSetHelixState(int unsin, int chip, int value) {
	 if (prepare != 1) return(0); // prepare before
	 if ((unsin > -1) && (unsin < MAXUNSIN) && 
	     (chip  > -1) && (chip  < MAXHELIX+1)) {
	   helixstat[unsin][chip] = value;
	 }
	 return(1);
       }

       /* ==================================================================
        * == fGetHelixState : Get the helix state table                   ==
        * == needs  : Evaluate or init has to have been run before        ==
        * == pars   : int UNSIN [0..MAXUNSIN], chip [0..MAXHELIX+1]       ==
        * == return : -1: if error, helix state else                      ==
        * == Note   : the 11th column of a side flags the whole side!     ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fGetHelixState(int unsin, int chip) {
	 if (evaluate != 1) return(-1); // prepare before
	 if ((unsin > -1) && (unsin < MAXUNSIN) && 
	     (chip  > -1) && (chip  < MAXHELIX)) {
	   return(helixstat[unsin][chip]); // okay here comes the result!
	 } 
	 return(-1);
       }

       /* ==================================================================
        * == fSetOccupancy : Set the occupancy to the internal structures ==
        * == needs  : Class prepared                                      ==
        * == pars   : int UNSIN [0..MAXUNSIN], chip [0..MAXHELIX],        ==
        * ==          value to SET int. chanal to                         ==
        * == return : 1 if already inited                                 ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fSetOccupancy(int unsin, int chip, int value) {
	 if (prepare != 1) return(0); // prepare before
	 if ((unsin > -1) && (unsin < MAXUNSIN) && 
	     (chip  > -1) && (chip  < MAXHELIX)) {
	   ocche[unsin][chip] = value;
	   accumulate+=value;
	 }
	 return(1);
       }
 
       /* ==================================================================
        * == fClearHelixes : reset hit-maps                               ==
        * == needs  : Class prepared                                      ==
        * == pars   : --                                                  ==
        * == return : 0 if not prepared                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fClearHelixes ();

       /* ==================================================================
        * == fGetOccupancy : Get the Occupancy in HELIX bins              ==
        * == needs  : Evaluated                                           ==
        * == pars   : int UNSIN [0..MAXUNSIN], chip [0..MAXHELIX]         ==
        * ==                                                              ==
        * == return : -1: if error, occupancy else                        ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fGetOccupancy(int unsin, int chip) {
	 if (accumulate < MIN_ACCU) return(-1); // accumulate before
	 if ((unsin > -1) && (unsin < MAXUNSIN) && 
	     (chip  > -1) && (chip  < MAXHELIX)) {
	   return(ocche[unsin][chip]);
	 }
	 return(-1);
       }

       /* ====================================================================
	* == fAccuHit : Insert a hit into the int. occ. maps                ==
	* == needs    : Class prepared                                      ==
	* == pars     : side=USID of hit, cog cog of hit                    ==
	* == return   : 1 if OK, 0 if not prepared or invalid side-code     ==
	* == Note     : This works with usid and side id !!!                ==
	* == Note     : Use THIS always to store. A NULL in the contructor  ==
	* ==             will lead to an internal store, a function pointer ==
	* ==             to external storage. (YOU have to copy back then!) ==
	* ==                                                                ==
	* ==================================================================== 
	*/
       int fAccuHit(int usid,float cog);

       /* ====================================================================
        * == fAccuHits: Insert num raw hits into the int. occ. maps         ==
        * == needs    : Class prepared                                      ==
        * == pars     : gedechan=USID of hits, cog=c.o.g. of all hits,      ==
        * ==            num= number to store                                ==
        * == return   : 1 if prepared                                       ==
        * == Note     : This works with usid and side id !!!                ==
	* == Note     : Use THIS always to store. A NULL in the contructor  ==
	* ==             will lead to an internal store, a function pointer ==
	* ==             to external storage. (YOU have to copy back then!) ==
        * ==                                                                ==
        * ==================================================================== 
        */
       int fAccuHits(int side,float cog, int num);


       /* ==================================================================
	* == fAcceptHit: check that a hit does not contain bad chips      ==
	* == needs    : preppare has to run before                        ==
	* == pars     : side ID and cog of hit                            ==
	* == return   : 1 hit OK, 0 if not prepared or hit bad            ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fAcceptHit(int side,float cog);


       /* ===================================================================
        * == fCorrectGeometry: scale the occupancy matrix of a side to     ==
	* ==                   correct for special geometries (50x!!)      ==
        * == needs     : fAccuHit has to have been run, called before eval ==
        * == iSide     : ID of side to correct                             ==
        * == fFact     : Arrey of factros to correct with (NO check done!) ==
	* == return    : 1: OK, 0: wrong side                              ==
        * ==                                                               ==
        * =================================================================== 
        */
       int fCorrectGeometry(      int   iSide,
			    const float fFact[MAXHELIX] );

       /* ===================================================================
        * == fComputeRefs: Get a histo+sigmas of each helix-pos in a set of==
	* ==               Superlayer                                      ==
        * == needs     : fAccuHit has to have been run                     ==
        * == iSuperlayer1 : ID of first SL (1..8)                          ==
        * == iSuperlayer2 : ID of first SL (1..8)                          ==
        * == fValues      : Values found (order: N1, N2, P1, MAXHELIX each)==
        * == fSigmas      : corresponing sigmas                            ==
	* == return       : 1: OK, 0: wrong SL's                           ==
        * ==                                                               ==
        * =================================================================== 
        */
       int fComputeRefs(int   iSuperlayer1,
			int   iSuperlayer2,
			float fValues[3*MAXHELIX],
			float fSigmas[3*MAXHELIX] );

       /* ===================================================================
        * == fEvaluate1: evaluate the occupancy matrices to derive helixes ==
        * == needs     : fAccuHit has to have been run                     ==
        * == pars      : evdone = number of processed events               ==
	* == return    : 1: no change, 0: chips changed, -1: error         ==
        * ==                                                               ==
        * =================================================================== 
        */
       int fEvaluate1(int evdone);

       /* ===================================================================
        * == fEvaluate2: evaluate the occupancy matrices to derive helixes ==
        * == needs     : fAccuHit has to have been run                     ==
        * == pars      : evdone = number of processed events               ==
	* == return    : 1: no change, 0: chips changed, -1: error,        ==
	* ==             NOT_READY: statistics needed                      ==
        * ==                                                               ==
        * =================================================================== 
        */
       int fEvaluate2(int evdone);

       /* ===================================================================
        * == fEvaluate2: evaluate the occupancy matrices to derive helixes ==
        * == needs     : fAccuHit has to have been run                     ==
        * == pars      : evdone = number of processed events               ==
	* == return    : 1: no change, 0: chips changed, -1: error,        ==
	* ==             NOT_READY: statistics needed                      ==
        * ==                                                               ==
        * =================================================================== 
        */
       int fEvaluate3(int evdone); // MEDIAN and rob. estimates

       /* ==================================================================
        * == fResult  : Textoutput of result                              ==
        * == needs    : fEvaluate1 has to run before                      ==
        * == pars     :                                                   ==
        * == return   : 1 if OK, 0 if not evaluated before                ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fResult();

       /* ==================================================================
        * == fGetWorking : returns the number of working helixes          ==
        * == needs    : fEvaluate1 has to run before                      ==
        * == pars     :                                                   ==
        * == return   : -1: if error, number otherwise                    ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fGetWorking();

       /* ==================================================================
	* == fGetNonWorking : returns the number of not working helixes   ==
	* == needs    : fEvaluate1 has to run before                      ==
	* == pars     :                                                   ==
	* == return   : -1: if error, number otherwise                    ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fGetNotWorking();

//------------------------------------------------------------------------------
       
       /* ==================================================================
	* == fseReset : Reset the single event cut structs                ==
	* == needs    : --                                                ==
	* == pars     : --                                                ==
	* == return   : --                                                ==
	* ==                                                              ==
	* ================================================================== 
	*/
       void fseReset();

       
       /* ==================================================================
	* == fseSetPars : Set the actual parameter for the se-functs      ==
	* == needs    : --                                                ==
	* == pars     : maxhitChip : Max num. of       hits/chip to accept==
	* ==            maxwronghit: Max num. of wrong-hits/chip to accept==
	* ==            minhits    : Min number of hits/event to process  ==
	* == return   : --                                                ==
	* ==                                                              ==
	* ================================================================== 
	*/
       void fseSetPars(int maxhitChip, int maxwronghit, int minhits);

       
       /* ==================================================================
	* == fseGetPars : Get the actual parameter for the se-functs      ==
	* == needs    : --                                                ==
	* == pars     : maxhitChip* : Max num. of hits/chip to accept     ==
	* ==            maxwronghit*: Max num. of wrong-hits/chip to accept=
	* ==            minhits*    : Min number of hits/event to process ==
	* == return   : --                                                ==
	* ==                                                              ==
	* ================================================================== 
	*/
       void fseGetPars(int* maxhitChip, int *maxwronghit, int* minhits);

       
       /* ==================================================================
	* == fseAccuHit : Accumulate a hit                                ==
	* == needs    : class inited, Reset done in time                  ==
	* == pars     : side (USID or side ID), cog of hit                ==
	* ==            iSign: 1 if right sign hit, -1 else. Default: 1   ==
	* == return   : 1 of stored                                       ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fseAccuHit(int side,float cog, int iSign = 1);

       /* ==================================================================
	* == fseAccuHit : Accumulate a hit                                ==
	* == needs    : class inited, Reset done in time                  ==
	* == pars     : superlayer, sector, layer, cog                    ==
	* ==            iSign: 1 if right sign hit, -1 else. Default: 1   ==
	* == return   : 1 if stored                                       ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fseAccuHit(int sl, int sect, int layer, float cog, int iSign = 1);

       // The public state table: 
       // 0 : BAD, 1 : OK
       signed char seHelixStat[MAXUNSIN][MAXHELIX];

       /* ==================================================================
	* == fseEvaluate : See if event is OK, reset done                 ==
	* == needs    : at least seMinHits accumulated hits               ==
	* == pars     : --                                                ==
	* == return   : -1 if not enough statis,                          ==
        * ==            or number of too hot guys                         ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fseEvaluate();

       /* ==================================================================
	* == fseHitOkay : See if hit is OK, reset done                    ==
	* == needs    : NO accumulated hits                               ==
	* == pars     : side (USID or side ID), cog of hit                ==
	* == return   : -1 if no evaluation availible,                    ==
	* ==            -2 if resulting layer is incorrect                ==
	* ==             1 if Hit is OK, 0 else                           ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fseHitOkay(int side,float cog);
       int fseHitOkay(int side,int   iChip);
       
       /* ==================================================================
	* == fseHitOkay : See if hit is OK, reset done                    ==
	* == needs    : NO accumulated hits                               ==
	* == pars     : superlayer, sector, layer, cog                    ==
	* == return   : -1 if no evaluation availible,                    ==
	* ==            -2 if resulting layer is incorrect                ==
	* ==             1 if Hit is OK, 0 else                           ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fseHitOkay(int sl, int sec, int lay, float cog);

       int gr(int i, int j) {return(     seOccHe[i][j]);};
       int gw(int i, int j) {return(seOccHeWrong[i][j]);};

  // Destructor:
       ~HCheck_class();

  // data public members
    // NONE there!
}; // ============== END of Class_HCeck def. ! ==============================

#endif
