// =================================================================
//
// silarte_updateGEDEeff.C
//
// G. Sciacca - MPI-K Heidelberg
//
// Called by: dgsil.C
// 
// Retrieve VDS efficencies from Database /MC_MASEF if not in
// CnA mode and update efficiency entries in GEDE banks
//
// Return values:
//		0 => OK
//		1 => error
//
// DBS: default is /MC_MASEF - Can be changed by setenv DB_VDS_EFF
// RunN: def is 0 (ARTE def efficiencies) - Can be changed by setenv VDSrunnumber
//
//		***** IMPORTANT *****
// NOT mounted planes are HARCODED in vdsmc_usid_to_arteid(int)
// since they are not included in GEDE
// MUST UPDATE routine if detector CHANGES
//
// 15.03.01 - (GS) Ready: on hold because of the RunN issue
// 06.04.01 - (GS) RunN fetched from env variable. Not released yet
// 09.04.01 - (GS) Returns if RunN = 0 (uses default efficiencies). Released
// 12.04.01 - (GS) Uses default efficiencies if no entries in DB (no ARTE_exit)
// 02.05.01 - (GS) Will retrieve latest minor rev. if more than one is present in DBS
// 15.11.01 - (GS) Now called by acReadVDS (if in no CnA mode)
//		   vdsmc_usid_to_arteid(int)updated for y2001 detector
// =================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <ctype.h>
#include <sys/time.h>
#include "database/table.h"
#include "database/keytable.h"
#include "arte/arte.hh"
#include "farm/farm.hh"
#include "grr.h"   
#include "silarte/silarte.hh"
   
#define MAXPLANES 128
   
int vdsmc_usid_to_arteid( int );

int silarte_updateGEDEeff()
{
   int rc = 0;	//  OK
   int error = 0;
   char *routine = "silarte_updateGEDEeff";

   Tables x;
   int t;
   int tmpid;
   int sl, sec, lay;
   time_t stime;
   int major = 0;
   int minor = 0;
//   int run = 0;   

   extern int CnAkey;
   extern int RunNoToSimu;

   struct VDSeff_struct
   {       
     int usid;
     float eff;   
     float chipeff0;
     float chipeff1;
     float chipeff2;
     float chipeff3;
     float chipeff4;
     float chipeff5;
     float chipeff6;
     float chipeff7;
     float chipeff8;
     float chipeff9;
   };
      
   struct VDSeff_struct VDSefficiency[MAXPLANES];

   // ArtePointer < EVHD > evhd (1);
   // int run = evhd->run;

   char *message = 0;
   if (!message) message = new char[512];

   cout<<""<<endl;
   cout<<"######## silarte_updateGEDEeff development version ########"<<endl;

   printf("extern int CnAkey: %i\n", CnAkey);
   printf("extern int RunNoToSimu: %i\n", RunNoToSimu);

// ------------------------------------------------------------------
// RunN whose VDS efficiencies refer to
// it is specified by setting runnumber in the kumac
// (default: 0 -> Use ARTE default (0.98 all planes)) ??? check this
// ------------------------------------------------------------------
   major = RunNoToSimu;   
   if (major==0 && minor==0) {
     sprintf(message,
	"%s: No run specified, using VDS default ARTE efficiencies (0.98 all planes)\n",
	routine);
     cout<<message<<endl;
     return rc;
   } else {
     sprintf(message,
        "%s: No CnA mode: VDSefficiencies forRun %i will be retrived",
        routine,RunNoToSimu);
     cout<<message<<endl;
   }								
   
   //get current time stamp
   time(&stime);
//   printf("Time: %d \n", (int) stime);

// ------------------------------------------------------------------
// database server (default: /MC_MASEF)
//                  can be changed by setting DB_VDS_EFF
// ------------------------------------------------------------------
  char defaultserver[11];
  sprintf(defaultserver,"/MC_MASEF");
  char * dbserver;

  // static bool firsttime = true;

  if (getenv("DB_VDS_EFF") == NULL){
    dbserver = defaultserver;
  } else {
    dbserver = getenv("DB_VDS_EFF");
  };

// ------------------------------------------------------------------
// Looking up and retrieving VDSefficncy
// ------------------------------------------------------------------
   
   // Opening and booking tables
   sprintf(message,
	"%s: Opening MC Masking & Efficiency DB server %s", routine,dbserver);
   cout<<message<<endl;

   if(!(x=OpenTables(dbserver,TABLEreadonly))) {
     sprintf(message,
	"%s: Error opening DB server %s\n", routine,dbserver);
     ARTE_exit(GRR_SFATAL, message, 1);	//return 1
   }
   if((t=BookTable(x,"VDSefficiency",0,0))==0) {
     sprintf(message,
	"%s: Error booking table VDSefficiency\n", routine);
     ARTE_exit(GRR_SFATAL, message, 1);	//return 1
   }

   // Setting major and minor
   if (major == 0) {				// getting last revision if major=0
     TableStamp(x,TABLEmaxrev,TABLEmaxrev);
     if (!(MatchTable(x,t,TABLElt))) {          // goes to the last revision of table
       sprintf(message,
	  "%s: Error matching table VDSefficiency\n", routine);
       ARTE_exit(GRR_SFATAL, message, 1);	//return 1
     }
     major = TableMajor(x);
     minor = TableMinor(x);
   } else {					// getting table .le. specified revision
     TableStamp(x,major,TABLEmaxrev);
     if (!(MatchTable(x,t,TABLEle))) {
       sprintf(message,
	  "%s: Error matching table VDSefficiency\n", routine);
       ARTE_exit(GRR_SFATAL, message, 1);	//return 1
     }
     major = TableMajor(x);
     minor = TableMinor(x);
   }
   TableStamp(x,major,minor);
   if (major == 1) {  
     sprintf(message,
	"%s: WARNING: No entries found in database for table VDSefficiency",
	routine);
     cout<<message<<endl;
     sprintf(message,
	"%s: Retrieving default ARTE VDSefficiency (0.98 all planes)", routine);
     cout<<message<<endl;
   } else if(major != RunNoToSimu) {
	    sprintf(message,
		"%s: No entries found in DB for Run %i; retrieving VDSefficiencies for Run %i",routine, RunNoToSimu, major);
	    cout<<message<<endl;

	  } else {
	      sprintf(message,
		"%s: Retrieving VDS efficiencies for (major, minor) (%i, %i)",
                        routine,major,minor);
	      cout<<message<<endl;
   }
   // Retrieving tables
   if(FetchTable(x,t,TABLEeq)==0) {
     sprintf(message,
	"%s: Error fetching table VDSefficiency\n", routine);
     ARTE_exit(GRR_SFATAL, message, 1);	//return 1
   }

   int structsize = sizeof(struct VDSeff_struct);

   SkipTableField(x,1);
   rc = GetTableStruct(x,TABLEint,structsize,MAXPLANES,&VDSefficiency[0].usid);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i usid elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].eff);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i eff elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff0);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff0 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff1);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff1 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff2);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff2 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff3);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff3 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff4);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff4 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff5);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff5 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff6);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff6 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff7);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff7 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff8);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff8 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   rc = GetTableStruct(x,TABLEfloat,structsize,MAXPLANES,&VDSefficiency[0].chipeff9);
   if (rc != MAXPLANES) {
     sprintf(message,
	"%s: Error retrieving data: only %i chipeff9 elements retrieved, %i expected\n",
		 routine, rc, MAXPLANES);
     ARTE_exit(GRR_SFATAL, message, 1);		//return 1
   }
   
// --------------------------------------------------------
// Unbooking and closing tables
// -------------------------------------------------------- 
   sprintf(message,
	"%s: Closing MC Masking & Efficiency DB server %s", routine,dbserver);
   cout<<message<<endl;

   if(!UnbookTable(x,t)) {
     sprintf(message,
	"%s: Error unbooking table VDSefficiency\n", routine);
     ARTE_error(GRR_SERROR, message, 0);
   }
   if(!CloseTables(x)) {
     sprintf(message,
	"%s: Error closing table VDSefficiency\n", routine);
     ARTE_error(GRR_SERROR, message, 0);
   }

// --------------------------------------------------------
// Update efficiency entry in GEDE bank 
// --------------------------------------------------------
   sprintf(message,
	"%s: Loadind VDS efficiency entries on to GEDE bank", routine);
   cout<<message<<endl;

   // loop over efficiency entries retrieved from DB
   for (int i=0; i<MAXPLANES; i++) {

if (i==5) cout<<"### plane 5 ###"<<endl;
     tmpid = vdsmc_usid_to_arteid( i );
if (i==5) cout<<"### tmpid = "<<tmpid<<endl;
     if (tmpid == 0) continue;		// not mounted plane
     if (tmpid < 0) {
       sprintf(message,
	   "%s: %i is not a valid VDS plane\n", routine, tmpid);
       ARTE_error(GRR_SERROR, message, 0);
     }
     // extract sl=A, sec=B, lay=C from tmpid=ABC
     sl = int(floor(tmpid/100));
     tmpid -= sl*100; 
     sec = int(floor(tmpid/10));
     lay = tmpid-(10*sec);
//printf("sl=%i, sec=%i, lay=%i - usid = %i \n", sl, sec, lay,i);

     if (  sl  < 1 || sl  > 8
        && lay < 1 || lay > 4
        && sec < 1 || sec > 4 ) {
        sprintf(message,                   
		"sl=%d, sec=%d, lay=%d is NOT a valid VDS plane\n",
                 sl, sec, lay);
        ARTE_error(GRR_SERROR, message, 0);
     }
     // loop over ARTE GEDE table to get corresponding GEDE row
     int found = 0;

     for ( ArteTable<GEDE>::iterator igede = ArteTable<GEDE>::begin();
          igede != ArteTable<GEDE>::end(); igede++ ) {
       ArtePointer<GEDE> gede = *igede;  

       if (   gede->cmp == 1
          && gede->gesl->nsl == sl
          && gede->lay == lay
          && gede->sec == sec ) {	// VDS plane?

	  // update eff entry value
	  gede->eff = VDSefficiency[i].eff;

	  found = 1;

       } // no VDS plane

     }	// end loop over GEDE rows

    // complain in case no corresponding ARTE GEDE row was found
    if ( !found ) {   
      sprintf(message,     
		"VDS side %i (sl=%i, sec=%i, lay=%i) not found in GEDE table!\n",
                 i, sl, sec, lay);
      ARTE_error(GRR_SERROR, message, 0);
      error++;
    }
   } // end of loop over efficiency entries retrieved from DB

  if ( error ) {
    sprintf(message,
	"%s: *** WARNING *** VDS efficiencies NOT succesfully loaded on to GEDE:",
	 routine);
    cout<<message<<endl;
    sprintf(message,
	" %i GEDE rows not found!\n", error);
    cout<<message<<endl;
    return 1;
  }

  sprintf(message,
	"%s: VDS efficiencies succesfully loaded on to GEDE\n", routine);
  cout<<message<<endl;
  return GRR_OK;
}

int vdsmc_usid_to_arteid( int plane ) {

//-------------------------------------------------
//	returns ABC -> A=sl, B=sec, C=lay
//	returns 0 if not mounted plane
//	returns -1 if plane.lt.0 || plane.gt.127
//-------------------------------------------------

  extern int CnAkey;
  extern int RunNoToSimu;
  char *routine = "vdsmc_usid_to_arteid";
  char *error = 0;
  if (!error) error = new char[512];

  switch (plane) {		
    case 0:     return 111;
    case 1:	return 0;
    case 2:     return 113;
    case 3:     return 114;
    case 16:    return 211;
    case 17:	return 0;
    case 18:    return 213;
    case 19:    return 214;
    case 32:    return 311;
    case 33:	return 0;
    case 34:    return 313;
    case 35:    return 314;
    case 4:     return 121;
    case 5:
	if (RunNoToSimu < 17629) {	// y2k detector?
//  cout<<"plane 5 returning 0"<<endl;
		return 0;
	} else	{			// or y2002 detector?
//  cout<<"plane 5 returning 122"<<endl;
		return 122;
	}
    case 6:     return 123;
    case 7:     return 124;
    case 20:    return 221;
    case 21:
        if (RunNoToSimu < 17629) {	// y2k detector?
//  cout<<"plane 21 returning 0"<<endl;
                return 0;
        } else  {			// or y2002 detector?
//  cout<<"plane 21 returning 222"<<endl;
                return 222;
	}
    case 22:    return 223;
    case 23:    return 224;
    case 36:    return 321;
    case 37:
        if (RunNoToSimu < 17629) {	// y2k detector?
//  cout<<"plane 37 returning 0"<<endl;
                return 0;
        } else  {			// or y2002 detector?
//  cout<<"plane 37 returning 322"<<endl;
                return 322;
	}
    case 38:    return 323;
    case 39:    return 324;
    case 8:     return 131;
    case 9:	return 0;
    case 10:    return 133;
    case 11:    return 134;
    case 24:    return 231;
    case 25:	return 0;
    case 26:    return 233;
    case 27:    return 234;
    case 40:    return 331;
    case 41:	return 0;
    case 42:    return 333;
    case 43:    return 334;
    case 12:    return 141;
    case 13:	return 0;
    case 14:    return 143;
    case 15:    return 144;
    case 28:    return 241;
    case 29:	return 0;
    case 30:    return 243;
    case 31:    return 244;
    case 44:    return 341;
    case 45:	return 0;
    case 46:    return 343;
    case 47:    return 344;
    case 48:    return 411;
    case 49:    return 412;
    case 50:    return 413;
    case 51:    return 414;
    case 52:    return 421;
    case 53:    return 422;
    case 54:    return 423;
    case 55:    return 424;
    case 56:    return 431;
    case 57:    return 432;
    case 58:    return 433;
    case 59:    return 434;
    case 60:    return 441;
    case 61:    return 442;
    case 62:    return 443;
    case 63:    return 444;
    case 64:    return 511;
    case 65:    return 512;
    case 66:    return 513;
    case 67:    return 514;
    case 68:    return 521;
    case 69:    return 522;
    case 70:    return 523;
    case 71:    return 524;
    case 72:    return 531;
    case 73:    return 532;
    case 74:    return 533;
    case 75:    return 534;
    case 76:    return 541;
    case 77:    return 542;
    case 78:    return 543;
    case 79:    return 544;
    case 80:    return 611;
    case 81:    return 612;
    case 82:    return 613;
    case 83:    return 614;
    case 84:    return 621;
    case 85:    return 622;
    case 86:    return 623;
    case 87:    return 624;
    case 88:    return 631;
    case 89:    return 632;
    case 90:    return 633;
    case 91:    return 634;
    case 92:    return 641;
    case 93:    return 642;
    case 94:    return 643;
    case 95:    return 644;
    case 96:    return 711;
    case 97:    return 712;
    case 98:    return 713;
    case 99:    return 714;
    case 100:   return 721;
    case 101:   return 722;
    case 102:   return 723;
    case 103:   return 724;
    case 104:   return 731;
    case 105:   return 732;
    case 106:   return 733;
    case 107:   return 734;
    case 108:   return 741;
    case 109:   return 742;
    case 110:   return 743;
    case 111:   return 744;
    case 112:   return 811;
    case 113:   return 812;
    case 114:   return 813;
    case 115:   return 814;
    case 116:   return 821;
    case 117:   return 822;
    case 118:   return 823;
    case 119:   return 824;
    case 120:   return 831;
    case 121:   return 832;
    case 122:   return 833;
    case 123:   return 834;
    case 124:   return 841;
    case 125:   return 842;
    case 126:   return 843;
    case 127:   return 844;
  }
    
  printf(error,
	"%s: %i is not a valid VDS plane (usid) \n",routine,plane);
  cout<<error<<endl;

  return -1;             // not valid plane
}
    

