// *****************************************************************************
//
// source:    L3_SpacePoint.C
//
// type:      source code
//
// created:   06 Dec 1999
//
// author:    Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// symbol:    
//
// changes:
//
// *****************************************************************************
#include <iostream>
#include <iomanip>
#include <cmath>
#include <algorithm>
#include "L3_Geometry.hh"
#include "L3_Parameter.hh"
#include "L3_SpacePoint.hh"

vector<HepMatrix> L3_SpacePoint::FitMatrizen;
vector<HepMatrix> L3_SpacePoint::MessMatrizen;

/* **************************

   The table with the fit matrices for each sl/sec looks like this:
   //2er
   0 ... 0 + Stereo(0) 
   1 ... 1 + Stereo(1) 
   2 ... Stereo(0) + 0
   3 ... Stereo(1) + 1  
   //3er
   4 ... 1 2 3
   5 ... 0 2 3
   6 ... 0 1 3
   7 ... 0 1 2
   //4er
   8 ... 0 1 2 3
   //again 2er 
   9 ... 0 + Orthog(0) 
   10... 1 + Orthog(1) 
   11... Orthog(0) + 0
   12... Orthog(1) + 1  
   --------------------
   13 matrices in total per sl and sec (*2 if you count measurement matrices ;-))
   -> 728 matrices. Great.
   
   ************************** */
   

void L3_SpacePoint::initFitMatrizen() {
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_SpacePoint::initFitMatrizen() has been called" << endl;
#endif
  for (int SL=0; SL<7; SL++) {
    for (int Sec=0; Sec<4; Sec++) {
      vector<L3_Plane> Plane;
      for (int Pl=0; Pl<4; Pl++) {
	Plane.push_back(L3_Geometry::plane(SL,Sec,Pl));
      }
      float Dist=0.;
      //generate matrices for two hits, the one with lower z first -> 2 matrices/sec/sl
      //angle between hits is 5 degree
      for (int l=0; l<=1; l++) { 
	int front = l;
	int back  = L3_Geometry::getPlaneIndex(SL,Sec,l,0);
 	if ( (Plane[front].l3nsl() != -1) && (Plane[back].l3nsl() != -1) ) { 
	  HepMatrix A(2, 2);
	  Dist = ((Plane[front].zfw(0) + Plane[back].zfw(0))* 0.5) - L3_Geometry::ztarget();
	  A[0][0]= ((Plane[front].zfw(0)-L3_Geometry::ztarget()) / Dist) * Plane[front].cos();
	  A[0][1]= -1. * ((Plane[front].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[front].sin());
	  A[1][0]= ((Plane[back].zfw(0)-L3_Geometry::ztarget()) / Dist) * Plane[back].cos();
	  A[1][1]= -1. * ((Plane[back].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[back].sin());
	  int ierr=0;
	  HepMatrix B(2, 2);
	  B = A.T() * A;
	  HepMatrix C(2, 2);
	  C = B.inverse(ierr);
	  HepMatrix Fit(2, 2);
	  Fit = C * A.T();
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
 	} else {
 	  HepMatrix Fit(2, 2, 0);
 	  HepMatrix A(2, 2, 0);
 	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
 	}
      }
      //generate matrices for two hits, the one with higher z first -> 2 matrices/sl/sec
      for (int l=0; l<=1; l++) { 
	int front = l;
	int back  = L3_Geometry::getPlaneIndex(SL,Sec,l,0);
	if ( (Plane[front].l3nsl() != -1) && (Plane[back].l3nsl() != -1) ) {
	  HepMatrix A(2, 2);
	  Dist = ((Plane[front].zfw(0) + Plane[back].zfw(0))* 0.5) - L3_Geometry::ztarget();
	  A[0][0]= ((Plane[back].zfw(0)-L3_Geometry::ztarget()) / Dist) * Plane[back].cos();
	  A[0][1]= -1. * ((Plane[back].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[back].sin());
	  A[1][0]= ((Plane[front].zfw(0)-L3_Geometry::ztarget()) / Dist) * Plane[front].cos();
	  A[1][1]= -1. * ((Plane[front].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[front].sin());
	  int ierr=0;
	  HepMatrix B(2, 2);
	  B = A.T() * A;
	  HepMatrix C(2, 2);
	  C = B.inverse(ierr);
	  HepMatrix Fit(2, 2);
	  Fit = C * A.T();
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	} else {
	  HepMatrix Fit(2, 2, 0);
	  HepMatrix A(2, 2, 0);
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	}
      }
      //generate matrices for 3 hits, nom denotes MISSING layer -> 4 matrices/sec/sl
      for (int nom=1; nom<=4; nom++) {  
	HepMatrix A(3, 2);
	if (nom == 1) {
	  Dist= ((Plane[1].zfw(0) 
		  + Plane[2].zfw(0) 
		  + Plane[3].zfw(0)) / 3.) - L3_Geometry::ztarget();
	  A[0][0]= ((Plane[1].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[1].cos();
	  A[0][1]= -1. * ((Plane[1].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[1].sin());
	  A[1][0]= ((Plane[2].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[2].cos();
	  A[1][1]= -1. * ((Plane[2].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[2].sin());
	  A[2][0]= ((Plane[3].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[3].cos();
	  A[2][1]= -1. * ((Plane[3].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[3].sin());
	}
	if (nom == 2) {
	  Dist= ((Plane[0].zfw(0)
		  + Plane[2].zfw(0)
		  + Plane[3].zfw(0)) / 3.) - L3_Geometry::ztarget();
	  A[0][0]= ((Plane[0].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[0].cos();
	  A[0][1]= -1. * ((Plane[0].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[0].sin());
	  A[1][0]= ((Plane[2].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[2].cos();
	  A[1][1]= -1. * ((Plane[2].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[2].sin());
	  A[2][0]= ((Plane[3].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[3].cos();
	  A[2][1]= -1. * ((Plane[3].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[3].sin());
	}
	if (nom == 3) {
	  Dist= ((Plane[0].zfw(0)
		  + Plane[1].zfw(0)
		  + Plane[3].zfw(0)) / 3.) - L3_Geometry::ztarget();
	  A[0][0]= ((Plane[0].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[0].cos();
	  A[0][1]= -1. * ((Plane[0].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[0].sin());
	  A[1][0]= ((Plane[1].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[1].cos();
	  A[1][1]= -1. * ((Plane[1].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[1].sin());
	  A[2][0]= ((Plane[3].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[3].cos();
	  A[2][1]= -1. * ((Plane[3].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[3].sin());
	}
	if (nom == 4) {
	  Dist= ((Plane[0].zfw(0)
		  + Plane[1].zfw(0)
		  + Plane[2].zfw(0)) / 3.) - L3_Geometry::ztarget();
	  A[0][0]= ((Plane[0].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[0].cos();
	  A[0][1]= -1. * ((Plane[0].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[0].sin());
	  A[1][0]= ((Plane[1].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[1].cos();
	  A[1][1]= -1. * ((Plane[1].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[1].sin());
	  A[2][0]= ((Plane[2].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[2].cos();
	  A[2][1]= -1. * ((Plane[2].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[2].sin());
	}
	HepMatrix B(3, 3);
	B = A.T() * A;
	int ierr=0;
	HepMatrix C(3, 3);
	C = B.inverse(ierr);
	if ( (Plane[1].l3nsl() != -1) && (Plane[2].l3nsl() != -1)
	     && (Plane[3].l3nsl() != -1) && (nom == 1) ) {
	  HepMatrix Fit(3, 3);
	  Fit = C * A.T();
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	} else if ( (Plane[0].l3nsl() != -1) && (Plane[2].l3nsl() != -1)
	     && (Plane[3].l3nsl() != -1) && (nom == 2) ) {
	  HepMatrix Fit(3, 3);
	  Fit = C * A.T();
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	} else if ( (Plane[0].l3nsl() != -1) && (Plane[1].l3nsl() != -1)
	     && (Plane[3].l3nsl() != -1) && (nom == 3) ) {
	  HepMatrix Fit(3, 3);
	  Fit = C * A.T();
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	} else if ( (Plane[0].l3nsl() != -1) && (Plane[1].l3nsl() != -1)
	     && (Plane[2].l3nsl() != -1) && (nom == 4) ) {
	  HepMatrix Fit(3, 3);
	  Fit = C * A.T();
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	} else {
	  HepMatrix Fit(3, 3, 0);
	  HepMatrix A_tmp(3, 2, 0);
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A_tmp);
	}
      }
      //generate matrix for spacepoints out of 4 hits -> 1 matrix/sec/sl
      if ( (Plane[0].l3nsl() != -1) && (Plane[1].l3nsl() != -1)
	   && (Plane[2].l3nsl() != -1) && (Plane[3].l3nsl() != -1) ) {
	HepMatrix A(4, 2);
	Dist= ((Plane[0].zfw(0) + Plane[1].zfw(0)
		+ Plane[2].zfw(0) + Plane[3].zfw(0)) / 4.) - L3_Geometry::ztarget();
	for (int i=0; i<=3; i++) {
	  A[i][0]= ((Plane[i].zfw(0) - L3_Geometry::ztarget()) / Dist) * Plane[i].cos();
	  A[i][1]= -1. * ((Plane[i].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[i].sin());
	}
	HepMatrix B(4, 4);
	B = A.T() * A;
	HepMatrix C(4, 4);
	int ierr=0;
	C = B.inverse(ierr);
	HepMatrix Fit(4, 4);
	Fit = C * A.T();
	FitMatrizen.push_back(Fit);
	MessMatrizen.push_back(A);
      } else {
	HepMatrix A(4, 2, 0);
	HepMatrix Fit(4, 4, 0);
	FitMatrizen.push_back(Fit);
	MessMatrizen.push_back(A);
      }
      //generate another permutation of 2 hit spacepoints (90 degree between hits)
      for (int l=0; l<=1; l++) { 
	int front = 0+2*l;
	int back  = 1+2*l;
	if ( (Plane[front].l3nsl() != -1) && (Plane[back].l3nsl() != -1) ) {
	  HepMatrix A(2, 2);
	  Dist = ((Plane[front].zfw(0) + Plane[back].zfw(0))* 0.5) - L3_Geometry::ztarget();
	  A[0][0]= ((Plane[front].zfw(0)-L3_Geometry::ztarget()) / Dist) * Plane[front].cos();
	  A[0][1]= -1. * ((Plane[front].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[front].sin());
	  A[1][0]= ((Plane[back].zfw(0)-L3_Geometry::ztarget()) / Dist) * Plane[back].cos();
	  A[1][1]= -1. * ((Plane[back].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[back].sin());
	  int ierr=0;
	  HepMatrix B(2, 2);
	  B = A.T() * A;
	  HepMatrix C(2, 2);
	  C = B.inverse(ierr);
	  HepMatrix Fit(2, 2);
	  Fit = C * A.T();
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	} else {
	  HepMatrix Fit(2, 2, 0);
	  HepMatrix A(2, 2, 0);
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	}
      }
      //generate another permutation of 2 hit spacepoints (90 degree between hits)
      for (int l=0; l<=1; l++) { 
	int front = 0+2*l;
	int back  = 1+2*l;
	if ( (Plane[front].l3nsl() != -1) && (Plane[back].l3nsl() != -1) ) {
	  HepMatrix A(2, 2);
	  Dist = ((Plane[front].zfw(0) + Plane[back].zfw(0))* 0.5) - L3_Geometry::ztarget();
	  A[0][0]= ((Plane[back].zfw(0)-L3_Geometry::ztarget()) / Dist) * Plane[back].cos();
	  A[0][1]= -1. * ((Plane[back].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[back].sin());
	  A[1][0]= ((Plane[front].zfw(0)-L3_Geometry::ztarget()) / Dist) * Plane[front].cos();
	  A[1][1]= -1. * ((Plane[front].zfw(0) - L3_Geometry::ztarget()) / Dist)
	    * (Plane[front].sin());
	  int ierr=0;
	  HepMatrix B(2, 2);
	  B = A.T() * A;
	  HepMatrix C(2, 2);
	  C = B.inverse(ierr);
	  HepMatrix Fit(2, 2);
	  Fit = C * A.T();
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	} else {
	  HepMatrix Fit(2, 2, 0);
	  HepMatrix A(2, 2, 0);
	  FitMatrizen.push_back(Fit);
	  MessMatrizen.push_back(A);
	}
      }
    } //sectors
  } //superlayer
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_SpacePoint::initFitMatrizen() is done" << endl;
#endif
}

void L3_SpacePoint::setFitMatrizen() {
  FitMatrizen.erase(FitMatrizen.begin(), FitMatrizen.end() );
  MessMatrizen.erase(MessMatrizen.begin(), MessMatrizen.end() );
  L3_SpacePoint::initFitMatrizen();
}

int L3_SpacePoint::gesl() const
{
  #define NN "L3_SpacePoint::gesl(): "
  if( hitList.size() < 1 ) 
    {
      cout << NN"You are asking a good question\n" << endl;
      return -1;
    }
  #ifdef HREF
  int rc = (hitList[0])->gesl();
  //be suspicious for the moment, check whether all hitbs have the same gesl
  for(vector<Hitb>::const_iterator i = hitList.begin()+1;i!=hitList.end();i++)
    {
      if( rc != (*i)->gesl() ) 
	{
	  cout << NN"gesls do not match" << endl;
	  for(vector<Hitb>::const_iterator ii = hitList.begin();ii!=hitList.end();ii++)
	    cout  << (*ii)->gesl() << ' ';
	  cout << endl;
	  break;
	}
    }
  #else
  int rc = (hitList[0]).gesl();
  //be suspicious for the moment, check whether all hitbs have the same gesl
  for(vector<Hitb>::const_iterator i = hitList.begin()+1;i!=hitList.end();i++)
    {
      if( rc != i->gesl() ) 
	{
	  cout << NN"gesls do not match" << endl;
	  for(vector<Hitb>::const_iterator ii = hitList.begin();ii!=hitList.end();ii++)
	    cout  << ii->gesl() << ' ';
	  cout << endl;
	  break;
	}
    }
  #endif
  #undef NN
  return rc;
}

int L3_SpacePoint::sec() const
{
  #define NN "L3_SpacePoint::sec(): "
  if( hitList.size() < 1 ) 
    {
      cout << NN"You are asking a good question\n" << endl;
      return -1;
    }
  #ifdef HREF
  int rc = (hitList[0])->sec();
  //be suspicious for the moment, check whether all hitbs have the same sector
  for(vector<Hitb>::const_iterator i = hitList.begin()+1;i!=hitList.end();i++)
    {
      if( rc != (*i)->sec() ) 
	{
	  cout << NN"secs do not match" << endl;
	  for(vector<Hitb>::const_iterator ii = hitList.begin();ii!=hitList.end();ii++)
	    cout << (*ii)->sec() << ' ';
	  cout << endl;
	  break;
	}
    }
  #else
  int rc = (hitList[0]).sec();
  //be suspicious for the moment, check whether all hitbs have the same sector
  for(vector<Hitb>::const_iterator i = hitList.begin()+1;i!=hitList.end();i++)
    {
      if( rc != i->sec() ) 
	{
	  cout << NN"secs do not match" << endl;
	  for(vector<Hitb>::const_iterator ii = hitList.begin();ii!=hitList.end();ii++)
	    cout << ii->sec() << ' ';
	  cout << endl;
	  break;
	}
    }
  #endif
  return rc;
  #undef NN
}

L3_SpacePoint::L3_SpacePoint(vector<L3_Hitb> &v){
  //cout << "L3_SpacePoint::L3_SpacePoint: " << v.size() << endl;
  vector<L3_Hitb> vx,vy;
  for(vector<L3_Hitb>::iterator i=v.begin();i!=v.end();i++){
    const L3_Plane& plane = i->getPlane();
    if( plane.l3nsl() != -1 ){
      if( fabs(plane.cos()) > 0.5 )
	vy.push_back(*i);
      else
	vx.push_back(*i);
    }
  }
  int nx = vx.size();
  int ny = vy.size();
  //cout << "L3_SpacePoint::L3_SpacePoint: " << nx << ' ' << ny << endl;
  if( nx==2 && ny==2 ){
    //cout << "x " << vx[0].lay() << ' ' << vx[1].lay() << endl;
    L3_SpacePoint point(vx[0],vx[1]);
    //cout << "y " << vy[0].lay() << ' ' << vy[1].lay() << endl;
    point.add(vy[0],vy[1]);
    //cout << "copy" << endl;
    *this = point;
    //cout << "done" << endl;
  } else if( nx==2 && ny==1 ){
    L3_SpacePoint point(vx[0],vx[1]);
    point.add(vy[0]);
    *this = point;
  } else if( ny==2 && nx==1 ){
    L3_SpacePoint point(vy[0],vy[1]);
    point.add(vx[0]);
    *this = point;
  } else if( nx==1 && ny==1 ){
    L3_SpacePoint point(vx[0],vy[0],0);
    *this = point;
  } else if( nx==2 && ny==0 ) {
    L3_SpacePoint point(vx[0],vx[1]);
    *this = point;
  } else if( ny==2 && nx==0 ) {
    L3_SpacePoint point(vy[0],vy[1]);
    *this = point;
  } else {
    if(L3_Parameter::Debug()) cout << "L3_SpacePoint::L3_SpacePoint(vector<L3_Hitb> &v) with nx=" << nx << " and ny=" << ny << endl;
    derPunkt.setX(invalidX);
    derPunkt.setY(invalidY);
    derPunkt.setZ(invalidZ);
    the_chiSquare = invalidChiSquare;
    hitList = v;
  }
  //cout << "L3_SpacePoint::L3_SpacePoint: done" << endl;
}






