#ifndef __WAX_PARIZATION_HH__
#define __WAX_PARIZATION_HH__

#ifndef __WAX_FUNCTIONVALUE_HH__
#include "wax/FunctionValue.hh"
#endif
#ifndef __SGI_STL_VECTOR
#include <vector>
#endif
#ifndef _Vector_H_
#include "CLHEP/Matrix/Vector.h"
#endif
#ifndef _SYMMatrix_H_
#include "CLHEP/Matrix/SymMatrix.h"
#endif

/** Base class for parameterization on [xmin,xmax) **/
class TH1 ;

class waxParization
{
public:
  // typedefs, enumerates
  typedef float XValue ;
  typedef float YValue ;
  enum EParType { kSpline=1, kChebychev=2 } ;
  
  // destructor 
  static waxParization* Create() ;
  static waxParization* Create(XValue xmin, XValue xmax, 
			       int degree=0,int npars=0,YValue y=0) ;
  virtual ~waxParization() {}

  // setters and getters
  const HepVector& GetPars() const { return fPars ; }
  HepVector& GetPars() { return fPars ; }
  const HepSymMatrix& GetCov() const { return fCov ; }
  void SetPars(const HepVector& pars,const HepSymMatrix& cov=0 ) ;
  int GetNPars() const { return fPars.num_row() ; }

  XValue GetXMin() const { return fKnots[0] ; }
  XValue GetXMax() const { return fKnots[fKnots.num_row()-1] ; }
  void SetKnots(const HepVector& knots) { fKnots = knots ; }
  const HepVector& GetKnots() const { return fKnots ; }
  float FitParsFromArray(int npoints, float* xvec,float* yvec, float* yerrvec) ;
  
  // implemented by derived class
  virtual void SetXBounds(XValue xmin, XValue xmax) = 0 ;
  virtual int GetDegree() const { return 0; }
#if !defined(__CINT__) && !defined(CINTDICT)
  virtual std::vector<waxFunctionValue> CalcBasicFuncVector(XValue x) const = 0 ;
#endif
  virtual HepVector CalcValueVector(XValue x) const = 0 ;
  virtual void ChangeNPars(int npar, int degree=-1) = 0;
  virtual void CalcParsFromFunction(int, float*, float*) = 0 ;
  virtual void SetToConstant(const float x) = 0 ;
  virtual HepVector GetSchoenbergKnots() const= 0 ;
  virtual void ChangeKnotsFromDistribution(XValue,XValue,int,float*,
					   int fixmin=0,int fixmax=0) = 0;

  // IO
  virtual void Read(int format,std::istream&) ;
  virtual void Write(int format,std::ostream&) const ;
  virtual void Read(XValue,XValue,int,std::istream&) = 0 ; //obsolete

  // other methods
  YValue CalcValue(XValue x) { return dot(CalcValueVector(x),fPars) ;}
  YValue CalcVariance(XValue x) { return fCov.similarity(CalcValueVector(x)) ;}
  waxFunctionValue CalcFunction(XValue x) ;
  waxFunctionValue CalcInverse(YValue y, YValue precision) ; // inverse function
protected:
  waxParization(int npars=0) : fKnots(npars+1,0),fPars(npars,0), fCov(npars,0) {}
  waxParization(const HepVector& knots, const HepVector& pars, 
		const HepSymMatrix& cov) : fKnots(knots),fPars(pars), fCov(cov) {}
private:
  HepVector fKnots ;
  HepVector fPars ;
  HepSymMatrix fCov ;
};

class waxSplineParization : public waxParization
{
public:
  // constructors, destructors
  waxSplineParization(XValue xmin, XValue xmax,int degree,int npars,YValue y) ;
  //   waxSplineParization(XValue xmin, XValue xmax,int degree,
  // 		      const HepVector& pars,const HepSymMatrix& cov) ;
  //   waxSplineParization(XValue xmin, XValue xmax, int degree, int npars,
  // 		      int npoints, float* xpoints, float* ypoints) ;
  //   waxSplineParization(HepVector knots, int degree, 
  // 		      int npoints, float* xpoints, float* ypoints) ;
  // setters and getters
  
  // virtual inlines
  virtual HepVector GetSchoenbergKnots() const { 
    return CalcSchoenbergKnots(GetDegree(),GetKnots()) ; }
  virtual int GetDegree() const { 
    return GetDegree(GetKnots().num_row(),GetPars().num_row()) ; }
  
  // declared by base class
  virtual void Read(XValue,XValue,int,std::istream&) ; //obsolete
  virtual void SetXBounds(XValue xmin, XValue xmax) ;
  //virtual int GetDegree() const { return fDegree ; }
#if !defined(__CINT__) && !defined(CINTDICT)
  virtual std::vector<waxFunctionValue> CalcBasicFuncVector(XValue x) const ;
#endif
  virtual HepVector CalcValueVector(XValue x) const ;
  virtual void ChangeNPars(int n, int degree=-1) ;
  virtual void CalcParsFromFunction(int, float*, float*) ;
  virtual void ChangeKnotsFromDistribution(XValue,XValue,int,float*,int,int) ;
  virtual void SetToConstant(const float x)  ;
public: // statics
  static int GetDegree(int nknots,int npars) { return nknots-npars-1; }
  static HepVector CalcKnotsFromDistribution(int k,int n,XValue xmin,
					     XValue xmax,int nbins,float* entries,
					     int fixmin, int fixmax) ;
  static HepVector CalcEquidistantKnots(int degree, int npar, 
					XValue xmin,XValue xmax) ;
  static HepVector CalcEquidistantKnots(int degree, int npar, 
					XValue xmin,XValue xmax,int fixmin,int fixmax) ;
  static HepVector CalcSchoenbergKnots(int degree,const HepVector& knots) ;
  static double EvaluateBSpline(const HepVector&,
				const int, const int, const double) ;
  static double EvaluateBSplineDerivative(const HepVector&,
					  const int, const int, const double) ;
  static double EvaluateBSplineSecondDerivative(const HepVector&,
						const int, const int, const double) ;
private:
  void SchoenbergUpdate(const HepVector& knots_new, int npar_new) ;
private:
  //int       fDegree ;
};

#endif //__WAX_PARIZATION_HH__
