#include <iostream.h>
#include <iomanip.h>
#include "wax/Parization.hh"

/********** waxParization *******************/

waxParization* waxParization::Create()
{
  waxParization* par(0);
  par = new waxSplineParization(-1.,1.,0,0,0) ;
  return par;
}

waxParization*
waxParization::Create(XValue xmin, XValue xmax, int degree,
		      int npars, YValue y)
{
  waxParization* par(0);
  par = new waxSplineParization(xmin,xmax,degree,npars,y) ;
  return par;
}

void
waxParization::SetPars(const HepVector& pars, const HepSymMatrix& cov) 
{
  fPars = pars ; 
  if(cov.num_row() == fPars.num_row()) fCov = cov ;
  else fCov = HepSymMatrix(fPars.num_row(),0) ;
}

waxFunctionValue
waxParization::CalcFunction(XValue x) 
{
  if( x<GetXMin() ) x = GetXMin() ;
  if( x>GetXMax() ) x = GetXMax() ;
  std::vector<waxFunctionValue> bfv = CalcBasicFuncVector(x) ;
  waxFunctionValue rv;
  for(int i=0; i<fPars.num_row(); i++) rv += fPars[i]*bfv[i] ;
  rv.fX = bfv[0].fX ;
  return rv ;
}

waxFunctionValue 
waxParization::CalcInverse(YValue y, YValue precision)
{
  //  std::cout << std::endl << "waxParization::CalcInverse(...)" << std::endl ;
  // calculate inverse using newton
  const int maxNSteps = 5 ;
  XValue x = (GetXMax()-GetXMin())/2 ;
  waxFunctionValue fvinx=CalcFunction(x) ;
  YValue deltay = fvinx.fY-y ;
  int nsteps = 0;
  while(fabs(deltay)>precision && nsteps < maxNSteps && fvinx.fdYdX!=0) {
    x -= deltay/fvinx.fdYdX ;
    fvinx=CalcFunction(x) ;
    deltay = fvinx.fY-y ;
    nsteps++;
  }
  //fvinx.fX = x ;
  return fvinx ;
} 


void waxParization::Read(int format, std::istream& ifs) 
{
  //std::cout << "waxParization::Read " << format << std::endl ;
  if(format>=4) {
    int nknots,npars ;
    ifs >> nknots >> npars ;
    fKnots = HepVector(nknots) ;
    for(int row=0; row<nknots; row++) ifs >> fKnots[row] ;
    fPars = HepVector(npars) ;
    for(int row=0; row<npars; row++) ifs >> fPars[row] ;
    fCov = HepSymMatrix(npars) ;
    for(int row=0; row<npars; row++)
      for(int col=0; col<=row; col++) ifs >> fCov[row][col] ;
  }else {
    std::cout << "waxParization::Read(int format, istream& ifs) : " << std::endl 
	 << "Error: file format " << format << " not supported." << std::endl ; 
  }
  //std::cout << "end of waxParization::Read " << std::endl ;
}

void waxParization::Write(int format, std::ostream& ofs) const
{
  //std::cout << "waxParization::Write " << format << std::endl ;
  if(format>=4) {
    int nknots = fKnots.num_row() ;
    int npars = fPars.num_row() ;
    ofs << setw(5) << setprecision(3) << nknots 
	<< setw(5) << setprecision(3) << npars << std::endl ;
    for(int row=0; row<nknots; row++) 
      ofs << setw(12) << setprecision(5) << fKnots[row] ;
    ofs << std::endl ;
    for(int row=0; row<npars; row++)
      ofs << setw(12) << setprecision(5) << fPars[row] ;
    ofs << std::endl ;
    for(int row=0; row<npars; row++) {
      for(int col=0; col<=row; col++)
	ofs << setw(12) << setprecision(5) << fCov[row][col] ;
      ofs << std::endl ;
    }
  } else {
    std::cout << "waxParization::Write(int format, istream& ifs) : " << std::endl 
	 << "Error: file format " << format << " not supported." << std::endl ; 
  }
  // std::cout << "end of  waxParization::Write " << std::endl ;
}

/********* one leftover IO from waxSplineParization *********/

void waxSplineParization::Read(XValue xmin, XValue xmax, int format, std::istream& ifs) 
{
  //std::cout << "waxSplineParization::Read(....)" << std::endl ;
  // This routine is obsolete, but needs to stay here till all rtrs are of format >=4
  if(format >=4 ) {
    std::cout << "waxSplineParization::Read(XValue xmin, XValue xmax, int format, istream& ifs) " << std::endl
	 << "This routine is obsolete, do not call for format >= 4" << std::endl ;
    waxParization::Read(format,ifs) ;
  } else {
    int npars, degree ;
    if(format==2 || format==3) {
      ifs >> degree >> npars ;
    } else {
      ifs >> npars ;
      degree = min(npars-1,3) ;
    }
    HepVector pars = HepVector(npars) ;
    for(int row=0; row<npars; row++) ifs >> pars[row] ;
    HepSymMatrix cov = HepSymMatrix(npars) ;
    for(int row=0; row<npars; row++)
      for(int col=0; col<=row; col++) ifs >> cov[row][col] ;
    SetPars(pars,cov) ;
    SetKnots(CalcEquidistantKnots(degree,npars,xmin,xmax)) ;
  }
  //std::cout << "end of waxSplineParization::Read(....)" << std::endl ;
}

/********** waxSplineParization *******************/

waxSplineParization::waxSplineParization(XValue xmin, XValue xmax, int degree,
					 int npars, YValue y)
  : waxParization()
{
  if(npars < degree+1) npars = degree+1 ;
  SetKnots(CalcEquidistantKnots(degree,npars,xmin,xmax)) ;
  HepVector pars(npars) ;
  for(int i=0; i<npars; i++) pars[i]=y;
  SetPars(pars) ;
}


std::vector<waxFunctionValue> 
waxSplineParization::CalcBasicFuncVector(XValue x) const
{
  int npar = GetNPars() ;
  std::vector<waxFunctionValue> vec(npar) ;
  XValue xtmp = x;
  if(xtmp >= GetXMax()) xtmp = GetXMax()-0.0001*(GetXMax()-GetXMin()) ;
  else if(xtmp < GetXMin()) xtmp = GetXMin() ;
  for(int ipar=0; ipar<npar; ipar++) {
    vec[ipar].fY    = EvaluateBSpline(GetKnots(), ipar, GetDegree(), xtmp) ;
    vec[ipar].fdYdX = EvaluateBSplineDerivative(GetKnots(), ipar, GetDegree(), xtmp) ;
    vec[ipar].fY += (x-xtmp)*vec[ipar].fdYdX ;
    //    vec[i].d2dx2 =
    //       EvaluateBSplineSecondDerivative(fKnots, ipar, degree, X) ;
    vec[ipar].fX    = x;
  }
  return vec ;
}

HepVector 
waxSplineParization::CalcValueVector(XValue x) const 
{
  if(x<GetXMin() )  x = GetXMin() ;
  if(x>=GetXMax() ) x = GetXMax()-0.0001*(GetXMax()-GetXMin()) ;
  HepVector Y(GetNPars()) ;
  for(int ipar=0; ipar<Y.num_row(); ipar++)
    Y[ipar] = EvaluateBSpline(GetKnots(),ipar,GetDegree(),x) ;
  return Y;
}

void
waxSplineParization::CalcParsFromFunction(int npoints, float* xpoints,float* ypoints)
{
  //std::cout << "waxSplineParization::CalcParsFromFunction(...)" << std::endl ;
  // Calc pars from function given by (xpoints,ypoints). xpoints must
  // be sorted x[i]<x[i+1], but not necessarily
  // equidistant. interpolation performed
  
  HepVector schoenbergknots = CalcSchoenbergKnots(GetDegree(),GetKnots()) ;
 
  HepVector newpars(GetNPars()) ;
  for(int j=0; j< GetNPars() ; j++) {
    float x = schoenbergknots[j] ;
    if(xpoints[0] >= x ) newpars[j] = ypoints[0] ;
    else if(xpoints[npoints-1] <= x ) newpars[j] = ypoints[npoints-1] ;
    else { 
      int i ;
      for(i=1; i<npoints; i++) if(xpoints[i]>x) break ;
      newpars[j] = ypoints[i-1] + (ypoints[i] - ypoints[i-1])*
	(x - xpoints[i-1])/(xpoints[i] - xpoints[i-1]) ;
    }
  } 
  SetPars(newpars) ;
}

void
waxSplineParization::ChangeNPars(int n_new, int k_new)
{
  if( (k_new<0) || (k_new+1>n_new)) k_new = min(n_new-1,3) ;
  if( (n_new != GetNPars()) || (k_new != GetDegree()) ) {
    HepVector knots_new = CalcEquidistantKnots(k_new,n_new,GetXMin(),GetXMax()) ;
    SchoenbergUpdate(knots_new,n_new) ;
  }
  //std::cout << "end of waxSplineParization::ChangeNPars" << endl ;
}

void
waxSplineParization::SetXBounds(XValue xmin, XValue xmax) 
{
  if(xmin != GetXMin() || xmax != GetXMax()) {
    // first scale knots to new interval
    int nknots = GetKnots().num_row() ;
    HepVector knots_new(nknots) ;
    const HepVector& knots_old = GetKnots();
    float dxdk = (xmax-xmin)/(knots_old[nknots-1]-knots_old[0]) ;
    for(int j=0; j<nknots; j++)
      knots_new[j] = xmin + dxdk * (knots_old[j]-knots_old[0]) ;
    // update parameters
    SchoenbergUpdate(knots_new,GetNPars()) ;
  }
}

void
waxSplineParization::ChangeKnotsFromDistribution(XValue xmin, XValue xmax,
						 int nbins, float* entries,
						 int fixmin, int fixmax)
{
  HepVector knots_new = 
    CalcKnotsFromDistribution(GetDegree(),GetNPars(),xmin,xmax,nbins,
			      entries,fixmin,fixmax) ;
  if(knots_new.num_row()==GetKnots().num_row()) SetKnots(knots_new) ;
  else SchoenbergUpdate(knots_new,GetNPars()) ;
}

void
waxSplineParization::SchoenbergUpdate(const HepVector& knots_new, int npar_new)
{
  const float dx = 0.001* (GetXMax() - GetXMin()) ;
  if( (npar_new != GetNPars()) || (&(knots_new) != &(GetKnots())) ) {
    //   std::cout << "waxSplineParization::UpdateTo" << npar_new << " "
    // 	 << knots_new.num_row() << std::endl ;
    int k_new = GetDegree(knots_new.num_row(),npar_new) ;
    HepVector schoenbergknots =CalcSchoenbergKnots(k_new, knots_new) ;
    HepVector par_new = HepVector(npar_new,0) ;
    for(int i=0; i<npar_new; i++) {
      float x = schoenbergknots[i] ;
      if( x>GetXMax()-dx ) x = GetXMax()-dx ;
      if( x<GetXMin()+dx ) x = GetXMin()+dx;
      float y(0),dydx(0) ;
      for(int j=0; j<GetNPars(); j++) {
	y    += GetPars()[j]*EvaluateBSpline(GetKnots(),j,GetDegree(),x) ;
	dydx += GetPars()[j]*EvaluateBSplineDerivative(GetKnots(),j,GetDegree(),x) ;
      }
      par_new[i] = y + (schoenbergknots[i]-x)*dydx ;
      //cout << "i,s,p: " << i << " " << schoenbergknots[i] << " "
      //	   << par_new[i] << endl ;
    }
    SetPars(par_new) ;
    SetKnots(knots_new) ;
  }
}



/* These are static functions. They just calculate some general spline
   things ...  */

HepVector
waxSplineParization::CalcSchoenbergKnots(const int k, 
					 const HepVector& knots)
{   
  int npar = knots.num_row() - k - 1 ;
  HepVector schoenbergknots(npar) ;
  if(k==0) for(int i=0; i<npar; i++)
    schoenbergknots[i] = (0.5*(knots[i]+knots[i+1])) ;
  else for(int i=0; i< npar ; i++) {
    double sum(0.) ;
    for(int j = i+1 ; j < i + k + 1; j++) sum += knots[j];
    schoenbergknots[i] = sum/k ;
  }
  return schoenbergknots ;
}

#define FUF 1e-16

HepVector
waxSplineParization::CalcEquidistantKnots(const int k /* degree */, 
					  const int n /* number of parameters */,
					  XValue xmin,XValue xmax)
{
  // Calculates equidistant knots belonging to a spline of
  // degree k, with n parameters, and width boudaries {-1,1].
  // The knots on the boundaries are of multiplicity k.

  int m = n + k + 1; // # knots 
  HepDouble deltaXi = (xmax-xmin)/(m - 2*k - 1) ;
  HepVector knots(m) ;
  int i;
  for(i=0; i<k+1; i++)     knots[i] = xmin ;
  for(i=k+1; i<m-k-1; i++) knots[i] = xmin + (i - k) * deltaXi ;
  for(i=m-k-1; i<m; i++)   knots[i] = xmax ;
  return knots ;
}
HepVector

waxSplineParization::CalcEquidistantKnots(const int k /* degree */, 
					  const int n /* number of parameters */,
					  XValue xmin,XValue xmax,
					  int fixmin, int fixmax)
{
  // Calculates equidistant knots belonging to a spline of degree k,
  // with n parameters, and width boudaries {-1,1].  The knots on the
  // boundaries are of multiplicity k, unless fixmin or
  // fixmax=specified
  int m     = n + k + 1; // # knots 
  int mmin  = max(k+1 - fixmin,1) ;
  int mmax  = max(k+1 - fixmax,1) ;
  HepDouble deltaXi = (xmax-xmin)/(m - mmin - mmax + 1) ;
  HepVector knots(m) ;
  int i;
  for(i=0; i<mmin; i++)      knots[i] = xmin ;
  for(i=mmin; i<m-mmax; i++) knots[i] = xmin + (i - mmin + 1) * deltaXi ;
  for(i=m-mmax; i<m; i++)    knots[i] = xmax ;
  return knots ;
}

HepVector
waxSplineParization::CalcKnotsFromDistribution(int k, int n,
					       XValue xmin, XValue xmax,
					       int nbins, float* entries, 
					       int fixmin, int fixmax)
{
  // calc bins in flat space
  HepVector xiknots = CalcEquidistantKnots(k,n,0,1,fixmin,fixmax) ;
  // check whether xmin <= xmax
  if(xmin > xmax) swap(xmin,xmax) ;
  // normalize spectrum
  float norm = 0;
  for(int i=0; i<nbins; i++) norm += entries[i] ;
  // now transform xiknots to xknots:
  HepVector xknots(xiknots.num_row()) ;
  for(int j=0; j<xknots.num_row(); j++)
    if(xiknots[j]<=0) xknots[j] = xmin ;
    else if(xiknots[j]>=1) xknots[j] = xmax ;
    else {
      float sum=0;
      int i=0;
      for(i=0; i<nbins; i++) {
	sum += entries[i] ;
	if(sum >= norm*xiknots[j]) break ; }
      float x1 = xmin + (i-0.5)*(xmax-xmin)/nbins ;
      float x2 = xmin + (i+0.5)*(xmax-xmin)/nbins ;
      xknots[j] = x2 + (norm*xiknots[j]-sum)/entries[i] * (x2-x1) ;
    }
  //  std::cout << "waxSplineParization::CalcKnotsFromDistribution()" 
  //	    << k << " " << n << " " << xmin << " " << xmax << " "
  //	    << " xi knots: " << xiknots
  //	    << " x knots:  " << xknots << std::endl ;
  return xknots ;
}

float
waxParization::FitParsFromArray(int npoints, float* xvec,float* yvec, float* yerrvec)
{
  HepVector    hdChi2dPar(GetNPars(),0) ;
  HepSymMatrix hd2Chi2dPar2(GetNPars(),0) ;
  double chi2 = 0 ;

  const float dx = 0.00001*(GetXMax()-GetXMin()) ;
  for(int i=0; i<npoints; i++) {
    float x = xvec[i] ;
    if(x>=GetXMax()) x -= dx ;
    if(x>=GetXMin() && x<GetXMax()) {
      HepVector basisfuncvalues = CalcValueVector(x) ;	
      double residual = yvec[i] - dot(basisfuncvalues,fPars) ;
      double weight   = 1/(yerrvec[i]*yerrvec[i]) ;
      chi2 += weight*residual*residual ;
      for(int row=0; row<hdChi2dPar.num_row(); row++) {
	hdChi2dPar[row] += - weight * residual * basisfuncvalues[row] ;
	for(int col=0; col<=row; col++)
	  hd2Chi2dPar2[row][col] += weight * basisfuncvalues[row] * basisfuncvalues[col] ;
      }
    } else {
      cout << "not using point: " << i << " " << x << " " 
	   << yvec[i] << " " << yerrvec[i] << endl ;
    }
  }
  
  int ierr=0;
  HepSymMatrix hd2Chi2dPar2Inv = hd2Chi2dPar2.inverse(ierr) ;
  if(ierr) {
    cout << "Inversion error" << endl ;
    for(int i=0; i<npoints; i++) 
       cout << "i,x,y,yerr: " << i << " " << xvec[i] << " " 
	    << yvec[i] << " " << yerrvec[i] << endl ;
    return -1;
  }
  
  chi2  += - hd2Chi2dPar2Inv.similarity(hdChi2dPar) ;
  fPars +=  -hd2Chi2dPar2Inv*hdChi2dPar ;
  fCov  = hd2Chi2dPar2Inv ;

  return chi2 ;
}

void waxSplineParization::SetToConstant(const float x) {
  HepVector newpars = GetPars() ;
  for(int row=0; row<newpars.num_row() ; row++) newpars[row]=x ;
  SetPars(newpars) ;
}

//Evaluate j'th basis spline of degree k in x on knots t
double
waxSplineParization::EvaluateBSpline(const HepVector& t,
				     const int j, const int k, 
				     const double x)
{
  //std::cout << "SplineRTRelation::EvaluateBSpline()" << std::endl;
  if(j < 0 || j+k+1 >= t.num_row() ) return 0 ;
  //if(t[j+k+1] <= x || t[j] > x) return 0 ;
  if( (t[j]>x) || (t[j+k+1]<x) || ( (j+k+1<t.num_row()-1) && t[j+k+1]<=x ) ) return 0 ;

  if(k==0) return 1 ;
  
  double v ;
  if(fabs(t[j+k]-t[j])< FUF) v = 0 ;
  else v = (x-t[j])/(t[k+j] - t[j]) * EvaluateBSpline(t, j, k-1, x) ;

  double w ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) w = 0 ;
  else w = (t[j+k+1]-x)/(t[j+k+1]-t[j+1]) * EvaluateBSpline(t, j+1, k-1, x) ;
  
  double rv = v + w ; 
  if(fabs(rv) < FUF) rv = 0;
  return rv ;
};

double
waxSplineParization::EvaluateBSplineDerivative(const HepVector& t,
					       const int j, const int k, 
					       const double x) 
{
  if(k==0) return 0 ;
  
  double vbar ;
  if(fabs(t[j+k]-t[j])< FUF) vbar = 0 ;
  else vbar = 1./(t[k+j] - t[j]) * EvaluateBSpline(t, j, k-1, x) ;

  double wbar ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) wbar = 0 ;
  else wbar = 1./(t[j+k+1]-t[j+1]) * EvaluateBSpline(t, j+1, k-1, x) ; 

  double rv = k*(vbar-wbar) ; 
  if(fabs(rv) < FUF) rv = 0;
  return rv ;
}


double
waxSplineParization::EvaluateBSplineSecondDerivative(const HepVector& t,
						     const int j, const int k, 
						     const double x) 
{
  if(k<=1) return 0 ;
  
  double vhat ;
  if(fabs(t[j+k]-t[j])< FUF) vhat = 0 ;
  else vhat = 1./(t[k+j] - t[j]) * 
         EvaluateBSplineDerivative(t, j, k-1, x) ;

  double what ;
  if(fabs(t[j+k+1]-t[j+1])< FUF) what = 0 ;
  else what = 1./(t[j+k+1]-t[j+1]) * 
         EvaluateBSplineDerivative(t, j+1, k-1, x) ;
  double rv = k*(vhat - what) ;
  if(fabs(rv) < FUF) rv = 0;
  return rv ; 
}

#undef FUF

